﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
//using Zcu.CurvatureBenchmark.DataSources;
using Zcu.CurvatureBenchmark.Framework;
using Zcu.CurvatureBenchmark.Framework.Evaluators;
using Zcu.CurvatureBenchmark.Framework.Estimators;
using System.IO;
using System.Xml;
using System.Diagnostics;
using System.Reflection;
using System.Threading;

namespace Zcu.CurvatureBenchmark.BatchTest
{
    public class BatchTest
    {
        static TestScenario modules;
        static Log log;

        static List<String> errors = new List<string>();
        static List<String> critical = new List<string>();

        public static string currentSource, currentEstimator, currentDistorter;
        public static double sourceProgress, meshProgress, distorterProgress, estimatorProgress;

        public static void Main(string[] args)
        {
            foreach (string s in args)
                if (s == "-p")
                {
                    ParallelExecute(args);
                    return;
                }

            bool dontWait = false;
            foreach (string s in args)
                if (s == "-d")
                    dontWait = true;

            Output.SetDefault();

            Output.OnError += Output_OnError;

            Output.OnCritical += Output_OnCritical;

            string testFileName;
            
            if (args.Length == 0)
            {
                testFileName = TestFilename();
            }
            else
            {
                testFileName = args[0];
            }

            RunTest(dontWait, testFileName, module_Loaded);
        }

        public static void RunTest(bool dontWait, string testFileName, NotifyModuleLoaded notification)
        {
            RandomExt.Seed = 5;
            modules = new TestScenario(testFileName, notification);
            log = new Log(testFileName);

            Run();

            if (errors.Count > 0)
            {
                Output.Warning((IModule)null, "There were the following errors:\n\t{0}", string.Join("\n\t", errors));
            }

            if (critical.Count > 0)
            {
                Output.Warning((IModule)null, "There were the following critical errors:\n\t{0}", string.Join("\n\t", critical));
            }

            if (dontWait)
            {
                Output.Success((IModule)null, "All done.");
            }
            else
            {
                Output.Success((IModule)null, "All done. Press any key...");
                Console.ReadKey();
            }
        }

        private static void ParallelExecute(string[] args)
        {
            Output.SetDefault();

            string testFileName;
            if (args.Length == 1)
            {
                testFileName = TestFilename();
            }
            else
            {
                testFileName = args[0];
            }

            XmlDocument test = new XmlDocument();
            test.Load(testFileName);
            XmlNode root = test.DocumentElement;
            List<XmlDocument> parts = new List<XmlDocument>();
            foreach(XmlNode child in root.ChildNodes)
            {
                if (child.Name == "Sources")
                {
                    foreach (XmlNode sourceNode in child.ChildNodes)
                    {
                        if (sourceNode.NodeType == XmlNodeType.Comment)
                            continue;
                        XmlDocument part = (XmlDocument)test.Clone();
                        XmlNode partRoot = part.DocumentElement;
                        bool found = false;
                        foreach (XmlNode partChild in partRoot.ChildNodes)
                        {
                            if (partChild.Name == "Sources")
                            {
                                partChild.RemoveAll();
                                if (!found)
                                {
                                    found = true;

                                    XmlNode partSource = part.ImportNode(sourceNode, true);
                                    partChild.AppendChild(partSource);
                                }
                            }
                        }
                        parts.Add(part);
                    }
                }
                
            }

            int procCount = Environment.ProcessorCount;

            string[] partNames = new string[parts.Count];

            int index = 0;
            Process[] workers = new Process[procCount];
            while (index < parts.Count)
            {
                Output.Info((IModule)null, "Starting part " + index.ToString());
                string partFN = String.Format("{0}_pt{1}.xml", Path.GetFileNameWithoutExtension(testFileName), index);
                partNames[index] = partFN;
                parts[index].Save(partFN);

                bool started = false;
                while(!started)
                {
                    for (int i = 0;i<workers.Length;i++)
                    {
                        if ((workers[i] == null)||(workers[i].HasExited))
                        {
                            ProcessStartInfo psi = new ProcessStartInfo(Assembly.GetEntryAssembly().Location, partFN + " -d");
                            workers[i] = Process.Start(psi);
                            Thread.Sleep(1000); // don't start all the processes at once, so that they will output in different directories
                            started = true;
                            Output.Info((IModule)null, String.Format("Part {0} started", index));
                            break;
                        }
                    }
                    Thread.Sleep(1000);
                }
                index++;
            }
            foreach (Process p in workers)
                p.WaitForExit();

            foreach (string s in partNames)
                File.Delete(s);

            Output.Success((IModule)null, "All done. Press any key...");
            Console.ReadKey();
        }

        private static void Output_OnCritical(OutputEventArguments args)
        {
            critical.Add(string.Format("{0}: {1}", args.Module, args.Message));    
        }

        private static void Output_OnError(OutputEventArguments args)
        {
            errors.Add(string.Format("{0}: {1}", args.Module, args.Message));
        }

        static public void Run()
        {
            int meshIndex = 0;
            int sourceIndex = 0;
            foreach (var source in modules.sources)
            {
				if (source.Active == false) 
				{
					continue;
				}

                Console.WriteLine("Datasource: {0} {1}", source.Name, source.ToString());
                log.LogValue("SourceType", source.Name);
                log.LogValue("Source", source);
                log.LogValue("SourceIndex", sourceIndex);
                sourceIndex++;
                sourceProgress = (double)sourceIndex / modules.sources.Count;
                currentSource = source.ToString();
                
                TriangleMesh[] data = source.GenerateData();

                int imesh = 0;
                foreach (var mesh in data)
                {
                    
                    log.LogValue("MeshIndex", meshIndex);
                    meshIndex++;

                    /*// let the evaluators know that a new dataset is coming
                    foreach (ICurvatureEvaluator e in modules.evaluators)
                        e.NewDataset();*/

                    log.LogValue("DataIndex", imesh);
                    Console.WriteLine("\tMesh: {0}", imesh++);
                    meshProgress = (double)imesh / data.Length;

                    int idistorter = 0;
                    foreach (var distorter in modules.distorters)
                    {
                        idistorter++;
                        distorterProgress = (double)idistorter / modules.distorters.Count;
                        currentDistorter = distorter.ToString();

                        if (distorter.Active == false)
                        {
                            continue;
                        }

                        log.LogValue("Distorter", distorter);
                        Console.WriteLine("\tDistorter: {0}", distorter);
                        var distMesh = distorter.Distort(mesh);

                        TriangleMeshStat stats = new TriangleMeshStat(distMesh, true, true, 2, 2, source, distorter);
                        stats.DumpStatsToFile(log.StatsFileName);

                        int iestimator = 0;
                        foreach (var estimator in modules.estimators)
                        {
                            iestimator++;
                            estimatorProgress = (double)iestimator / modules.estimators.Count;
                            currentEstimator = estimator.ToString();

                            if (estimator.Active == false)
                            {
                                continue;
                            }
                            log.LogValue("Estimator", estimator);

                            Console.WriteLine("\t\tEstimator: {0}", estimator);
                            estimator.EstimateCurvatures(distMesh);
                            foreach (var evaluator in modules.evaluators)
                            {
                                if (evaluator.Active == false)
                                {
                                    continue;
                                }

                                Console.WriteLine("\t\tEvaluator: {0}", evaluator);
                                log.LogValues(evaluator.Evaluate(distMesh, estimator)); // it does not matter whether mesh or distMesh is passed, since distorter does not change the exact curvatures.
                            }
                            log.NewLine();
                        }
                    }
                }
            }
            log.WriteSummary();
        }

        static void module_ProgressChanged(IModule module, NotifyProgressChangedEventArgs args)
        {
            Console.CursorLeft = 0;
            Console.Write("{0}%", args.Progress);
            if (args.Progress == 100) Console.WriteLine();
        }

        static void module_Loaded(IModule module)
        {
            module.ProgressChanged += module_ProgressChanged;
        }

        static string TestFilename()
        {
            var files = Directory.GetFiles("./", "*.xml");
            if (files.Length == 0)
            {
                Console.WriteLine("No test file were found.");
                Environment.Exit(0);
            }
            for (int i = 0; i < files.Length; i++)
            {
                Console.WriteLine("{0}: {1}", i, files[i]);
            }
            Console.Write("Choose test file: ");
            int selection;
            int.TryParse(Console.ReadLine(), out selection);
            return (files[selection]);
        }
    }
}
