﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Zcu.Mve.ZMalik.SlimDXRenderer;
using Zcu.CurvatureBenchmark.DataSources;
using Zcu.CurvatureBenchmark.DataSources.Explicit;
using Zcu.CurvatureBenchmark.Framework;
using Zcu.CurvatureBenchmark.Framework.Estimators;
using Zcu.CurvatureBenchmark.Estimators;
using System.Xml;

using RenderMgr = SlimDXRenderSystem.Managers.RenderManager;
using System.IO;

namespace CurvatureVisualizer
{


	/// <summary>
	/// Visualisator
	/// Messy class, but it is enough for basic usage
	/// </summary>
    public class Visualizer
    {
		//==============================================================================================
		// Enums
		//==============================================================================================

		/// <summary>
		/// Enum with curvature type - do not modify, since no other curvatures are not supported
		/// </summary>
		enum CurvatureToVisualise
		{
			K1 = 0,
			K2 = 1,
			Mean = 2,
			Gauss = 3
		};

		/// <summary>
		/// Enum with curvature coloring
		/// If you add color mapping, modify this to add new mapping and make it accessible from XML
		/// </summary>
		enum CurvatureColorMapping
		{
			Constant = 0,
			PositiveNegative = 1,
			Gradient = 2,
			GradientFromExact = 3
		};

		//==============================================================================================
		// Public section
		//==============================================================================================

		/// <summary>
		/// Main method for visualizer
		/// </summary>
		/// <param name="args"></param>
        static void Main(string[] args)
        {

            Output.SetDefault();
            string testFileName;

            RandomExt.Seed = 5;
            if (args.Length == 0)
            {
                testFileName = TestFilename();
            } else
            {
                testFileName = args[0];
            }

            Output.Info((IModule)null, "Test file: {0}", testFileName);

			CurvatureToVisualise c = CurvatureToVisualise.Mean;
			CurvatureColorMapping col = CurvatureColorMapping.Gradient;
			bool enableExact = true;

            TestScenario modules = new TestScenario(testFileName, ModuleLoaded);
            LoadXMLConfig(testFileName, ref c, ref col, ref enableExact);

           // if (enableExact) modules.estimators.Insert(0, new ExactEstimator());


            List<TriangleMesh> meshes = new List<TriangleMesh>();
            List<ICurvatureEstimator> estimators = new List<ICurvatureEstimator>();
			List<string> desc = new List<string>();

            TriangleMesh meshCopy = null;

            foreach (var source in modules.sources)
            {
				if (source.Active == false)
				{
					continue;
				}

				var data = source.StartGenerateData();

                foreach (var dato in data)
                {
                    

                    string sourceDesc = source.ToString();
                    sourceDesc = sourceDesc.Replace("Zcu.CurvatureBenchmark.DataSources.", "");


                    if (enableExact)
                    {
                        ExactEstimator exact = new ExactEstimator();
                        meshCopy = dato.DeepCopyOfColors();
                        meshCopy.Estimator = exact;

                        exact.StartEstimateCurvatures(meshCopy);
                        estimators.Add(exact);


                        meshes.Add(meshCopy);

                        desc.Add(" Exact || " + sourceDesc + " || " + c.ToString());
                    }

                    foreach (var dist in modules.distorters)
                    {
                        if (dist.Active == false)
                        {
                            continue;
                        }

                        var distDato = dist.Distort(dato);

                        // TODO what to do with stats? Use log files?
                        TriangleMeshStat meshstat = new TriangleMeshStat(distDato, false, false, 0, 0, source, dist);
                        foreach (var est in modules.estimators)
                        {
                            if (est.Active == false)
                            {
                                continue;
                            }

                            //we need to create copy, because otherwise, colors and curvatures are overwritten
                            //by next estimator
                            meshCopy = distDato.DeepCopyOfColors();                            

                            est.StartEstimateCurvatures(meshCopy);
                            estimators.Add(est.ShallowCopy());

                            meshCopy.Estimator = estimators[estimators.Count - 1];

                            meshes.Add(meshCopy);

                            string estDesc = est.ToString();

                            var mi = est.GetType().GetMethod("ToString");
                            if ((mi != null) && (mi.DeclaringType == est.GetType()))
                            {
                                estDesc += " ( ";
                                estDesc += est.ToString();
                                estDesc += ")";
                            }

                            desc.Add("" + estDesc + " || " + dist + " || " + sourceDesc + " || " + c.ToString());
                            foreach (var evaluator in modules.evaluators)
                            {
                                if (evaluator.Active == false)
                                {
                                    continue;
                                }
                                evaluator.Evaluate(meshCopy, est);
                            }
                        }
                        ObjSaver saver = new ObjSaver();
                        saver.FileName = "mesh1.obj";
                        saver.Execute(dato);
                    }
                }
            }

           

            Colorize(meshes.ToArray(), estimators.ToArray(), c, col);
            
			
            SlimDXRenderer renderer = new SlimDXRenderer();
			//renderer.BackgroundColor = System.Drawing.Color.LightGray;			
            renderer.ShowColorAttribute = true;
			
			renderer.ShowRenderWindow(meshes.ToArray(), desc.ToArray());
			
        }

		//==============================================================================================
		// Private section
		//==============================================================================================

        static string TestFilename()
        {
            var files = Directory.GetFiles("./","*.xml");
            if (files.Length == 0) {
                Console.WriteLine("No test file were found.");
                Environment.Exit(0);
            }
            for(int i=0; i<files.Length; i++)
            {
                Console.WriteLine("{0}: {1}", i, files[i]);
            }
            Console.Write("Choose test file: ");
            int selection;
            int.TryParse(Console.ReadLine(), out selection);
            return (files[selection]);
        }


		/// <summary>
		/// Load XML from args[0] and parse nodes specific for visualizer
		/// Those nodes are not used in BatchTest, so no need to parse them globally
		/// </summary>
		/// <param name="configFileName"></param>
		/// <param name="c">ref to enum of visualised curvature</param>
		/// <param name="col">ref to enum of color mapping function</param>
		private static void LoadXMLConfig(string configFileName, 
			ref CurvatureToVisualise c, ref CurvatureColorMapping col,
			ref bool enableExact)
		{
			System.Threading.Thread.CurrentThread.CurrentCulture = new System.Globalization.CultureInfo("en-US");            
            XmlDocument configXml = new XmlDocument();
            configXml.Load(Path.GetFileNameWithoutExtension(configFileName) + ".xml");
            XmlNode root = configXml.DocumentElement;

            foreach (XmlNode node in root.ChildNodes)
            {
                switch (node.Name)
                {
					case "Visualisator": 
						foreach (XmlNode child in node.ChildNodes)
						{
                            if (child.Attributes == null) 
                            {
                                continue;
                            }

							foreach (XmlAttribute atr in child.Attributes)
							{
								if ((child.Name == "curvature") && (atr.Name == "type"))
								{
									c = (CurvatureToVisualise)Enum.Parse(typeof(CurvatureToVisualise), atr.Value, true);
								}

								if ((child.Name == "curvature") && (atr.Name == "color"))
								{
									col = (CurvatureColorMapping)Enum.Parse(typeof(CurvatureColorMapping), atr.Value, true);
								}

								if ((child.Name == "curvature") && (atr.Name == "exact"))
								{
									enableExact = bool.Parse(atr.Value);
								}
							}							
						}						
						break;                    
                }
            }


			if ((enableExact == false) && (col == CurvatureColorMapping.GradientFromExact))
			{
				throw new Exception("Can not map gradient. GradientFromExact can be used only with exact curvature enabled ");
            }

		}

		/// <summary>
		/// Colorize mesh curvature based on input enums from XML
		/// </summary>
		/// <param name="mesh">mesh</param>
		/// <param name="est">estimator</param>
		/// <param name="c">enum with type of curvature</param>
		/// <param name="col">enum with color mapping</param>
		private static void Colorize(TriangleMesh[] meshes, ICurvatureEstimator[] est, CurvatureToVisualise c, CurvatureColorMapping col)
		{
			switch (col)
			{
				case CurvatureColorMapping.Constant: Map.Constant(meshes, null, new ColorRGBA[] { new ColorRGBA(1f,0.8f,0) }); break;
				case CurvatureColorMapping.PositiveNegative: Map.PositiveNegative(meshes, GetCurvatures(est, c)); break;
				case CurvatureColorMapping.Gradient: Map.Gradient(meshes, GetCurvatures(est, c)); break;
				case CurvatureColorMapping.GradientFromExact: Map.GradientFromExact(meshes, est, GetCurvatures(est, c)); break;
			}
			
		}

		/// <summary>
		/// Get array of data with selected curvature
		/// Return directly K1 or K2, Gauss and Mean are mapped using Lambda function and Linq
		/// </summary>
		/// <param name="est">estimator</param>
		/// <param name="c">enum with type of curvature</param>
		/// <returns>array with curvature data</returns>
		private static double[] GetCurvature(ICurvatureEstimator est, CurvatureToVisualise c) 
		{
			if (c == CurvatureToVisualise.K1)
			{
				return est.K1;
			}

			if (c == CurvatureToVisualise.K2)
			{
				return est.K2;
			}

			if (c == CurvatureToVisualise.Mean)
			{
				return est.K2.Select((x, index) => x + est.K1[index]).ToArray();
			}

			if (c == CurvatureToVisualise.Gauss)
			{
				return est.K2.Select((x, index) => x * est.K1[index]).ToArray();
			}

			return new double[1];
		}

        private static double[][] GetCurvatures(ICurvatureEstimator[] ests, CurvatureToVisualise c)
        {
            double[][] result = new double[ests.Length][];
            for (int i = 0; i < result.Length; i++)
			{
                result[i] = GetCurvature(ests[i], c);
			}
            return result;
        }

        static void ModuleLoaded(IModule module)
        {
            module.Success("Loaded");
            module.ProgressChanged += (m, args) => 
            {

                if (args.Progress == 0)
                {
                    if (Console.CursorLeft != 0) Console.WriteLine();
                    Console.WriteLine(m.GetType());
                }
                Console.CursorLeft = 0;
                Console.Write(args.Progress + "%");
            };
        }


    }
}
