﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Zcu.CurvatureBenchmark.Framework;
using Zcu.CurvatureBenchmark.Framework.DataSources.Explicit;
using MathNet.Symbolics;
using Expr = MathNet.Symbolics.Expression;
using Eval = MathNet.Symbolics.Evaluate;


namespace Zcu.CurvatureBenchmark.DataSources.Explicit
{
    /// <summary>
    /// Symbolic function is IExplcitFunction implementation using MathNet.Symbolics.
    /// </summary>
    /// <example>
    /// More examples in !TestData/Examples-ExplicitFunc.xml
    /// </example>
    /// <author>Petr Vanecek</author>
    public class SymbolicFunction : AbstractModule, IExplicitFunction
    {
        /// <summary>
        /// Dictionary with predefined functions. The names could be used in config file.
        /// </summary>
        /// <remarks>
        /// Currently, these functions are defined:
        /// <list type="bullet">
        /// <item>sphere - the top half-sphere, radius is controlled by parameter r.</item>
        /// <item>pyramid - pyramid like surface, the sharpness of edges is controlled by parameter p.</item>
        /// <item>gauss - 2D Gaussian function with height controlled by parameter A and sigma controlled by parameters sigmaX and sigmaY.</item>
        /// </list>
        /// </remarks>
        Dictionary<string, Tuple<string, string>> predefinedFunctions = new Dictionary<string, Tuple<string, string>>()
        {
            {"sphere", Tuple.Create("(r*r-(x*x)-(y*y))^(1/2)", "r=1")},
            {"pyramid",Tuple.Create("(1+x^2*10^(2*p))^(1/2)/10^p+(1+y^2*10^(2*p))^(1/2)/10^p","p=10")},
            {"gauss", Tuple.Create("A * exp(-(x^2/(2*sigmaX^2)+y^2/(2*sigmaY^2)))","A=1; sigmaX=0.1; sigmaY=0.1")}
        };


        /// <summary>
        /// <see cref="Function"/>
        /// </summary>
        protected string function;
       
        /// <summary>
        /// Function string. The syntax is defined by MathNet.Symbolics. To be able to compute the curvatures and normals,
        /// the function must be of class C2.
        /// </summary>
        public string Function 
        { 
            get { return function; }
            set
            {
                value = value.Trim().ToLower();
                if (predefinedFunctions.ContainsKey(value))
                {
                    function = predefinedFunctions[value].Item1;
                    Params = predefinedFunctions[value].Item2;
                }
                else
                    function = value;
                Build();
            }
        }

        /// <summary>
        /// <see cref="Params"/>
        /// </summary>
        protected string _params;
        /// <summary>
        /// Function parameters given as a string in a form param1=value1; param2=value2...
        /// </summary>
        public string Params
        {
            get { return _params; }
            set
            {
                _params = value;
                foreach (var param in _params.Split(';'))
                {
                    var expr = param.Split('=');
                    paramIntervals[expr[0].Trim()] = new Tuple<double, double>(Convert.ToDouble(expr[1]), Convert.ToDouble(expr[1]));
                }
            }
        }

        /// <summary>
        /// <see cref="RandomParams"/>
        /// </summary>
        protected string _randomParams;
        /// <summary>
        /// Random parameters given as a string in a form param1=from1:to1; param2...
        /// </summary>
        public string RandomParams
        {
            get { return _randomParams; }
            set
            {
                _randomParams = value;
                foreach (var param in _randomParams.Split(';'))
                {
                    var expr = param.Split('=');
                    var range = expr[1].Split(':');
                    double min = Convert.ToDouble(range[0]);
                    double max = Convert.ToDouble(range[1]);
                    paramIntervals[expr[0].Trim()] = new Tuple<double, double>(min,max);
                }
            }
        }

        /// <summary>
        /// Function sets the parameters for current dataset. For random parameters, new values are generated.
        /// </summary>
        public void SetParams()
        {
            foreach (var param in paramIntervals)
            {
                values[param.Key] = param.Value.Item1 + Rand.NextDouble() * (param.Value.Item2 - param.Value.Item1);
            }
        }

        /// <summary>
        /// Parameters intervals in dictionary  [parameter, [from, to]]. For non-randomized parameters from==to.
        /// </summary>
        Dictionary<string, Tuple<double, double>> paramIntervals = new Dictionary<string, Tuple<double, double>>();

        Dictionary<string, FloatingPoint> values = new Dictionary<string, FloatingPoint>();
        private Expr functionExpr;
        private Expr xExpr;
        private Expr yExpr;
        private Expr dxExpr;
        private Expr dyExpr;
        private Expr dxdxExpr;
        private Expr dydyExpr;
        private Expr dxdyExpr;


        /// <summary>
        /// Preevaluation of function at x,y. Evaluates only function value and non-normalized normal vector.
        /// </summary>
        /// <param name="x">X coordinate</param>
        /// <param name="y">Y coordinate</param>
        public void EvaluateGradient(double x, double y)
        {
            values["x"] = x;
            values["y"] = y;
            var feval = Eval.Evaluate(values, functionExpr);
            if (!feval.IsReal)
            {
                F = double.NaN;
                K1 = double.NaN;
                K2 = double.NaN;
                Normal = new Point3D();
                this.Warning("NaN at {0}; {1}", x, y);
                return;
            }
            F = feval.RealValue;
            double dx = Eval.Evaluate(values, dxExpr).RealValue;
            double dy = Eval.Evaluate(values, dyExpr).RealValue;

            Point3D normal = new Point3D(dx, dy, -1);            
            Normal = normal;

        }


        /// <summary>
        /// Evaluates explicit function at x,y.
        /// </summary>
        /// <param name="x">X coordinate</param>
        /// <param name="y">Y coordinate</param>
        public void Evaluate(double x, double y)
        {

            EvaluateGradient(x, y);
            if (double.IsNaN(F)) return;

            double dx = Normal.X;
            double dy = Normal.Y;
            Point3D normal = Normal;
            normal.Divide(normal.Abs());
            Normal = normal;

            double dxdx = Eval.Evaluate(values, dxdxExpr).RealValue;
            double dydy = Eval.Evaluate(values, dydyExpr).RealValue;
            double dxdy = Eval.Evaluate(values, dxdyExpr).RealValue;
            double dx2 = dx * dx;
            double dy2 = dy * dy;

            double K = (dxdx * dydy - dxdy * dxdy) / Math.Pow(1 + dx2 + dy2, 2);
            double H = - ((1 + dx2) * dydy + (1 + dy2) * dxdx - 2 * dx * dy * dxdy) / (2 * Math.Pow(1 + dx2 + dy2, 3.0 / 2.0));

            double H2 = H * H;

            if (H2<K)
            {
                K1 = K2 = H;
            }
            else
            {
                double sqrt = Math.Sqrt(H2 - K);
                K1 = H + sqrt;
                K2 = H - sqrt;
            }

        }

        /// <summary>
        /// Builds MathNet expressions from given string.
        /// </summary>
        protected virtual void Build()
        {
            xExpr = Expr.Symbol("x");
            yExpr = Expr.Symbol("y");
            functionExpr = Infix.ParseOrThrow(function);
            dxExpr = Calculus.Differentiate(xExpr, functionExpr);
            dxdxExpr = Calculus.Differentiate(xExpr, dxExpr);
            dyExpr = Calculus.Differentiate(yExpr, functionExpr);
            dydyExpr = Calculus.Differentiate(yExpr, dyExpr);
            dxdyExpr = Calculus.Differentiate(yExpr, dxExpr);
        }

        /// <inheritdoc/>
        public double F { get; set;}

        /// <inheritdoc/>
        public double K1 { get; set;}

        /// <inheritdoc/>
        public double K2 { get; set;}

        /// <inheritdoc/>
        public Point3D Normal { get; set;}

        /// <inheritdoc/>
        public override string ToString()
        {
            string p = ";" + string.Join(",", values.Where(k => k.Key != "x" && k.Key != "y").Select(k => string.Format("{0}={1}", k.Key, k.Value.RealValue)));
            return this.function + p;
        }

    }
}
