﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Zcu.CurvatureBenchmark.Framework;
using Zcu.CurvatureBenchmark.Framework.DataSources.Explicit;

namespace Zcu.CurvatureBenchmark.DataSources.Explicit
{
    /// <summary>
    /// Random sampler driven by curvature and normal direction.
    /// </summary>
    /// <example>
    /// More examples in !TestData/Examples-ExplicitSamplers.xml
    /// </example>
    /// <author>Petr Vanecek</author>
    public class CurvatureDependentRandomSampler : AbstractSampler
    {

        /// <summary>
        /// Number of samples
        /// </summary>
        public virtual int Samples { get; set; }

        /// <summary>
        /// Percentage of random samples vs. curvature dependent samples. Default value is 10%.
        /// </summary>
        public virtual int Ratio { get; set; }

        /// <summary>
        /// Default values.
        /// </summary>
        public CurvatureDependentRandomSampler() : base()
        {
            Samples = 1000;
            Ratio = 10;
        }

        /// <summary>
        /// Maximal mean curvature in presampled data.
        /// </summary>
        double maxMean = 0;

        /// <summary>
        /// Minimal mean curvature in presampled data.
        /// </summary>
        double minMean = double.PositiveInfinity;

        /// <summary>
        /// Maximal normal dot baseNormal in presampled data.
        /// </summary>
        double maxNorm = 0;

        /// <summary>
        /// Minimal normal dot baseNormal in presampled data.
        /// </summary>
        double minNorm = double.PositiveInfinity;

        /// <summary>
        /// Base normal (normal of Z=0 plane).
        /// </summary>
        Point3D baseNormal = new Point3D(0, 0, -1);

        /// <inheritdoc/>
        public override void Sample(IExplicitFunction function)
        {
            Progress = 0;
            Function = function;

            Values = new double[Samples];
            Points = new Point3D[Samples];
            K1 = new double[Samples];
            K2 = new double[Samples];
            Normals = new Point3D[Samples];

            int presamples = (Samples * Ratio) / 100;

            Presample(presamples);
            Sample(presamples);

            Progress = 100;

        }

        /// <summary>
        /// Sample function using presampled values to accept/reject current sample.
        /// </summary>
        /// <param name="presamples">Number of samples for presmapling.</param>
        private void Sample(int presamples)
        {
            for (int i = presamples; i < Samples; i++)
            {
                double x, y;
                while (true)
                {
                    x = Rand.NextDouble(MinX, MaxX);
                    y = Rand.NextDouble(MinY, MaxY);
                    Function.Evaluate(x, y);
                    Rand.NextDouble(minMean, maxMean);
                    //if (Math.Abs(Function.K1 + Function.K2) > Rand.NextDouble(minMean, maxMean)) break;

                    double dotnormal = (1 - Function.Normal.DotProduct(ref baseNormal) - minNorm) / (maxNorm - minNorm);
                    dotnormal = dotnormal * 0.9 + 0.1;

                    double mean = (Math.Abs(Function.K1 + Function.K2) - minMean) / (maxMean - minMean);
                    mean = mean * 0.9 + 0.1;

                    // Accept sample
                    if (dotnormal * mean > Rand.NextDouble()) break;
                }
                Points[i] = new Point3D(x, y, Function.F);
                Values[i] = Function.F;
                K1[i] = Function.K1;
                K2[i] = Function.K2;
                Normals[i] = Function.Normal;
                Progress = (byte)(100 * i / Samples);
            }
            Distort();
        }

        /// <summary>
        /// Presample data set and estimate mean curvature and normal.
        /// </summary>
        /// <param name="presamples">Number of samples for presmapling.</param>
        private void Presample(int presamples)
        {
            for (int i = 0; i < presamples; i++)
            {
                double x = Rand.NextDouble(MinX, MaxX);
                double y = Rand.NextDouble(MinY, MaxY);
                Function.Evaluate(x, y);
                Points[i] = new Point3D(x, y, Function.F);
                Values[i] = Function.F;
                K1[i] = Function.K1;
                K2[i] = Function.K2;
                double mean = Math.Abs(Function.K1 + Function.K2);
                maxMean = Math.Max(mean, maxMean);
                minMean = Math.Min(mean, minMean);
                Normals[i] = Function.Normal;
                double dotnormal = 1 - Function.Normal.DotProduct(ref baseNormal);
                maxNorm = Math.Max(dotnormal, maxNorm);
                minNorm = Math.Min(dotnormal, minNorm);

                Progress = (byte)(100 * i / Samples);
            }
        }
    }
}
