﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Zcu.CurvatureBenchmark.Framework;
using Zcu.CurvatureBenchmark.Framework.DataSources.Explicit;

namespace Zcu.CurvatureBenchmark.DataSources.Explicit
{
    /// <summary>
    /// Sampler for equilateral triangles.
    /// </summary>
    /// <remarks>
    /// The size of area (MinX, MaxX, MinY, MaxY) is shrunk to be able to fit in the equilateral triangles.
    /// </remarks>
    /// <example>
    /// More examples in !TestData/Examples-ExplicitSamplers.xml
    /// </example>
    /// <author>Petr Vanecek</author>
    public class EquilateralTriangleSampler : RectangularGridSampler
    {
        /// <summary>
        /// Length of triangle edge.
        /// </summary>
        public double EdgeLength { get; private set; }
        double TriHeight { get; set; }        

        /// <summary>
        /// Computes edge length and change the sampling area according the size of triangle.
        /// </summary>
        protected void SetEdgeLength()
        {
            double ax = (MaxX - MinX) / (SamplesX - 1 / 2);
            double ay = (2 / Math.Sqrt(3)) * (MaxY - MinY) / (SamplesY - 1);

            EdgeLength = Math.Min(ax, ay);
            TriHeight = EdgeLength * Math.Sqrt(3) / 2;
            double midX = (MaxX + MinX) / 2;
            double midY = (MaxY + MinY) / 2;

            double lenx = EdgeLength * (SamplesX - 1 / 2);
            double leny = TriHeight * (SamplesY - 1);

            MinX = midX - lenx / 2;
            MaxX = midX + lenx / 2;
            MinY = midY - leny / 2;
            MaxY = midY + lenx / 2;
        }

        /// <inheritdoc/>
        public override void SetConfig(System.Xml.XmlNode configNode)
        {
            base.SetConfig(configNode);
            SetEdgeLength();
        }

        /// <summary>
        /// Sample function.
        /// </summary>
        /// <param name="function">IExplicit function</param>
        public override void Sample(IExplicitFunction function)
        {
            Progress = 0;
            base.Sample(function);

            int index = 0;

            for (int iy = 0; iy < SamplesY; iy++)
            {
                double y = MinY + iy * TriHeight;
                double ox = (1.0 / 2.0) * EdgeLength * (iy % 2);
                for (int ix = 0; ix < SamplesX; ix++)
                {
                    double x = MinX + ox + ix * EdgeLength;
                    double nx = 0;
                    double ny = 0;
                    if (Noise)
                    {
                        nx = EdgeLength * Rand.NextDouble(-1, 1) * NoiseStrength;
                        ny = EdgeLength * Rand.NextDouble(-1, 1) * NoiseStrength;
                    }

                    function.Evaluate(x+nx, y+ny);
                    Points[index] = new Point3D(x+nx, y+ny, function.F);
                    Values[index] = function.F;
                    K1[index] = function.K1;
                    K2[index] = function.K2;
                    Normals[index] = function.Normal;
                    index++;

                    Progress = (byte)(100 * index / count);
                }
            }
            Distort();
            Progress = 100;
        }
    }
}
