﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Zcu.CurvatureBenchmark.Framework;
using Zcu.CurvatureBenchmark.Framework.DataSources.Explicit;
using System.IO;

namespace Zcu.CurvatureBenchmark.DataSources.Explicit
{
    /// <summary>
    /// Samples non-linearly in X=pow(x, NonLinear), Y=pow(y, NonLinear) and skewed X=Skew*y, Y=Skew*x
    /// </summary>
    /// <example>
    /// More examples in !TestData/Examples-ExplicitSamplers.xml
    /// </example>
    /// <author>Petr Vanecek</author>
    public class NonLinearSkewGridSampler : RectangularGridSampler
    {
        /// <summary>
        /// Non-linearity factor
        /// </summary>
        public double NonLinear { get; set; }

        /// <summary>
        /// Skew factor
        /// </summary>
        public double Skew { get; set; }

        /// <summary>
        /// Set default values.
        /// </summary>
        public NonLinearSkewGridSampler() : base()
        {
            NonLinear = 1;
            Skew = 0;
        }


        /// <inheritdoc/>
        public override void Sample(IExplicitFunction function) {
            Progress = 0;
            base.Sample(function);

            double stepX = (MaxX - MinX) / (SamplesX - 1);
            double stepY = (MaxY - MinY) / (SamplesY - 1);
            double sizeX = (MaxX - MinX);
            double sizeY = (MaxY - MinY);

            double y = 0;
            double x = 0;

            double[] lastY = new double[SamplesX];
            for (int ix = 0; ix < SamplesX; ix++)
                lastY[ix] = double.NaN;

            int index = 0;

            for (int iy = 0; iy < SamplesY; iy++)
            {
                double sy = Math.Pow((double)iy / SamplesY, NonLinear);
                double lastX = double.NaN;
                for (int ix = 0; ix < SamplesX; ix++)
                {
                    double sx = Math.Pow((double)ix / SamplesX, NonLinear);
                    y = MinY + sizeY * sy + sizeY * Skew * sx;
                    x = MinX + sizeX * sx + sizeX * Skew * sy;

                    double nx = 0;
                    double ny = 0;

                    if (Noise)
                    {
                        if (double.IsNaN(lastX))
                            stepX = 0;
                        else
                            stepX = lastX - x;

                        if (double.IsNaN(lastY[ix]))
                            stepY = 0;
                        else
                            stepY = lastY[ix] - y;

                        nx = stepX * Rand.NextDouble(-1, 1) * NoiseStrength;
                        ny = stepY * Rand.NextDouble(-1, 1) * NoiseStrength;
                    }

                    lastX = x;
                    lastY[ix] = y;

                    function.Evaluate(x + nx, y + ny);
                    Points[index] = new Point3D(x + nx, y + ny, function.F);
                    Values[index] = function.F;
                    K1[index] = function.K1;
                    K2[index] = function.K2;
                    Normals[index] = function.Normal;
                    index++;
                }
            }
            Distort();
            Progress = 100;
        }
    }
}
