﻿using System;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Reflection;
using Zcu.CurvatureBenchmark.Framework;
using Zcu.CurvatureBenchmark.Framework.DataSources.Explicit;
using System.Threading;


namespace Zcu.CurvatureBenchmark.DataSources.Explicit
{
    /// <summary>
    /// Delaunay triangulation using external SW.
    /// </summary>
    /// <example>
    /// More examples in !TestData/Examples-ExplicitTriangulations.xml
    /// </example>
    /// <author>Petr Vanecek</author>      

    public class DelaunayTriangularizator : AbstractModule, IExplicitTriangularizator
    {
        /// <summary>
        /// Path to DT.exe
        /// </summary>
        string DTPath;

        /// <summary>
        /// Checks the presence of dt.exe and ShewLib.dll. 
        /// Please, download the complete Benchmark from http://graphics.zcu.cz/curvature.html.
        /// </summary>
        /// <returns>TRUE if files are present</returns>
        public override bool CheckPrerequisities()
        {
            bool check = true;
            if (!File.Exists(@"Explicit\Triangularizators\DT.exe")) check = false;
            if (!File.Exists(@"Explicit\Triangularizators\ShewLib.dll")) check = false;

            if (!check)
            {
                this.Error(@"Missing important files (Explicit\Triangularizators\DT.exe or ShewLib.dll). Please, download full installation of Benchmark from http://graphics.zcu.cz/curvature.html");
            }

            return check;            
        }

        /// <inheritdoc/>
        public void Triangulate(IExplicitSampler sampler)
        {
            Mutex mutex = new Mutex(false, "CurvatureBenchmarkDelaunay");
            mutex.WaitOne();

            DTPath = Path.GetDirectoryName((new System.Uri(Assembly.GetExecutingAssembly().CodeBase)).AbsolutePath) + @"\Explicit\Triangularizators\";

            string fileIn = "";
            try
            {
                fileIn = PrepareInput(sampler);

                var dtProc = new ProcessStartInfo();
                dtProc.UseShellExecute = false;
                dtProc.WorkingDirectory = DTPath;

                dtProc.FileName = DTPath + @"dt.exe";
                dtProc.CreateNoWindow = true;
                dtProc.Arguments = string.Format("{0}", fileIn);
                var process = Process.Start(dtProc);
                process.WaitForExit();
                ReadOutput(sampler);
            }
            catch (Exception e)
            {
                this.Critical("External DT failed: {0}", e.Message);
            }
            finally
            {
                CleanUp(fileIn);
                mutex.ReleaseMutex();
            }

        }

        /// <summary>
        /// Reads output file of DT.exe and converts it to triangle mesh.
        /// </summary>
        /// <param name="sampler"></param>
        private void ReadOutput(IExplicitSampler sampler)
        {
            var tris = File.ReadAllLines(DTPath + "data.dat").ToList();
            tris.RemoveAt(0); // first line contains the number of triangles

            TriangleMesh = new Framework.TriangleMesh();
            TriangleMesh.Points = sampler.Points;
            TriangleMesh.Normals = sampler.Normals;
            TriangleMesh.ExactK1 = sampler.K1;
            TriangleMesh.ExactK2 = sampler.K2;

            TriangleMesh.Triangles = tris.Select(line =>
            {
                var index = line.Split((char[])null, StringSplitOptions.RemoveEmptyEntries);
                return new Triangle()
                {
                    V1 = Convert.ToInt32(index[0]),
                    V2 = Convert.ToInt32(index[1]),
                    V3 = Convert.ToInt32(index[2]),
                };
            }).ToArray();
        }

        /// <summary>
        /// Deletes all temporary files.
        /// </summary>
        /// <param name="fileIn"></param>
        private void CleanUp(string fileIn)
        {
            File.Delete(DTPath + "data.dat");
            File.Delete(DTPath + fileIn);
        }

        /// <summary>
        /// Prepares input file for DT.exe.
        /// </summary>
        /// <param name="sampler"></param>
        /// <returns></returns>
        private string PrepareInput(IExplicitSampler sampler)
        {
            string fileIn = DateTime.Now.ToString("yyMMddhhmmss") + ".xyz";
            File.WriteAllText(DTPath + fileIn, 
                sampler.Points.Length + "\n" +
                string.Join("\n", sampler.Points.Select(p => string.Format("{0} {1} {2}", p.X, p.Y, p.Z)))
            );
            return fileIn;
        }

        /// <inheritdoc/>
        public TriangleMesh TriangleMesh { get; private set;}
    }
}
