﻿using System;
using Zcu.CurvatureBenchmark.Framework;
using Zcu.CurvatureBenchmark.Framework.DataSources.Explicit;

namespace Zcu.CurvatureBenchmark.DataSources.Explicit
{
    /// <summary>
    /// Control of row diagonals. 
    /// </summary>
    /// <remarks>
    /// For a regular grid sampler, it is necessary to choose the diagonal. The RectangularGridTriangularizator 
    /// can use uniformly oriented diagonals or varying normals for odd and even rows.
    /// </remarks>
    [Flags]
    public enum AlterDiagonal {
        /// <summary>
        /// Same orientation
        /// </summary>
        None = 0,
        /// <summary>
        /// <see cref="Odd"/>
        /// </summary>
        First = 1,
        /// <summary>
        /// Diagonals in odd rows are swapped
        /// </summary>
        Odd = 1,
        /// <summary>
        /// <see cref="Even"/>
        /// </summary>
        Second = 2,
        /// <summary>
        /// Diagonals in even rows are swapped
        /// </summary>
        Even = 2,
        /// <summary>
        /// All diagonals are swapped.
        /// </summary>
        Both = 3
    };

    /// <summary>
    /// Rectangular grid triangularizator takes samples from RectangularGridSampler and creates regular triangulation.
    /// </summary>
    /// <example>
    /// More examples in !TestData/Examples-ExplicitTriangulations.xml
    /// </example>
    /// <author>Petr Vanecek</author>      
    public class RectangularGridTriangularizator : AbstractModule, IExplicitTriangularizator
    {
        /// <summary>
        /// Alter the direction of diagonals in a row? <see cref="AlterDiagonal">AlterDiagonal enum</see>
        /// </summary>
        public AlterDiagonal AlterDiagonal { get; set; }

        RectangularGridSampler gridsampler;

        /// <inheritdoc/>
        public void Triangulate(IExplicitSampler sampler)
        {
            Progress = 0;
            if (!(sampler is RectangularGridSampler)) this.Critical("RectangularGridTriangularizator needs RectangularGridSampler");
            gridsampler = sampler as RectangularGridSampler;

            DiagonalFunc[] diagFunc = new DiagonalFunc[] { AntiDiagonal, AntiDiagonal };

            if((AlterDiagonal & AlterDiagonal.First)!=0)
                diagFunc[0] = MainDiagonal;
            if ((AlterDiagonal & AlterDiagonal.Second) != 0)
                diagFunc[1] = MainDiagonal;

            TriangleMesh = new TriangleMesh();
            TriangleMesh.Points = sampler.Points;
            TriangleMesh.ExactK1 = sampler.K1;
            TriangleMesh.ExactK2 = sampler.K2;
            TriangleMesh.Normals = sampler.Normals;

            int count = 2 * (gridsampler.SamplesX - 1) * (gridsampler.SamplesY - 1);

            TriangleMesh.Triangles = new Triangle[count];

            int index = 0;
            for (int y = 0; y < gridsampler.SamplesY-1; y++)
            {
                for (int x = 0; x < gridsampler.SamplesX-1; x++)
                {
                    index = diagFunc[y%2](index, y, x);
                    Progress = (byte)(100 * index / count);
                }
            }
            Progress = 100;
        }

        delegate int DiagonalFunc(int index, int y, int x);

        /// <summary>
        /// Adds triangles for quad x,y -> x+1, y+1 with diagonal x,y+1->x+1,y
        /// </summary>
        /// <param name="index">index of last generated triangle</param>
        /// <param name="y">index of top left corner</param>
        /// <param name="x">index of top left corner</param>
        /// <returns></returns>
        protected int AntiDiagonal(int index, int y, int x)
        {
            TriangleMesh.Triangles[index] = new Triangle(x + y * gridsampler.SamplesX, x + 1 + y * gridsampler.SamplesX, x + (y + 1) * gridsampler.SamplesX);
            index++;
            TriangleMesh.Triangles[index] = new Triangle(x + (y + 1) * gridsampler.SamplesX, x + 1 + y * gridsampler.SamplesX, x + 1 + (y + 1) * gridsampler.SamplesX);
            index++;
            return index;
        }

        /// <summary>
        /// Adds triangles for quad x,y -> x+1, y+1 with diagonal x,y->x+1,y+1
        /// </summary>
        /// <param name="index">index of last generated triangle</param>
        /// <param name="y">index of top left corner</param>
        /// <param name="x">index of top left corner</param>
        /// <returns></returns>
        protected int MainDiagonal(int index, int y, int x)
        {
            TriangleMesh.Triangles[index] = new Triangle(x + y * gridsampler.SamplesX, x + 1 + (y+1) * gridsampler.SamplesX, x + (y + 1) * gridsampler.SamplesX);
            index++;
            TriangleMesh.Triangles[index] = new Triangle(x + y * gridsampler.SamplesX, x + 1 + y * gridsampler.SamplesX, x + 1 + (y + 1) * gridsampler.SamplesX);
            index++;
            return index;
        }

        /// <inheritdoc/>
        public TriangleMesh TriangleMesh {get ;set;}
    }
}
