﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Zcu.CurvatureBenchmark.Framework;
using Zcu.CurvatureBenchmark.Framework.DataSources.Implicit;
using MathNet.Symbolics;
using Expr = MathNet.Symbolics.Expression;
using Eval = MathNet.Symbolics.Evaluate;
using MathNet.Numerics;
using MathNet.Numerics.LinearAlgebra;


namespace Zcu.CurvatureBenchmark.DataSources.Implicit
{
    /// <summary>
    /// Symbolic function is IImplicitFunction implementation using MathNet.Symbolics.
    /// </summary>
    public class SymbolicFunction : AbstractModule, IImplicitFunction
    {
        /// <summary>
        /// Dictionary with predefined functions. The names could be used in config file.
        /// </summary>
        /// <remarks>
        /// Currently, these functions are defined:
        /// <list type="bullet">
        /// <item>sphere - sphere, radius is controlled by parameter r.</item>
        /// </list>
        /// </remarks>
        Dictionary<string, Tuple<string, string>> predefinedFunctions = new Dictionary<string, Tuple<string, string>>()
        {
            {"sphere", Tuple.Create("(r*r)-(x*x)-(y*y)-(z*z)", "r=1")},
        };

        /// <summary>
        /// <see cref="Function"/>
        /// </summary>
        protected string function;

        /// <summary>
        /// Function string. The syntax is defined by MathNet.Symbolics. To be able to compute the curvatures and normals,
        /// the function must be of class C2.
        /// </summary>
        public string Function
        {
            get { return function; }
            set
            {
                value = value.Trim().ToLower();
                if (predefinedFunctions.ContainsKey(value))
                {
                    function = predefinedFunctions[value].Item1;
                    Params = predefinedFunctions[value].Item2;
                }
                else
                    function = value;
                Build();
            }
        }

        /// <summary>
        /// <see cref="Params"/>
        /// </summary>
        protected string _params;

        /// <summary>
        /// Function parameters given as a string in a form param1=value1; param2=value2...
        /// </summary>
        public string Params
        {
            get { return _params; }
            set
            {
                _params = value;
                foreach (var param in _params.Split(';'))
                {
                    var expr = param.Split('=');
                    paramIntervals[expr[0].Trim()] = new Tuple<double, double>(Convert.ToDouble(expr[1]), Convert.ToDouble(expr[1]));
                }
            }
        }

        /// <summary>
        /// <see cref="RandomParams"/>
        /// </summary>
        protected string _randomParams;
        /// <summary>
        /// Random parameters given as a string in a form param1=from1:to1; param2...
        /// </summary>
        public string RandomParams
        {
            get { return _randomParams; }
            set
            {
                _randomParams = value;
                foreach (var param in _randomParams.Split(';'))
                {
                    var expr = param.Split('=');
                    var range = expr[1].Split(':');
                    double min = Convert.ToDouble(range[0]);
                    double max = Convert.ToDouble(range[1]);
                    paramIntervals[expr[0].Trim()] = new Tuple<double, double>(min, max);
                }
            }
        }

        /// <summary>
        /// Function sets the parameters for current dataset. For random params, new values are generated.
        /// </summary>
        public void SetParams()
        {
            foreach (var param in paramIntervals)
            {
                values[param.Key] = param.Value.Item1 + Rand.NextDouble() * (param.Value.Item2 - param.Value.Item1);
            }
        }

        /// <summary>
        /// Dictionary with params and from,to intervals. For constant params, from==to.
        /// </summary>
        Dictionary<string, Tuple<double, double>> paramIntervals = new Dictionary<string, Tuple<double, double>>();

        private Expr functionExpr;
        private Expr xExpr;
        private Expr yExpr;
        private Expr zExpr;
        private Expr dxExpr;
        private Expr dyExpr;
        private Expr dzExpr;
        private Expr dxdxExpr;
        private Expr dxdyExpr;
        private Expr dxdzExpr;
        private Expr dydxExpr;
        private Expr dydyExpr;
        private Expr dydzExpr;
        private Expr dzdxExpr;
        private Expr dzdyExpr;
        private Expr dzdzExpr;
        Dictionary<string, FloatingPoint> values = new Dictionary<string, FloatingPoint>();

        /// <summary>
        /// Evaluation of function at x,y,z. Evaluates function value only.
        /// </summary>
        /// <param name="x">X coordinate</param>
        /// <param name="y">Y coordinate</param>
        /// <param name="z">Z coordinate</param>
		public void EvaluateValue(double x, double y, double z)
		{
			values["x"] = x;
			values["y"] = y;
			values["z"] = z;

			var feval = Eval.Evaluate(values, functionExpr);
			if (!feval.IsReal)
			{
				F = double.NaN;
                this.Warning("NaN at {0}; {1}; {2}", x, y, z );
                return;
			}
			F = feval.RealValue;
		}

        /// <summary>
        /// Preview evaluation computes only function value and non-normalized normal.
        /// </summary>
        /// <param name="x">X coordinate</param>
        /// <param name="y">Y coordinate</param>
        /// <param name="z">Z coordinate</param>
        public void EvaluateGradient(double x, double y, double z)
        {
			this.EvaluateValue(x, y, z);

			if (double.IsNaN(F)) return;
			
            K1 = double.NaN;
            K2 = double.NaN;
            Normal = new Point3D();

            double dx = Eval.Evaluate(values, dxExpr).RealValue;
            double dy = Eval.Evaluate(values, dyExpr).RealValue;
            double dz = Eval.Evaluate(values, dzExpr).RealValue;

            Point3D normal = new Point3D(dx, dy, dz);            
            Normal = normal;
            Grad = normal;
        }

        /// <summary>
        /// Computes full evaluation of implicit function.
        /// </summary>
        /// <param name="x">X coordinate</param>
        /// <param name="y">Y coordinate</param>
        /// <param name="z">Z coordinate</param>
        public void Evaluate(double x, double y, double z)
        {
            EvaluateGradient(x, y, z);
            if (double.IsNaN(F)) return;

            double dx = Normal.X;
            double dy = Normal.Y;
            double dz = Normal.Z;
            Point3D normal = Normal;
            normal.Divide(normal.Abs());
            Normal = normal;

            double dxdx = Eval.Evaluate(values, dxdxExpr).RealValue;
            double dxdy = Eval.Evaluate(values, dxdyExpr).RealValue;
            double dxdz = Eval.Evaluate(values, dxdzExpr).RealValue;
            double dydx = Eval.Evaluate(values, dydxExpr).RealValue;
            double dydy = Eval.Evaluate(values, dydyExpr).RealValue;
            double dydz = Eval.Evaluate(values, dydzExpr).RealValue;
            double dzdx = Eval.Evaluate(values, dzdxExpr).RealValue;
            double dzdy = Eval.Evaluate(values, dzdyExpr).RealValue;
            double dzdz = Eval.Evaluate(values, dzdzExpr).RealValue;

            Vector<double> gradF = Vector<double>.Build.Dense(new double[] { dx, dy, dz });
            Matrix<double> HF = Matrix<double>.Build.Dense(3, 3, new double[] 
                { dxdx, dydx, dzdx, 
                  dxdy, dydy, dzdy, 
                  dxdz, dydz, dzdz 
                });
            Matrix<double> adjHF = Matrix<double>.Build.Dense(3, 3, new double[] {
                dydy*dzdz - dydz*dzdy, dxdz*dzdy - dxdy*dzdz, dxdy*dydz - dxdz*dydy,
                dydz*dzdx - dydx*dzdz, dxdx*dzdz - dxdz*dzdx, dydx*dxdz - dxdx*dydz,
                dydx*dzdy - dydy*dzdx, dxdy*dzdx - dxdx*dzdy, dxdx*dydy - dxdy*dydx
            });

            double gradFnorm = gradF.L2Norm();
            double gradFnorm2 = gradFnorm*gradFnorm;
            double K = (gradF * adjHF * gradF)/(gradFnorm2*gradFnorm2);
            double H = (gradF * HF * gradF - gradFnorm2 * HF.Trace()) / (2 * gradFnorm * gradFnorm2);

            double H2 = H * H;

            if (H2 < K)
            {
                K1 = K2 = H;
            }
            else
            {
                double sqrt = Math.Sqrt(H2 - K);
                K1 = H + sqrt;
                K2 = H - sqrt;
            }
        }


        private void Build()
        {
            xExpr = Expr.Symbol("x");
            yExpr = Expr.Symbol("y");
            zExpr = Expr.Symbol("z");
            functionExpr = Infix.ParseOrThrow(function);
            dxExpr = Calculus.Differentiate(xExpr, functionExpr);
            dyExpr = Calculus.Differentiate(yExpr, functionExpr);
            dzExpr = Calculus.Differentiate(zExpr, functionExpr);
            dxdxExpr = Calculus.Differentiate(xExpr, dxExpr);
            dxdyExpr = Calculus.Differentiate(yExpr, dxExpr);
            dxdzExpr = Calculus.Differentiate(zExpr, dxExpr);
            dydxExpr = Calculus.Differentiate(xExpr, dyExpr);
            dydyExpr = Calculus.Differentiate(yExpr, dyExpr);
            dydzExpr = Calculus.Differentiate(zExpr, dyExpr);
            dzdxExpr = Calculus.Differentiate(xExpr, dzExpr);
            dzdyExpr = Calculus.Differentiate(yExpr, dzExpr);
            dzdzExpr = Calculus.Differentiate(zExpr, dzExpr);
        }

        /// <inheritdoc/>
        public double F { get; set;}

        /// <inheritdoc/>
        public double K1 { get; set;}

        /// <inheritdoc/>
        public double K2 { get; set;}

        /// <inheritdoc/>
        public Point3D Normal { get; set;}

        /// <inheritdoc/>
        public Point3D Grad { get; set; }

        /// <inheritdoc/>
        public override string ToString()
        {
            return this.function;
        }
    }
}
