﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Zcu.CurvatureBenchmark.Framework;
using Zcu.CurvatureBenchmark.Framework.DataSources.Implicit;

namespace Zcu.CurvatureBenchmark.DataSources.Implicit
{
    /// <summary>
    /// Regular rectangular box sampler.
    /// </summary>
    public class RegularRectangularBoxSampler : RectangularBoxSampler
    {
        /// <summary>
        /// Gets/sets distorsion that changes the function value. Typical values are 0-0.3.
        /// </summary>
        /// <remarks>
        /// <para>
        /// The distorsion modifies the function value as follows:
        /// <code>
        /// function.Grad.Length() * Rand.NextDouble(-1, 1) * Distorsion * minStep
        /// </code>
        /// where minStep is minimum from (maxA - minA)/(SamplesA - 1) for each axis.
        /// </para>
        /// <para>
        /// When using Distorsion != 0, it is not possible to find the exact surface, as this is usually done by 
        /// some iterative method, using gradient. 
        /// </para>
        /// </remarks>
        public double Distorsion { get; set; }

        /// <inheritdoc/>
        public override void Sample(IImplicitFunction function)
        {           

            Progress = 0;
            base.Sample(function);
            double stepX = (MaxX - MinX) / (SamplesX - 1);
            double stepY = (MaxY - MinY) / (SamplesY - 1);
            double stepZ = (MaxZ - MinZ) / (SamplesZ - 1);

            double minStep = Math.Min(stepX, Math.Min(stepY, stepZ));

            int index = 0;

            for (int iz = 0; iz < SamplesZ; iz++)
            {
                double z = MinZ + iz * stepZ;
                for (int iy = 0; iy < SamplesY; iy++)
                {
                    double y = MinY + iy * stepY;
                    for (int ix = 0; ix < SamplesX; ix++)
                    {
                        double x = MinX + ix * stepX;
                        function.EvaluateGradient(x, y, z);
                        Points[index] = new Point3D(x, y, z);

                        double distorsion = function.Grad.Length() * Rand.NextDouble(-1, 1) * Distorsion * minStep;

                        Values[index] = function.F + distorsion;
                        //K1[index] = function.K1;
                        //K2[index] = function.K2;
                        //Normals[index] = function.Normal;
                        index++;
                    }
                }
            }
            Progress = 100;

        }
    }
}
