﻿/*
 * This is free and unencumbered software released into the public domain.
 *
 * Anyone is free to copy, modify, publish, use, compile, sell, or
 * distribute this software, either in source code form or as a compiled
 * binary, for any purpose, commercial or non-commercial, and by any
 * means.
 *
 * In jurisdictions that recognize copyright laws, the author or authors
 * of this software dedicate any and all copyright interest in the
 * software to the public domain. We make this dedication for the benefit
 * of the public at large and to the detriment of our heirs and
 * successors. We intend this dedication to be an overt act of
 * relinquishment in perpetuity of all present and future rights to this
 * software under copyright law.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 *
 * For more information, please refer to <http://unlicense.org/>
 */

/*
 * https://github.com/tuckbone/DualContouringCSharp
 * 
 * Please do not use this code as the "Ideal way to code in C#"!!! 
 * It is by no means optimized for managed code. This is nothing but a direct port 
 * of NGildea's Dual Contouring Impl from his blog.
 * 
 */
using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Zcu.CurvatureBenchmark.Framework;
using MathNet.Numerics.LinearAlgebra;
using MathNet.Numerics.LinearAlgebra.Double;
using MathNet.Numerics.LinearAlgebra.Factorization;

namespace Zcu.CurvatureBenchmark.DataSources.Implicit
{

    internal static class MatUtils
    {
        public static double fnorm(Mat3 a)
        {
            return Math.Sqrt((a.m00 * a.m00) + (a.m01 * a.m01) + (a.m02 * a.m02)
                        + (a.m10 * a.m10) + (a.m11 * a.m11) + (a.m12 * a.m12)
                        + (a.m20 * a.m20) + (a.m21 * a.m21) + (a.m22 * a.m22));
        }

        public static double fnorm(SMat3 a)
        {
            return Math.Sqrt((a.m00 * a.m00) + (a.m01 * a.m01) + (a.m02 * a.m02)
                        + (a.m01 * a.m01) + (a.m11 * a.m11) + (a.m12 * a.m12)
                        + (a.m02 * a.m02) + (a.m12 * a.m12) + (a.m22 * a.m22));
        }

        public static double off(Mat3 a)
        {
            return Math.Sqrt((a.m01 * a.m01) + (a.m02 * a.m02) + (a.m10 * a.m10) + (a.m12 * a.m12) + (a.m20 * a.m20) + (a.m21 * a.m21));
        }

        public static double off(SMat3 a)
        {
            return Math.Sqrt(2 * ((a.m01 * a.m01) + (a.m02 * a.m02) + (a.m12 * a.m12)));
        }

        public static void mmul(out Mat3 Out, Mat3 a, Mat3 b)
        {
            Out = new Mat3();
            Out.set(a.m00 * b.m00 + a.m01 * b.m10 + a.m02 * b.m20,
                    a.m00 * b.m01 + a.m01 * b.m11 + a.m02 * b.m21,
                    a.m00 * b.m02 + a.m01 * b.m12 + a.m02 * b.m22,
                    a.m10 * b.m00 + a.m11 * b.m10 + a.m12 * b.m20,
                    a.m10 * b.m01 + a.m11 * b.m11 + a.m12 * b.m21,
                    a.m10 * b.m02 + a.m11 * b.m12 + a.m12 * b.m22,
                    a.m20 * b.m00 + a.m21 * b.m10 + a.m22 * b.m20,
                    a.m20 * b.m01 + a.m21 * b.m11 + a.m22 * b.m21,
                    a.m20 * b.m02 + a.m21 * b.m12 + a.m22 * b.m22);
        }

        public static void mmul_ata(out SMat3 Out, Mat3 a)
        {
            Out = new SMat3();

            Out.setSymmetric(a.m00 * a.m00 + a.m10 * a.m10 + a.m20 * a.m20,
                             a.m00 * a.m01 + a.m10 * a.m11 + a.m20 * a.m21,
                             a.m00 * a.m02 + a.m10 * a.m12 + a.m20 * a.m22,
                             a.m01 * a.m01 + a.m11 * a.m11 + a.m21 * a.m21,
                             a.m01 * a.m02 + a.m11 * a.m12 + a.m21 * a.m22,
                             a.m02 * a.m02 + a.m12 * a.m12 + a.m22 * a.m22);
        }

        public static void transpose(out Mat3 Out, Mat3 a)
        {
            Out = new Mat3();

            Out.set(a.m00, a.m10, a.m20, a.m01, a.m11, a.m21, a.m02, a.m12, a.m22);
        }

        public static void vmul(out Point3D Out, Mat3 a, Point3D v)
        {
            Out = new Point3D(
                (a.m00 * v.X) + (a.m01 * v.Y) + (a.m02 * v.Z),
                (a.m10 * v.X) + (a.m11 * v.Y) + (a.m12 * v.Z),
                (a.m20 * v.X) + (a.m21 * v.Y) + (a.m22 * v.Z));
        }

        public static void vmul_symmetric(out Point3D Out, SMat3 a, Point3D v)
        {
            Out = new Point3D(
                (a.m00 * v.X) + (a.m01 * v.Y) + (a.m02 * v.Z),
                (a.m01 * v.X) + (a.m11 * v.Y) + (a.m12 * v.Z),
                (a.m02 * v.X) + (a.m12 * v.Y) + (a.m22 * v.Z));
        }
    }
}
