﻿/*
 * This is free and unencumbered software released into the public domain.
 *
 * Anyone is free to copy, modify, publish, use, compile, sell, or
 * distribute this software, either in source code form or as a compiled
 * binary, for any purpose, commercial or non-commercial, and by any
 * means.
 *
 * In jurisdictions that recognize copyright laws, the author or authors
 * of this software dedicate any and all copyright interest in the
 * software to the public domain. We make this dedication for the benefit
 * of the public at large and to the detriment of our heirs and
 * successors. We intend this dedication to be an overt act of
 * relinquishment in perpetuity of all present and future rights to this
 * software under copyright law.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 *
 * For more information, please refer to <http://unlicense.org/>
 */

/*
 * https://github.com/tuckbone/DualContouringCSharp
 * 
 * Please do not use this code as the "Ideal way to code in C#"!!! 
 * It is by no means optimized for managed code. This is nothing but a direct port 
 * of NGildea's Dual Contouring Impl from his blog.
 * 
 */
using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Zcu.CurvatureBenchmark.Framework;
using MathNet.Numerics.LinearAlgebra;
using MathNet.Numerics.LinearAlgebra.Double;
using MathNet.Numerics.LinearAlgebra.Factorization;

namespace Zcu.CurvatureBenchmark.DataSources.Implicit
{

    internal class QefSolver
    {
        private QefData data;
        private SMat3 ata;
        private Point3D atb, massPoint, x;
        private bool hasSolution;

        public QefSolver()
        {
            data = new QefData();
            ata = new SMat3();
            atb = new Point3D();
            massPoint = new Point3D();
            x = new Point3D();
            hasSolution = false;
        }

        private QefSolver(QefSolver rhs)
        { }

        public Point3D getMassPoint()
        {
            return massPoint;
        }

        public void add(double px, double py, double pz, double nx, double ny, double nz)
        {
            hasSolution = false;

            Point3D tmpv = new Point3D(nx, ny, nz);
            tmpv.Normalize();
            nx = tmpv.X;
            ny = tmpv.Y;
            nz = tmpv.Z;

            data.ata_00 += nx * nx;
            data.ata_01 += nx * ny;
            data.ata_02 += nx * nz;
            data.ata_11 += ny * ny;
            data.ata_12 += ny * nz;
            data.ata_22 += nz * nz;
            double dot = nx * px + ny * py + nz * pz;
            data.atb_x += dot * nx;
            data.atb_y += dot * ny;
            data.atb_z += dot * nz;
            data.btb += dot * dot;
            data.massPoint_x += px;
            data.massPoint_y += py;
            data.massPoint_z += pz;
            ++data.numPoints;
        }

        public void add(Point3D p, Point3D n)
        {
            add(p.X, p.Y, p.Z, n.X, n.Y, n.Z);
        }

        public void add(QefData rhs)
        {
            hasSolution = false;
            data.add(rhs);
        }

        public QefData getData()
        {
            return data;
        }

        public double getError()
        {
            if (!hasSolution)
            {
                throw new ArgumentException("Qef Solver does not have a solution!");
            }

            return getError(x);
        }

        public double getError(Point3D pos)
        {
            if (!hasSolution)
            {
                setAta();
                setAtb();
            }

            Point3D atax;
            MatUtils.vmul_symmetric(out atax, ata, pos);
            return Point3D.DotProduct(pos, atax) - 2 * Point3D.DotProduct(pos, atb) + data.btb;
        }

        public void reset()
        {
            hasSolution = false;
            data.clear();
        }

        public double solve(out Point3D outx, double svd_tol, int svd_sweeps, double pinv_tol)
        {
            if (data.numPoints == 0)
            {
                throw new ArgumentException("...");
            }

            massPoint= new Point3D(data.massPoint_x, data.massPoint_y, data.massPoint_z);
            massPoint *= (1.0f / data.numPoints);
            setAta();
            setAtb();
            Point3D tmpv;
            MatUtils.vmul_symmetric(out tmpv, ata, massPoint);
            atb = atb - tmpv;
            x = new Point3D();
            double result = SVD.solveSymmetric(ata, atb, x, svd_tol, svd_sweeps, pinv_tol);
            x += massPoint * 1;
            setAtb();
            outx = x;
            hasSolution = true;
            return result;
        }

        private void setAta()
        {
            ata.setSymmetric(data.ata_00, data.ata_01, data.ata_02, data.ata_11, data.ata_12, data.ata_22);
        }

        private void setAtb()
        {
            atb = new Point3D(data.atb_x, data.atb_y, data.atb_z);
        }



    }
}
    
