﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Zcu.CurvatureBenchmark.Framework;

namespace Zcu.CurvatureBenchmark.DataSources.Implicit
{
   internal class MarchingCubes
    {

        public int count = 0;
        
        protected readonly double[] data = null;
        protected readonly Point3D[] pointsP = null;
        protected readonly int xSize, ySize, zSize;
        protected readonly int xInc, yInc, zInc;
        protected readonly double threshold;

        protected readonly Point3D[] cube = new Point3D[8];
        protected readonly int[] inc = new int[8];
        protected ArrayList   Verticies = new ArrayList();
        protected ArrayList TriangleIndicies = new ArrayList();        

        static protected readonly int[] vertIndex1 = new int[12] { 0, 1, 2, 3, 4, 5, 6, 7, 0, 1, 2, 3 };
        static protected readonly int[] vertIndex2 = new int[12] { 1, 2, 3, 0, 5, 6, 7, 4, 4, 5, 6, 7 };        
        static protected readonly double epsilon = Utils.EPSILON;
        RectangularBoxSampler gridsampler;

        public MarchingCubes(Point3D[] body,double[] data, double treshold, int[] parameters, RectangularBoxSampler gridsampler, bool exact, int iteration)
        {
            this.data = data;
            this.pointsP = body;
            this.gridsampler = gridsampler;

            xSize = parameters[1];
            ySize = parameters[2];
            zSize = parameters[3];
           

            this.threshold = treshold;

            xInc = 1;
            yInc = xSize;
            zInc = xSize * ySize;
                       
            inc[0] = 0;
            inc[1] = yInc;
            inc[2] = yInc + zInc;
            inc[3] = zInc;
            inc[4] = xInc;
            inc[5] = xInc + yInc;
            inc[6] = xInc + yInc + zInc;
            inc[7] = xInc + zInc;
                      
            double stepX = (gridsampler.MaxX - gridsampler.MinX) / (gridsampler.SamplesX - 1);
            double stepY = (gridsampler.MaxY - gridsampler.MinY) / (gridsampler.SamplesY - 1);
            double stepZ = (gridsampler.MaxZ - gridsampler.MinZ) / (gridsampler.SamplesZ - 1);

            //x,y,z
            cube[0] = new Point3D(0, 0, 0);
            cube[1] = new Point3D(0, stepY, 0);
            cube[2] = new Point3D(0, stepY, stepZ);
            cube[3] = new Point3D(0, 0, stepZ);
            cube[4] = new Point3D(stepX, 0, 0);
            cube[5] = new Point3D(stepX, stepY, 0);
            cube[6] = new Point3D(stepX, stepY, stepZ);
            cube[7] = new Point3D(stepX, 0, stepZ);


            TraceContours(exact,iteration);
        }

        protected void TraceContours(bool exact, int iteration)
        {
             ArrayList VerticiesCandidates = new ArrayList();
                        
            int xyzOffset = 0;
            double value = 0;
           
            for (int z = 0; z < zSize -1; z++)
            {
                for (int y = 0; y < ySize -1; y++)
                {
                    for (int x = 0; x < xSize -1; x++)
                    {
                        xyzOffset = z * zInc + y * yInc + x * xInc;

                        count++;
                        // get marching cubes case index
                        int cubeCase = 0;
                        for (int vertex = 0; vertex < 8; vertex++)
                        {
                            value = data[xyzOffset + inc[vertex]];
                            if (value < threshold+epsilon )
                            {
                                cubeCase |= (1 << vertex);
                                
                            }
                        }

                        // get number of tris in this case
                        int triangles = MarchingCubesTable.NumberOfTriangles[cubeCase];
                       
                        // if this case has triangles
                        if (triangles > 0)
                        {
                            // pocet++;
                            //Point3D ptOrigin = new Point3D(x, y, z);
                            Point3D ptOrigin = pointsP[xyzOffset];

                            int totalVerts = Verticies.Count;

                            // create new vertices
                            for (int edge = 0; edge < 12; edge++)
                            {
                                if ((MarchingCubesTable.Edges[cubeCase] & (1 << edge)) != 0)
                                {
                                    int v1 = vertIndex1[edge];
                                    int v2 = vertIndex2[edge];
                                    double w1 = data[xyzOffset + inc[v1]];
                                    double w2 = data[xyzOffset + inc[v2]];

                                    Point3D pt;

                                    if (Math.Abs(threshold - w1) < epsilon)
                                    {
                                        pt = cube[v1];
                                        pt = pt + ptOrigin;
                                    }
                                    else if (Math.Abs(threshold - w2) < epsilon)
                                    {
                                        pt = cube[v2];
                                        pt = pt + ptOrigin;
                                    }
                                    else if (Math.Abs(w1 - w2) < epsilon)
                                    {
                                        pt = (cube[v1] + cube[v2]) * 0.5f;
                                        pt = pt + ptOrigin;
                                    }
                                    else
                                    {
                                        if (exact)
                                        {
                                            pt = exactValue(v1, v2, w1, w2, iteration, ptOrigin);
                                        }
                                        else
                                        {
                                           
                                            double mu = (double)((threshold - w1) / (w2 - w1));                                            
                                            pt = cube[v1] + ((double)mu * (cube[v2] - cube[v1]));
                                            pt = pt + ptOrigin;
                                        }

                                    }

                                    Point3D resultingPoint = pt;
                                    Verticies.Add(resultingPoint);
                                   
                                    
                                }
                            }


                            //calculate indices of triangles
                            for (int i = 0; i < triangles; i++)
                            {

                                int j = i << 2;
                                TriangleIndicies.Add(totalVerts + MarchingCubesTable.TriangleIndicies[cubeCase, j + 0]);
                                TriangleIndicies.Add(totalVerts + MarchingCubesTable.TriangleIndicies[cubeCase, j + 1]);
                                TriangleIndicies.Add(totalVerts + MarchingCubesTable.TriangleIndicies[cubeCase, j + 2]);

                            }
                                                    
                        }
                        
                    }
                }
            }          
        }

        /// <summary>
        /// Bisection method
        /// </summary>
        /// <param name="v1">index of v1</param>
        /// <param name="v2">index of v2</param>
        /// <param name="w1">function value of of point with index v1</param>
        /// <param name="w2">function value of of point with index v2</param>
        /// <param name="iteration">count of iteration of bisection method</param>
        /// <param name="orig">point coordinates</param>
        /// <returns></returns>
        public Point3D exactValue(int v1, int v2, double w1, double w2, int iteration, Point3D orig) {
            Point3D exactpoint = new Point3D(0, 0, 0);
            double w0 = 0;
            Point3D a = cube[v1] + orig;
            Point3D b = cube[v2]+ orig;


            for (int i = 0; i < iteration; i++) {
                exactpoint = (a + b) / 2;
                gridsampler.Function.Evaluate(exactpoint.X, exactpoint.Y, exactpoint.Z);
                w0 = gridsampler.Function.F;

                if (w0 < threshold)
                {
                    if (w1<w2)
                    {
                        a = exactpoint;
                        w1 = w0;
                    }
                    else 
                    {
                        b = exactpoint;
                        w2 = w0;
                    }
                   
                }
                else {
                    if (w1 > w2)
                    {
                        a = exactpoint;
                        w1 = w0;
                    }
                    else
                    {
                        b = exactpoint;
                        w2 = w0;
                    }
                }
            }

            exactpoint = (a + b) / 2;
            return exactpoint;
        }
        
        public Point3D[] GVerticies
        {
            get { return ((Point3D[])Verticies.ToArray(typeof(Point3D))); }
        }
        public int[] GTriangleIndicies
        {
            get { return (int[])TriangleIndicies.ToArray(typeof(int)); }
        }
       

    }
}
