﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace Zcu.CurvatureBenchmark.DataSources.Implicit
{
    internal class MarchingCubesTable
    {
        ///                     v1-------v5  
        ///                    /|        /|  
        ///                   / |       / | y
        ///                  /  |      /  |
        ///                 v2-------v6   |
        ///                 |   v0----|--v4
        ///                 |  /      |  /
        ///                 | /       | / z
        ///                 |/        |/
        ///                 v3-------v7
        ///                  x

        static public readonly int[] Edges = new int[256] {
                  0x0  , 0x109, 0x203, 0x30a, 0x406, 0x50f, 0x605, 0x70c,
                  0x80c, 0x905, 0xa0f, 0xb06, 0xc0a, 0xd03, 0xe09, 0xf00,
                  0x190, 0x99 , 0x393, 0x29a, 0x596, 0x49f, 0x795, 0x69c,
                  0x99c, 0x895, 0xb9f, 0xa96, 0xd9a, 0xc93, 0xf99, 0xe90,
                  0x230, 0x339, 0x33 , 0x13a, 0x636, 0x73f, 0x435, 0x53c,
                  0xa3c, 0xb35, 0x83f, 0x936, 0xe3a, 0xf33, 0xc39, 0xd30,
                  0x3a0, 0x2a9, 0x1a3, 0xaa , 0x7a6, 0x6af, 0x5a5, 0x4ac,
                  0xbac, 0xaa5, 0x9af, 0x8a6, 0xfaa, 0xea3, 0xda9, 0xca0,
                  0x460, 0x569, 0x663, 0x76a, 0x66 , 0x16f, 0x265, 0x36c,
                  0xc6c, 0xd65, 0xe6f, 0xf66, 0x86a, 0x963, 0xa69, 0xb60,
                  0x5f0, 0x4f9, 0x7f3, 0x6fa, 0x1f6, 0xff , 0x3f5, 0x2fc,
                  0xdfc, 0xcf5, 0xfff, 0xef6, 0x9fa, 0x8f3, 0xbf9, 0xaf0,
                  0x650, 0x759, 0x453, 0x55a, 0x256, 0x35f, 0x55 , 0x15c,
                  0xe5c, 0xf55, 0xc5f, 0xd56, 0xa5a, 0xb53, 0x859, 0x950,
                  0x7c0, 0x6c9, 0x5c3, 0x4ca, 0x3c6, 0x2cf, 0x1c5, 0xcc ,
                  0xfcc, 0xec5, 0xdcf, 0xcc6, 0xbca, 0xac3, 0x9c9, 0x8c0,
                  0x8c0, 0x9c9, 0xac3, 0xbca, 0xcc6, 0xdcf, 0xec5, 0xfcc,
                  0xcc , 0x1c5, 0x2cf, 0x3c6, 0x4ca, 0x5c3, 0x6c9, 0x7c0,
                  0x950, 0x859, 0xb53, 0xa5a, 0xd56, 0xc5f, 0xf55, 0xe5c,
                  0x15c, 0x55 , 0x35f, 0x256, 0x55a, 0x453, 0x759, 0x650,
                  0xaf0, 0xbf9, 0x8f3, 0x9fa, 0xef6, 0xfff, 0xcf5, 0xdfc,
                  0x2fc, 0x3f5, 0xff , 0x1f6, 0x6fa, 0x7f3, 0x4f9, 0x5f0,
                  0xb60, 0xa69, 0x963, 0x86a, 0xf66, 0xe6f, 0xd65, 0xc6c,
                  0x36c, 0x265, 0x16f, 0x66 , 0x76a, 0x663, 0x569, 0x460,
                  0xca0, 0xda9, 0xea3, 0xfaa, 0x8a6, 0x9af, 0xaa5, 0xbac,
                  0x4ac, 0x5a5, 0x6af, 0x7a6, 0xaa , 0x1a3, 0x2a9, 0x3a0,
                  0xd30, 0xc39, 0xf33, 0xe3a, 0x936, 0x83f, 0xb35, 0xa3c,
                  0x53c, 0x435, 0x73f, 0x636, 0x13a, 0x33 , 0x339, 0x230,
                  0xe90, 0xf99, 0xc93, 0xd9a, 0xa96, 0xb9f, 0x895, 0x99c,
                  0x69c, 0x795, 0x49f, 0x596, 0x29a, 0x393, 0x99 , 0x190,
                  0xf00, 0xe09, 0xd03, 0xc0a, 0xb06, 0xa0f, 0x905, 0x80c,
                  0x70c, 0x605, 0x50f, 0x406, 0x30a, 0x203, 0x109, 0x0
              };

         
        static public readonly int[,] TriangleIndicies = new int[256, 20] { {
               -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                0,  2,  1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                0,  1,  2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                0,  2,  1, -1,  3,  2,  0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                0,  1,  2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                0,  4,  3, -1,  1,  2,  5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                2,  1,  3, -1,  0,  1,  2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                0,  2,  1, -1,  0,  4,  2, -1,  4,  3,  2, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
               1,  2,  0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
               0,  3,  1, -1,  2,  3,  0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
               1,  4,  0, -1,  2,  3,  5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                0,  4,  1, -1,  0,  3,  4, -1,  3,  2,  4, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
               1,  2,  0, -1,  3,  2,  1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                0,  3,  1, -1,  0,  2,  3, -1,  2,  4,  3, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                1,  2,  0, -1,  1,  4,  2, -1,  4,  3,  2, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                1,  0,  2, -1,  2,  0,  3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                0,  1,  2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                2,  1,  0, -1,  3,  1,  2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                0,  1,  5, -1,  4,  2,  3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                2,  0,  4, -1,  2,  3,  0, -1,  3,  1,  0, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                0,  1,  5, -1,  4,  2,  3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                3,  4,  5, -1,  3,  0,  4, -1,  1,  2,  6, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                5,  1,  6, -1,  5,  0,  1, -1,  4,  2,  3, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                0,  5,  4, -1,  0,  4,  3, -1,  0,  3,  1, -1,  3,  4,  2, -1, -1, -1, -1, -1}, {
                4,  2,  3, -1,  1,  5,  0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                4,  2,  3, -1,  4,  1,  2, -1,  1,  0,  2, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                7,  0,  1, -1,  6,  4,  5, -1,  2,  3,  8, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                2,  3,  5, -1,  4,  2,  5, -1,  4,  5,  1, -1,  4,  1,  0, -1, -1, -1, -1, -1}, {
                1,  5,  0, -1,  1,  6,  5, -1,  3,  4,  2, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                1,  5,  4, -1,  1,  2,  5, -1,  1,  0,  2, -1,  3,  5,  2, -1, -1, -1, -1, -1}, {
                2,  3,  4, -1,  5,  0,  7, -1,  5,  7,  6, -1,  7,  0,  1, -1, -1, -1, -1, -1}, {
                0,  1,  4, -1,  0,  4,  2, -1,  2,  4,  3, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                2,  1,  0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                5,  3,  2, -1,  0,  4,  1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                0,  3,  2, -1,  1,  3,  0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                4,  3,  2, -1,  4,  1,  3, -1,  1,  0,  3, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                0,  1,  5, -1,  4,  3,  2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                3,  0,  6, -1,  1,  2,  8, -1,  4,  7,  5, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                3,  1,  4, -1,  3,  2,  1, -1,  2,  0,  1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                0,  5,  3, -1,  1,  0,  3, -1,  1,  3,  2, -1,  1,  2,  4, -1, -1, -1, -1, -1}, {
                4,  3,  2, -1,  0,  1,  5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                0,  6,  1, -1,  0,  4,  6, -1,  2,  5,  3, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                0,  5,  4, -1,  0,  1,  5, -1,  2,  3,  6, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                1,  0,  3, -1,  1,  3,  4, -1,  1,  4,  5, -1,  2,  4,  3, -1, -1, -1, -1, -1}, {
                5,  1,  6, -1,  5,  0,  1, -1,  4,  3,  2, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                2,  5,  3, -1,  0,  4,  1, -1,  4,  6,  1, -1,  4,  7,  6, -1, -1, -1, -1, -1}, {
                3,  2,  0, -1,  3,  0,  5, -1,  3,  5,  4, -1,  5,  0,  1, -1, -1, -1, -1, -1}, {
                1,  0,  2, -1,  1,  2,  3, -1,  3,  2,  4, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                3,  1,  2, -1,  0,  1,  3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                4,  1,  0, -1,  4,  2,  1, -1,  2,  3,  1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                0,  3,  4, -1,  0,  1,  3, -1,  1,  2,  3, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                0,  2,  1, -1,  1,  2,  3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                5,  3,  4, -1,  5,  2,  3, -1,  6,  0,  1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                7,  1,  2, -1,  6,  4,  0, -1,  4,  3,  0, -1,  4,  5,  3, -1, -1, -1, -1, -1}, {
                4,  0,  1, -1,  4,  1,  2, -1,  4,  2,  3, -1,  5,  2,  1, -1, -1, -1, -1, -1}, {
                0,  4,  2, -1,  0,  2,  1, -1,  1,  2,  3, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                3,  5,  2, -1,  3,  4,  5, -1,  1,  6,  0, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                4,  2,  3, -1,  4,  3,  1, -1,  4,  1,  0, -1,  1,  3,  5, -1, -1, -1, -1, -1}, {
               2,  3,  7, -1,  0,  1,  6, -1,  1,  5,  6, -1,  1,  4,  5, -1, -1, -1, -1, -1}, {
               4,  1,  0, -1,  4,  0,  3, -1,  3,  0,  2, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
               5,  2,  4, -1,  4,  2,  3, -1,  6,  0,  1, -1,  6,  1,  7, -1, -1, -1, -1, -1}, {
               2,  3,  0, -1,  2,  0,  4, -1,  3,  6,  0, -1,  1,  0,  5, -1,  6,  5,  0, -1}, {
               6,  5,  0, -1,  6,  0,  1, -1,  5,  2,  0, -1,  4,  0,  3, -1,  2,  3,  0, -1}, {
               3,  2,  0, -1,  1,  3,  0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                2,  1,  0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                0,  4,  1, -1,  2,  5,  3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
               4,  0,  1, -1,  2,  5,  3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                0,  4,  1, -1,  0,  5,  4, -1,  2,  6,  3, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                0,  3,  2, -1,  1,  3,  0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                1,  5,  4, -1,  1,  2,  5, -1,  3,  0,  6, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                4,  3,  2, -1,  4,  0,  3, -1,  0,  1,  3, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                2,  5,  4, -1,  2,  4,  0, -1,  2,  0,  3, -1,  1,  0,  4, -1, -1, -1, -1, -1}, {
                0,  1,  5, -1,  4,  3,  2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                6,  0,  4, -1,  6,  1,  0, -1,  5,  3,  2, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                0,  1,  6, -1,  2,  3,  8, -1,  4,  7,  5, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                2,  6,  3, -1,  0,  5,  1, -1,  5,  7,  1, -1,  5,  4,  7, -1, -1, -1, -1, -1}, {
               3,  1,  4, -1,  3,  2,  1, -1,  2,  0,  1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                0,  4,  5, -1,  0,  5,  2, -1,  0,  2,  1, -1,  2,  5,  3, -1, -1, -1, -1, -1}, {
                1,  5,  3, -1,  0,  1,  3, -1,  0,  3,  2, -1,  0,  2,  4, -1, -1, -1, -1, -1}, {
                1,  0,  3, -1,  1,  3,  4, -1,  4,  3,  2, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                1,  5,  2, -1,  0,  3,  4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                2,  1,  0, -1,  2,  5,  1, -1,  4,  3,  6, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                1,  7,  0, -1,  3,  8,  4, -1,  6,  2,  5, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                7,  4,  3, -1,  0,  6,  5, -1,  0,  5,  1, -1,  5,  6,  2, -1, -1, -1, -1, -1}, {
                4,  0,  1, -1,  4,  3,  0, -1,  2,  5,  6, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                1,  2,  5, -1,  5,  2,  6, -1,  3,  0,  4, -1,  3,  4,  7, -1, -1, -1, -1, -1}, {
                6,  2,  5, -1,  7,  0,  3, -1,  0,  4,  3, -1,  0,  1,  4, -1, -1, -1, -1, -1}, {
                5,  1,  6, -1,  5,  6,  2, -1,  1,  0,  6, -1,  3,  6,  4, -1,  0,  4,  6, -1}, {
                1,  8,  0, -1,  5,  6,  2, -1,  7,  4,  3, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                3,  6,  4, -1,  2,  5,  1, -1,  2,  1,  0, -1,  1,  5,  7, -1, -1, -1, -1, -1}, {
                0,  1,  9, -1,  4,  7,  8, -1,  2,  3, 11, -1,  5, 10,  6, -1, -1, -1, -1, -1}, {
                6,  1,  0, -1,  6,  8,  1, -1,  6,  2,  8, -1,  5,  8,  2, -1,  3,  7,  4, -1}, {
                6,  2,  5, -1,  1,  7,  3, -1,  1,  3,  0, -1,  3,  7,  4, -1, -1, -1, -1, -1}, {
                3,  1,  6, -1,  3,  6,  4, -1,  1,  0,  6, -1,  5,  6,  2, -1,  0,  2,  6, -1}, {
                0,  3,  7, -1,  0,  4,  3, -1,  0,  1,  4, -1,  8,  4,  1, -1,  6,  2,  5, -1}, {
                2,  1,  4, -1,  2,  4,  5, -1,  0,  3,  4, -1,  3,  5,  4, -1, -1, -1, -1, -1}, {
                3,  0,  2, -1,  1,  0,  3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                2,  6,  3, -1,  2,  5,  6, -1,  0,  4,  1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                4,  0,  1, -1,  4,  3,  0, -1,  3,  2,  0, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                4,  1,  0, -1,  4,  0,  3, -1,  4,  3,  2, -1,  3,  0,  5, -1, -1, -1, -1, -1}, {
                0,  2,  4, -1,  0,  1,  2, -1,  1,  3,  2, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                3,  0,  6, -1,  1,  2,  7, -1,  2,  4,  7, -1,  2,  5,  4, -1, -1, -1, -1, -1}, {
                0,  1,  2, -1,  2,  1,  3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                4,  1,  0, -1,  4,  0,  2, -1,  2,  0,  3, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                5,  2,  4, -1,  5,  3,  2, -1,  6,  0,  1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                0,  4,  1, -1,  1,  4,  7, -1,  2,  5,  6, -1,  2,  6,  3, -1, -1, -1, -1, -1}, {
                3,  7,  2, -1,  0,  1,  5, -1,  0,  5,  4, -1,  5,  1,  6, -1, -1, -1, -1, -1}, {
                3,  2,  0, -1,  3,  0,  5, -1,  2,  4,  0, -1,  1,  0,  6, -1,  4,  6,  0, -1}, {
                4,  3,  2, -1,  4,  1,  3, -1,  4,  0,  1, -1,  5,  3,  1, -1, -1, -1, -1, -1}, {
                4,  6,  1, -1,  4,  1,  0, -1,  6,  3,  1, -1,  5,  1,  2, -1,  3,  2,  1, -1}, {
                1,  4,  3, -1,  1,  3,  0, -1,  0,  3,  2, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                1,  0,  2, -1,  3,  1,  2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                1,  4,  0, -1,  1,  2,  4, -1,  2,  3,  4, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                0,  3,  1, -1,  0,  5,  3, -1,  0,  4,  5, -1,  2,  3,  5, -1, -1, -1, -1, -1}, {
                5,  2,  3, -1,  1,  5,  3, -1,  1,  3,  4, -1,  1,  4,  0, -1, -1, -1, -1, -1}, {
                4,  2,  3, -1,  4,  3,  0, -1,  0,  3,  1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
               0,  1,  2, -1,  0,  2,  4, -1,  0,  4,  5, -1,  4,  2,  3, -1, -1, -1, -1, -1}, {
               2,  4,  6, -1,  2,  6,  1, -1,  4,  5,  6, -1,  0,  6,  3, -1,  5,  3,  6, -1}, {
                3,  4,  0, -1,  3,  0,  2, -1,  2,  0,  1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
               3,  1,  0, -1,  2,  3,  0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
               0,  1,  7, -1,  6,  2,  4, -1,  6,  4,  5, -1,  4,  2,  3, -1, -1, -1, -1, -1}, {
                1,  0,  3, -1,  1,  3,  6, -1,  0,  4,  3, -1,  2,  3,  5, -1,  4,  5,  3, -1}, {
                1,  6,  0, -1,  1,  5,  6, -1,  1,  7,  5, -1,  4,  5,  7, -1,  2,  3,  8, -1}, {
                5,  1,  0, -1,  5,  0,  3, -1,  4,  2,  0, -1,  2,  3,  0, -1, -1, -1, -1, -1}, {
                4,  5,  2, -1,  4,  2,  3, -1,  5,  0,  2, -1,  6,  2,  1, -1,  0,  1,  2, -1}, {
                0,  4,  1, -1,  5,  2,  3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                3,  4,  0, -1,  3,  0,  2, -1,  1,  5,  0, -1,  5,  2,  0, -1, -1, -1, -1, -1}, {
                1,  2,  0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                1,  0,  2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                1,  0,  4, -1,  5,  3,  2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                0,  1,  4, -1,  5,  3,  2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                4,  0,  5, -1,  4,  1,  0, -1,  6,  3,  2, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                4,  0,  1, -1,  2,  5,  3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                1,  2,  7, -1,  3,  0,  6, -1,  4,  8,  5, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                1,  4,  0, -1,  1,  5,  4, -1,  2,  6,  3, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                2,  7,  3, -1,  0,  6,  1, -1,  6,  4,  1, -1,  6,  5,  4, -1, -1, -1, -1, -1}, {
                3,  0,  1, -1,  2,  0,  3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                3,  0,  4, -1,  3,  2,  0, -1,  2,  1,  0, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                2,  5,  4, -1,  2,  3,  5, -1,  0,  1,  6, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                0,  2,  1, -1,  0,  4,  2, -1,  0,  5,  4, -1,  4,  3,  2, -1, -1, -1, -1, -1}, {
                4,  3,  2, -1,  4,  0,  3, -1,  0,  1,  3, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                5,  3,  2, -1,  1,  3,  5, -1,  1,  4,  3, -1,  1,  0,  4, -1, -1, -1, -1, -1}, {
                0,  1,  3, -1,  0,  3,  5, -1,  0,  5,  4, -1,  2,  5,  3, -1, -1, -1, -1, -1}, {
                1,  0,  4, -1,  1,  4,  2, -1,  2,  4,  3, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                1,  2,  0, -1,  3,  2,  1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                1,  3,  4, -1,  1,  0,  3, -1,  0,  2,  3, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                4,  3,  6, -1,  4,  2,  3, -1,  5,  0,  1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                4,  2,  3, -1,  4,  3,  1, -1,  4,  1,  0, -1,  5,  1,  3, -1, -1, -1, -1, -1}, {
                3,  4,  2, -1,  3,  6,  4, -1,  1,  5,  0, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                1,  2,  6, -1,  3,  0,  7, -1,  0,  5,  7, -1,  0,  4,  5, -1, -1, -1, -1, -1}, {
                2,  7,  4, -1,  2,  3,  7, -1,  0,  1,  5, -1,  1,  6,  5, -1, -1, -1, -1, -1}, {
                5,  4,  1, -1,  5,  1,  0, -1,  4,  2,  1, -1,  6,  1,  3, -1,  2,  3,  1, -1}, {
                4,  0,  1, -1,  4,  2,  0, -1,  2,  3,  0, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                0,  2,  1, -1,  2,  3,  1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                1,  7,  0, -1,  2,  3,  4, -1,  2,  4,  5, -1,  4,  3,  6, -1, -1, -1, -1, -1}, {
                0,  4,  2, -1,  0,  2,  1, -1,  1,  2,  3, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                4,  0,  1, -1,  4,  3,  0, -1,  4,  2,  3, -1,  3,  5,  0, -1, -1, -1, -1, -1}, {
                4,  1,  0, -1,  4,  0,  3, -1,  3,  0,  2, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                2,  3,  1, -1,  2,  1,  4, -1,  3,  6,  1, -1,  0,  1,  5, -1,  6,  5,  1, -1}, {
                3,  2,  0, -1,  1,  3,  0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                0,  4,  1, -1,  3,  2,  5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                0,  6,  1, -1,  2,  7,  3, -1,  8,  5,  4, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                3,  0,  1, -1,  3,  2,  0, -1,  5,  4,  6, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                7,  5,  4, -1,  6,  1,  2, -1,  1,  3,  2, -1,  1,  0,  3, -1, -1, -1, -1, -1}, {
                6,  3,  2, -1,  7,  0,  1, -1,  5,  4,  8, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                6, 11,  7, -1,  1,  2, 10, -1,  0,  8,  3, -1,  4,  9,  5, -1, -1, -1, -1, -1}, {
                5,  4,  7, -1,  3,  2,  6, -1,  2,  1,  6, -1,  2,  0,  1, -1, -1, -1, -1, -1}, {
                1,  2,  6, -1,  1,  3,  2, -1,  1,  0,  3, -1,  7,  3,  0, -1,  8,  5,  4, -1}, {
                5,  0,  1, -1,  5,  4,  0, -1,  3,  2,  6, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                7,  3,  2, -1,  0,  6,  4, -1,  0,  4,  1, -1,  4,  6,  5, -1, -1, -1, -1, -1}, {
                3,  6,  2, -1,  3,  7,  6, -1,  1,  5,  0, -1,  5,  4,  0, -1, -1, -1, -1, -1}, {
                4,  1,  6, -1,  4,  6,  5, -1,  1,  0,  6, -1,  2,  6,  3, -1,  0,  3,  6, -1}, {
                6,  3,  2, -1,  7,  0,  4, -1,  0,  5,  4, -1,  0,  1,  5, -1, -1, -1, -1, -1}, {
                1,  4,  8, -1,  1,  5,  4, -1,  1,  0,  5, -1,  6,  5,  0, -1,  7,  3,  2, -1}, {
                2,  0,  6, -1,  2,  6,  3, -1,  0,  1,  6, -1,  4,  6,  5, -1,  1,  5,  6, -1}, {
                3,  2,  5, -1,  3,  5,  4, -1,  1,  0,  5, -1,  0,  4,  5, -1, -1, -1, -1, -1}, {
                1,  3,  0, -1,  1,  4,  3, -1,  4,  2,  3, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                1,  3,  5, -1,  0,  3,  1, -1,  0,  2,  3, -1,  0,  4,  2, -1, -1, -1, -1, -1}, {
                0,  5,  4, -1,  0,  2,  5, -1,  0,  1,  2, -1,  2,  3,  5, -1, -1, -1, -1, -1}, {
                3,  4,  1, -1,  3,  1,  2, -1,  2,  1,  0, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                0,  1,  6, -1,  5,  2,  7, -1,  5,  7,  4, -1,  7,  2,  3, -1, -1, -1, -1, -1}, {
                0,  8,  3, -1,  0,  5,  8, -1,  0,  6,  5, -1,  4,  5,  6, -1,  1,  2,  7, -1}, {
                6,  4,  2, -1,  6,  2,  3, -1,  4,  0,  2, -1,  5,  2,  1, -1,  0,  1,  2, -1}, {
                3,  5,  1, -1,  3,  1,  2, -1,  0,  4,  1, -1,  4,  2,  1, -1, -1, -1, -1, -1}, {
                2,  4,  5, -1,  2,  0,  4, -1,  2,  3,  0, -1,  1,  4,  0, -1, -1, -1, -1, -1}, {
                4,  2,  3, -1,  4,  3,  0, -1,  0,  3,  1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                1,  4,  6, -1,  1,  6,  0, -1,  4,  5,  6, -1,  3,  6,  2, -1,  5,  2,  6, -1}, {
                0,  2,  3, -1,  1,  0,  3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                0,  1,  3, -1,  0,  3,  6, -1,  1,  4,  3, -1,  2,  3,  5, -1,  4,  5,  3, -1}, {
                5,  1,  0, -1,  5,  0,  3, -1,  4,  2,  0, -1,  2,  3,  0, -1, -1, -1, -1, -1}, {
                0,  1,  4, -1,  2,  3,  5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                2,  0,  1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                3,  0,  2, -1,  1,  0,  3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                6,  2,  5, -1,  6,  3,  2, -1,  4,  1,  0, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                2,  6,  3, -1,  2,  5,  6, -1,  1,  4,  0, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                6,  3,  2, -1,  6,  7,  3, -1,  5,  4,  0, -1,  4,  1,  0, -1, -1, -1, -1, -1}, {
                4,  0,  1, -1,  4,  3,  0, -1,  3,  2,  0, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                0,  6,  3, -1,  1,  2,  5, -1,  1,  5,  4, -1,  5,  2,  7, -1, -1, -1, -1, -1}, {
                4,  3,  2, -1,  4,  1,  3, -1,  4,  0,  1, -1,  1,  5,  3, -1, -1, -1, -1, -1}, {
                3,  2,  0, -1,  3,  0,  6, -1,  2,  5,  0, -1,  1,  0,  4, -1,  5,  4,  0, -1}, {
                0,  2,  4, -1,  0,  1,  2, -1,  1,  3,  2, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                4,  1,  0, -1,  4,  2,  1, -1,  4,  3,  2, -1,  5,  1,  2, -1, -1, -1, -1, -1}, {
                6,  0,  1, -1,  4,  7,  3, -1,  4,  3,  5, -1,  3,  7,  2, -1, -1, -1, -1, -1}, {
                5,  4,  1, -1,  5,  1,  0, -1,  4,  3,  1, -1,  6,  1,  2, -1,  3,  2,  1, -1}, {
                0,  1,  2, -1,  1,  3,  2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                0,  4,  3, -1,  0,  3,  1, -1,  1,  3,  2, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                4,  0,  1, -1,  4,  1,  2, -1,  2,  1,  3, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                3,  2,  1, -1,  0,  3,  1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                1,  2,  0, -1,  1,  3,  2, -1,  3,  4,  2, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                3,  0,  2, -1,  3,  5,  0, -1,  3,  4,  5, -1,  5,  1,  0, -1, -1, -1, -1, -1}, {
                0,  1,  5, -1,  4,  2,  6, -1,  4,  6,  7, -1,  6,  2,  3, -1, -1, -1, -1, -1}, {
                5,  6,  2, -1,  5,  2,  3, -1,  6,  1,  2, -1,  4,  2,  0, -1,  1,  0,  2, -1}, {
                1,  3,  0, -1,  1,  4,  3, -1,  1,  5,  4, -1,  2,  3,  4, -1, -1, -1, -1, -1}, {
                0,  4,  6, -1,  0,  6,  3, -1,  4,  5,  6, -1,  2,  6,  1, -1,  5,  1,  6, -1}, {
                0,  1,  3, -1,  0,  3,  5, -1,  1,  6,  3, -1,  2,  3,  4, -1,  6,  4,  3, -1}, {
                4,  2,  3, -1,  0,  5,  1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                0,  3,  5, -1,  1,  3,  0, -1,  1,  2,  3, -1,  1,  4,  2, -1, -1, -1, -1, -1}, {
                3,  4,  1, -1,  3,  1,  2, -1,  2,  1,  0, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                3,  8,  2, -1,  3,  5,  8, -1,  3,  6,  5, -1,  4,  5,  6, -1,  0,  1,  7, -1}, {
                3,  5,  1, -1,  3,  1,  2, -1,  0,  4,  1, -1,  4,  2,  1, -1, -1, -1, -1, -1}, {
                4,  2,  3, -1,  4,  3,  1, -1,  1,  3,  0, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                0,  2,  3, -1,  1,  0,  3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                4,  2,  3, -1,  4,  3,  1, -1,  5,  0,  3, -1,  0,  1,  3, -1, -1, -1, -1, -1}, {
                2,  0,  1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                0,  4,  1, -1,  0,  2,  4, -1,  2,  3,  4, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                0,  4,  1, -1,  2,  5,  3, -1,  5,  7,  3, -1,  5,  6,  7, -1, -1, -1, -1, -1}, {
                1,  4,  5, -1,  1,  5,  2, -1,  1,  2,  0, -1,  3,  2,  5, -1, -1, -1, -1, -1}, {
                1,  0,  2, -1,  1,  2,  4, -1,  0,  5,  2, -1,  3,  2,  6, -1,  5,  6,  2, -1}, {
                2,  5,  3, -1,  4,  5,  2, -1,  4,  1,  5, -1,  4,  0,  1, -1, -1, -1, -1, -1}, {
                7,  5,  4, -1,  7,  8,  5, -1,  7,  1,  8, -1,  2,  8,  1, -1,  0,  6,  3, -1}, {
                4,  3,  2, -1,  4,  2,  1, -1,  1,  2,  0, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                5,  3,  2, -1,  5,  2,  0, -1,  4,  1,  2, -1,  1,  0,  2, -1, -1, -1, -1, -1}, {
                0,  4,  5, -1,  0,  3,  4, -1,  0,  1,  3, -1,  3,  2,  4, -1, -1, -1, -1, -1}, {
                5,  6,  3, -1,  5,  3,  2, -1,  6,  1,  3, -1,  4,  3,  0, -1,  1,  0,  3, -1}, {
                3,  5,  6, -1,  3,  6,  2, -1,  5,  4,  6, -1,  1,  6,  0, -1,  4,  0,  6, -1}, {
                0,  5,  1, -1,  4,  3,  2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                2,  4,  0, -1,  2,  0,  3, -1,  3,  0,  1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                2,  5,  1, -1,  2,  1,  3, -1,  0,  4,  1, -1,  4,  3,  1, -1, -1, -1, -1, -1}, {
                2,  0,  1, -1,  3,  2,  1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                0,  2,  1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                1,  2,  0, -1,  2,  3,  0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                1,  0,  2, -1,  1,  2,  4, -1,  4,  2,  3, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                0,  1,  3, -1,  0,  3,  2, -1,  2,  3,  4, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                1,  0,  2, -1,  3,  1,  2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                0,  1,  4, -1,  0,  4,  3, -1,  3,  4,  2, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                3,  0,  4, -1,  3,  4,  5, -1,  1,  2,  4, -1,  2,  5,  4, -1, -1, -1, -1, -1}, {
                0,  1,  3, -1,  2,  0,  3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                1,  0,  2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
               0,  1,  2, -1,  0,  2,  4, -1,  4,  2,  3, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                2,  3,  1, -1,  0,  2,  1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                2,  3,  4, -1,  2,  4,  5, -1,  0,  1,  4, -1,  1,  5,  4, -1, -1, -1, -1, -1}, {
                0,  2,  1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                0,  1,  2, -1,  3,  0,  2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                0,  2,  1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
                0,  1,  2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}, {
               -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
          };

        static public readonly int[] NumberOfTriangles = new int[256] {
              0, 1, 1, 2, 1, 2, 2, 3, 1, 2, 2, 3, 2, 3, 3, 2,
              1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 3,
              1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 3,
              2, 3, 3, 2, 3, 4, 4, 3, 3, 4, 4, 3, 4, 5, 5, 2,
              1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 3,
              2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 4,
              2, 3, 3, 4, 3, 4, 2, 3, 3, 4, 4, 5, 4, 5, 3, 2,
              3, 4, 4, 3, 4, 5, 3, 2, 4, 5, 5, 4, 5, 2, 4, 1,
              1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 3,
             2, 3, 3, 4, 3, 4, 4, 5, 3, 2, 4, 3, 4, 3, 5, 2,
              2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 4,
              3, 4, 4, 3, 4, 5, 5, 4, 4, 3, 5, 2, 5, 4, 2, 1,
              2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 2, 3, 3, 2,
              3, 4, 4, 5, 4, 5, 5, 2, 4, 3, 5, 4, 3, 2, 4, 1,
              3, 4, 4, 5, 4, 5, 3, 4, 4, 5, 5, 2, 3, 4, 2, 1,
              2, 3, 3, 2, 3, 4, 2, 1, 3, 2, 4, 1, 2, 1, 1, 0
          };

    }
}
