﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Zcu.CurvatureBenchmark.Framework;
using Zcu.CurvatureBenchmark.Framework.DataSources.Implicit;

namespace Zcu.CurvatureBenchmark.DataSources.Implicit
{

    /// <summary>
    /// Marching tetrahedra 6 implementation. The module uses the samples produced by RectagularBoxSampler and extracts 0 iso-value.
    /// </summary>
    /// <author>Petr Martinek</author>      
    public class RGTriangularizator_MT : AbstractModule, IImplicitTriangularizator
    {
        /// <summary>
        /// Gets/sets whether to try to find the exact surface.
        /// </summary>
        /// <remarks>
        /// Usually, marching tetrahedra uses linear interpolation to estimate the surface position on the edges of the cells.
        /// As there is a continuous function, it is possible to try to find the exact surface position by an iterative method.
        /// To avoid some very long computations, there is a limit of iterations (<see cref="Iteration"/>).
        /// </remarks>
        public virtual bool Exact { get; set; }

        /// <summary>
        /// Gets/sets maximal number of iterations for exact surface computation.
        /// </summary>
        public virtual int Iteration { get; set; }

        /// <summary>
        /// Initialization.
        /// </summary>
        public RGTriangularizator_MT()
        {
            Exact = false;
            Iteration = 10;
        }

        RectangularBoxSampler gridsampler;

        /// <inheritdoc/>
        public void Triangulate(IImplicitSampler sampler)
        {
            Progress = 0;
            if (!(sampler is RectangularBoxSampler)) throw new ApplicationException("RectangularBoxTriangularizator needs RectangularBoxSampler");
            gridsampler = sampler as RectangularBoxSampler;

            
            TriangleMesh = new TriangleMesh();
            //TriangleMesh.Points = sampler.Points;
            
            int[] parameters = new int[4] {0, gridsampler.SamplesX, gridsampler.SamplesY, gridsampler.SamplesZ};


            MT6 mt = new MT6(sampler.Points, sampler.Values, (double)0, parameters, gridsampler, Exact, Iteration);
            Point3D[] rvVerticies = mt.GVerticies;
            Triangle[] triang = mt.GTriangleIndicies;

            for (int i = 0; i < triang.Length; i++) {
                triang[i] = checkOrientation(triang[i], rvVerticies, sampler);

            }
            
                   
            double[] k1s = new double[rvVerticies.Length];
            double[] k2s = new double[rvVerticies.Length];
            Point3D[] normalS = new Point3D[rvVerticies.Length];

            for (int i = 0; i < rvVerticies.Length; i++) {
                sampler.Function.Evaluate(rvVerticies[i].X, rvVerticies[i].Y, rvVerticies[i].Z);
                k1s[i] = sampler.Function.K1;
                k2s[i] = sampler.Function.K2;
                normalS[i] = sampler.Function.Normal;
            }


            TriangleMesh.Points = rvVerticies;
            TriangleMesh.Triangles = triang;
            TriangleMesh.ExactK1 = k1s;
            TriangleMesh.ExactK2 = k2s;
            TriangleMesh.Normals = normalS;

            //ObjSaver saver = new ObjSaver();

            //saver.FileName = "mesh1.obj";

            //saver.Execute(TriangleMesh);
                       
            Progress = 100;
        }


        //unification of orientation
        private Triangle checkOrientation(Triangle t, Point3D[] vertices, IImplicitSampler sampler)
        {
            Triangle g = t;
            Point3D a = vertices[t.V1];
            Point3D b = vertices[t.V2];
            Point3D c = vertices[t.V3];

            // sampler.Function.Evaluate((a.X+b.X+c.X)/3, (a.Y + b.Y + c.Y) / 3, (a.Z + b.Z + c.Z) / 3);
            sampler.Function.Evaluate(a.X, a.Y, a.Z);
            Point3D exactNorm = sampler.Function.Normal;

            //Point3D normala1 = Point3D.CrossProduct((b - a), (c - a));
            //Point3D normala2 = Point3D.CrossProduct((c - b), (a - b));
            //Point3D normala3 = Point3D.CrossProduct((a - c), (b - c));

            Point3D normal = Point3D.CrossProduct((c - a), (b - a));

            double test = Point3D.DotProduct(exactNorm, normal);
            if (test < 0)
            {
                t.V2 = g.V3;
                t.V3 = g.V2;
            }


            return t;
        }

        /// <inheritdoc/>
        public TriangleMesh TriangleMesh {get ;set;}

       


    }
}
