﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Zcu.CurvatureBenchmark.Framework;
using MathNet.Symbolics;
using System.Xml;
using Zcu.CurvatureBenchmark.DataSources.Explicit;
using Zcu.CurvatureBenchmark.Framework.DataSources;
using Zcu.CurvatureBenchmark.Framework.DataSources.Explicit;

namespace Zcu.CurvatureBenchmark.DataSources
{
    /// <summary>
    /// Module generates NURBS surface. The surface is sampled by any <see cref="IExplicitSampler"/>
    /// and triangulated by <see cref="IExplicitTriangularizator"/>. The sampler samples the parameter domain
    /// instead of E2.
    /// </summary>
    public class NURBSSource : AbstractDataSource
    {
        /// <summary>
        /// dummy function, used to sample the parameter domain 
        /// </summary>
        IExplicitFunction function;

        /// <summary>
        /// sampler for the parameter domain 
        /// </summary>
        IExplicitSampler sampler;

        /// <summary>
        /// triangularizator for the parameter domain
        /// </summary>
        IExplicitTriangularizator triangularizator;

        /// <summary>
        /// 2D basis functions and their derivatives 
        /// </summary>
        Expression[,] R, Ru, Rv, Ruu, Rvv, Ruv;

        /// <summary>
        /// Sum of basis functions and its derivatives 
        /// </summary>
        Expression S, Su, Sv, Suu, Suv, Svv;

        /// <summary>
        /// 3D control points of the surface
        /// </summary>
        Point3D[,] mesh;

        int currentSet;

        Dictionary<string, FloatingPoint> values;

        /// <inheritdoc/>
        public override TriangleMesh[] GenerateData()
        {
            TriangleMesh[] result = new TriangleMesh[DatasetCount];
            Progress = 0;
            
            for (currentSet = 0; currentSet < DatasetCount; currentSet++)
            {
                // init random patch
                initialize();

                // init function
                function.SetParams();

                // sample UV
                sampler.Sample(function);

                // triangulate UV
                triangularizator.Triangulate(sampler);
                TriangleMesh uvMesh = triangularizator.TriangleMesh;

                // for all UVs, get their corresponding NURBS vertices
                result[currentSet] = NurbsFromUV(uvMesh);
            }
            Progress = 100;
            return result;
        }

        /// <summary>
        /// takes a mesh, where X and Y coordinates correspond with U and V parameters, and produces corresponding NURBS patch
        /// </summary>
        /// <param name="uvMesh"></param>
        /// <returns></returns>
        private TriangleMesh NurbsFromUV(TriangleMesh uvMesh)
        {
            TriangleMesh result = new TriangleMesh();
            // the triangles remain the same
            result.Triangles = uvMesh.Triangles;


            result.ExactK1 = new double[uvMesh.Points.Length];
            result.ExactK2 = new double[uvMesh.Points.Length];
            result.Points = new Point3D[uvMesh.Points.Length];
            result.Normals = new Point3D[uvMesh.Points.Length];
            for (int i = 0; i < uvMesh.Points.Length; i++)
            {
                Eval(uvMesh.Points[i].X, uvMesh.Points[i].Y, result, i);
                Progress = (byte)(100 * i / uvMesh.Points.Length);
            }
            return result;
        }

        

        /// <inheritdoc/>
        public override void SetConfig(XmlNode configNode)
        {
            DatasetCount = 1;
            base.SetConfig(configNode);

            // read configuration for sampler and triangularizator, will be used in UV
            foreach (XmlNode node in configNode.ChildNodes)
            {
                switch (node.Name)
                {
                    case "Sampler":
                        sampler = TestScenario.GetModule<IExplicitSampler>(node.ChildNodes[0]);
                        sampler.ProgressChanged += sampler_ProgressChanged;
                        break;
                    case "Triangularizator":
                        triangularizator = TestScenario.GetModule<IExplicitTriangularizator>(node.ChildNodes[0]);
                        triangularizator.ProgressChanged += triangularizator_ProgressChanged;
                        break;
                    default: break;
                }
            }

            // the function is actually not used, it is just a placeholder for sampling UV
            SymbolicFunction f = new SymbolicFunction();
            f.Function = "0";
            this.function = f;
        }

        void triangularizator_ProgressChanged(IModule module, NotifyProgressChangedEventArgs args)
        {
            Progress = (byte)(100 * currentSet / DatasetCount + (50 + args.Progress / 2) / DatasetCount);
        }

        void sampler_ProgressChanged(IModule module, NotifyProgressChangedEventArgs args)
        {
            Progress = (byte)(100 * currentSet / DatasetCount + (0 + args.Progress / 2) / DatasetCount);
        }

        
        /// <summary>
        /// initializes a random NURBS patch
        /// </summary>
        private void initialize()
        {
            mesh = new Point3D[4, 4];
            values = new Dictionary<string, FloatingPoint>();
            Expression[,] w = new Expression[4, 4];
            for (int i = 0; i < 4; i++)
                for (int j = 0; j < 4; j++)
                {
                    mesh[i, j] = new Point3D(i + Rand.NextDouble(), j + Rand.NextDouble(), Rand.NextDouble() * 4);
                    w[i, j] = Infix.ParseOrThrow(Rand.NextDouble().ToString());
                }

            // basis functions
            var N4u = Infix.ParseOrThrow("(u^3)/6");
            var N3u = Infix.ParseOrThrow("1/6+u/2+u*u/2-u*u*u/2");
            var N2u = Infix.ParseOrThrow("1/6+(1-u)/2+((1-u)^2)/2-((1-u)^3)/2");
            var N1u = Infix.ParseOrThrow("((1-u)^3)/6");

            var N4v = Infix.ParseOrThrow("(v^3)/6");
            var N3v = Infix.ParseOrThrow("1/6+v/2+v*v/2-v*v*v/2");
            var N2v = Infix.ParseOrThrow("1/6+(1-v)/2+((1-v)^2)/2-((1-v)^3)/2");
            var N1v = Infix.ParseOrThrow("((1-v)^3)/6");

            // sum of basis functions
            S =
                N1u * N1v * w[0, 0] +
                N1u * N2v * w[0, 1] +
                N1u * N3v * w[0, 2] +
                N1u * N4v * w[0, 3] +

                N2u * N1v * w[1, 0] +
                N2u * N2v * w[1, 1] +
                N2u * N3v * w[1, 2] +
                N2u * N4v * w[1, 3] +

                N3u * N1v * w[2, 0] +
                N3u * N2v * w[2, 1] +
                N3u * N3v * w[2, 2] +
                N3u * N4v * w[2, 3] +

                N4u * N1v * w[3, 0] +
                N4u * N2v * w[3, 1] +
                N4u * N3v * w[3, 2] +
                N4u * N4v * w[3, 3];

            // partial differentials of the sum of basis functions
            Su = Calculus.Differentiate(Expression.Symbol("u"), S);
            Sv = Calculus.Differentiate(Expression.Symbol("v"), S);
            Suu = Calculus.Differentiate(Expression.Symbol("u"), Su);
            Suv = Calculus.Differentiate(Expression.Symbol("v"), Su);
            Svv = Calculus.Differentiate(Expression.Symbol("v"), Sv);

            // 2- dismensional basis functions
            R = new Expression[4, 4];
            R[0, 0] = (N1u * N1v * w[0, 0]);
            R[0, 1] = (N1u * N2v * w[0, 1]);
            R[0, 2] = (N1u * N3v * w[0, 2]);
            R[0, 3] = (N1u * N4v * w[0, 3]);

            R[1, 0] = (N2u * N1v * w[1, 0]);
            R[1, 1] = (N2u * N2v * w[1, 1]);
            R[1, 2] = (N2u * N3v * w[1, 2]);
            R[1, 3] = (N2u * N4v * w[1, 3]);

            R[2, 0] = (N3u * N1v * w[2, 0]);
            R[2, 1] = (N3u * N2v * w[2, 1]);
            R[2, 2] = (N3u * N3v * w[2, 2]);
            R[2, 3] = (N3u * N4v * w[2, 3]);

            R[3, 0] = (N4u * N1v * w[3, 0]);
            R[3, 1] = (N4u * N2v * w[3, 1]);
            R[3, 2] = (N4u * N3v * w[3, 2]);
            R[3, 3] = (N4u * N4v * w[3, 3]);

            // partial derivatives of the 2-dimensional basis functions
            Ru = new Expression[4, 4];
            Rv = new Expression[4, 4];
            Ruu = new Expression[4, 4];
            Rvv = new Expression[4, 4];
            Ruv = new Expression[4, 4];
            for (int i = 0; i < 4; i++)
                for (int j = 0; j < 4; j++)
                {
                    Ru[i, j] = Calculus.Differentiate(Expression.Symbol("u"), R[i, j]);
                    Rv[i, j] = Calculus.Differentiate(Expression.Symbol("v"), R[i, j]);
                    Ruu[i, j] = Calculus.Differentiate(Expression.Symbol("u"), Ru[i, j]);
                    Rvv[i, j] = Calculus.Differentiate(Expression.Symbol("v"), Rv[i, j]);
                    Ruv[i, j] = Calculus.Differentiate(Expression.Symbol("v"), Ru[i, j]);

                    // this is probably not needed
                    Rational.Reduce(Ru[i, j]);
                    Rational.Reduce(Rv[i, j]);
                    Rational.Reduce(Ruu[i, j]);
                    Rational.Reduce(Ruv[i, j]);
                    Rational.Reduce(Rvv[i, j]);
                }
        }


        /// <summary>
        /// evaluates the NURBS at given u and v, together with normal and curvature, and stores to mesh as point with given index
        /// </summary>
        /// <param name="u">U-coordinate</param>
        /// <param name="v">V-coordinate</param>
        /// <param name="target">target mesh</param>
        /// <param name="index">target index</param>
        private void Eval(double u, double v, TriangleMesh target, int index)
        {
            values["u"] = u;
            values["v"] = v;

            // partial derivatives of the sum (denominator)
            double vS = Evaluate.Evaluate(values, S).RealValue;
            double vSu = Evaluate.Evaluate(values, Su).RealValue;
            double vSv = Evaluate.Evaluate(values, Sv).RealValue;
            double vSuu = Evaluate.Evaluate(values, Suu).RealValue;
            double vSuv = Evaluate.Evaluate(values, Suv).RealValue;
            double vSvv = Evaluate.Evaluate(values, Svv).RealValue;


            Point3D vertex = new Point3D();
            Point3D Tu = new Point3D();
            Point3D Tv = new Point3D();
            Point3D Tuu = new Point3D();
            Point3D Tvv = new Point3D();
            Point3D Tuv = new Point3D();

            // sum over all 2D basis functions
            for (int i = 0; i < 4; i++)
                for (int j = 0; j < 4; j++)
                {
                    // vertex position
                    double vR = Evaluate.Evaluate(values, R[i, j]).RealValue;
                    vertex += mesh[i, j] * ( vR / vS);

                    // u-tangent
                    double vRu = Evaluate.Evaluate(values, Ru[i, j]).RealValue;
                    Tu += mesh[i, j] * ((vRu*vS-vR*vSu)/(vS*vS));

                    // v-tangent
                    double vRv = Evaluate.Evaluate(values, Rv[i, j]).RealValue;
                    Tv += mesh[i, j] * ((vRv * vS - vR * vSv) / (vS * vS));

                    // second order partial derivatives. Will be needed for curvature
                    double vRuu = Evaluate.Evaluate(values, Ruu[i, j]).RealValue;
                    double vRuv = Evaluate.Evaluate(values, Ruv[i, j]).RealValue;
                    double vRvv = Evaluate.Evaluate(values, Rvv[i, j]).RealValue;
                    Tuu += mesh[i, j] * (-(vR * vSuu) / (vS * vS) + (2 * vR * vSu * vSu) / (vS * vS * vS) - (2 * vRu * vSu) / (vS * vS) + (vRuu / vS));
                    Tvv += mesh[i, j] * (-(vR * vSvv) / (vS * vS) + (2 * vR * vSv * vSv) / (vS * vS * vS) - (2 * vRv * vSv) / (vS * vS) + (vRvv / vS));
                    Tuv += mesh[i, j] * ((2 * vR * vSu * vSv) / (vS * vS * vS) - (vRu * vSv) / (vS * vS) - (vR * vSuv) / (vS * vS) - (vRv * vSu) / (vS * vS) + vRuv / vS);
                }

            // normal
            Point3D normal = Tv.CrossProduct(Tu);
            normal = normal.Normalize();

            // first fundamental form
            double E = Tu.DotProduct(Tu);
            double F = Tu.DotProduct(Tv);
            double G = Tv.DotProduct(Tv);

            // second fundamental form
            double e = Tuu.DotProduct(normal);
            double f = Tuv.DotProduct(normal);
            double g = Tvv.DotProduct(normal);

            double gaussian = (e * g - f * f) / (E * G - F * F);
            double mean = 0.5 * (e * G + g * E - 2 * f * F) / (E * G - F * F);

            double dif = Math.Sqrt(Math.Max(0, mean * mean - gaussian));

            double k1 = mean + dif;
            double k2 = mean - dif;
            
            target.Points[index] = vertex;
            target.ExactK1[index] = k1;
            target.ExactK2[index] = k2;
            target.Normals[index] = normal;
        }
    }
}
