﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Xml;
using Zcu.CurvatureBenchmark.Framework;
using MathNet.Numerics.LinearAlgebra;
using MathNet.Numerics.LinearAlgebra.Double;
using MathNet.Numerics.LinearAlgebra.Factorization;

namespace Zcu.CurvatureBenchmark.Distorters
{
    public class FrankensteinForNormals : AbstractDistorter
    {
        //vertex neighborhood which is used to compute curvature estimation
        public int neighborhood { get; set; }

        //number of iterations in optimization 
        public int iterations { get; set; }

        public override TriangleMesh Distort(TriangleMesh mesh)
        {
            int n = mesh.Points.Length;
            Point3D[] newNormals = new Point3D[n];
            if (neighborhood == 0)
            {
                neighborhood = 1;
            }
            if (iterations == 0)
            {
                iterations = 10;
            }
            CornerTable t = new CornerTable(mesh);
            for (int i = 0; i < n; i++)
            {
                newNormals[i] = QuadraticSurfaceApproximate(mesh, i, t);
                if ((i % 1000) == 0)
                {
                    Progress = (byte)(100 * i / mesh.Points.Length);
                }
            }
            for (int i = 0; i < newNormals.Length; i++)
            {
                newNormals[i] = newNormals[i].Normalize();
            }
            TriangleMesh result = mesh.ShallowCopy();
            result.Normals = newNormals;
            return result;

        }
        /// <summary>
        /// Create and optimizes quadratic surface and estimate curvature
        /// </summary>
        /// <param name="mesh"></param>
        /// <param name="pointIndex">point index</param>
        /// <param name="t"></param>
        public Point3D QuadraticSurfaceApproximate(TriangleMesh mesh, int pointIndex, CornerTable t)
        {
            Point3D normal = mesh.Normals[pointIndex];
            try
            {
                int[] incidentVertices = t.VVK(pointIndex, neighborhood);
                Point3D[] xi = new Point3D[incidentVertices.Length]; //points in lokal coordinates

                Matrix<double> U = DenseMatrix.Create(incidentVertices.Length, 3, 0); // (nx3) matrix 

                Vector<double> x = DenseVector.Create(3, 0); //solve of linear system
                Vector<double> x2 = DenseVector.Create(3, 0); //solve of linear system
                Vector<double> d = DenseVector.Create(incidentVertices.Length, 0); // (n)-vector
                Plane tangentPlane = new Plane(ref normal, ref mesh.Points[pointIndex]);
                Point3D projectionVj = tangentPlane.ProjectPoint(ref mesh.Points[incidentVertices[1]]);

                //new lokal basis
                Point3D[] newBase = new Point3D[3];
                newBase[0] = normal / normal.Abs();
                newBase[1] = (projectionVj - mesh.Points[pointIndex]) / (projectionVj - mesh.Points[pointIndex]).Abs();
                newBase[2] = Point3D.CrossProduct(newBase[1], newBase[0]);
                newBase[2] = newBase[2] / newBase[2].Abs();

                Matrix<double> transformation = Matrix<double>.Build.Dense(3, 3);
                transformation[0, 0] = newBase[1].X;
                transformation[0, 1] = newBase[2].X;
                transformation[0, 2] = newBase[0].X;
                transformation[1, 0] = newBase[1].Y;
                transformation[1, 1] = newBase[2].Y;
                transformation[1, 2] = newBase[0].Y;
                transformation[2, 0] = newBase[1].Z;
                transformation[2, 1] = newBase[2].Z;
                transformation[2, 2] = newBase[0].Z;

                //inverse of transforming matrix
                Matrix<double> mInverse = transformation.Inverse();
                for (int i = 1; i < incidentVertices.Length; i++)
                {
                    //compute the lokal coordinates of each vertex incident to Vi
                    xi[i].X = mInverse[0, 0] * (mesh.Points[incidentVertices[i]].X - mesh.Points[pointIndex].X) +
                       mInverse[0, 1] * (mesh.Points[incidentVertices[i]].Y - mesh.Points[pointIndex].Y) +
                       mInverse[0, 2] * (mesh.Points[incidentVertices[i]].Z - mesh.Points[pointIndex].Z);
                    xi[i].Y = mInverse[1, 0] * (mesh.Points[incidentVertices[i]].X - mesh.Points[pointIndex].X) +
                        mInverse[1, 1] * (mesh.Points[incidentVertices[i]].Y - mesh.Points[pointIndex].Y) +
                        mInverse[1, 2] * (mesh.Points[incidentVertices[i]].Z - mesh.Points[pointIndex].Z);
                    xi[i].Z = mInverse[2, 0] * (mesh.Points[incidentVertices[i]].X - mesh.Points[pointIndex].X) +
                        mInverse[2, 1] * (mesh.Points[incidentVertices[i]].Y - mesh.Points[pointIndex].Y) +
                        mInverse[2, 2] * (mesh.Points[incidentVertices[i]].Z - mesh.Points[pointIndex].Z);
                }

                int it = 0;
                double newE = 0;
                double previousE = Double.MaxValue;
                double newESquared = 0;
                double previousESquared = Double.MaxValue;
                while (it < iterations)
                {
                    for (int i = 1; i < incidentVertices.Length; i++)
                    {
                        //create linear system of equations
                        U[i, 0] = xi[i].X * xi[i].X;
                        U[i, 1] = xi[i].X * xi[i].Y;
                        U[i, 2] = xi[i].Y * xi[i].Y;
                        d[i] = xi[i].Z;
                    }
                    //solve linear system of equations
                    x2 = (U.Transpose() * U).Inverse() * U.Transpose() * d;
                    //points on quadratic surface
                    Point3D[] xiSurface = new Point3D[incidentVertices.Length];
                    ComputePointsOnSurface(xiSurface, xi, x2);
                    FindTheNearestPointOnSurface(xiSurface, xi, x2);

                    newE = ComputeE(xi, xiSurface);
                    newESquared = ComputeESquared(xi, xiSurface);

                    //stopping condition 
                    if (Math.Abs(newESquared) > Math.Abs(previousESquared))
                    {
                        break;
                    }
                    x = x2;
                    previousE = newE;
                    previousESquared = newESquared;


                    Matrix<double> M = DenseMatrix.Create(incidentVertices.Length, incidentVertices.Length, 0);
                    //matrix of xi - points in lokal coordinates
                    Matrix A = DenseMatrix.Create(incidentVertices.Length, 3, 0);
                    //matrix of xParaboloid - nearest points on Paraboloid
                    Matrix B = DenseMatrix.Create(incidentVertices.Length, 3, 0);
                    for (int l = 1; l < incidentVertices.Length; l++)
                    {
                        A[l, 0] = xi[l].X;
                        A[l, 1] = xi[l].Y;
                        A[l, 2] = xi[l].Z;
                        B[l, 0] = xiSurface[l].X;
                        B[l, 1] = xiSurface[l].Y;
                        B[l, 2] = xiSurface[l].Z;
                    }
                    M = B.Transpose() * A;
                    // orthogonal procrustes problem and solve matrices U and V
                    Svd<double> svd = M.Svd();
                    //rotate matrix
                    Matrix<double> R = svd.VT.Transpose() * svd.U.Transpose();
                    //rotate the lokal system
                    Matrix<double> AR = A * R;
                    for (int l = 1; l < incidentVertices.Length; l++)
                    {
                        xi[l].X = AR[l, 0];
                        xi[l].Y = AR[l, 1];
                        xi[l].Z = AR[l, 2];

                    }
                    Vector<double> normalPom = DenseVector.Create(3, 0);
                    normalPom[0] = normal.X;
                    normalPom[1] = normal.Y;
                    normalPom[2] = normal.Z;
                    normalPom = normalPom * R;
                    normal.X = normalPom[0];
                    normal.Y = normalPom[1];
                    normal.Z = normalPom[2];
                    it++;
                }
            }
            catch (Exception e)
            {
                normal = mesh.Normals[pointIndex];
            }
            return normal;
        }
        /// <summary>
        /// Return the nearest points on quadratic surface corresponding to xi
        /// </summary>
        /// <param name="xiSurface">points on quadratic surface</param>
        /// <param name="xi">points in lokal coordinate system.</param>
        /// <param name="x">coeficient of quadratic surface</param>
        public void FindTheNearestPointOnSurface(Point3D[] xiSurface, Point3D[] xi, Vector<double> x)
        {
            for (int i = 1; i < xi.Length; i++)
            {
                Point3D D = xi[i] - xiSurface[i];
                Point3D newPoint = new Point3D(0, 0, 0);
                Point3D previousPoint = new Point3D(0, 0, 0);
                //gradient of quadratic surface
                Point3D fugrad = new Point3D(1, 0, 2 * xiSurface[i].X * x[0] + x[1] * xiSurface[i].Y);
                Point3D fvgrad = new Point3D(0, 1, 2 * xiSurface[i].Y * x[2] + x[1] * xiSurface[i].X);
                double du = Point3D.DotProduct(fugrad, D);
                double dv = Point3D.DotProduct(fvgrad, D);
                int it = 0;
                double k = 1;
                previousPoint = xiSurface[i];
                newPoint.X = xiSurface[i].X + du * k;
                newPoint.Y = xiSurface[i].Y + dv * k;
                newPoint.Z = x[0] * newPoint.X * newPoint.X + x[2] * newPoint.Y * newPoint.Y + x[1] * newPoint.X * newPoint.Y;

                while (!((previousPoint - newPoint).Abs() == 0) && it < 50)
                {
                    while ((xi[i] - newPoint).Abs() > D.Abs())
                    {
                        k = k / 2;
                        newPoint.X = xiSurface[i].X + du * k;
                        newPoint.Y = xiSurface[i].Y + dv * k;
                        newPoint.Z = x[0] * newPoint.X * newPoint.X + x[2] * newPoint.Y * newPoint.Y + x[1] * newPoint.X * newPoint.Y;

                    }
                    previousPoint = xiSurface[i];
                    xiSurface[i].X = xiSurface[i].X + du * k;
                    xiSurface[i].Y = xiSurface[i].Y + dv * k;
                    xiSurface[i].Z = x[0] * xiSurface[i].X * xiSurface[i].X + x[2] * xiSurface[i].Y * xiSurface[i].Y + x[1] * xiSurface[i].X * xiSurface[i].Y;
                    newPoint = xiSurface[i];
                    D = xi[i] - xiSurface[i];
                    fugrad = new Point3D(1, 0, 2 * xiSurface[i].X * x[0] + x[1] * xiSurface[i].Y);
                    fvgrad = new Point3D(0, 1, 2 * xiSurface[i].Y * x[2] + x[1] * xiSurface[i].X);
                    du = Point3D.DotProduct(fugrad, D);
                    dv = Point3D.DotProduct(fvgrad, D);
                    it++;
                }
            }
        }
        /// <summary>
        /// Compute points on quadratic surface corresponding to xi
        /// </summary>
        /// <param name="xiSurface">points on quadratic surface</param>
        /// <param name="xi">points in lokal coordinate system.</param>
        /// <param name="x">coeficient of quadratic surface</param>
        public void ComputePointsOnSurface(Point3D[] xiSurface, Point3D[] xi, Vector<double> x)
        {
            for (int i = 1; i < xiSurface.Length; i++)
            {
                xiSurface[i].X = xi[i].X;
                xiSurface[i].Y = xi[i].Y;
                xiSurface[i].Z = x[0] * xi[i].X * xi[i].X + x[2] * xi[i].Y * xi[i].Y + x[1] * xi[i].X * xi[i].Y;
            }
        }
        /// <summary>
        /// Compute sum of distance xi from xiSurface
        /// </summary>
        /// <param name="xiSurface">points on quadratic surface</param>
        /// <param name="xi">points in lokal coordinate system.</param>
        /// <returns></returns>
        public double ComputeE(Point3D[] xi, Point3D[] xiSurface)
        {
            double E = 0;
            for (int i = 0; i < xi.Length; i++)
            {
                E += (xi[i] - xiSurface[i]).Abs();
            }
            return E;
        }
        /// <summary>
        /// Compute sum of squared distance xi from xiSurface
        /// </summary>
        /// <param name="xiSurface">points on quadratic surface</param>
        /// <param name="xi">points in lokal coordinate system.</param>
        /// <returns></returns>
        public double ComputeESquared(Point3D[] xi, Point3D[] xiSurface)
        {
            double E = 0;
            for (int i = 1; i < xi.Length; i++)
            {
                E += (xi[i] - xiSurface[i]).Abs() * (xi[i] - xiSurface[i]).Abs();
            }
            return E;
        }
        public override string ToString()
        {
            return base.ToString() + string.Format("(neighborhood {0})", neighborhood);
        }
    }
}
