﻿using MathNet.Numerics.LinearAlgebra;
using MathNet.Numerics.LinearAlgebra.Double;
using MathNet.Numerics.LinearAlgebra.Factorization;
using System;
using System.Xml;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Zcu.CurvatureBenchmark.Framework;
using Zcu.CurvatureBenchmark.Framework.Estimators;
using Zcu.CurvatureBenchmark.Framework.Distorters;

using System.Numerics;

namespace Zcu.CurvatureBenchmark.Estimators
{
    /// <summary>
    /// Implementation of Cohen-Stainer and Morvan curvature estimator<cite>csm03</cite>.
    /// </summary>
    /// <remarks>
    /// The method is able to estimate normals, it is also possible to use the module as normal distorter.
    /// </remarks>
    public class CSMEstimator : AbstractEstimator, IDistorter
    {
        /// <summary>
        /// Gets/sets euclidean range over which the curvature is estimated.
        /// </summary>
        /// <value>
        /// Default value is 0.
        /// </value>
        public double Range { get; set; } = 0;

        /// <summary>
        /// Gets/sets the distance over which the curvature is estimated is set as the average edge length multiplied 
        /// by the RangeKEdge parameter. This value is used if Range is 0.
        /// </summary>
        /// <value>
        /// Default value is 6.0;
        /// </value>
        public double RangeKEdge { get; set; } = 6;

        CornerTable ct;
        TriangleMesh mesh;

        // Euclidean range over which the curvature is estimated 
        // - computed based on Range or RangeKEdge (whichever is set in the input XML)
        double r;   

        Plane[] planes;
        Matrix<double>[] normalizedEdgeTensors; 
        Point3D[] normalizedEdge;
        double[] angles;    // angle over the edge defined by the corner
        double[] edgeLengths;
        double[] edgeLengthsCut;
        Point3D[] normals;

        double[] k1, k2;

        /// <inheritdoc/>
        override public double[] K1
        {
            get { return k1; }
        }

        /// <inheritdoc/>
        override public double[] K2
        {
            get { return k2; }
        }

        /// <summary>
        /// Calculates the plane for each mesh face
        /// </summary>
        private void InitPlanes()
        {
            planes = new Plane[mesh.Triangles.Length];

            for (int i = 0; i < planes.Length; i++)
            {
                Triangle t = mesh.Triangles[i];
                planes[i] = new Plane(ref mesh.Points[t.V1], ref mesh.Points[t.V2], ref mesh.Points[t.V3]);
            }
        }

        /// <summary>
        /// Calculates the angles between the neighboring faces over a common edges.
        /// The edge is defined by the corresponding corner in the CornerTable data structure.
        /// </summary>
        private void InitAngles()
        {
            angles = new double[mesh.Triangles.Length * 3];

            for (int i = 0; i < angles.Length; i++)
            {
                if (ct.opposite[i] == -1)
                {
                    // for an open edge the angle is set to 0
                    angles[i] = 0;
                    continue;
                }

                int t1 = ct.t(i);
                int t2 = ct.t(ct.opposite[i]);


                Point3D n1 = planes[t1].Normal;
                Point3D n2 = planes[t2].Normal;
                double n1n2 = (n1.CrossProduct(n2)).DotProduct(normalizedEdge[i]);
                n1n2 = Math.Max(Math.Min(1.0, n1n2), -1.0);
                double cos = Math.Max(Math.Min(1.0, n1.DotProduct(n2)), -1.0);
                double beta = Math.Atan2(n1n2, cos);
                angles[i] = -beta;
            }
        }

        /// <summary>
        /// Inits necessary edge properties
        /// </summary>
        private void InitEdges()
        {
            edgeLengths = new double[mesh.Triangles.Length * 3];
            edgeLengthsCut = new double[mesh.Triangles.Length * 3];
            normalizedEdgeTensors = new Matrix<double>[mesh.Triangles.Length * 3];
            normalizedEdge = new Point3D[mesh.Triangles.Length * 3];

            for (int i = 0; i < edgeLengths.Length; i++)
            {
                Point3D edge = mesh.Points[ct.v(ct.n(i))] - mesh.Points[ct.v(ct.p(i))];
                edgeLengths[i] = edge.Abs();
                if (edgeLengths[i] != 0) edge /= edgeLengths[i];

                Matrix<double> ev = DenseMatrix.Create(3, 1, 1);
                ev[0, 0] = edge.X;
                ev[1, 0] = edge.Y;
                ev[2, 0] = edge.Z;

                normalizedEdgeTensors[i] = ev.Multiply(ev.Transpose());
                normalizedEdge[i] = edge;
            }
        }

        private void Init()
        {
            ct = new CornerTable(mesh);
            k1 = new double[mesh.Points.Length];
            k2 = new double[mesh.Points.Length];
            normals = new Point3D[mesh.Normals.Length];

            InitPlanes();
            InitEdges();
            InitAngles();

            // init range over which the curvature is estimated
            if (Range != 0) r = Range;
            else r = RangeKEdge * mesh.AverageEL();
        }

        /// <summary>
        /// Cuts edges that go beyond the specified range.
        /// </summary>
        /// <param name="edges">Edges located in range from the point p</param>
        /// <param name="p">Index of the vertex for which the curvature is calculated</param>
        private void CutEdges(Dictionary<int, bool> edges, int p)
        {
            Point3D c = mesh.Points[p];

            foreach (KeyValuePair<int, bool> e in edges)
            {
                int key = e.Key;
                int op = ct.opposite[key];

                // if the edge lies completely inside the range, no need to cut it
                if (e.Value == false)
                {
                    edgeLengthsCut[key] = edgeLengths[key];
                    if (op != -1) edgeLengthsCut[op] = edgeLengths[key];
                    continue;
                }

                Point3D start = mesh.Points[ct.v(ct.p(key))];
                Point3D end = mesh.Points[ct.v(ct.n(key))];

                // check which edge point lies inside the range
                if (c.DistanceTo2(ref start) > c.DistanceTo2(ref end))
                {
                    Point3D temp = start;
                    start = end;
                    end = temp;
                }

                Point3D l = end - start;
                Point3D oc = start - c;

                double len = l.Abs();
                if (len != 0) l /= len;

                // calculate the intersection between the line and a sphere
                double loc = l.DotProduct(oc);
                double sqrt = loc * loc - oc.Abs2() + r * r;

                if (sqrt < 0)
                {
                    // if no intersection is found, the edge lenght is set to 0
                    // this should not happen and implies a possible bug
                    this.Warning("CSMEstimator::CutEdges edge does not intersect the sphere");
                    edgeLengthsCut[key] = 0; 
                    if (op != -1) edgeLengthsCut[op] = 0;
                    continue;
                }

                sqrt = Math.Sqrt(sqrt);
                double t1 = -loc + sqrt;
                double t2 = -loc - sqrt;

                double mD = 2 * start.DotProduct(l) - 2 * l.DotProduct(c);
                mD *= mD;
                mD -= 4 * l.DotProduct(l) * (c.DotProduct(c) - r * r - 2 * start.DotProduct(c) - start.DotProduct(start));

                double mt = -2 * start.DotProduct(l) + 2 * l.DotProduct(c) + Math.Sqrt(mD);
                mt /= 2 * l.DotProduct(l);

                if (t1 > len)
                    this.Warning("CSMEstimator::Intersection outside edge");

                // the parameter with the possitive value corresponds to the intersection point
                // the value of the parameter equals to the cut edge length 
                if (t1 > 0)
                {
                    if (t1 > edgeLengths[key])
                        this.Warning("CSMEstimator::Intersection outside edge");
                    edgeLengthsCut[key] = t1;
                    if (op != -1) edgeLengthsCut[op] = t1;
                }
                else
                {
                    edgeLengthsCut[key] = t2;
                    if (op != -1) edgeLengthsCut[op] = t2;
                }
            }
        }

        /// <summary>
        /// Calculates the area of the triangles within the range.
        /// </summary>
        /// <param name="visitedTriangles">The set of faces located within the range from the point p</param>
        /// <param name="distances">Contains the distance to point p for each vertex within the range</param>
        /// <param name="p">Index of the vertex for which the curvature is calculated</param>
        /// <returns>The area of the triangles within the range</returns>
        private double CalculateArea(HashSet<int> visitedTriangles, Dictionary<int, double> distances, int p)
        {
            double area = 0;
            Point3D c = mesh.Points[p];
            double r2 = r * r;
            if (!distances.ContainsKey(p)) distances.Add(p, 0);

            foreach (int triangle in visitedTriangles)
            {
                int i1 = mesh.Triangles[triangle].V1;
                int i2 = mesh.Triangles[triangle].V2;
                int i3 = mesh.Triangles[triangle].V3;

                if (!distances.ContainsKey(i1))
                    distances.Add(i1, c.SquareDistanceTo(mesh.Points[i1]));

                if (!distances.ContainsKey(i2))
                    distances.Add(i2, c.SquareDistanceTo(mesh.Points[i2]));

                if (!distances.ContainsKey(i3))
                    distances.Add(i3, c.SquareDistanceTo(mesh.Points[i3]));

                Point3D p1 = mesh.Points[i1];
                Point3D p2 = mesh.Points[i2];
                Point3D p3 = mesh.Points[i3];

                double d1, d2, d3;
                d1 = distances[i1];
                d2 = distances[i2];
                d3 = distances[i3];

                // count how many vertices of the faces are located outside the range
                int outCnt = 0;
                if (d1 > r2) outCnt++;
                if (d2 > r2) outCnt++;
                if (d3 > r2) outCnt++;

                // all points inside => calculate the area of the triangle
                if (outCnt == 0)
                {
                    Point3D v1 = p2 - p1;
                    Point3D v2 = p3 - p1;
                    area += (v1.CrossProduct(v2)).Length();
                }
                // two points outside => calculate the area of a smaller triangle defined by the two edges 
                // intersecting at the single interior vertex
                else if (outCnt == 2)
                {
                    if (d1 < r2)
                    {
                        Point3D v1 = (p2 - p1).Normalize() * edgeLengthsCut[triangle * 3 + 2];
                        Point3D v2 = (p3 - p1).Normalize() * edgeLengthsCut[triangle * 3 + 1];
                        area += (v1.CrossProduct(v2)).Length();
                    }
                    if (d2 < r2) 
                    {
                        Point3D v1 = (p1 - p2).Normalize() * edgeLengthsCut[triangle * 3 + 2];
                        Point3D v2 = (p3 - p2).Normalize() * edgeLengthsCut[triangle * 3];
                        area += (v1.CrossProduct(v2)).Length();
                    }
                    if (d3 < r2) 
                    {
                        Point3D v1 = (p1 - p3).Normalize() * edgeLengthsCut[triangle * 3 + 1];
                        Point3D v2 = (p2 - p3).Normalize() * edgeLengthsCut[triangle * 3];
                        area += (v1.CrossProduct(v2)).Length();
                    }
                }
                // one point outside => the area can be calculated as the area of the original triangle
                // minus the area of a smaller triangle defined by the two edges 
                // intersecting at the single outerior vertex
                else if (outCnt == 1)
                {
                    if (d1 > r2)
                    {
                        Point3D v1 = p2 - p1;
                        Point3D v2 = p3 - p1;
                        double areaTriangle = (v1.CrossProduct(v2)).Length();

                        double l1 = v1.Length() - edgeLengthsCut[triangle * 3 + 2];
                        double l2 = v2.Length() - edgeLengthsCut[triangle * 3 + 1];
                        Point3D v3 = v1.Normalize() * l1;
                        Point3D v4 = v2.Normalize() * l2;
                        area += areaTriangle - (v3.CrossProduct(v4)).Length();
                    }
                    if (d2 > r2)
                    {
                        Point3D v1 = p1 - p2;
                        Point3D v2 = p3 - p2;
                        double areaTriangle = (v1.CrossProduct(v2)).Length();

                        double l1 = v1.Length() - edgeLengthsCut[triangle * 3 + 2];
                        double l2 = v2.Length() - edgeLengthsCut[triangle * 3];
                        Point3D v3 = v1.Normalize() * l1;
                        Point3D v4 = v2.Normalize() * l2;
                        area += areaTriangle - (v3.CrossProduct(v4)).Length();
                    }
                    if (d3 > r2)
                    {
                        Point3D v1 = p1 - p3;
                        Point3D v2 = p2 - p3;
                        double areaTriangle = (v1.CrossProduct(v2)).Length();

                        double l1 = v1.Length() - edgeLengthsCut[triangle * 3 + 1];
                        double l2 = v2.Length() - edgeLengthsCut[triangle * 3];
                        Point3D v3 = v1.Normalize() * l1;
                        Point3D v4 = v2.Normalize() * l2;
                        area += areaTriangle - (v3.CrossProduct(v4)).Length();
                    }
                }
            }

            // all the areas are calculated using the cross product of two vectors
            // to get the correct area, we have to divide by 2
            area /= 2;
            return area;
        }

        /// <inheritdoc/>
        public override string ToString()
        {
            if (Range != 0) return base.ToString() + string.Format("(Range {0})", Range);
            else return base.ToString() + string.Format("(RangeKEdge {0})", RangeKEdge);
        }

        private void CalculateCurvature(bool distortNormals)
        {
            Init();

            for (int i = 0; i < k1.Length; i++)
            {
                if (ct.IsOnBorder(i))
                    continue;

                HashSet<int> visitedTriangles = new HashSet<int>();
                Dictionary<int, double> distances = new Dictionary<int, double>();

                Dictionary<int, bool> edges = ct.VEE(i, r, visitedTriangles, distances);
                CutEdges(edges, i);
                Matrix<double> tensor = DenseMatrix.Create(3, 3, 0);

                // calculate the tensor
                foreach (KeyValuePair<int, bool> e in edges)
                {
                    double a = angles[e.Key];
                    double ee = edgeLengthsCut[e.Key];
                    Matrix<double> mm = normalizedEdgeTensors[e.Key];
                    tensor += a * ee * mm;
                }

                double area = CalculateArea(visitedTriangles, distances, i);
                if (area == 0)
                {
                    k1[i] = 0;
                    k2[i] = 0;
                    continue;
                }
                tensor /= area;

                Evd<double> evd = tensor.Evd();
                Vector<Complex> eigenvalues = evd.EigenValues;

                double[] absValues = { Math.Abs(eigenvalues[0].Real), Math.Abs(eigenvalues[1].Real), Math.Abs(eigenvalues[2].Real) };

                if (!distortNormals)
                {
                    // the principal curvatures are the two bigger eigenvalues from the tensor
                    k2[i] = Math.Max(Math.Min(absValues[0], absValues[1]), Math.Min(Math.Max(absValues[0], absValues[1]), absValues[2])); ;
                    k1[i] = Math.Max(absValues[0], Math.Max(absValues[1], absValues[2]));

                    // restore the sign
                    for (int j = 0; j < 3; j++)
                    {
                        if (k1[i] == Math.Abs(eigenvalues[j].Real))
                        {
                            k1[i] *= Math.Sign(eigenvalues[j].Real);
                            break;
                        }
                    }

                    // restore the sign
                    for (int j = 0; j < 3; j++)
                    {
                        if (k2[i] == Math.Abs(eigenvalues[j].Real))
                        {
                            k2[i] *= Math.Sign(eigenvalues[j].Real);
                            break;
                        }
                    }

                    // k1 should be the larger value
                    if (k1[i] < k2[i])
                    {
                        double tmp = k1[i];
                        k1[i] = k2[i];
                        k2[i] = tmp;
                    }
                }
                else
                {
                    // calculate normal at vertex from min eigenvalue
                    double min = Math.Min(absValues[0], Math.Min(absValues[1], absValues[2]));

                    for (int j = 0; j < 3; j++)
                    {
                        if (min == absValues[j])
                        {
                            Matrix<double> eigenvectors = evd.EigenVectors;
                            Vector<double> n = eigenvectors.Column(j);
                            normals[i] = new Point3D(n.ElementAt(0), n.ElementAt(1), n.ElementAt(2));
                            normals[i].Normalize();

                            if (normals[i].DotProduct(mesh.Normals[i]) < 0)
                            {
                                normals[i] *= -1;
                            }

                            break;
                        }
                    }
                }

                if ((i % 100) == 0)
                    Progress = (byte)(100 * i / k1.Length);
            }

        }

        /// <inheritdoc/>
        override public void EstimateCurvatures(TriangleMesh mesh)
        {
            this.mesh = mesh;
            CalculateCurvature(false);
        }

        /// <inheritdoc/>
        public TriangleMesh Distort(TriangleMesh input)
        {
            this.mesh = input.ShallowCopy();
            CalculateCurvature(true);
            mesh.Normals = normals;

            return mesh;
        }
    }
}
