﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Zcu.CurvatureBenchmark.Framework;
using Zcu.CurvatureBenchmark.Framework.Estimators;
using Zcu.CurvatureBenchmark.Framework.Distorters;


namespace Zcu.CurvatureBenchmark.Estimators
{
    /// <summary>
    /// Control points for one edge.
    /// </summary>
    internal struct EdgePatch
    {
        public Point3D b300 { get; set; }
        public Point3D bp1_210 { get; set; }
        public Point3D bp2_120 { get; set; }
        public Point3D b030 { get; set; }
        public Point3D bp1_201 { get; set; }
        public Point3D bp1_111 { get; set; }
        public Point3D bp1_021 { get; set; }
        public Point3D bp2_201 { get; set; }
        public Point3D bp2_111 { get; set; }
        public Point3D bp2_021 { get; set; }

    }

    /// <summary>
    /// Implementation of G-1 continuous surface patches by Fünfzig et al. <cite>fmhf08</cite>.
    /// </summary>
    public class FunfzigEstimator : AbstractEstimator /*, IDistorter */
    {
        CornerTable ct;
        TriangleMesh mesh;

        Plane[] planes;  // tangent planes
        EdgePatch[] edgePatches;
        PNG1Patch_gen trianglePatch;

        double[] k1, k2;

        /// <inheritdoc/>
        override public double[] K1
        {
            get { return k1; }
        }

        /// <inheritdoc/>
        override public double[] K2
        {
            get { return k2; }
        }

        /// <summary>
        /// Calculates the tangent plane for each mesh vertex
        /// </summary>
        private void InitPlanes()
        {
            planes = new Plane[mesh.Points.Length];

            for (int i = 0; i < planes.Length; i++)
            {
                planes[i] = new Plane(ref mesh.Normals[i], ref mesh.Points[i]);
            }
        }

        /// <summary>
        /// Transforms a point to the tangent plane. Corresponds to Figure 6 in the article
        /// "PNG1 triangles for tangent plane continuous surfaces on the GPU".
        /// The points are labeled according to the nomeclature used in the paper.
        /// </summary>
        /// <param name="planeNormal">Normal of the tangent plane</param>
        /// <returns></returns>
        Point3D TransformToPlane(Point3D p1, Point3D p210, Point3D p201,
            Point3D p2, Point3D p120, Point3D planeNormal, bool flip)
        {
            // compute angle in the source plane
            double angleP1 = Point3D.GetAngle(p210 - p1, p201 - p1);
            if (flip) angleP1 *= -1;

            // create rotation matrix around point p120
            Matrix4x4 inRot = Matrix4x4.Translation(-p120); //translate back
            inRot *= Matrix4x4.Rotation(planeNormal, angleP1); ///rotate
	        inRot *= Matrix4x4.Translation(p120); //translate to origin

            // rotate p2 using the rotation matrix
            Point3D temp = Point3D.TransformCoordinate(p2, inRot);

            Point3D p021 = temp - p120; // get direction of the vector
            p021.Normalize(); 
            return p120 + p021 * (p201 - p1).Length(); // get the point in the same distance as in the source plane
        }

        /// <summary>
        /// Calculates the control points for an edge following the approach
        /// from Section 3.2 of the paper "PNG1 triangles for tangent plane continuous surfaces on the GPU".
        /// </summary>
        /// <param name="i">The corner defining the edge</param>
        /// <returns>Control points for the edge</returns>
        private EdgePatch CalculateEdgePatch(int i)
        {
            EdgePatch patch = new EdgePatch();

            int pi1 = ct.v(ct.n(i));
            int pi2 = ct.v(ct.p(i));
            Point3D p1 = mesh.Points[pi1];
            Point3D p2 = mesh.Points[pi2];
            Point3D p3 = mesh.Points[ct.v(i)];
            Point3D dir = p2 - p1;

            patch.b300 = p1;
            patch.b030 = p2;
            
            // tangent planes
            Plane tangentPlaneP1 = planes[pi1];
            Plane tangentPlaneP2 = planes[pi2];
            // edge lengths
            double lenght1 = (p3 - p2).Length() / 3;
            double lenght2 = (p3 - p1).Length() / 3;
            double lenght3 = (p2 - p1).Length() / 3;
            // project points to corresponding tangent plane
            patch.bp1_210 = p1 + (tangentPlaneP1.ProjectPoint(p2) - p1).Normalize() * lenght3; 
            patch.bp2_120 = p2 + (tangentPlaneP2.ProjectPoint(p1) - p2).Normalize() * lenght3;
            patch.bp1_201 = p1 + (tangentPlaneP1.ProjectPoint(p3) - p1).Normalize() * lenght2;
            patch.bp2_021 = p2 + (tangentPlaneP2.ProjectPoint(p3) - p2).Normalize() * lenght1;

            // project points to opposite tangent plane
            patch.bp1_021 = TransformToPlane(p1, patch.bp1_210, patch.bp1_201, p2, patch.bp2_120, tangentPlaneP2.Normal, false);
            patch.bp2_201 = TransformToPlane(p2, patch.bp2_120, patch.bp2_021, p1, patch.bp1_210, tangentPlaneP1.Normal, true);

            Point3D p12;
            if (ct.opposite[i] != -1) p12 = mesh.Points[ct.v(ct.opposite[i])];
            else
            {
                // calculate auxiliary point on open edges
                Point3D middle = p1 + (p2 - p1) * 0.5;
                p12 = middle - (p3 - middle);
            }

            // calculate the middle "merged" plane
            Plane t1 = new Plane(ref p1, ref p2, ref p3);
            Plane t2 = new Plane(ref p2, ref p1, ref p12);
            Point3D e1 = patch.bp2_120 - patch.bp1_210;
            Point3D e2 = (t1.Normal + t2.Normal).CrossProduct(e1);
            Point3D n12 = e1.CrossProduct(e2).Normalize();

            // and transform the points to it
            patch.bp1_111 = TransformToPlane(p1, patch.bp1_210, patch.bp1_201, patch.bp2_120, patch.bp1_210, n12, false);
            patch.bp2_111 = TransformToPlane(p2, patch.bp2_120, patch.bp2_021, patch.bp1_210, patch.bp2_120, n12, true);

            return patch;
        }

        /// <summary>
        /// Calculates the control points for each edge
        /// </summary>
        private void InitEdgePatches()
        {
            edgePatches = new EdgePatch[mesh.Triangles.Length * 3];

            for (int i = 0; i < edgePatches.Length; i++)
            {
                edgePatches[i] = CalculateEdgePatch(i);
            }
        }

        private void Init()
        {
            ct = new CornerTable(mesh);

            InitPlanes();
            InitEdgePatches();

            trianglePatch = new PNG1Patch_gen(edgePatches);

            k1 = new double[mesh.Points.Length];
            k2 = new double[mesh.Points.Length];

        }

        /// <summary>
        /// Subdivides the mesh using the PNG1 patches.
        /// Each face is subdivided into 10 new faces,
        /// with vertices lying on the corresponding PNG1 patch.
        /// </summary>
        /// <param name="mesh">Mesh to be subdivided</param>
        private void SubdivideMesh(TriangleMesh mesh)
        {
            Point3D[] points = new Point3D[mesh.Triangles.Length * 10];
            Triangle[] triangles = new Triangle[mesh.Triangles.Length * 9];
            double t = 1.0 / 3.0;
            int cntTriangles = 0;
            int cntPoints = 0;

            // create the new triangles
            for (int i = 0; i < mesh.Triangles.Length; i++)
            {
                triangles[cntTriangles++] = new Triangle(cntPoints, cntPoints + 1, cntPoints + 4);
                triangles[cntTriangles++] = new Triangle(cntPoints + 1, cntPoints + 2, cntPoints + 5);
                triangles[cntTriangles++] = new Triangle(cntPoints + 2, cntPoints + 3, cntPoints + 6);
                triangles[cntTriangles++] = new Triangle(cntPoints + 1, cntPoints + 5, cntPoints + 4);
                triangles[cntTriangles++] = new Triangle(cntPoints + 2, cntPoints + 6, cntPoints + 5);
                triangles[cntTriangles++] = new Triangle(cntPoints + 4, cntPoints + 5, cntPoints + 7);
                triangles[cntTriangles++] = new Triangle(cntPoints + 5, cntPoints + 6, cntPoints + 8);
                triangles[cntTriangles++] = new Triangle(cntPoints + 5, cntPoints + 8, cntPoints + 7);
                triangles[cntTriangles++] = new Triangle(cntPoints + 7, cntPoints + 8, cntPoints + 9);

                points[cntPoints++] = mesh.Points[mesh.Triangles[i][0]];
                points[cntPoints++] = trianglePatch.EvaluatePoint(2 * t, t, 0, i);
                points[cntPoints++] = trianglePatch.EvaluatePoint(t, 2 * t, 0, i);
                points[cntPoints++] = mesh.Points[mesh.Triangles[i][1]];
                points[cntPoints++] = trianglePatch.EvaluatePoint(2 * t, 0, t, i);
                points[cntPoints++] = trianglePatch.EvaluatePoint(t, t, t, i);
                points[cntPoints++] = trianglePatch.EvaluatePoint(0, 2 * t, t, i);
                points[cntPoints++] = trianglePatch.EvaluatePoint(t, 0, 2 * t, i);
                points[cntPoints++] = trianglePatch.EvaluatePoint(0, t, 2 * t, i);
                points[cntPoints++] = mesh.Points[mesh.Triangles[i][2]];
            }

            // fill the normal and curvature fields with dummy values
            // the values are not correct, however, they are necessary 
            // for the visualizator to work correctly
            Point3D[] n = new Point3D[points.Length];
            k1 = new double[n.Length];
            k2 = new double[n.Length];
            for (int i = 0; i < n.Length; i++)
            {
                n[i] = new Point3D(0, 1, 0);
                k1[i] = 1;
                k2[i] = 1;
            }

            mesh.Points = points;
            mesh.Normals = n;
            mesh.Triangles = triangles;

            ObjSaver saver = new ObjSaver();
            saver.Execute(mesh);
        }

        /// <summary>
        /// Estimates the curvature of the mesh
        /// </summary>
        /// <param name="mesh">Input mesh</param>
        override public void EstimateCurvatures(TriangleMesh mesh)
        {
            this.mesh = mesh;

            Init();

            const int GAUS = 0;
            const int MEAN = 1;
            const int CURVATURES_COUNT = 2;

            int[] counts = new int[mesh.Points.Length * CURVATURES_COUNT];
            double[] vals = new double[mesh.Points.Length * CURVATURES_COUNT];

            for (int i = 0; i < mesh.Triangles.Length; i++)
            {
                Triangle tri = mesh.Triangles[i];

                double u = 0.98;
                double v = 0.01;

                double gauss = 0;
                double mean = 0;

                double wt = 1;

                trianglePatch.Evaluate(u, v, 1 - u - v, i, out mean, out gauss);
                vals[tri[0] * CURVATURES_COUNT + GAUS] += gauss * wt;
                counts[tri[0] * CURVATURES_COUNT + GAUS]++;
                vals[tri[0] * CURVATURES_COUNT + MEAN] += mean * wt;
                counts[tri[0] * CURVATURES_COUNT + MEAN]++;

                trianglePatch.Evaluate(1 - u - v, u, v, i, out mean, out gauss);
                vals[tri[1] * CURVATURES_COUNT + GAUS] += gauss * wt;
                counts[tri[1] * CURVATURES_COUNT + GAUS]++;
                vals[tri[1] * CURVATURES_COUNT + MEAN] += mean * wt;
                counts[tri[1] * CURVATURES_COUNT + MEAN]++;

                trianglePatch.Evaluate(v, 1 - u - v, u, i, out mean, out gauss);
                vals[tri[2] * CURVATURES_COUNT + GAUS] += gauss * wt;
                counts[tri[2] * CURVATURES_COUNT + GAUS]++;
                vals[tri[2] * CURVATURES_COUNT + MEAN] += mean * wt;
                counts[tri[2] * CURVATURES_COUNT + MEAN]++;

            }


            //Calculate curvature at vertex as average from all summed curvatures at the point
            // and clamp values to desired interval

            this.k1 = new double[mesh.Points.Length];
            this.k2 = new double[mesh.Points.Length];

            int index = 0;
            for (int i = 0; i < vals.Length; i += CURVATURES_COUNT)
            {
                for (int j = 0; j < CURVATURES_COUNT; j++)
                {
                    if (counts[i + j] != 0)
                    {
                        vals[i + j] /= counts[i + j];
                    }
                }

                double s = vals[i + MEAN] * vals[i + MEAN] - vals[i + GAUS];
                if (s < 0) s = 0;

                //why the "-" sign ???
                //without it, curvatures are "flipped" 

                this.k1[index] = -(vals[i + MEAN] - Math.Sqrt(s));
                this.k2[index] = -(vals[i + MEAN] + Math.Sqrt(s));


                index++;
            }
        }

        /// <summary>
        /// Calculates the normals using the PNG1 patches.
        /// </summary>
        /// <param name="input">Input mesh</param>
        /// <returns>Mesh with distorted normals</returns>
        public TriangleMesh Distort(TriangleMesh input)
        {
            this.mesh = input.ShallowCopy();

            Init();

            Point3D[] normals = new Point3D[mesh.Normals.Length];
            for (int i = 0; i < mesh.Normals.Length; i++)
            {
                normals[i] = new Point3D(0, 0, 0);
            }


            for (int i = 0; i < mesh.Triangles.Length; i++)
            {
                Triangle tri = mesh.Triangles[i];

                double u = 0.98;
                double v = 0.01;

                normals[tri[0]] += trianglePatch.EvaluateNormal(u, v, 1 - u - v, i);
                normals[tri[1]] += trianglePatch.EvaluateNormal(1 - u - v, u, v, i);
                normals[tri[2]] += trianglePatch.EvaluateNormal(v, 1 - u - v, u, i);
            }


            //Calculate normal at vertex as average from all adjacent triangles
            for (int i = 0; i < normals.Length; i++)
            {
                normals[i].Normalize();
            }

            mesh.Normals = normals;

            return mesh;
        }

        
    }
}
