﻿using System;

using Zcu.CurvatureBenchmark.Framework;
using Zcu.CurvatureBenchmark.Framework.Estimators;
using MathNet.Numerics.LinearAlgebra;
using MathNet.Numerics.LinearAlgebra.Double;
using MathNet.Numerics.LinearAlgebra.Factorization;

namespace Zcu.CurvatureBenchmark.Estimators
{
    /// <summary>
    /// Principal direction approximation methods enumeration
    /// </summary>
    public enum GoldfeatherInterranteEstimatorMethods
    {
        /// <summary>
        /// The Normal Curvature Approximation Method
        /// </summary>
        NormalCurvature = 1,
        /// <summary>
        /// The Quadratic Surface Approximation Method
        /// </summary>
        QuadraticSurface = 2,
        /// <summary>
        /// The Adjacent-Normal Cubic Approximation Method
        /// </summary>
        AdjacentNormalCubic = 3
    }

    /// <summary>
    /// Implementation of Cubic-order algorithm for approximating principal direction vectors by Goldfeather and Interrante <cite>gi04</cite>.
    /// </summary>
    public class GoldfeatherInterranteEstimator: AbstractEstimator
    {
        //estimation of curvature
        double[] k1, k2;


        /// <summary>
        /// Gets/sets method used for curvature estimating
        /// </summary>
        /// <value>Default value is <see cref="GoldfeatherInterranteEstimatorMethods.NormalCurvature"/>.</value>
        public GoldfeatherInterranteEstimatorMethods Method { get; set; }

        /// <summary>
        /// Gets/sets vertex neighborhood which is used to compute curvature estimation
        /// </summary>
        /// <value>
        /// Default value is 1.
        /// </value>
        public int Neighborhood { get; set; }

        /// <inheritdoc/>
        override public double[] K1
        {
            get { return k1; }
        }

        /// <inheritdoc/>
        override public double[] K2
        {
            get { return k2; }
        }

        /// <summary>
        /// Gets/sets reporting of method parameters
        /// </summary>
        internal bool Verbose
        {
            get
            {
                return verbose;
            }

            set
            {
                verbose = value;
            }
        }

        bool verbose = false;

        /// <summary>
        /// Initialization.
        /// </summary>
        public GoldfeatherInterranteEstimator()
        {
            Neighborhood = 1;
            Method = GoldfeatherInterranteEstimatorMethods.NormalCurvature;
            ToStringFormat = "$Name$ (method $Method$, neighborhood $Neighborhood$)";
        }

        /// <inheritdoc/>
        override public void EstimateCurvatures(TriangleMesh mesh)
        {
            Progress = 0;
            int n = mesh.Points.Length;
            if (Neighborhood == 0)
            {
                Neighborhood = 1;
            }
            if (verbose)
            {
                this.Info("Vertices: " + n);
                if (Method == GoldfeatherInterranteEstimatorMethods.NormalCurvature)
                {
                    this.Info("TheNormalCurvatureApproximation");
                }
                else if (Method == GoldfeatherInterranteEstimatorMethods.QuadraticSurface)
                {
                    this.Info("TheQuadraticSurfaceApproximation");
                }
                else if (Method == GoldfeatherInterranteEstimatorMethods.AdjacentNormalCubic)
                    this.Info("TheAdjacentNormalCubicApproximation");
            }
            CornerTable t = new CornerTable(mesh);
            double[] Curvature = new double[2];
            k1 = new double[mesh.Points.Length];
            k2 = new double[mesh.Points.Length];
            
            for (int i = 0; i < mesh.Points.Length; i++)
            {
                if (Method == GoldfeatherInterranteEstimatorMethods.QuadraticSurface)
                    Curvature = TheQuadraticSurfaceApproximation(i, t, mesh);
                else if (Method == GoldfeatherInterranteEstimatorMethods.AdjacentNormalCubic)
                    Curvature = TheAdjacentNormalCubicApproximation(i, t, mesh);
                else
                    Curvature = TheNormalCurvatureApproximation(i, t, mesh);
                if (Curvature[1] > (Curvature[0]))
                {
                    k1[i] = Curvature[1];
                    k2[i] = Curvature[0];
                }else
                {
                    k1[i] = Curvature[0];
                    k2[i] = Curvature[1];
                }
                if ((i % 1000) == 0)
                {
                    Progress = (byte)(100 * i / mesh.Points.Length);
                }
            }
        }

        private double[] TheNormalCurvatureApproximation(int pointIndex, CornerTable t, TriangleMesh mesh)
        {
            double[] Curvature = new double[2];
            int[] incidentVertices = t.VVK(pointIndex, Neighborhood);
            Point3D normal = mesh.Normals[pointIndex];
            Plane tangentPlane = new Plane(ref normal, ref mesh.Points[pointIndex]);
            Matrix<double> U = DenseMatrix.Create(incidentVertices.Length, 3, 0);// (nx3) matrix 
            Vector<double> x = DenseVector.Create(3, 0); //solve of linear system
            Vector<double> d = DenseVector.Create(incidentVertices.Length, 0); // (n)-vector

            //projection of point vj onto the plane to determine new baze
            Point3D projectionVj = tangentPlane.ProjectPoint(ref mesh.Points[incidentVertices[1]]);

            //new local basis
            Point3D[] newBase = new Point3D[3];
            newBase[0] = normal / normal.Abs();
            newBase[1] = (projectionVj - mesh.Points[pointIndex]) / (projectionVj - mesh.Points[pointIndex]).Abs();
            newBase[2] = Point3D.CrossProduct(newBase[1], newBase[0]);
            newBase[2] = newBase[2] / newBase[2].Abs();

            //transforming matrix
            Matrix<double> transformation = Matrix<double>.Build.Dense(3, 3); 
            transformation[0, 0] = newBase[1].X;
            transformation[0, 1] = newBase[2].X;
            transformation[0, 2] = newBase[0].X;
            transformation[1, 0] = newBase[1].Y;
            transformation[1, 1] = newBase[2].Y;
            transformation[1, 2] = newBase[0].Y;
            transformation[2, 0] = newBase[1].Z;
            transformation[2, 1] = newBase[2].Z;
            transformation[2, 2] = newBase[0].Z;
            Matrix<double> mInverse = transformation.Inverse(); //inverse of transforming matrix
            Point3D xi = new Point3D(0,0,0); //points in local coordinates

            
            for (int j = 1; j < incidentVertices.Length; j++)
            {
                //compute the local coordinates of each vertex incident to Vi
                Point3D y_j = mesh.Points[incidentVertices[j]] -mesh.Points[pointIndex];
                xi.X = mInverse[0, 0] * (y_j.X) +mInverse[0, 1] * (y_j.Y) +mInverse[0, 2] * (y_j.Z);
                xi.Y = mInverse[1, 0] * (y_j.X) + mInverse[1, 1] * (y_j.Y) + mInverse[1, 2] * (y_j.Z);
                xi.Z = mInverse[2, 0] * (y_j.X) +mInverse[2, 1] * (y_j.Y) +mInverse[2, 2] * (y_j.Z);
                xi = (xi) / (xi).Abs();

                //create a linear system of equations
                U[j, 0] = xi.X * xi.X;
                U[j, 1] = 2 *xi.X * xi.Y;
                U[j, 2] = xi.Y *xi.Y;
                d[j] = 2 * Point3D.DotProduct(mesh.Points[pointIndex] - mesh.Points[incidentVertices[j]], normal) /
                    Point3D.DotProduct(mesh.Points[pointIndex] - mesh.Points[incidentVertices[j]], mesh.Points[pointIndex] - mesh.Points[incidentVertices[j]]);


            }
            //solve linear system of equations
            x = (U.Transpose() * U).Inverse() * U.Transpose() * d;

            //Weingarten curvature matrix
            Matrix<double> W = DenseMatrix.Create(2, 2, 0);
            W[0, 0] = x[0];
            W[0, 1] = x[1];
            W[1, 0] = x[1];
            W[1, 1] = x[2];
            Evd<double> evd = W.Evd();

            //eigenvalues of Weingarten curvature matrix
            Vector<System.Numerics.Complex> eigenValues = evd.EigenValues;
            Curvature[0] = -(float)(eigenValues[0].Real);
            Curvature[1] = -(float)(eigenValues[1].Real);
            if (double.IsNaN(Curvature[0])) Curvature[0] = 0;
            if (double.IsNaN(Curvature[1])) Curvature[1] = 0;
            return Curvature;
        }

        private double[] TheQuadraticSurfaceApproximation(int pointIndex, CornerTable t, TriangleMesh mesh)
        {
            double[] Curvature = new double[2];
            int[] incidentVertices = t.VVK(pointIndex, Neighborhood);
            Point3D normal = mesh.Normals[pointIndex];
            Plane tangentPlane = new Plane(ref normal, ref mesh.Points[pointIndex]);
            Matrix<double> U = DenseMatrix.Create(incidentVertices.Length, 3, 0); // (nx3) matrix 
            Vector<double> x = DenseVector.Create(3, 0); //solve of linear system
            Vector<double> d = DenseVector.Create(incidentVertices.Length, 0); // (n)-vector
            //projection of point vj onto the plane to determine new base
            Point3D projectionVj = tangentPlane.ProjectPoint(ref mesh.Points[incidentVertices[1]]);

            //new local basis
            Point3D[] newBase = new Point3D[3];
            newBase[0] = normal / normal.Abs();
            newBase[1] = (projectionVj - mesh.Points[pointIndex]) / (projectionVj - mesh.Points[pointIndex]).Abs();
            newBase[2] = Point3D.CrossProduct(newBase[1], newBase[0]);
            newBase[2] = newBase[2] / newBase[2].Abs();

            //transforming matrix
            Matrix<double> tranformation = Matrix<double>.Build.Dense(3, 3); 
            tranformation[0, 0] = newBase[1].X;
            tranformation[0, 1] = newBase[2].X;
            tranformation[0, 2] = newBase[0].X;
            tranformation[1, 0] = newBase[1].Y;
            tranformation[1, 1] = newBase[2].Y;
            tranformation[1, 2] = newBase[0].Y;
            tranformation[2, 0] = newBase[1].Z;
            tranformation[2, 1] = newBase[2].Z;
            tranformation[2, 2] = newBase[0].Z;
            Matrix<double> mInverse = tranformation.Inverse();  //inverse of transforming matrix
            Point3D xi = new Point3D(0, 0, 0); //points in local coordinates

            //for each of the n neighbors incident to Vi
            for (int j = 1; j < incidentVertices.Length; j++)
            {
                //compute the local coordinates of each vertex incident to Vi
                xi.X = mInverse[0, 0] * (mesh.Points[incidentVertices[j]].X - mesh.Points[pointIndex].X) +
                    mInverse[0, 1] * (mesh.Points[incidentVertices[j]].Y - mesh.Points[pointIndex].Y) +
                    mInverse[0, 2] * (mesh.Points[incidentVertices[j]].Z - mesh.Points[pointIndex].Z);
                xi.Y = mInverse[1, 0] * (mesh.Points[incidentVertices[j]].X - mesh.Points[pointIndex].X) +
                    mInverse[1, 1] * (mesh.Points[incidentVertices[j]].Y - mesh.Points[pointIndex].Y) +
                    mInverse[1, 2] * (mesh.Points[incidentVertices[j]].Z - mesh.Points[pointIndex].Z);
                xi.Z = mInverse[2, 0] * (mesh.Points[incidentVertices[j]].X - mesh.Points[pointIndex].X) +
                    mInverse[2, 1] * (mesh.Points[incidentVertices[j]].Y - mesh.Points[pointIndex].Y) +
                    mInverse[2, 2] * (mesh.Points[incidentVertices[j]].Z - mesh.Points[pointIndex].Z);

                //create linear system of equations
                double ki = xi.X / Math.Sqrt(xi.X * xi.X + xi.Y * xi.Y);
                U[j, 0] = 0.5 * xi.X * xi.X;
                U[j, 1] = xi.X * xi.Y;
                U[j, 2] = 0.5 * xi.Y * xi.Y;
                d[j] = xi.Z;

            }
            //solve linear system of equations
            x = (U.Transpose() * U).Inverse() * U.Transpose() * d;

            //Weingarten curvature matrix
            Matrix<double> W = DenseMatrix.Create(2, 2, 0); 
            W[0, 0] = x[0];
            W[0, 1] = x[1];
            W[1, 0] = x[1];
            W[1, 1] = x[2];
            Evd<double> evd = W.Evd();

            //eigenvalues of Weingarten matrix
            Vector<System.Numerics.Complex> eigenValues = evd.EigenValues;
            Curvature[0] = (eigenValues[0].Real);
            Curvature[1] = (eigenValues[1].Real);
            if (double.IsNaN(Curvature[0])) Curvature[0] = 0;
            if (double.IsNaN(Curvature[1])) Curvature[1] = 0;
            return Curvature;
        }
        private double[] TheAdjacentNormalCubicApproximation(int pointIndex, CornerTable t, TriangleMesh mesh)
        {
            if (t.IsOnBorder(pointIndex))
                return new double[] { 0, 0 };

            double[] Curvature = new double[2];
            int[] incidentVertices = t.VVK(pointIndex, Neighborhood);
            Point3D normal = mesh.Normals[pointIndex];
            Plane tangentPlane = new Plane(ref normal, ref mesh.Points[pointIndex]);
            Matrix<double> U = DenseMatrix.Create(3 * incidentVertices.Length, 7, 0); //(3nx7) matrix
            Vector<double> x = DenseVector.Create(7, 0); //solve of linear system
            Vector<double> d = DenseVector.Create(3 * incidentVertices.Length, 0); //(3n)-vector

            //transformation of normal vector
            Point3D normalT = new Point3D(0, 0, 0);
            //projection of point vj onto the plane to determine new base
            Point3D projectionVj = tangentPlane.ProjectPoint(ref mesh.Points[incidentVertices[1]]);

            Point3D[] newBase = new Point3D[3];//new base
            newBase[0] = normal / normal.Abs();
            newBase[1] = (projectionVj - mesh.Points[pointIndex]) / (projectionVj - mesh.Points[pointIndex]).Abs();
            newBase[2] = Point3D.CrossProduct(newBase[1], newBase[0]);
            newBase[2] = newBase[2] / newBase[2].Abs();

            //transforming matrix
            Matrix<double> transformation = Matrix<double>.Build.Dense(3, 3); 
            transformation[0, 0] = newBase[1].X;
            transformation[0, 1] = newBase[2].X;
            transformation[0, 2] = newBase[0].X;
            transformation[1, 0] = newBase[1].Y;
            transformation[1, 1] = newBase[2].Y;
            transformation[1, 2] = newBase[0].Y;
            transformation[2, 0] = newBase[1].Z;
            transformation[2, 1] = newBase[2].Z;
            transformation[2, 2] = newBase[0].Z;
            Matrix<double> mInverse = transformation.Inverse();//inverse of transforming matrix
            Point3D xi = new Point3D(0, 0, 0);  //points in local coordinates
            
            //for each of the n neighbors incident to Vi
            for (int j = 1; j < incidentVertices.Length; j++)
            {
                //compute the local coordinates of each vertex incident to Vi
                xi.X = mInverse[0, 0] * (mesh.Points[incidentVertices[j]].X -mesh.Points[pointIndex].X) +
                  mInverse[0, 1] * (mesh.Points[incidentVertices[j]].Y -mesh.Points[pointIndex].Y) +
                  mInverse[0, 2] * (mesh.Points[incidentVertices[j]].Z - mesh.Points[pointIndex].Z);
                xi.Y = mInverse[1, 0] * (mesh.Points[incidentVertices[j]].X - mesh.Points[pointIndex].X) +
                    mInverse[1, 1] * (mesh.Points[incidentVertices[j]].Y - mesh.Points[pointIndex].Y) +
                    mInverse[1, 2] * (mesh.Points[incidentVertices[j]].Z - mesh.Points[pointIndex].Z);
                xi.Z = mInverse[2, 0] * (mesh.Points[incidentVertices[j]].X - mesh.Points[pointIndex].X) +
                    mInverse[2, 1] * (mesh.Points[incidentVertices[j]].Y - mesh.Points[pointIndex].Y) +
                    mInverse[2, 2] * (mesh.Points[incidentVertices[j]].Z - mesh.Points[pointIndex].Z);

                //compute local coordinates of each normal vector
                Point3D normalXi = mesh.Normals[incidentVertices[j]]; //normal vector at incident vertices
                normalT.X = mInverse[0, 0] * (normalXi.X) +mInverse[0, 1] * (normalXi.Y) +mInverse[0, 2] * (normalXi.Z);
                normalT.Y = mInverse[1, 0] * (normalXi.X ) +mInverse[1, 1] * (normalXi.Y ) +mInverse[1, 2] * (normalXi.Z);
                normalT.Z = mInverse[2, 0] * (normalXi.X ) +mInverse[2, 1] * (normalXi.Y) +mInverse[2, 2] * (normalXi.Z);
                normalT = normalT / normalT.Abs(); 
               
                //create linear system of equations
                U[j, 0] = 0.5*xi.X*xi.X;
                U[j, 1] = xi.X*xi.Y;
                U[j, 2] = 0.5 * xi.Y * xi.Y;
                U[j, 3] = xi.X * xi.X * xi.X;
                U[j, 4] = xi.X* xi.X* xi.Y;
                U[j, 5] = xi.X * xi.Y * xi.Y;
                U[j, 6] = xi.Y * xi.Y * xi.Y;
                U[j+1, 0] = xi.X;
                U[j+1, 1] = xi.Y;
                U[j+1, 2] = 0;
                U[j + 1, 3] = 3 * xi.X * xi.X;
                U[j+1, 4] = 2 * xi.X * xi.Y;
                U[j+1, 5] = xi.Y * xi.Y;
                U[j+1, 6] = 0;
                U[j+2, 0] = 0;
                U[j+2, 1] = xi.X;
                U[j+2, 2] = xi.Y;
                U[j+2, 3] = 0;
                U[j+2, 4] = xi.X * xi.X;
                U[j+2, 5] = 2*xi.X* xi.Y;
                U[j+2, 6] = 3*xi.Y * xi.Y;
                d[j] = xi.Z;
                d[j + 1] = -normalT.X / normalT.Z;
                d[j+2] = -normalT.Y /normalT.Z;

            }
            //solve linear system of equations
            x = (U.Transpose() * U).Inverse() * U.Transpose() * d;
            Matrix<double> W = DenseMatrix.Create(2, 2, 0);
            
            //Weingarten curvature matrix
            W[0, 0] = x[0];
            W[0, 1] = x[1];
            W[1, 0] = x[1];
            W[1, 1] = x[2];
            Evd<double> evd = W.Evd();

            //eigenvalues of Weingarten matrix
            Vector<System.Numerics.Complex> eigenValues = evd.EigenValues; 
            Curvature[0] = (eigenValues[0].Real);
            Curvature[1] = (eigenValues[1].Real);
            if (double.IsNaN(Curvature[0])) Curvature[0] = 0;
            if (double.IsNaN(Curvature[1])) Curvature[1] = 0;
            return Curvature;
        }
    }
}