﻿using System;
using System.IO;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Xml;
using Zcu.CurvatureBenchmark.Framework;
using Zcu.CurvatureBenchmark.Framework.Estimators;
using Zcu.CurvatureBenchmark.Framework.Distorters;

using MathNet.Numerics.LinearAlgebra;
using MathNet.Numerics.LinearAlgebra.Double;
using MathNet.Numerics.LinearAlgebra.Factorization;
using System.Globalization;

namespace Zcu.CurvatureBenchmark.Estimators
{
    /// <summary>
    /// Shape operators
    /// </summary>
    public enum HildebrandtEstimatorOperator
    {
        /// <summary>
        /// Operator S^hat
        /// </summary>
        Shat = 1,
        /// <summary>
        /// Operator S^bar
        /// </summary>
        Sbar = 2
    }

    /// <summary>
    /// Estimates curvatures or normals based on Generalized Shape Operators on Polyhedral Surfaces by Hildebrandt and Polthier <cite>hp11</cite>.
    /// </summary>
    /// <remarks>
    /// The method is able to estimate normals, it is also possible to use the module as normal distorter.
    /// </remarks>

    public class HildebrandtEstimator : AbstractEstimator, IDistorter
    {

        /// <summary>
        /// Gets/sets mesh size multiplicator that is used to compute neighborhood radius.
        /// </summary>
        /// <remarks>
        /// From Definition 1.23 (Mesh size) (in Hildebrandt dissertation): 
        /// For a triangle Th of a polyhedral surface Mh, let rcirc(Th)
        /// denote circumradius of Th. We define mesh size h as
        /// <code>
        ///     h = max{rcirc(Th) for Th in Mh}.
        /// </code>    
        /// <para>
        /// However this does't give good results and causes
        /// problems with long triangles, hence we use average
        /// edge length.
        /// </para>
        /// <para>
        /// The neighborhood radius is computed as: 
        /// <code>
        /// MeshSizeMul * mesh.AverageEL
        /// </code>
        /// </para>
        /// </remarks>
        /// <value>
        /// Default value is 4.0.
        /// </value>
        public double MeshSizeMul
        {
            get { return mesh_size_mult; }
            set { mesh_size_mult = value; }
        }

        /// <summary>
        /// Gets/sets whether to use operator  for IDistorter normal computation.
        /// </summary>
        /// <value>
        /// Default value is TRUE.
        /// </value>
        public bool UseOperator { set; get; }

        HildebrandtEstimatorOperator _Operator;
        /// <summary>
        /// Gets/sets shape operator type.
        /// </summary>
        /// <value>
        /// Default value is HildebrandtEstimatorOperator.Shat.
        /// </value>
        public HildebrandtEstimatorOperator Operator
        {
            set
            {
                _Operator = value;
                switch (value)
                {
                    case HildebrandtEstimatorOperator.Shat:
                        // Hat version
                        set_curvatures = setCurvaturesFromHatS;
                        gradient_contribution = hatSContribution;
                        break;
                    case HildebrandtEstimatorOperator.Sbar:
                        // Bar version
                        set_curvatures = setCurvaturesFromBarS;
                        gradient_contribution = barSContribution;
                        break;
                    default:
                        break;
                }
            }
            get { return _Operator; }
        }

        int _LocalFunction;
        internal int LocalFunction
        {
            set
            {
                _LocalFunction = value;
                switch (value)
                {
                    case 1:
                        loc_fun = local_func1;
                        loc_fun_grad = d_loc_func1;
                        break;
                    case 2:
                        loc_fun = local_func2;
                        loc_fun_grad = d_loc_func2;
                        throw new NotImplementedException("Caculation of loc_func2 support is incorrect, please don't use it");
                        break;
                    default:
                        break;
                }
            }
            get { return _LocalFunction; }

        }

        /// <summary>
        /// We need to create new triangles 
        /// with new points which are
        /// not vertices of the mesh
        /// </summary>
        class MyTriangle
        {
            public Point3D V1;
            public Point3D V2;
            public Point3D V3;
            public Point3D Normal;

            /// <summary>
            /// Takes first three Point3Ds in verts and creates triangle from them
            /// </summary>
            /// <param name="verts"></param>
            public MyTriangle(Point3D[] verts)
            {
                if(verts.Length < 3)
                {
                    throw new ArgumentOutOfRangeException("verts", "verts must contain at least three points to create triangle");
                }
                V1 = verts[0];
                V2 = verts[1];
                V3 = verts[2];
                Normal = new Plane(ref V1, ref V2, ref V3).Normal;
                
            }

            public MyTriangle(Point3D v1, Point3D v2, Point3D v3, Point3D N)
            {
                V1 = v1;
                V2 = v2;
                V3 = v3;
                Normal = N;
            }

            public MyTriangle(int t, TriangleMesh m)
            {
                V1 = m.Points[m.Triangles[t].V1];
                V2 = m.Points[m.Triangles[t].V2];
                V3 = m.Points[m.Triangles[t].V3];
                Normal = new Plane(ref V1, ref V2, ref V3).Normal;
            }

            public MyTriangle(Triangle t, TriangleMesh m)
            {
                V1 = m.Points[t.V1];
                V2 = m.Points[t.V2];
                V3 = m.Points[t.V3];
                Normal = new Plane(ref m.Points[t.V1], ref m.Points[t.V2], ref m.Points[t.V3]).Normal;
            }

            
            public override string ToString()
            {
                return V1.ToString() + "\n" +
                        V2.ToString() + "\n" +
                        V3.ToString() + "\n\n";
            }
        }

        /// <inheritdoc/>
        public override double[] K1
        {
            get
            {
                return this.k1;
            }
        }

        /// <inheritdoc/>
        public override double[] K2
        {
            get
            {
                return this.k2;
            }
        }

        double[] k1;
        double[] k2;
        Point3D[] newNormals;

        /// <summary>
        /// From Definition 1.23 (Mesh size) (in Hildebrandt dissertation): 
        /// For a triangle Th of a polyhedral surface Mh, let rcirc(Th)
        /// denote circumradius of Th. We define mesh size h as
        ///     
        ///     h = max{rcirc(Th) for Th in Mh}.
        /// 
        /// See HildebrandtExtimator.computeMaxCirlceRad.
        /// 
        /// However this does't give good results and causes
        /// problems with long triangles, hence we use average
        /// edge length:
        /// mesh.AverageEL()
        /// </summary>
        double mesh_size;
        double mesh_size_mult;
        /// <summary>
        /// nbrhd_radius = mesh_size_mult * mesh_size;
        /// </summary>
        double nbrhd_radius;

        private bool distortNormals;

        TriangleMesh mesh;
        CornerTable ct;

        delegate double local_function(Point3D X, Point3D v);
        delegate Point3D local_function_gradient(Point3D X, Point3D v);
        delegate void set_curvatures_from_op(int i, Matrix<double> op, Vector N);
        delegate Point3D operator_contribution(Point3D grad, Point3D N, Point3D[] e, int jj, double face_vol);

        local_function loc_fun;
        local_function_gradient loc_fun_grad;
        set_curvatures_from_op set_curvatures;
        operator_contribution gradient_contribution;

        /// <summary>
        /// Initialization.
        /// </summary>
        public HildebrandtEstimator()
        {
            ToStringFormat = "$Name$ (Op $Operator$, Mul $MeshSizeMul$)";
            //default setting
            LocalFunction = 1;
            MeshSizeMul = 4.0;
            Operator = HildebrandtEstimatorOperator.Shat;
            distortNormals = false;
            UseOperator = true;
        }

        private Point3D hatSContribution(Point3D grad, Point3D N, Point3D[] e, int jj, double face_vol)
        {
            return - N * grad.CrossProduct(e[jj]).DotProduct(N) * face_vol;
        }

        private Point3D barSContribution(Point3D grad, Point3D N, Point3D[] e, int jj, double face_vol)
        {
            return N * grad[jj] * face_vol;
        }

        /// <summary>
        /// If distort normals is set to true calculates normals from used operator, either hatS or barS
        /// </summary>
        /// <param name="mesh">Input</param>
        public override void EstimateCurvatures(TriangleMesh mesh)
        {
            Point3D[] e = { new Point3D(1, 0, 0), new Point3D(0, 1, 0), new Point3D(0, 0, 1) };

            this.k1 = new double[mesh.Points.Length];
            this.k2 = new double[mesh.Points.Length];

            this.mesh = mesh;
            this.ct = new CornerTable(mesh);

            mesh_size = mesh.AverageEL();
            nbrhd_radius = mesh_size_mult * mesh_size;//TODO for local_func2 nbrhd_radius is different, see its definition

            for (int i = 0; i < mesh.Points.Length; i++)
            {
                if ((i % 100) == 0)
                {
                    Progress = (byte)(100 * i / mesh.Points.Length);
                }
                /*                          
                Calculate Shat or Sbar:                    
                   sigmabar_j = int {N div loc_fun * e_j}dvol = int{N * 6psi/6x_j}dvol 

                   sigmahat_j = int {N curl loc_fun * e_j}dvol

                   Sbar_ij = ei . bar_sigma_j / leb_norm(local_func1)
                   
                   Shat_ij = ei . hat_sigma_j / leb_norm(local_func1)
                
                Optionally shape operator S:

                   S = (id - NN^T).Sbar.(id - NN^T)
                   or
                   S = Shat.C (for definition of C see (67) in article)

                From Shat, Sbar or S determine curvature:
                    Eigen values of Sbar correspond to k1, k2 and -H

                    S is singular (rank 2) - one eigen value is equal to zero, 
                    the other two are principal curvatures
                */
                Matrix Sest = DenseMatrix.Create(3, 3, 0);
                List<MyTriangle> triangle_neighbr = getRNbrhd(i, nbrhd_radius);
           
                double leb1_norm = 0;

                Point3D[] grad_psi = new Point3D[triangle_neighbr.Count];
                double[] face_vol = new double[triangle_neighbr.Count];

                for (int k = 0; k < triangle_neighbr.Count; k++)
                {
                    MyTriangle t = triangle_neighbr[k];
                    //grad_psi is constant on face - we can use its value in centre
                    grad_psi[k] = loc_fun_grad(mesh.Points[i], getFaceCenter(t));
                    face_vol[k] = calcFaceVol(t);
                    leb1_norm += calcLocFunIntoverTriangle(i, t, face_vol[k]);
                }

                for (int jj = 0; jj < 3; jj++)
                {
                    Point3D sig_jj = new Point3D(0, 0, 0);

                    for (int k = 0; k < triangle_neighbr.Count; k++)
                    { 
                        MyTriangle t = triangle_neighbr[k];
                        Point3D N = -t.Normal;
                        sig_jj += gradient_contribution(grad_psi[k], N, e, jj, face_vol[k]);
                    }

                    for (int ii = 0; ii < 3; ii++)
                    {   /*
                        According to definition (60) - Extrinsic hat local function must be normalized with regard to L^1 norm
                         */
                        Sest[ii, jj] = sig_jj[ii] / leb1_norm;
                    }

                }

                Vector N_Mh = new DenseVector(new double[] { mesh.Normals[i].X, mesh.Normals[i].Y, mesh.Normals[i].Z });
                if (distortNormals)
                {

                    newNormals[i] = getNormalfromOperator(Sest, N_Mh);
                }
                else
                {
                    set_curvatures(i, Sest, N_Mh);
                } 
            }
        }

        /// <summary>
        /// Calculates shape operator S and gets curvatures from it
        /// </summary>
        /// <param name="i"></param>
        /// <param name="hatS"></param>
        /// <param name="N_Mh"></param>
        private void setCurvaturesFromHatS(int i, Matrix<double> hatS, Vector N_Mh)
        {
            Matrix<double> S = calcShapeOpfromHatS(hatS, N_Mh);
            setCurvaturesfromShapeOp(i, S);
        }

        /// <summary>
        /// Eigen values of Sbar correspond to k1, k2 and -H
        /// </summary>
        /// <param name="i"></param>
        /// <param name="barS"></param>
        /// <param name="N_Mh"></param>
        private void setCurvaturesFromBarS(int i, Matrix<double> barS, Vector N_Mh)
        {
            Evd<double> eigS;
            try
            {
                eigS = barS.Evd();
            }
            catch (Exception)
            {
                k1[i] = 0;
                k2[i] = 0;
                return;
            }
            var dps = eigS.EigenVectors.Transpose() * N_Mh;

            double max = 0;
            for (int k = 0; k < 3; k++)
            {
                double pk1 = eigS.EigenValues[(k + 1) % 3].Real;
                double pk2 = eigS.EigenValues[(k + 2) % 3].Real;
                if (Math.Abs(dps[k]) > max)
                    {
                        max = Math.Abs(dps[k]);
                        k1[i] = Math.Max(pk1, pk2);
                        k2[i] = Math.Min(pk1, pk2);                        
                    }
            }
        }

        private void setCurvaturesfromShapeOp(int i, Matrix<double> shapeOp)
        {   /*
             S is singular (rank 2) - one eigen value is equal to zero,
             the other two are principal curvatures.
            */
            Evd<double> eigSP;
            try
            {
                eigSP = shapeOp.Evd();
            }
            catch(Exception)
            {
                k1[i] = 0;
                k2[i] = 0;
                return;
            }
            double[] eigVReAbs = { Math.Abs(eigSP.EigenValues[0].Real), Math.Abs(eigSP.EigenValues[1].Real), Math.Abs(eigSP.EigenValues[2].Real)};
            double[] eigVRe = { eigSP.EigenValues[0].Real, eigSP.EigenValues[1].Real, eigSP.EigenValues[2].Real };
            Array.Sort(eigVReAbs, eigVRe);

            k1[i] = eigVRe[2];
            k2[i] = eigVRe[1];
        }

        /// <summary>
        /// Calculates normal at vertex as eigen vector of arbitrary
        /// operator using estimated normal as "selector" 
        /// and orienntation check.
        /// 
        /// Works for hatS too, why?
        /// </summary>
        /// <param name="barS"></param>
        /// <param name="Nest">estimated normal</param>
        /// <returns></returns>
        private Point3D getNormalfromOperator(Matrix<double> barS, Vector Nest)
        {
            Evd<double> eigS = barS.Evd();
            var dps = eigS.EigenVectors.Transpose() * Nest;

            Vector<double> N = null;
            double max = -1;
            for (int k = 0; k < 3; k++)
            {
                if (Math.Abs(dps[k]) > max)
                {
                    max = Math.Abs(dps[k]);
                    N = eigS.EigenVectors.Column(k);
                    if (dps[k] < 0)
                    {
                        N *= -1;
                    }  
                }
            }
            return new Point3D(N[0], N[1], N[2]).Normalize();
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="hatS"></param>
        /// <param name="N_Mh">normal in mesh vertex</param>
        /// <returns></returns>
        private Matrix<double> calcShapeOpfromHatS(Matrix<double> hatS, Vector N_Mh)
        {   //S = Shat.C, see p. 29 equation (67) and following
            Matrix<double> C = DenseMatrix.OfRowVectors(
                new DenseVector(new double[] {   0,      N_Mh[2], -N_Mh[1] }),
                new DenseVector(new double[] {-N_Mh[2],    0,      N_Mh[0] }),
                new DenseVector(new double[] { N_Mh[1], -N_Mh[0],    0 }));
            return hatS * C;
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="hatS"></param>
        /// <param name="N_Mh">normal in mesh vertex</param>
        /// <returns></returns>
        private Matrix<double> calcShapeOpfromBarS(Matrix<double> hatS, Vector N_Mh)
        {
            Matrix id = DenseMatrix.CreateIdentity(3);
            /*
            S = (id - N.N^T).Sbar.(id - N.N^T)
            tran = (id - N.N^T) is singular (N.OuterProduct(N) is rank-1 matrix, id - N.N^T
            is rank 2: its determinant D = 1 - (a^2 + b^2 + c^2), where a, b, c are elemnets of 
            N_Mh since it is unit vector D = 1 - 1 = 0 holds. Hence S is also rank 2
            */
            Matrix<double> tran = id - N_Mh.OuterProduct(N_Mh);
            return tran* hatS *tran;
        }
        
        /// <summary>
        /// Local function is linear on triangles so in order to calculate
        /// integral we can calculate volume between original trianagle and 
        /// triangle whose vertices are vertices of the original triangle
        /// tanslated by loc. fun. values in direction of the original 
        /// triangle normal.
        /// 
        /// Or we can use some variant of mean value theorem for definite integrals
        /// </summary>
        /// <param name="x"></param>
        /// <param name="t"></param>
        /// <param name="face_vol"></param>
        /// <returns></returns>
        private double calcLocFunIntoverTriangle(int x, MyTriangle t, double face_vol)
        {
            return face_vol * (loc_fun(mesh.Points[x], t.V1)+ loc_fun(mesh.Points[x], t.V2)+ loc_fun(mesh.Points[x], t.V3)) / 3;
        }

        /// <summary>
        /// Returns all trinagles or their triangulized parts which are 
        /// in r-neigborouhood of x-th vertex.
        /// Trinagles aren't in any particular order. 
        /// </summary>
        /// <param name="x">number of the vertex</param>
        /// <param name="r">radius of the neigborouhood</param>
        /// <returns></returns>
        private List<MyTriangle> getRNbrhd(int x, double r)
        {
            Dictionary<int, double> dists = new Dictionary<int, double>();
            HashSet<int> r_nbrhd_triangles = ct.VFE(x, r, dists);
            List<MyTriangle> nbrhs = new List<MyTriangle>();

            double r2 = r * r;

            foreach (int k in r_nbrhd_triangles)
            {
                getFaceParts(x, r2, k, dists, nbrhs);                
            }
            return nbrhs;
        }

        /// <summary>
        /// Return part of the face within r-neighborouhood of x i.e. within ball
        /// of radius r and center x
        /// </summary>
        /// <param name="x">currently processed point of the mesh</param>
        /// <param name="sqr">radius of cutting sphere</param>
        /// <param name="k">face number</param>
        /// <param name="dists">squared distances to vertices</param>        
        /// <param name="triangulazied_face_part">List of triangles to fill</param>
        void getFaceParts(int x, double sqr, int k, Dictionary<int, double> dists, List<MyTriangle> triangulazied_face_part)
        {
            Point3D X = mesh.Points[x];
            Triangle t = mesh.Triangles[k];
            if (!dists.ContainsKey(t.V1))
                dists[t.V1] = X.SquareDistanceTo(mesh.Points[t.V1]);
            if (!dists.ContainsKey(t.V2))
                dists[t.V2] = X.SquareDistanceTo(mesh.Points[t.V2]);
            if (!dists.ContainsKey(t.V3))
                dists[t.V3] = X.SquareDistanceTo(mesh.Points[t.V3]);
            double[] w = { dists[t.V1], dists[t.V2], dists[t.V3] }; 

            Point3D[] outer = new Point3D[3];
            Point3D[] inner = new Point3D[3];
            int ic = 0;
            int oc = 0;            
            
            /*
            Here we need to calculace intersections p of face edges with neighborouhood border.
           
            There are always two of these instersections for either:
            1) two t.VA points within sqrt(h)-neigborouhood of x (then the face part is NOT triangle) or
            2) two t.VB points out of it (then the face part is triangle)
            */
            for (int i = 0; i < w.Length; i++)
            {
                if (w[i] >= sqr)
                {
                    outer[oc] = mesh.Points[t[i]];
                    oc++;
                }
                else
                {
                    inner[ic] = mesh.Points[t[i]];
                    ic++;
                }
            }
            switch(oc)
            {
                case 0:
                    // whole triangle is inside
                    triangulazied_face_part.Add(new MyTriangle(t, mesh));
                    break;

                case 1:
                    // one vertex outside. Must split triangle into two parts
                    Point3D o = outer[0];
                    Point3D in1 = inner[0];
                    Point3D in2 = inner[1];
                    Point3D p1 = sphereLineSegInter(X, sqr, in1, o);
                    Point3D p2 = sphereLineSegInter(X, sqr, in2, o);

                triangulazied_face_part.Add(new MyTriangle(p1, in1, in2, getFaceNormal(t)));//new triangles must inherit normal from "parent"
                triangulazied_face_part.Add(new MyTriangle(p1, p2, in2, getFaceNormal(t))); //otherwise we can't be sure about their orientation
                    break;

                case 2:
                    // two vertices outside. Triangle is reduced.
                    Point3D o1 = outer[0];
                    Point3D o2 = outer[1];
                    Point3D in0 = inner[0];
                    Point3D p1p = sphereLineSegInter(X, sqr, in0, o1);
                    Point3D p2p = sphereLineSegInter(X, sqr, in0, o2);

                    triangulazied_face_part.Add(new MyTriangle(p1p, p2p, in0, getFaceNormal(t)));
                    break;
            }
        }

        /// <summary>
        /// I'm not sure about numerical stability of this
        /// 
        /// If the intersections doesn't exist, returns point not
        /// on the line segment.
        /// </summary>
        /// <param name="sp_center">sphere centre</param>
        /// <param name="sp_sq_radius">sphere radius</param>
        /// <param name="line_start">start of line segment</param>
        /// <param name="line_end">end of line segment</param>
        /// <returns>Sphere and linesegment instersection</returns>
        private Point3D sphereLineSegInter(Point3D sp_center, double sp_sq_radius, Point3D line_start, Point3D line_end)
        {
            /*
            Solving 

                 dot(C - X, C - X) == h <=> || C - X || == Sqrt(h) , // Sphere equation
                 X := A + (B - A)*a ,                               //Line seg. parametrization        

            where X = [xc, yc, zc] is currently processed point,  A is point of the line seg. 
            within sphere, B is out of it, and a is parameter in <0, 1>,
            yields:
            */

            Point3D CA = line_start - sp_center;
            Point3D AB = line_end - line_start;
            double ABabs2 = AB.Abs2();
            double negb = -2*CA.DotProduct(AB);
            double D = negb * negb - 4 * (-sp_sq_radius + CA.Abs2()) * ABabs2;
            double sqrtD = Math.Sqrt(D);
            

            double a = (negb + sqrtD) / (2*ABabs2);
            if (a < 0 || 1 < a)
            {
                a = (negb - sqrtD) / (2*ABabs2);
            }

            Point3D p = line_start + (line_end - line_start) * a;
            return p;
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="t"></param>
        /// <returns>triangle center</returns>
        private Point3D getFaceCenter(Triangle t)
        {
            return (mesh.Points[t.V1] + mesh.Points[t.V2]  + mesh.Points[t.V3]) / 3;
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="t"></param>
        /// <returns>triangle center</returns>
        private Point3D getFaceCenter(MyTriangle t)
        {
            return (t.V1 + t.V2 + t.V3) / 3;
        }

        /// <summary>
        /// Uses the cross product formula
        /// </summary>
        /// <param name="t">face</param>
        /// <returns>volume of face</returns>
        private double calcFaceVol(Triangle t)
        {
            Point3D v1 = mesh.Points[t.V2] - mesh.Points[t.V1];
            Point3D v2 = mesh.Points[t.V3] - mesh.Points[t.V1];
            return v1.CrossProduct(v2).Abs() / 2;
        }

        /// <summary>
        /// Uses the cross product formula
        /// </summary>
        /// <param name="t">face</param>
        /// <returns>volume of face</returns>
        private double calcFaceVol(MyTriangle t)
        {
            Point3D v1v = t.V2 - t.V1;
            Point3D v2v = t.V3 - t.V1;
            double vol2 = v1v.CrossProduct(v2v).Abs() / 2;
            return vol2;
        }

        /// <summary>
        /// Uses class Plane, which returns unit normal vector of a plane
        /// How do we know that the normal of the face respects orientation of the surface?
        /// </summary>
        /// <param name="t">face</param>
        /// <returns>face unit normal vector</returns>
        private Point3D getFaceNormal(Triangle t)
        {
            Plane p = new Plane(ref mesh.Points[t.V1], ref mesh.Points[t.V2], ref mesh.Points[t.V3]);
            return p.Normal;
        }

        /// <summary>
        /// Uses class Plane, which returns unit normal vector of a plane
        /// </summary>
        /// <param name="t">face</param>
        /// <returns>face unit normal vector</returns>
        private Point3D getFaceNormal(MyTriangle t)
        {
            Plane p = new Plane(ref t.V1, ref t.V2, ref t.V3);
            return p.Normal;
        }

        /// <summary>
        /// Psi local function
        /// </summary>
        /// <param name="X">Estiamted point</param>
        /// <param name="v">Mesh point</param>
        /// <returns></returns>
        private double local_func1(Point3D X, Point3D v)
        {
            double w = 1.0 - (X - v).Abs() / nbrhd_radius;

            if (w < 0)
                return 0;
            else
                return w;
        }

        /// <summary>
        /// Psi* local function
        /// </summary>
        /// <param name="X">Estiamted point</param>
        /// <param name="v">Mesh point</param>
        /// <returns></returns>
        private double local_func2(Point3D X, Point3D v)
        {
            /*
            According to p. 27 (62) "e is fixed unit vector in R^3" but isn't further specified
            */
            Point3D e = new Point3D(1 / Math.Sqrt(3), 1 / Math.Sqrt(3), 1 / Math.Sqrt(3));
            //Point3D e = new Point3D(0, 1, 0);
            double w = 1.0 - (X + nbrhd_radius/20 * e - v).Abs() / nbrhd_radius;
            /*
            (X + nbrhd_radius/20 * e) can be interpreted as moving point X, hence it makes
            support of loc_func2 different from nbrhd_radius-neighborouhood of the original X!
            */
            if (w < 0)
                return 0;
            else
                return w;
        }

        /// <summary>
        /// Psi tilda used in calculation of normals in vertices 
        /// according to article p. 29 expression (65)
        /// 
        /// For now it is the same as psi (local_func1) but in article 
        /// it is different
        /// 
        /// </summary>
        /// <param name="X"></param>
        /// <param name="v"></param>
        /// <returns></returns>
        private double local_func3(Point3D X, Point3D v)
        {
            double w = 1.0 - (X - v).Abs() /(nbrhd_radius); 
            //in the paper this function uses 2*h in denominator

            if (w < 0)
                return 0;
            else
                return w;
        }

        private Point3D d_loc_func1(Point3D X, Point3D v)
        {
            double denom = nbrhd_radius * (X - v).Abs();            
            return (X-v) / denom;
        }

        private Point3D d_loc_func2(Point3D X, Point3D v)
        { 
            /*
            According to (62): "e is fixed unit vector in R^3" but isn't further specified
            */
            Point3D e = new Point3D(1 / Math.Sqrt(3), 1 / Math.Sqrt(3), 1 / Math.Sqrt(3));
            //Point3D e = new Point3D(0, 1, 0);
            double denom = nbrhd_radius * (X + nbrhd_radius / 20 * e - v).Abs();
            //is this really desired gradient of loc_fun2? - check it!
            return (X - v) / denom; 
        }

      
        /// <inheritdoc/>
        public TriangleMesh Distort(TriangleMesh input)
        {          
            newNormals = new Point3D[input.Points.Length];
            if (UseOperator)
            {
                distortNormals = true;
                EstimateCurvatures(input);
                distortNormals = false;
            }
            else
            {
                computeNormals(input);
            }

            TriangleMesh result = input.ShallowCopy();
            result.Normals = newNormals;
            return result;
        }

        /// <summary>
        /// Sets newNormals[]  
        /// </summary>
        /// <param name="input"></param>
        private void computeNormals(TriangleMesh input)
        {
            this.mesh = input;
            this.ct = new CornerTable(mesh);
            mesh_size = mesh.AverageEL();
            nbrhd_radius = mesh_size_mult * mesh_size;
            local_function old_loc_fun = loc_fun; //maybe not neccesary
            /*
            N^psi_Mh = int{N_Mh * psi} dvol
            */
            for (int i = 0; i < mesh.Points.Length; i++)
            {
                if ((i % 100) == 0)
                {
                    Progress = (byte)(100 * i / mesh.Points.Length);
                }

                List<MyTriangle> triangle_neighbr = getRNbrhd(i, nbrhd_radius);
                for (int k = 0; k < triangle_neighbr.Count; k++)
                {
                    double dvol = calcFaceVol(triangle_neighbr[k]);
                    /*
                    On triangle normal is constant, we can write:
                    N^psi_N_Mh = N_Mh * int{psi}dvol
                    which gives
                    */
                    newNormals[i] += triangle_neighbr[k].Normal * calcLocFunIntoverTriangle(i, triangle_neighbr[k], calcFaceVol(triangle_neighbr[k]));
                }

                newNormals[i].Normalize();
                if (newNormals[i].DotProduct(mesh.Normals[i]) < 0)
                {
                    newNormals[i] *= -1;
                }
            }
            loc_fun = old_loc_fun;
        }

    }
}