﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Zcu.CurvatureBenchmark.Framework;
using Zcu.CurvatureBenchmark.Framework.Estimators;
using Zcu.CurvatureBenchmark.Framework.Distorters;

using System.Threading;

namespace Zcu.CurvatureBenchmark.Estimators
{
    /// <summary>
    /// Estimates curvatures using a reference implementation of "Robust statistical estimation of curvature on discretized surfaces" by Kalogerakis et al. <cite>ksns07</cite>.
    /// </summary>
    /// <remarks>
    /// <para>
    /// The method is able to estimate normals, it is also possible to use the module as normal distorter.
    /// </para>
    /// <para>
    /// This module is a wrapper that calls the reference implementation (CurvatureX3.exe).
    /// </para>
    /// </remarks>
    public class KalogerakisEstimator : AbstractEstimator, IDistorter
    {
        private double[] k1, k2;

        /// <inheritdoc/>
        public override double[] K1
        {
            get
            {
                return k1;
            }
        }

        /// <inheritdoc/>
        public override double[] K2
        {
            get
            {
                return k2;
            }
        }

        int runningTime = 0;

        /// <inheritdoc/>
        public override int RunningTime
        {
            get
            {
                return runningTime;
            }
        }

        /// <summary>
        /// Checks the presence of CurvatureX3.exe, BLAS.dll and LAPACK.dll. 
        /// These files can be obtained from https://people.cs.umass.edu/~kalo/papers/curvature/index.html
        /// </summary>
        /// <returns>TRUE if files are present</returns>
        public override bool CheckPrerequisities()
        {
            bool check = true;
            if (!File.Exists("CurvatureX3.exe")) check = false;
            if (!File.Exists("BLAS.dll")) check = false;
            if (!File.Exists("LAPACK.dll")) check = false;

            if(!check)
            {
                this.Error("Missing important files (CurvatureX3.exe,BLAS.dll or LAPACK.dll). These files can be obtained from https://people.cs.umass.edu/~kalo/papers/curvature/index.html");
            }

            return check;
        }

        /// <inheritdoc/>
        public override void EstimateCurvatures(TriangleMesh mesh)
        {
            string dirName, current;
            RunKalogerakis(mesh, out dirName, out current);

            // reader to parse the data
            StreamReader sr = new StreamReader("curv.txt");

            System.Globalization.NumberFormatInfo nfi = new System.Globalization.NumberFormatInfo();
            nfi.NumberDecimalSeparator = ".";

            k1 = new double[mesh.Points.Length];
            k2 = new double[mesh.Points.Length];

            string line = sr.ReadLine();

            // read two header lines
            line = sr.ReadLine();
            line = sr.ReadLine();
            int i = 0;
            while (line != null)
            {
                // parse a single line
                string[] parts = line.Split(new char[] { ' ' }, StringSplitOptions.RemoveEmptyEntries);
                double k1p = double.Parse(parts[0], nfi);
                double k2p = double.Parse(parts[1], nfi);

                if (k1p > k2p)
                {
                    k1[i] = k1p;
                    k2[i] = k2p;
                }
                else
                {
                    k1[i] = k2p;
                    k2[i] = k1p;
                }

                line = sr.ReadLine();
                i++;
            }
            sr.Close();
            Cleanup(dirName, current);
        }

        private static void Cleanup(string dirName, string current)
        {
            File.Delete("CurvatureX3.exe");
            File.Delete("BLAS.dll");
            File.Delete("LAPACK.dll");
            File.Delete("curv.txt");
            File.Delete("temp.off");
            Directory.SetCurrentDirectory(current);
            Directory.Delete(dirName);
        }

        private void RunKalogerakis(TriangleMesh mesh, out string dirName, out string current)
        {
            Random rnd = new Random();
            dirName = "KE" + DateTime.Now.ToString("yyMMddHHmmss_") + rnd.Next(100000);
            Directory.CreateDirectory(dirName);
            File.Copy("CurvatureX3.exe", dirName + "/CurvatureX3.exe");
            File.Copy("BLAS.dll", dirName + "/BLAS.dll");
            File.Copy("LAPACK.dll", dirName + "/LAPACK.dll");
            current = Directory.GetCurrentDirectory();
            Directory.SetCurrentDirectory("./" + dirName);

            // saves mesh to an OFF file
            OFFSaver saver = new OFFSaver();
            saver.FileName = "temp.off";
            saver.Execute(mesh);

            // process that computes the curvature using the reference implementation
            ProcessStartInfo psi = new ProcessStartInfo("CurvatureX3.exe", "temp.off");
            psi.WindowStyle = ProcessWindowStyle.Hidden;

            DateTime start = DateTime.Now;
            Process curvatureProcess = Process.Start(psi);

            curvatureProcess.WaitForExit();
            TimeSpan time = DateTime.Now - start;
            this.runningTime = (int)time.TotalMilliseconds;
        }

        /// <inheritdoc/>
        public TriangleMesh Distort(TriangleMesh input)
        {
            string dirName, current;
            RunKalogerakis(input, out dirName, out current);

            TriangleMesh result = input.ShallowCopy();
            result.Normals = new Point3D[input.Points.Length];

            // reader to parse the data
            StreamReader sr = new StreamReader("curv.txt");

            System.Globalization.NumberFormatInfo nfi = new System.Globalization.NumberFormatInfo();
            nfi.NumberDecimalSeparator = ".";            

            string line = sr.ReadLine();

            // read two header lines
            line = sr.ReadLine();
            line = sr.ReadLine();
            int i = 0;
            while (line != null)
            {
                // parse a single line
                string[] parts = line.Split(new char[] { ' ' }, StringSplitOptions.RemoveEmptyEntries);
                Point3D v1 = new Point3D(double.Parse(parts[2], nfi), double.Parse(parts[3], nfi), double.Parse(parts[4], nfi));
                Point3D v2 = new Point3D(double.Parse(parts[5], nfi), double.Parse(parts[6], nfi), double.Parse(parts[7], nfi));
                Point3D n = v1.CrossProduct(v2);
                n = n.Normalize();
                if (n.DotProduct(input.Normals[i]) > n.DotProduct(-input.Normals[i]))
                    result.Normals[i] = n;
                else
                    result.Normals[i] = -n;

                line = sr.ReadLine();
                i++;
            }
            sr.Close();
            Cleanup(dirName, current);

            return result;
        }
    }
}
