﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Xml;
using Zcu.CurvatureBenchmark.Framework;
using Zcu.CurvatureBenchmark.Framework.Estimators;
using MathNet.Numerics.LinearAlgebra;
using MathNet.Numerics.LinearAlgebra.Double;
using MathNet.Numerics.LinearAlgebra.Factorization;

namespace Zcu.CurvatureBenchmark.Estimators
{
    /// <summary>
    /// Estimates curvatures based on discrete differential-geometry operators by Meyer et al. <cite>mdsb02</cite>.
    /// </summary>
    public class MeyerEstimator : AbstractEstimator
    {
        //curvature estimation
        double[] k1, k2;

        //method used for curvature estimating 
        internal int method { get; set; }

        /// <inheritdoc/>
        override public double[] K1
        {
            get { return k1; }
        }

        /// <inheritdoc/>
        override public double[] K2
        {
            get { return k2; }
        }

        /// <inheritdoc/>
        override public void EstimateCurvatures(TriangleMesh mesh)
        {
            Progress = 0;
            int weights = 1;
            if (method == 2)
            {
                //it does not work!!
                weights = 2;
                Console.WriteLine("Mean Value Coordinate Weights");
            }

            CornerTable t = new CornerTable(mesh);
            k1 = new double[mesh.Points.Length];
            k2 = new double[mesh.Points.Length];
            double[] GAMCurvature = new double[2]; //array of gauss (0) and mean curvature (1)
            for (int i = 0; i < mesh.Points.Length; i++)
            {
                GAMCurvature = MeanAndGaussCurvature(i, t, mesh, weights);
                double diff = (GAMCurvature[1] * GAMCurvature[1] - GAMCurvature[0]);
                if (diff <= 0) diff = 0;
                diff = (float)Math.Sqrt(diff);
                k1[i] = Math.Round(GAMCurvature[1] + diff,6);
                k2[i] = Math.Round(GAMCurvature[1] - diff,6);
                if ((i % 1000) == 0)
                {
                    Progress = (byte)(100 * i / mesh.Points.Length);
                }
            }
        }
        /// <summary>
        /// Compute mean curvature and Gauss curvature
        /// </summary>
        /// <param name="pointIndex">vertex index</param>
        /// <param name="t">corner table</param>
        /// <param name="mesh">triangle mesh</param>
        /// <param name="weights">which weights should be used. 1 for cotan, 2 for mean value</param>
        /// <returns></returns>
        private double[] MeanAndGaussCurvature(int pointIndex, CornerTable t, TriangleMesh mesh, int weights)
        {
            double[] GAMCurvature = new double[2];
            double angleAtVertex = 0;
            double mixedArea = 0;
            Point3D laplaceVector = new Point3D(0, 0, 0);
            Point3D edge, a, b, c, d = new Point3D(0, 0, 0);
            //for every faces incident with vertex 
            int[] incidentTriangles = t.VF(pointIndex);
            int[] incidentVertices = t.VV(pointIndex);
            if (incidentTriangles.Length != incidentVertices.Length || incidentTriangles.Length < 0)
            {
                GAMCurvature[0] = 0;
                GAMCurvature[1] = 0;
            }
            else
            {
                int vi = pointIndex;
                for (int i = 0; i < incidentVertices.Length; i++)
                {
                    int vj = incidentVertices[i];
                    int Alpha, Beta = -1;
                    if (i < incidentVertices.Length - 1 && i > 0)
                    {
                        Alpha = incidentVertices[i - 1];
                        Beta = incidentVertices[i + 1];
                    }
                    else if (i == 0)
                    {
                        Alpha = incidentVertices[incidentVertices.Length - 1];
                        Beta = incidentVertices[i + 1];
                    }
                    else
                    {
                        Alpha = incidentVertices[i - 1];
                        Beta = incidentVertices[0];
                    }
                    //calculate vectors
                    edge = mesh.Points[vj] - mesh.Points[vi];
                    a = mesh.Points[vj] - mesh.Points[Alpha];
                    b = mesh.Points[vi] - mesh.Points[Alpha];
                    c = mesh.Points[vj] - mesh.Points[Beta];
                    d = mesh.Points[vi] - mesh.Points[Beta];

                    //calculate cot(angle), which is opposite to edge
                    double cotAlpha = Point3D.DotProduct(a, b) / (Point3D.CrossProduct(a, b)).Abs();
                    double cotBeta = Point3D.DotProduct(c, d) / (Point3D.CrossProduct(c, d)).Abs();
                    //calculate sum of mixed area
                    double cosAtVi = (Point3D.DotProduct(edge, -b) / (edge.Abs() * (-b).Abs()));
                    double cosAtVi2 = (Point3D.DotProduct(edge, -d) / (edge.Abs() * (-d).Abs()));
                    double cosAtAplha = (Point3D.DotProduct(b, a) / (b.Abs() * a.Abs()));
                    double cosAtVj = (Point3D.DotProduct(-edge, -a) / ((-edge).Abs() * (-a).Abs()));

                    //add value of actual edge
                    if (weights == 1)
                    {//calculate with cotangent weights
                        laplaceVector += ((cotAlpha + cotBeta) * edge);
                    }
                    else if (weights == 2) //calculate with mean value coordinates
                    {
                        laplaceVector += (edge * (Math.Tan(Math.Acos(cosAtVi) / 2) + Math.Tan(Math.Acos(cosAtVi2) / 2)) / edge.Abs());
                    }
                    // non-obtuse triangle -> voronoi region
                    if ((cosAtVi > 0) && (cosAtAplha > 0) && (cosAtVj > 0))
                    {
                        //cot(angle)
                        double cotAtVj = (Point3D.DotProduct(-edge, -a) / Math.Sqrt((-edge).Abs2() * (-a).Abs2() -
                            Math.Pow((Point3D.DotProduct(-edge, -a)), 2)));
                        cotAtVj = (Point3D.DotProduct(-edge, -a) / (Point3D.CrossProduct((-edge), (-a))).Abs());

                        //calculate sum of voronoi area
                        double voronoiArea = ((1.0f / 8) * (edge.Abs() * edge.Abs() * cotAlpha +
                            (-b).Abs() * (-b).Abs() * cotAtVj));
                        mixedArea += voronoiArea;
                    }
                    else // obtuse triangle
                    {
                        double areaOfTriangle = (Point3D.CrossProduct(edge, -b).Abs() / 2);
                        if (cosAtVi <= 0)
                            mixedArea += (areaOfTriangle / 2);
                        else
                            mixedArea += (areaOfTriangle / 4);
                    }
                    //add inner angle of actual triangle
                    angleAtVertex += Math.Acos(cosAtVi);
                }

                GAMCurvature[0] = ((2 * Math.PI - angleAtVertex) / mixedArea);

                //compute sign of mean curvature by dotproduct with normal vector
                if (Point3D.DotProduct(mesh.Normals[pointIndex], laplaceVector) < 0)
                {
                    GAMCurvature[1] = ((1.0f / (4 * mixedArea)) * -laplaceVector.Abs());
                }
                else { GAMCurvature[1] = ((1.0f / (4 * mixedArea)) * laplaceVector.Abs()); }
            }

            return GAMCurvature;
        }
    }
}
