﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Zcu.CurvatureBenchmark.Framework;
using Zcu.CurvatureBenchmark.Framework.Estimators;
using System.Threading;
using System.Globalization;

namespace Zcu.CurvatureBenchmark.Estimators
{
    /// <summary>
    /// Estimates curvature using a reference implementation of Integral invariants for robust geometry processing by Pottmann et al. <cite>pwhy09</cite>.
    /// </summary>
    /// <remarks>
    /// <para>
    /// This module is a wrapper that calls the reference implementation (BallPCACurvature.exe) implemented by Yongliang Yang.
    /// </para>
    /// </remarks>
    public class PottmannEstimator : AbstractEstimator
    {
        private double[] k1, k2;

        /// <inheritdoc/>
        public override double[] K1
        {
            get
            {
                return k1;
            }
        }

        /// <inheritdoc/>
        public override double[] K2
        {
            get
            {
                return k2;
            }
        }

        /// <summary>
        /// Gets/sets the discretization quality (relative size of grid).
        /// </summary>
        /// <remarks>
        /// The smaller the GridRatio is, the better the result will be. But this also means more memory cost.
        /// </remarks>
        /// <value>
        /// Default value is 0.005.
        /// </value>
        public double GridRatio { get; set; } = 0.005;

        /// <summary>
        /// Gets/sets the scale at which the curvature is estimated.
        /// </summary>
        /// <remarks>
        /// A good choice of ScaleRatio is > 3*GridRatio.
        /// </remarks>
        /// <value>
        /// Default value is 0.1.
        /// </value>
        public double ScaleRatio { get; set; } = 0.1;

        /// <inheritdoc/>
        public override string ToStringFormat { get; set; } = "$Name$ (GridRatio $GridRatio$, ScaleRatio $ScaleRatio$)";

        int runningTime = 0;

        /// <inheritdoc/>
        public override int RunningTime
        {
            get
            {
                return runningTime;
            }
        }

        /// <summary>
        /// Checks the presence of CurvatureX3.exe, BLAS.dll and LAPACK.dll. 
        /// Please, contact the author Dr. Yongliang Yang http://www.bath.ac.uk/comp-sci/contacts/academics/yongliang_yang/index.html.
        /// </summary>
        /// <returns>TRUE if files are present</returns>
        public override bool CheckPrerequisities()
        {
            bool check = true;
            if (!File.Exists("BallPCACurvature.exe")) check = false;
            if (!File.Exists("fftw3.dll")) check = false;

            if (!check)
            {
                this.Error("Missing important files (BallPCACurvature.exe or fftw3.dll). Please, contact the author Dr. Yongliang Yang http://www.bath.ac.uk/comp-sci/contacts/academics/yongliang_yang/index.html");
            }

            return check;
        }

        /// <inheritdoc/>
        public override void EstimateCurvatures(TriangleMesh mesh)
        {
            Random rnd = new Random();
            string dirName = "PE"+DateTime.Now.ToString("yyMMddHHmmss_")+rnd.Next(100000);
            Directory.CreateDirectory(dirName);
            File.Copy("BallPCACurvature.exe", dirName + "/BallPCACurvature.exe");
            File.Copy("fftw3.dll", dirName + "/fftw3.dll");            
            string current = Directory.GetCurrentDirectory();
            Directory.SetCurrentDirectory("./" + dirName);

            // saves mesh to an OFF file
            ObjSaver saver = new ObjSaver();
                saver.FileName = "temp.obj";
                saver.Execute(mesh);

            // process that computes the curvature using the reference implementation            
            ProcessStartInfo psi = new ProcessStartInfo("BallPCACurvature.exe", String.Format("temp.obj {0} {1} curv.txt", GridRatio, ScaleRatio));
            psi.WindowStyle = ProcessWindowStyle.Hidden;

                DateTime start = DateTime.Now;
                Process curvatureProcess = Process.Start(psi);

                curvatureProcess.WaitForExit();
                TimeSpan time = DateTime.Now - start;
                this.runningTime = (int)time.TotalMilliseconds;

                // reader to parse the data
                StreamReader sr = new StreamReader("curv.txt");

                System.Globalization.NumberFormatInfo nfi = new System.Globalization.NumberFormatInfo();
                nfi.NumberDecimalSeparator = ".";

                k1 = new double[mesh.Points.Length];
                k2 = new double[mesh.Points.Length];

                string line = sr.ReadLine();


            // read one header line
            line = sr.ReadLine();
                int i = 0;
                while (line != null)
                {
                    // parse a single line
                    string[] parts = line.Split(new char[] { ' ' }, StringSplitOptions.RemoveEmptyEntries);
                    double k1p = double.Parse(parts[0], nfi);
                    double k2p = double.Parse(parts[1], nfi);

                    if (k1p > k2p)
                    {
                        k1[i] = k1p;
                        k2[i] = k2p;
                    } else
                    {
                        k1[i] = k2p;
                        k2[i] = k1p;
                    }

                    line = sr.ReadLine();
                    i++;
                }
                sr.Close();

            File.Delete("BallPCACurvature.exe");
            File.Delete("fftw3.dll");
            File.Delete("curv.txt");
            File.Delete("temp.obj");
            Directory.SetCurrentDirectory(current);
            Directory.Delete(dirName);
        }
    }
}
