﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Zcu.CurvatureBenchmark.Framework;
using Zcu.CurvatureBenchmark.Framework.Estimators;

using LR = MathNet.Numerics.LinearRegression;

namespace Zcu.CurvatureBenchmark.Estimators
{
    /// <summary>
    /// Curvature estimator based on article Estimating Curvatures and Their Derivatives on Triangle Meshes by Rusinkiewicz <cite>rus04</cite>.
    /// </summary>
    /// <remarks>
    /// Official implementation of algorithm:
    /// http://gfx.cs.princeton.edu/gfx/proj/trimesh2/
    ///
    /// same as above, but in one file and source can be seen directly in browser
    /// https://code.google.com/p/extend-mesh/source/browse/trunk/GraphicsLibrary/Curvature.cpp
    ///
    /// part of the code used from this:
    /// <list type="bullet">
    /// <item>rotation of system</item>
    /// <item>Voronoi area</item>
    /// <item>final curvatures</item>
    /// </list>
    /// </remarks>
    public class RusinkiewiczEstimator : AbstractEstimator
    {
		//==================================================================
		// Properties and variables
		//==================================================================
		TriangleMesh mesh;
        Point3D[] cornerAreas;
        double[] pointAreas;

        double[] k1, k2;

        /// <inheritdoc/>
        override public double[] K1
        {
            get { return k1; }
        }

        /// <inheritdoc/>
        override public double[] K2
        {
            get { return k2; }
        }

		//==================================================================
		// Methods
		//==================================================================

		//==================================================================
		// Least square solver based on original implementation
		// (not used)
		//==================================================================
		
		/// <summary>
		/// Get Previous index
		/// </summary>
		/// <param name="i"></param>
		/// <returns></returns>
        private int Prev(int i) 
        {
            return ((i) > 0 ? (i) - 1 : (i) + 2);
        }

		/// <summary>
		/// Get Next index
		/// </summary>
		/// <param name="i"></param>
		/// <returns></returns>
        private int Next(int i) 
        {
            return ((i) < 2 ? (i) + 1 : (i) - 2);
        }

       
        bool ldltdc(double[,] A, double[] rdiag)
        {
                double[] v = new double[rdiag.Length - 1];
                for (int i = 0; i < rdiag.Length; i++) 
                {
                        for (int k = 0; k < i; k++)
                        {
                                v[k] = A[i,k] * rdiag[k];
                        }
                        for (int j = i; j < rdiag.Length; j++) 
                        {
                                double sum = A[i,j];
                                for (int k = 0; k < i; k++)
                                {
                                        sum -= v[k] * A[j,k];
                                }
                                if (i == j) 
                                {
                                        if (sum <= (double)(0))
                                        {
                                                return false;
                                        }
                                        rdiag[i] = (double)(1) / sum;
                                }
                                else 
                                {
                                        A[j,i] = sum;
                                }
                        }
                }

                return true;
        }

        /// <summary>
		/// Solve Ax=B after ldltdc        
        /// </summary>
        /// <param name="A"></param>
        /// <param name="rdiag"></param>
        /// <param name="B"></param>
        /// <param name="x"></param>
        void ldltsl(double[,] A, double[] rdiag, double[] B, double[] x)
        {
                int i;
                for (i = 0; i < rdiag.Length; i++) 
                {
                    double sum = B[i];
                    for (int k = 0; k < i; k++)
                    {
                        sum -= A[i,k] * x[k];
                    }
                    x[i] = sum * rdiag[i];
                }
                for (i = rdiag.Length - 1; i >= 0; i--) 
                {
                    double sum = 0;
                    for (int k = i + 1; k < rdiag.Length; k++)
                    {
                        sum += A[k,i] * x[k];
                    }
                    x[i] -= sum * rdiag[i];
                }
        }

		 //==================================================================
		 //==================================================================
		 //==================================================================

		/// <inheritdoc/>
        override public void EstimateCurvatures(TriangleMesh mesh)
        {
			this.mesh = mesh;
           
            this.ComputePointAreas();

            this.k1 = new double[mesh.Points.Length];
            this.k2 = new double[mesh.Points.Length];

            double[] curv1 = new double[mesh.Points.Length];
            double[] curv2 = new double[mesh.Points.Length];
            double[] curv12 = new double[mesh.Points.Length];


            Point3D[] pdir1 = new Point3D[mesh.Points.Length];
            Point3D[] pdir2 = new Point3D[mesh.Points.Length];
            

            // Set up an initial coordinate system per vertex
            for (int i = 0; i < mesh.Triangles.Length; i++)
            {

                Triangle tri = mesh.Triangles[i];

                pdir1[tri[0]] = mesh.Points[tri[1]] - mesh.Points[tri[0]];
                pdir1[tri[1]] = mesh.Points[tri[2]] - mesh.Points[tri[1]];
                pdir1[tri[2]] = mesh.Points[tri[0]] - mesh.Points[tri[2]];
            }

            for (int i = 0; i < mesh.Points.Length; i++)
            {

                Point3D n = mesh.Normals[i];

                pdir1[i] = Point3D.CrossProduct(ref n, ref  pdir1[i]);
                pdir1[i].Normalize();

                pdir2[i] = Point3D.CrossProduct(ref pdir1[i], ref  n);
            }

            //-------------------------------------------

			//prealocate storage for matrices and vectors 
			//for solving system
            double[][] Am = new double[6][];
            for (int i = 0; i < 6; i++)
            {
                Am[i] = new double[] { 0, 0, 0 };
                
            }

            double[] bv = new double[6] {0, 0, 0, 0, 0, 0};

			//compute curvature
            for (int i = 0; i < mesh.Triangles.Length; i++)
            {

                Triangle tri = mesh.Triangles[i];

                Point3D[] e = new Point3D[3];               
                e[0] = mesh.Points[tri[2]] - mesh.Points[tri[1]];
                e[1] = mesh.Points[tri[0]] - mesh.Points[tri[2]];
                e[2] = mesh.Points[tri[1]] - mesh.Points[tri[0]];

                
                //compute tangent space (TBN) (orthonormal)
                Point3D t = e[0];
                Point3D n = Point3D.CrossProduct(ref e[0], ref e[1]);
                Point3D b = Point3D.CrossProduct(ref n, ref t);
                
                t.Normalize();                
                //n.Normalize();
                b.Normalize();

				double[] res = this.SolveSystem(e, t, b, tri, Am, bv);
                
                
                for (int j = 0; j < 3; j++)
                {
                    int vj = tri[j];
                        
                    double c1, c12, c2;
                    this.ProjCurv(ref t, ref b, 
                        res[0], res[1], res[2], 
                        ref pdir1[vj], ref pdir2[vj], 
                        out c1, out c12, out c2);

                    double wt = this.cornerAreas[i][j] / this.pointAreas[vj];

                    curv1[vj] += wt * c1;

                    curv12[vj] += wt * c12;

                    curv2[vj] += wt * c2;


                }                              
            }



            for (int i = 0; i < mesh.Points.Length; i++)
            {
                double ku = curv1[i];
                double kuv = curv12[i];
                double kv = curv2[i];

                double tt = 0;

                if (kuv != 0.0)
                {
                    // Jacobi rotation to diagonalize
                    double h = 0.5 * (kv - ku) / kuv;
                    tt = (h < 0.0) ? (1.0 / (h - Math.Sqrt(1.0 + h * h))) : (1.0 / (h + Math.Sqrt(1.0 + h * h)));                                        
                }
                               

                double kk1 = (ku - tt * kuv);
                double kk2 = (kv + tt * kuv);
				                
                if (kk1 < kk2)
                {
                    double tmp = kk1;
                    kk1 = kk2;
                    kk2 = tmp;
                }
                
               
                k1[i] = kk1;
                k2[i] = kk2;
               
            }


           

        }

		/// <summary>
		/// Solve system in tangent space
		/// 
		/// Estimate curvature based on variation of normals along edges
		/// 
		/// Using dummy implementation with Math.net solver
		/// Instead of original article least-squares implementation
		/// (It is commented out)
		/// 
		/// 
		/// Right figure on page 3
		/// </summary>
		/// <param name="e">triangle edges</param>
		/// <param name="t">tangent</param>
		/// <param name="b">bitangent</param>
		/// <param name="tri">triangle</param>
		/// <param name="Am">prealocated matrix (just to spare some alocation)</param>
		/// <param name="bv">prealocated vector (just to spare some alocation)</param>
		/// <returns></returns>
		double[] SolveSystem(Point3D[] e, Point3D t, Point3D b, Triangle tri, double[][] Am, double[] bv)
		{
			//--------------------------------------------

			/*
			 //from original implementation
			 //works same as math.net solver

			double[] res = new double[]{ 0, 0, 0 };
			double[, ] w = new double[,]{ {0,0,0}, {0,0,0}, {0,0,0} };
			for (int j = 0; j < 3; j++) 
			{
					double u = Point3D.DotProduct(e[j], t);
					double v = Point3D.DotProduct(e[j], b);
					w[0,0] += u*u;
					w[0,1] += u*v;
					//w[1][1] += v*v + u*u; 
					//w[1][2] += u*v; 
					w[2,2] += v*v;
					//Point3D dn = mesh.Normals[faces[i]->VIndex(PREV_Index(j))] - normals[faces[i]->VIndex(NEXT_Index(j))];
					Point3D dnor = mesh.Normals[tri[Prev(j)]] - mesh.Normals[tri[Next(j)]];
					double dnu = Point3D.DotProduct(dnor, t);
					double dnv = Point3D.DotProduct(dnor, b);
					res[0] += dnu * u;
					res[1] += dnu * v + dnv * u;
					res[2] += dnv * v;
			}
			w[1,1] = w[0,0] + w[2,2];
			w[1,2] = w[0,1];

			// Least squares solution
			double[] diag = new double[3];
			if (!ldltdc(w, diag)) 
			{
					//dprintf("ldltdc failed!\n");
					continue;
			}
			ldltsl(w, diag, res, res);
			*/
			
			//-------
			//-------

			Point3D[] dn = new Point3D[3];
			
			dn[0] = mesh.Normals[tri[1]] - mesh.Normals[tri[2]];
			dn[1] = mesh.Normals[tri[2]] - mesh.Normals[tri[0]];
			dn[2] = mesh.Normals[tri[0]] - mesh.Normals[tri[1]];

			//-------
			//-------

			Am[0][0] = Point3D.DotProduct(ref e[0], ref t);
			Am[0][1] = Point3D.DotProduct(ref e[0], ref b);

			Am[1][1] = Point3D.DotProduct(ref e[0], ref t);
			Am[1][2] = Point3D.DotProduct(ref e[0], ref b);
			//----
			Am[2][0] = Point3D.DotProduct(ref e[1], ref t);
			Am[2][1] = Point3D.DotProduct(ref e[1], ref b);

			Am[3][1] = Point3D.DotProduct(ref e[1], ref t);
			Am[3][2] = Point3D.DotProduct(ref e[1], ref b);
			//----
			Am[4][0] = Point3D.DotProduct(ref e[2], ref t);
			Am[4][1] = Point3D.DotProduct(ref e[2], ref b);

			Am[5][1] = Point3D.DotProduct(ref e[2], ref t);
			Am[5][2] = Point3D.DotProduct(ref e[2], ref b);

			//-------
			//-------

			bv[0] = Point3D.DotProduct(ref dn[0], ref t);
			bv[1] = Point3D.DotProduct(ref dn[0], ref b);

			bv[2] = Point3D.DotProduct(ref dn[1], ref t);
			bv[3] = Point3D.DotProduct(ref dn[1], ref b);

			bv[4] = Point3D.DotProduct(ref dn[2], ref t);
			bv[5] = Point3D.DotProduct(ref dn[2], ref b);



			//double[] res = LR.MultipleRegression.QR(Am, bv);
			double[] res = LR.MultipleRegression.DirectMethod(Am, bv);

			return res;
		}

		/// <summary>
		/// Implements Equation 7) from Section 3.2 in the article
		/// </summary>
		/// <param name="uf"></param>
		/// <param name="vf"></param>
		/// <param name="e"></param>
		/// <param name="f"></param>
		/// <param name="g"></param>
		/// <param name="up"></param>
		/// <param name="vp"></param>
		/// <param name="newE"></param>
		/// <param name="newF"></param>
		/// <param name="newG"></param>
        void ProjCurv( ref Point3D uf, ref Point3D vf, 
                        double e, double f, double g, 
                        ref Point3D up, ref Point3D vp, 
                        out double newE, out double newF, out double newG )
        {

			Point3D tmp_up, tmp_vp;               
            this.RotateCoordSystem(ref up, ref vp, Point3D.CrossProduct(uf, vf), out tmp_up, out tmp_vp);
  
          
            double u1 = Point3D.DotProduct(tmp_up, uf);
            double v1 = Point3D.DotProduct(tmp_up, vf);


            double u2 = Point3D.DotProduct(tmp_vp, uf);
            double v2 = Point3D.DotProduct(tmp_vp, vf);
                        
            newE = e * u1*u1 + f * (2.0  * u1*v1) + g * v1*v1;
            newF = e * u1*u2 + f * (u1*v2 + u2*v1) + g * v1*v2;
            newG = e * u2*u2 + f * (2.0  * u2*v2) + g * v2*v2;
        }
		
		/// <summary>
		/// Rotation of coordinate system
		/// </summary>
		/// <param name="old_u"></param>
		/// <param name="old_v"></param>
		/// <param name="new_norm"></param>
		/// <param name="new_u"></param>
		/// <param name="new_v"></param>
        void RotateCoordSystem( ref Point3D old_u, ref Point3D old_v, 
            Point3D new_norm,
            out Point3D new_u, out Point3D new_v)
        {
            new_u = old_u;
            new_v = old_v;
            Point3D old_norm = Point3D.CrossProduct(old_u, old_v);
            double ndot = Point3D.DotProduct(old_norm, new_norm);

            if (ndot <= -1.0f) 
            {
                new_u = -new_u;
                new_v = -new_v;
                return;
            }

            Point3D perp_old = new_norm - ndot * old_norm;
            Point3D dperp = 1.0f / (1 + ndot) * (old_norm + new_norm);
            new_u -= dperp * Point3D.DotProduct(new_u, perp_old);
            new_v -= dperp * Point3D.DotProduct(new_v, perp_old);
        }


		/// <summary>
		/// Compute Voronoi areas of triangles
		/// (Taken from original implementation)
		/// </summary>
		/// <param name="mesh"></param>
        void ComputePointAreas()
        {

            this.cornerAreas = new Point3D[mesh.Triangles.Length];
            this.pointAreas = new double[mesh.Points.Length];

            for (int i = 0; i < mesh.Triangles.Length; i++) 
            {
                Triangle tri = mesh.Triangles[i];

                // Edges
                Point3D[] e = new Point3D[3];
                e[0] = mesh.Points[tri[2]] - mesh.Points[tri[1]];
                e[1] = mesh.Points[tri[0]] - mesh.Points[tri[2]];
                e[2] = mesh.Points[tri[1]] - mesh.Points[tri[0]];                                

                //area of triangle
                double area = 0.5f * (Point3D.CrossProduct(ref e[0], ref e[1])).Length();

                // Compute corner weights

                //calc test if triangle is obtuse
                //see http://mathworld.wolfram.com/ObtuseTriangle.html
                double[] l2 =  new double[3];
                l2[0] = e[0].LengthSquared();
                l2[1] = e[1].LengthSquared();
                l2[2] = e[2].LengthSquared();

                double[] ew = new double[3];
                ew[0] = l2[0] * (l2[1] + l2[2] - l2[0]);
                ew[1] = l2[1] * (l2[2] + l2[0] - l2[1]);
                ew[2] = l2[2] * (l2[0] + l2[1] - l2[2]);

                if (ew[0] <= 0.0f) 
                {
                    //obtuse angle at [0]
                    this.cornerAreas[i][1] = -0.25f * l2[2] * area / (Point3D.DotProduct(e[0], e[2]));
                    this.cornerAreas[i][2] = -0.25f * l2[1] * area / (Point3D.DotProduct(e[0], e[1]));
                    this.cornerAreas[i][0] = area - this.cornerAreas[i][1] - this.cornerAreas[i][2];
                } 
                else if (ew[1] <= 0.0f) 
                {
                    //obtuse angle at [1]
                    this.cornerAreas[i][2] = -0.25f * l2[0] * area / (Point3D.DotProduct(e[1], e[0]));
                    this.cornerAreas[i][0] = -0.25f * l2[2] * area / (Point3D.DotProduct(e[1], e[2]));
                    this.cornerAreas[i][1] = area - this.cornerAreas[i][2] - this.cornerAreas[i][0];
                } 
                else if (ew[2] <= 0.0f) 
                {
                    //obtuse angle at [2]
                    this.cornerAreas[i][0] = -0.25f * l2[1] * area / (Point3D.DotProduct(e[2], e[1]));
                    this.cornerAreas[i][1] = -0.25f * l2[0] * area / (Point3D.DotProduct(e[2], e[0]));
                    this.cornerAreas[i][2] = area - this.cornerAreas[i][0] - this.cornerAreas[i][1];
                } 
                else 
                {
                    double ewscale = 0.5 * area / (ew[0] + ew[1] + ew[2]);
                    for (int j = 0; j < 3; j++)
                    {
                        this.cornerAreas[i][j] = ewscale * (ew[(j + 1) % 3] + ew[(j + 2) % 3]);
                    }
                }


                this.pointAreas[tri[0]] += this.cornerAreas[i][0];
                this.pointAreas[tri[1]] += this.cornerAreas[i][1];
                this.pointAreas[tri[2]] += this.cornerAreas[i][2];
           }
             
        }



    }

}
