﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Zcu.CurvatureBenchmark.Framework;

namespace Zcu.CurvatureBenchmark.Estimators
{

    /// <summary>
    /// Bezier surface "helper" class for Zhihong, but can also be used elsewhere    
    /// 
    /// Perry
    /// </summary>    
    internal class BezierSurface
    {
		//==================================================================================
		// Variables
		//==================================================================================
        const int N = 3;

        TriangleMesh mesh;
               
		//cornes points
		Point3D b300;
		Point3D b030;
		Point3D b003;

		//edge points
		Point3D b021;
		Point3D b012;

		Point3D b102;
		Point3D b201;

		Point3D b210;
		Point3D b120;

		//center point
		Point3D b111;


		//normals at corner points
		Point3D n300;
		Point3D n030;
		Point3D n003;

		//==================================================================================
		// ctor
		//==================================================================================

        public BezierSurface(ref Triangle tri, TriangleMesh mesh)
        {
            
	       
            this.mesh = mesh;

	        this.CalcBezierSurface(ref tri);
        }

		
		//==================================================================================
        // Calculate curvature 
		//==================================================================================

        public void CalcCurvature(double u, double v, out double gauss, out double mean)
        {
	        Point3D Fu = this.GetDu(u, v);
	        Point3D Fv = this.GetDv(u, v);

	        Point3D Fuu = this.GetDuu(u, v);
	        Point3D Fvv = this.GetDvv(u, v);
	        Point3D Fuv = this.GetDuv(u, v);
        
	        double E = Point3D.DotProduct(ref Fu, ref Fu);
	        double F = Point3D.DotProduct(ref Fu, ref Fv);
	        double G = Point3D.DotProduct(ref Fv, ref Fv);

	        double tmp = Point3D.CrossProduct(ref Fu, ref Fv).Abs();

                        
            if (Utils.IsZero(tmp))
	        {
		        gauss = 0;
		        mean = 0;
		        return;
	        }

	        //double L = Point3D.DotProduct(ref Fu, Point3D.CrossProduct(ref Fv, ref Fuu)) / tmp;
	        //double M = Point3D.DotProduct(ref Fu, Point3D.CrossProduct(ref Fv, ref Fuv)) / tmp;
	        //double N = Point3D.DotProduct(ref Fu, Point3D.CrossProduct(ref Fv, ref Fvv)) / tmp;

            double L = Point3D.DotProduct(Point3D.CrossProduct(ref Fu, ref Fv), ref Fuu) / tmp;
            double M = Point3D.DotProduct(Point3D.CrossProduct(ref Fu, ref Fv), ref Fuv) / tmp;
            double N = Point3D.DotProduct(Point3D.CrossProduct(ref Fu, ref Fv), ref Fvv) / tmp;

	        tmp = (E * G - F * F);

            if (Utils.IsZero(tmp))
	        {
		        gauss = 0;
		        mean = 0;
		        return;
	        }

	        double K = (L * N - M * M) / (tmp); //gauss
	        double H = (G * L - 2 * F * M + E * N) / (2 * tmp); //mean

	        gauss = K;
	        mean = H;
        }

		//==================================================================================
        // Calculate derivatives 
		//==================================================================================

        //1st

        Point3D GetDu(double u, double v)
        {
	        double w = 1 - u - v;
	        return this.GetDu(u, v, w);
        }

        Point3D GetDu(double u, double v, double w)
        {
	        //http://www.gamasutra.com/view/feature/131389/b%C3%A9zier_triangles_and_npatches.php?print=1
	        //http://www.gamedev.net/topic/624397-bezier-triangle-partial-derivatives/

	        //for every i,j,k
	        //i + j + k = n - 1

	        //B[i,j,k](n-1) * (3 * (b[i+1, j, k] - b[i, j, k+1]))
	        //possible i,j,k
	        //200 => b[300] - b[201]; (2!) / (2!0!0!) = 1; uu
	        //020 => b[120] - b[021]; (2!) / (0!2!0!) = 1; vv
	        //002 => b[102] - b[003]; (2!) / (0!0!2!) = 1; ww
	        //110 => b[210] - b[111]; (2!) / (1!1!0!) = 2; uv
	        //011 => b[111] - b[012]; (2!) / (0!1!1!) = 2; vw
	        //101 => b[201] - b[102]; (2!) / (1!0!1!) = 2; uw


	        return 
			           u * u * (3.0 * (this.b300 - this.b201)) +
			           v * v * (3.0 * (this.b120 - this.b021)) +
			           w * w * (3.0 * (this.b102 - this.b003)) +
		        2.0 * u * v * (3.0 * (this.b210 - this.b111)) +
		        2.0 * v * w * (3.0 * (this.b111 - this.b012)) + 
		        2.0 * u * w * (3.0 * (this.b201 - this.b102));
			

        }

        Point3D GetDv(double u, double v)
        {
	        double w = 1 - u - v;
	        return this.GetDv(u, v, w);
        }

        Point3D GetDv(double u, double v, double w)
        {
	        //http://www.gamasutra.com/view/feature/131389/b%C3%A9zier_triangles_and_npatches.php?print=1
	        //http://www.gamedev.net/topic/624397-bezier-triangle-partial-derivatives/

	        //for every i,j,k
	        //i + j + k = n - 1

	        //B[i,j,k](n-1) * (3 * (b[i, j+1, k] - b[i, j, k+1]))
	        //possible i,j,k
	        //200 => b[210] - b[201]; (2!) / (2!0!0!); uu
	        //020 => b[030] - b[021]; (2!) / (0!2!0!); vv
	        //002 => b[012] - b[003]; (2!) / (0!0!2!); ww
	        //110 => b[120] - b[111]; (2!) / (1!1!0!); uv
	        //011 => b[021] - b[012]; (2!) / (0!1!1!); vw
	        //101 => b[111] - b[102]; (2!) / (1!0!1!); uw
	
	        return
			           u * u * (3.0 * (this.b210 - this.b201)) +
			           v * v * (3.0 * (this.b030 - this.b021)) +
			           w * w * (3.0 * (this.b012 - this.b003)) +
		        2.0 * u * v * (3.0 * (this.b120 - this.b111)) +		
		        2.0 * v * w * (3.0 * (this.b021 - this.b012)) +
		        2.0 * u * w * (3.0 * (this.b111 - this.b102));
        }

        //2nd

        Point3D GetDuu(double u, double v)
        {
	        double w = 1 - u - v;
	        return this.GetDuu(u, v, w);
        }

        Point3D GetDuu(double u, double v, double w)
        {
	        //for every i,j,k
	        //i + j + k = n - 2

	        //B[i,j,k](n-2) * (6 * (b[i+2, j, k] - 2 * b[i+1, j, k+1] + b[i, j, k+2]))
	        //possible i,j,k
	        //100 => b[300] - b[201] + b[102]; (1!) / (1!0!0!) = 1; u
	        //010 => b[210] - b[111] + b[012]; (1!) / (0!1!0!) = 1; v
	        //001 => b[201] - b[102] + b[003]; (1!) / (0!0!1!) = 1; w
		
	        return
		        u * (6.0 * (this.b300 - 2.0 * this.b201 + this.b102)) +
		        v * (6.0 * (this.b210 - 2.0 * this.b111 + this.b012)) +
		        w * (6.0 * (this.b201 - 2.0 * this.b102 + this.b003));	
        }

        Point3D GetDvv(double u, double v)
        {
	        double w = 1 - u - v;
	        return this.GetDvv(u, v, w);
        }

        Point3D GetDvv(double u, double v, double w)
        {
	
	        //for every i,j,k
	        //i + j + k = n - 2

	        //B[i,j,k](n-2) * (6 * (b[i, j+2, k] - 2 * b[i, j+1, k+1] + b[i, j, k+2]))
	        //possible i,j,k
	        //100 => b[120] - b[111] + b[102]; (1!) / (1!0!0!) = 1; u
	        //010 => b[030] - b[021] + b[012]; (1!) / (0!1!0!) = 1; v
	        //001 => b[021] - b[012] + b[003]; (1!) / (0!0!1!) = 1; w

	        return
		        u * (6.0 * (this.b120 - 2.0 * this.b111 + this.b102)) +
		        v * (6.0 * (this.b030 - 2.0 * this.b021 + this.b012)) +
		        w * (6.0 * (this.b021 - 2.0 * this.b012 + this.b003));
        }

        Point3D GetDuv(double u, double v)
        {
	        double w = 1 - u - v;
	        return this.GetDuv(u, v, w);
        }

        Point3D GetDuv(double u, double v, double w)
        {

	        //for every i,j,k
	        //i + j + k = n - 2

	        //B[i,j,k](n-2) * (6 * (b[i+1, j+1, k] - b[i+1, j, k+1] - b[i, j+1, k+1] + b[i, j, k+2]))
	        //possible i,j,k
	        //100 => b[210] - b[201] - b[111] + b[102]; (1!) / (1!0!0!) = 1; u
	        //010 => b[120] - b[111] - b[021] + b[012]; (1!) / (0!1!0!) = 1; v
	        //001 => b[111] - b[102] - b[012] + b[003]; (1!) / (0!0!1!) = 1; w

	        return
		        u * (6.0 * (this.b210 - this.b201 - this.b111 + this.b102)) +
		        v * (6.0 * (this.b120 - this.b111 - this.b021 + this.b012)) +
		        w * (6.0 * (this.b111 - this.b102 - this.b012 + this.b003));

        }

		//==================================================================================
        // Calculate surface
		//==================================================================================
     
        /// <summary>
        /// Calc control points of Bezier surface
		/// See: http://ogldev.atspace.co.uk/www/tutorial31/tutorial31.html
        /// </summary>
        /// <param name="tri"></param>
        void CalcBezierSurface(ref Triangle tri)
        {
	        this.b030 = this.mesh.Points[tri.V1];
            this.b003 = this.mesh.Points[tri.V2];
            this.b300 = this.mesh.Points[tri.V3];

            this.n030 = this.mesh.Normals[tri.V1];
            this.n003 = this.mesh.Normals[tri.V2];
            this.n300 = this.mesh.Normals[tri.V3];


	        //edges - names are oposite vertex
	        Point3D edge_b300 = this.b003 - this.b030;
	        Point3D edge_b030 = this.b300 - this.b003;
	        Point3D edge_b003 = this.b030 - this.b300;

	        // Generate two midpoints on each edge
	        this.b021 = this.b030 + edge_b300 / 3.0;
	        this.b012 = this.b030 + (edge_b300 * 2.0) / 3.0;

	        this.b102 = this.b003 + edge_b030 / 3.0;
	        this.b201 = this.b003 + (edge_b030 * 2.0) / 3.0;

	        this.b210 = this.b300 + edge_b003 / 3.0;
	        this.b120 = this.b300 + (edge_b003 * 2.0) / 3.0;

	        // Project each midpoint on the plane defined by the nearest vertex and its normal
            
	        Plane p030 = new Plane(ref this.n030, ref this.b030);
	        Plane p003 = new Plane(ref this.n003, ref this.b003);
            Plane p300 = new Plane(ref this.n300, ref this.b300);
	
	        this.b021 = p030.ProjectPoint(ref this.b021);
	        this.b012 = p003.ProjectPoint(ref this.b012);
	

	        this.b102 = p003.ProjectPoint(ref this.b102);
	        this.b201 = p300.ProjectPoint(ref this.b201);

	        this.b210 = p300.ProjectPoint(ref this.b210);
	        this.b120 = p030.ProjectPoint(ref this.b120);
	        

           
            
	        // Handle the center
	        Point3D c = (this.b003 + this.b030 + this.b300) / 3.0;
	        this.b111 = (this.b021 + this.b012 + this.b102 + this.b201 + this.b210 + this.b120) / 6.0;
	        this.b111 += (this.b111 - c) / 2.0;
        }


    }
}
