﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Zcu.CurvatureBenchmark.Framework;
using Zcu.CurvatureBenchmark.Framework.Estimators;

namespace Zcu.CurvatureBenchmark.Estimators
{
    /// <summary>    
    /// Zhihong (Bezier) Estimator based on article "Curvature estimation for meshes based on vertex normal triangles" <cite>zgyl11</cite>.
    /// </summary>
    public class ZhihongEstimator : AbstractEstimator
    {
		//==================================================================
		// Properties and variables
		//==================================================================

        Point3D[] cornerAreas;
        double[] pointAreas;

        double[] k1, k2;

        /// <inheritdoc/>
        override public double[] K1
        {
            get { return k1; }
        }

        /// <inheritdoc/>
        override public double[] K2
        {
            get { return k2; }
        }

		//==================================================================
		// Methods
		//==================================================================


        /// <inheritdoc/>
        override public void EstimateCurvatures(TriangleMesh mesh)
        {
            this.ComputePointAreas(mesh);

            const int GAUS = 0;
            const int MEAN = 1;
            const int CURVATURES_COUNT = 2;

            int[] counts = new int[mesh.Points.Length * CURVATURES_COUNT];
            double[] vals = new double[mesh.Points.Length * CURVATURES_COUNT];

                       
            for (int i = 0; i < mesh.Triangles.Length; i++)
            {
                Triangle tri = mesh.Triangles[i];

                BezierSurface b = new BezierSurface(ref tri, mesh);

                double bx = 1.0 / 3.0;
                double by = 1.0 / 3.0;

                double gauss = 0;
                double mean = 0;

                b.CalcCurvature(bx, by, out gauss, out mean);


                for (int j = 0; j < 3; j++)
                {

                    double wt = this.cornerAreas[i][j] / this.pointAreas[tri[j]];

					vals[tri[j] * CURVATURES_COUNT + GAUS] += gauss * wt;
                    //counts[tri[j] * CURVATURES_COUNT + GAUS]++;

					vals[tri[j] * CURVATURES_COUNT + MEAN] += mean * wt;
                    //counts[tri[j] * CURVATURES_COUNT + MEAN]++;
                }

            }



            //Calculate curvature at point as average from all summed curvatures at the point
	        // and clamp values to desired interval

            this.k1 = new double[mesh.Points.Length];
            this.k2 = new double[mesh.Points.Length];

            int index = 0;
	        for (int i = 0; i < vals.Length; i += CURVATURES_COUNT)
	        {
		        for (int j = 0; j < CURVATURES_COUNT; j++)
		        {
			        if (counts[i + j] != 0)
			        {
				        vals[i + j] /= counts[i + j];
			        }
		        }
               
                double s  = vals[i + MEAN] * vals[i + MEAN] - vals[i + GAUS];
                if (s < 0) s = 0;

				//why the "-" sign ???
				//without it, curvatures are "flipped" 

                double kk1 = -(vals[i + MEAN] + Math.Sqrt(s));
                double kk2 = -(vals[i + MEAN] - Math.Sqrt(s));

                if (kk1 < kk2)
                {
                    double tmp = kk1;
                    kk1 = kk2;
                    kk2 = tmp;
                }

                this.k1[index] = kk1;
                this.k2[index] = kk2;

                index++;                
	        }

            

        }

		/// <summary>
		/// Compute Voronoi areas of triangles
		/// (Taken from Rusinkiewicz implementation)
		/// </summary>
		/// <param name="mesh"></param>
        void ComputePointAreas(TriangleMesh mesh)
        {

            this.cornerAreas = new Point3D[mesh.Triangles.Length];
            this.pointAreas = new double[mesh.Points.Length];

            for (int i = 0; i < mesh.Triangles.Length; i++)
            {
                Triangle tri = mesh.Triangles[i];

                // Edges
                Point3D[] e = new Point3D[3];
                e[0] = mesh.Points[tri[2]] - mesh.Points[tri[1]];
                e[1] = mesh.Points[tri[0]] - mesh.Points[tri[2]];
                e[2] = mesh.Points[tri[1]] - mesh.Points[tri[0]];

                //area of triangle
                double area = 0.5f * (Point3D.CrossProduct(ref e[0], ref e[1])).Length();

                // Compute corner weights

                //calc test if triangle is obtuse
                //see http://mathworld.wolfram.com/ObtuseTriangle.html
                double[] l2 = new double[3];
                l2[0] = e[0].LengthSquared();
                l2[1] = e[1].LengthSquared();
                l2[2] = e[2].LengthSquared();

                double[] ew = new double[3];
                ew[0] = l2[0] * (l2[1] + l2[2] - l2[0]);
                ew[1] = l2[1] * (l2[2] + l2[0] - l2[1]);
                ew[2] = l2[2] * (l2[0] + l2[1] - l2[2]);

                if (ew[0] <= 0.0f)
                {
                    //obtuse angle at [0]
                    this.cornerAreas[i][1] = -0.25f * l2[2] * area / (Point3D.DotProduct(e[0], e[2]));
                    this.cornerAreas[i][2] = -0.25f * l2[1] * area / (Point3D.DotProduct(e[0], e[1]));
                    this.cornerAreas[i][0] = area - this.cornerAreas[i][1] - this.cornerAreas[i][2];
                }
                else if (ew[1] <= 0.0f)
                {
                    //obtuse angle at [1]
                    this.cornerAreas[i][2] = -0.25f * l2[0] * area / (Point3D.DotProduct(e[1], e[0]));
                    this.cornerAreas[i][0] = -0.25f * l2[2] * area / (Point3D.DotProduct(e[1], e[2]));
                    this.cornerAreas[i][1] = area - this.cornerAreas[i][2] - this.cornerAreas[i][0];
                }
                else if (ew[2] <= 0.0f)
                {
                    //obtuse angle at [2]
                    this.cornerAreas[i][0] = -0.25f * l2[1] * area / (Point3D.DotProduct(e[2], e[1]));
                    this.cornerAreas[i][1] = -0.25f * l2[0] * area / (Point3D.DotProduct(e[2], e[0]));
                    this.cornerAreas[i][2] = area - this.cornerAreas[i][0] - this.cornerAreas[i][1];
                }
                else
                {
                    double ewscale = 0.5f * area / (ew[0] + ew[1] + ew[2]);
                    for (int j = 0; j < 3; j++)
                    {
                        this.cornerAreas[i][j] = ewscale * (ew[(j + 1) % 3] + ew[(j + 2) % 3]);
                    }
                }


                this.pointAreas[tri[0]] += this.cornerAreas[i][0];
                this.pointAreas[tri[1]] += this.cornerAreas[i][1];
                this.pointAreas[tri[2]] += this.cornerAreas[i][2];
            }

        }

    }
}
