﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text.RegularExpressions;
using System.Xml;

namespace Zcu.CurvatureBenchmark.Framework
{
    /// <summary>
    /// AbstractModule implements <see cref="IModule"/> interface and it provides basic functionality.
    /// </summary>
    /// <remarks>
    /// The main functionality provided by AbstractModule is:
    /// <list type="bullet">
    /// <item>Registers the module to random generator system, so that each module has its own random generator.</item>
    /// <item>Provides SetConfig implementation that takes an attribute or element from config xml section
    /// and tries to find and set matching public R/W property.</item>
    /// <item>Implement <see cref="IModule.Progress"/> with <see cref="IModule.ProgressChanged"/> event.</item>
    /// </list>
    /// </remarks>
    public abstract class AbstractModule : IModule
    {
        /// <summary>
        /// Register module instance to random generator system.
        /// </summary>
        public AbstractModule()
        {
            //dummy call for random
            //init randoms in "singleton" class - without this, in 
            //multithreaded code, it will crash, because of used dictionary
            this.Rand.Next();
        }

        /// <summary>
        /// Checks all prerequisities (usually used for wrapper modules).
        /// </summary>
        /// <returns>TRUE if everything is OK.</returns>
        public virtual bool CheckPrerequisities()
        {
            return true;
        }
        
        /// <summary>
        /// Setting module parameters according to the content of xmlnode.
        /// It takes all xml attributes and elements and tries to find and set corresponding property (ignoring case).
        /// </summary>
        /// <param name="configNode">Part of xml setting</param>
        public virtual void SetConfig(XmlNode configNode)
        {
            var properties =
            this.GetType().GetProperties().Where(p => p.CanWrite).ToDictionary(p => p.Name.ToLower(), p => p);

            foreach (XmlAttribute atr in configNode.Attributes)
            {
                this.SetConfig(properties, atr.Name, atr.InnerText);
            }

            foreach (XmlNode node in configNode.ChildNodes)
            {
                this.SetConfig(properties, node.Name, node.InnerText);
            }
        }

        /// <summary>
        /// Set property values by property name and XML name
        /// Value is taken from XML
        /// </summary>
        /// <param name="properties"></param>
        /// <param name="atrName"></param>
        /// <param name="atrContent"></param>
        protected void SetConfig(Dictionary<string, System.Reflection.PropertyInfo> properties,
            string atrName, string atrContent)
        {
            if (properties.ContainsKey(atrName.ToLower()))
            {
                try
                {
                    var property = properties[atrName.ToLower()];
                    if (property.PropertyType.IsEnum)
                    {
                        property.SetValue(this, Enum.Parse(property.PropertyType, atrContent, true));
                    }
                    else
                    {
                        property.SetValue(this, Convert.ChangeType(atrContent, property.PropertyType));
                    }
                }
                catch (Exception e)
                {
                    this.Error(string.Format("Error setting config in node {0}\n\t{1}", atrName, e.Message));
                }
            }
        }

        byte progress = 100;
        /// <summary>
        /// Implementation of progress changed reporter. The default value of Progress is 100.
        /// </summary>
        public virtual byte Progress
        {
            get
            {
                return progress;
            }
            set
            {
                if (value != progress)
                {
                    progress = value;
                    var progressChanged = ProgressChanged;
                    if (progressChanged != null)
                        progressChanged(this, new NotifyProgressChangedEventArgs(progress));
                }
            }
        }

        /// <summary>
        /// Notifies Progress property change.
        /// </summary>
        public virtual event NotifyProgressChanged ProgressChanged;

        /// <summary>
        /// Shortcut for this.Rand()
        /// </summary>
        public Random Rand { get { return this.Rand(); } }

        /// <inheritdoc/>       
        public virtual bool Active { get; set; } = true;

        /// <inheritdoc/>
        public string Name { get { var name = this.GetType().Name; return name.Substring(name.LastIndexOf(".") + 1); } }

        /// <inheritdoc/>
        public string FullName { get { return this.GetType().Name; } }


        /// <inheritdoc/>
        /// <remarks>Default value is <c>$Name$</c></remarks>
        public virtual string ToStringFormat { get; set; } = "$Name$";

        /// <summary>
        /// Converts instance to its string representation using ToStringFormat.
        /// </summary>
        /// <returns>String representation of instance.</returns>
        public override string ToString()
        {
            return Regex.Replace(ToStringFormat, @"\$([^$]+)\$", ToStringReplace);
        }

        string ToStringReplace(Match match)
        {
            if (this.GetType().GetProperty(match.Groups[1].Value) != null)
                return this.GetType().GetProperty(match.Groups[1].Value).GetValue(this).ToString();
            else
                return "$" + match.Groups[1].Value + "$";
        }

    }

}
