﻿using System;
using System.Collections.Generic;
using System.Linq;

namespace Zcu.CurvatureBenchmark.Framework
{
    /// <summary>
    /// Implementation of Corner Table data structure <cite>ros01</cite>.
    /// </summary>
    public class CornerTable
    {
        TriangleMesh mesh;

        // these two fields represent the Corner table data structure

        /// <summary>
        /// Index of opposite corner
        /// </summary>
        public int[] opposite;

        /// <summary>
        /// Index of corner corresponding to vertex
        /// </summary>
        public int[] incidentCorner;

        // edge data structure
        struct Edge
        {
            // start and end vertex
            public int v1, v2;

            // hash code - needed for the dictionary
            public override int GetHashCode()
            {
                return (v1 + 10000 * v2);
            }

            //constructor - simplicity incarnate.
            public Edge(int v1, int v2)
            {
                this.v1 = v1;
                this.v2 = v2;
            }

            // equality operator. Edges are equal, when start and end point coiincide. 
            // Edge cannot be equal to anyhing else than an edge
            public override bool Equals(object obj)
            {
                if (obj is Edge)
                {
                    Edge other = (Edge)obj;
                    if ((this.v1 == other.v1) && (this.v2 == other.v2))
                        return (true);
                    else
                        return (false);
                }
                else
                    return (false);
            }
        }


        /// <summary>
        /// Constructs a corner table for a given mesh. It should take roughly linear time
        /// </summary>
        /// <param name="mesh">input mesh</param>
        public CornerTable(TriangleMesh mesh)
        {
            this.mesh = mesh;
            InitCornerTable();
        }

        /// <summary>
        /// Returns next corner in a triangle
        /// </summary>
        /// <param name="c">Corner index</param>
        /// <returns>Next corner</returns>
        public int n(int c)
        {
            if (c % 3 == 2)
                return (c - 2);
            else
                return (c + 1);
        }

        /// <summary>
        /// Returns previous corner in a triangle.
        /// </summary>
        /// <param name="c">Corner index</param>
        /// <returns>Previous corner</returns>
        public int p(int c)
        {
            if (c % 3 == 0)
                return (c + 2);
            else
                return (c - 1);
        }

        /// <summary>
        /// Returns vertex index incident with a corner.
        /// </summary>
        /// <param name="c">Given corner</param>
        /// <returns>Incident vertex</returns>
        public int v(int c)
        {
            int t = c / 3;
            if ((c % 3) == 0)
                return (mesh.Triangles[t].V1);
            if ((c % 3) == 1)
                return (mesh.Triangles[t].V2);
            return (mesh.Triangles[t].V3);
        }

        /// <summary>
        /// Returns triangle index corresponding to a given corner.
        /// </summary>
        /// <param name="c">Given corner</param>
        /// <returns>Corresponding triangle index</returns>
        public int t(int c)
        {
            return c / 3;
        }

        /// <summary>
        /// Returns vertices incident with a vertex
        /// </summary>
        /// <param name="vertex">Given vertex</param>
        /// <returns>Incident vertices</returns>
        public int[] VV(int vertex)
        {
            List<int> result = new List<int>();
            int initCorner = incidentCorner[vertex];

            result.Add(v(p(initCorner)));

            int o = opposite[n(initCorner)];
            while (o != -1)
            {
                if (n(o) == initCorner)
                    break;
                result.Add(v(o));
                o = opposite[p(o)];
            }

            if (o == -1)
            {
                // this seems to be a border vertex
                // lets assume it is not complex, i.e. that there are exactly two border edges
                // and lets move around the vertex in the opposite direction
                result.Add(v(n(initCorner)));

                o = opposite[p(initCorner)];
                while (o != -1)
                {
                    if (p(o) == initCorner)
                        break;
                    result.Add(v(o));
                    o = opposite[n(o)];
                }
            }

            return (result.ToArray());
        }


        /// <summary>
        /// Returns faces incident with a vertex
        /// </summary>
        /// <param name="vertex">Given vertex</param>
        /// <returns>Incident faces</returns>
        public int[] VF(int vertex)
        {
            List<int> result = new List<int>();
            int initCorner = incidentCorner[vertex];

            result.Add(initCorner / 3);

            int o = opposite[n(initCorner)];
            while (o != -1)
            {
                if (n(o) == initCorner)
                    break;
                result.Add(o / 3);
                o = opposite[p(o)];
            }

            if (o == -1)
            {
                // this seems to be a border vertex
                // lets assume it is not complex, i.e. that there are exactly two border edges
                // and lets move around the vertex in the opposite direction                

                o = opposite[p(initCorner)];
                while (o != -1)
                {
                    if (p(o) == initCorner)
                        break;
                    result.Add(o / 3);
                    o = opposite[n(o)];
                }
            }

            return (result.ToArray());
        }

        /// <summary>
        /// Determines whether a vertex lies on the border
        /// </summary>
        /// <param name="vertex">Given vertex</param>
        /// <returns>True if vertex lies on the border</returns>
        public bool IsOnBorder(int vertex)
        {
            // walk around the vertex and look for -1 in the corner table
            int initCorner = incidentCorner[vertex];
            int o = opposite[n(initCorner)];
            while (o != -1)
            {
                if (n(o) == initCorner)
                    break;
                o = opposite[p(o)];
            }

            return (o == -1);
        }

        /// <summary>
        /// Returns next vertex on the border.
        /// </summary>
        /// <param name="vertex">Vertex</param>
        /// <returns>Next vertex on the border</returns>
        private int NextVertexOnBorder(int vertex)
        {
            // walk around the vertex and look for -1 in the corner table
            int initCorner = incidentCorner[vertex];
            int o = opposite[n(initCorner)];

            if (o == -1)
                return (v(p(initCorner)));
            while (o != -1)
            {
                if (n(o) == initCorner)
                    break;
                int no = opposite[p(o)];
                if (no == -1)
                    return (v(o));
                o = no;
            }

            // this is not a border vertex!
            return (-1);
        }

        /// <summary>
        /// Initializes the cornertable data structure fills the incidentCorner and oposite.
        /// </summary>
        public void InitCornerTable()
        {
            Dictionary<Edge, int> dictionary = new Dictionary<Edge, int>();
            this.opposite = new int[mesh.Triangles.Length * 3];
            for (int i = 0; i < opposite.Length; i++)
            {
                opposite[i] = -1;
            }
            this.incidentCorner = new int[mesh.Points.Length];

            for (int i = 0; i < mesh.Triangles.Length; i++)
            {
                Triangle t = mesh.Triangles[i];

                Edge e = new Edge(t.V1, t.V2);
                int cornerIndex = 3 * i+2;
                processEdge(dictionary, e, cornerIndex);
                incidentCorner[t.V3] = cornerIndex;

                e = new Edge(t.V2, t.V3);
                cornerIndex = 3 * i;
                processEdge(dictionary, e, cornerIndex);
                incidentCorner[t.V1] = cornerIndex;

                e = new Edge(t.V3, t.V1);
                cornerIndex = 3 * i + 1;
                processEdge(dictionary, e, cornerIndex);
                incidentCorner[t.V2] = cornerIndex;
            }
        }

        /// <summary>
        /// this is an auxiliary function for the creation of the corner table
        /// sees whether the edge is already in the dictionary. If it isn't, then it is added.
        /// </summary>
        /// <param name="dictionary"></param>
        /// <param name="e"></param>
        /// <param name="cornerIndex"></param>
        private void processEdge(Dictionary<Edge, int> dictionary, Edge e, int cornerIndex)
        {
            if (dictionary.ContainsKey(e))
            {
                int oppositeCornerIndex = dictionary[e];
                opposite[cornerIndex] = oppositeCornerIndex;
                opposite[oppositeCornerIndex] = cornerIndex;
            }
            else
            {
                Edge reversed = new Edge(e.v2, e.v1);
                if (dictionary.ContainsKey(reversed))
                    throw new Exception("Non-manifold mesh at the input.");
                dictionary.Add(reversed, cornerIndex);
            }
        }

        /// <summary>
        /// Returns a list of neighbours in the K-ring.
        /// </summary>
        /// <param name="v">Vertex index</param>
        /// <param name="k">Neighbourhood width</param>        
        /// <returns>List of neighbours</returns>
        public int[] VVK(int v, int k)
        {
            if (k == 0) 
            {
                return new int[1] { v };
            }

            Dictionary<int, int> distances = new Dictionary<int, int>();
            HashSet<int> visitedTriangles = new HashSet<int>();
            HashSet<int> result = new HashSet<int>();
            result.Add(v);
            int start = incidentCorner[v];
            if (start < 0)
                return (null);
            Stack<int> workQueue = new Stack<int>();
            int vertex = this.v(n(start));
            result.Add(vertex);
            distances[vertex] = 1;
            visitedTriangles.Add(t(start));
            workQueue.Push(start);
            int current = -1;
            int op = opposite[p(start)];
            if (op>=0)
                current = p(op);
            while ((current != start) && (current >= 0))
            {
                vertex = this.v(n(current));
                result.Add(vertex);
                distances[vertex] = 1;
                visitedTriangles.Add(t(current));
                workQueue.Push(current);
                op = opposite[p(current)];
                if (op < 0)
                    current = -1;
                else
                    current = p(op);
            }

            // treatment of border. Only two border edges per vertex allowed!
            if (current < 0)
            {
                vertex = this.v(p(start));
                result.Add(vertex);
                distances[vertex] = 1;
                op = opposite[n(start)];
                if (op >= 0)
                    current = n(op);
                while (current >=0)
                {
                    vertex = this.v(p(current));
                    result.Add(vertex);
                    distances[vertex] = 1;
                    visitedTriangles.Add(t(current));
                    workQueue.Push(current);
                    op = opposite[n(current)];
                    if (op < 0)
                        current = -1;
                    else
                        current = n(op);
                }
            }            

            while (workQueue.Count > 0)
            {
                int c = workQueue.Pop();
                if (opposite[c] < 0)
                    continue;

                if (visitedTriangles.Contains(t(opposite[c])))
                    continue;

                int v1 = distances[this.v(p(c))];
                int v2 = distances[this.v(n(c))];

                if ((v1 == k) && (v2 == k))
                    continue;

                vertex = this.v(opposite[c]);
                if (!distances.ContainsKey(vertex))
                {
                    distances[vertex] = Math.Min(v1 + 1, v2 + 1);
                    result.Add(vertex);
                }
                visitedTriangles.Add(t(opposite[c]));

                workQueue.Push(n(opposite[c]));
                workQueue.Push(p(opposite[c]));
            }
            return (result.ToArray());
        }

        /// <summary>
        /// Returns a list of neighbouring vertices in given euclidean distance.
        /// </summary>
        /// <param name="v">Vertex index</param>
        /// <param name="e">Max euclidean distance</param>
        /// <returns>List of neighbouring vertices</returns>
        public int[] VVE(int v, double e)
        {
            double e2 = e * e;
            Point3D center = mesh.Points[v];
            Dictionary<int, double> distances = new Dictionary<int, double>();
            HashSet<int> visitedTriangles = new HashSet<int>();
            HashSet<int> result = new HashSet<int>();
            result.Add(v);
            int start = incidentCorner[v];
            if (start < 0)
                return (null);
            Stack<int> workQueue = new Stack<int>();
            int vertex = this.v(n(start));            
            distances[vertex] = center.DistanceTo2(ref mesh.Points[vertex]);
            if (distances[vertex]<e2)
                result.Add(vertex);
            visitedTriangles.Add(t(start));
            workQueue.Push(start);
            int current = -1;
            int op = opposite[p(start)];
            if (op >= 0)
                current = p(op);
            while ((current != start) && (current >= 0))
            {
                vertex = this.v(n(current));                
                distances[vertex] = center.DistanceTo2(ref mesh.Points[vertex]);
                if (distances[vertex]<e2)
                    result.Add(vertex);
                visitedTriangles.Add(t(current));
                workQueue.Push(current);
                op = opposite[p(current)];
                if (op < 0)
                    current = -1;
                else
                    current = p(op);
            }

            // treatment of border. Only two border edges per vertex allowed!
            if (current < 0)
            {
                vertex = this.v(p(start));                
                distances[vertex] = center.DistanceTo2(ref mesh.Points[vertex]);
                if (distances[vertex]<e2)
                    result.Add(vertex);
                op = opposite[n(start)];
                if (op >= 0)
                    current = n(op);
                while (current >= 0)
                {
                    vertex = this.v(p(current));                    
                    distances[vertex] = center.DistanceTo2(ref mesh.Points[vertex]);
                    if (distances[vertex]<e2)
                        result.Add(vertex);
                    visitedTriangles.Add(t(current));
                    workQueue.Push(current);
                    op = opposite[n(current)];
                    if (op < 0)
                        current = -1;
                    else
                        current = n(op);
                }
            }

            while (workQueue.Count > 0)
            {
                int c = workQueue.Pop();
                if (opposite[c] < 0)
                    continue;

                if (visitedTriangles.Contains(t(opposite[c])))
                    continue;

                double v1 = distances[this.v(p(c))];
                double v2 = distances[this.v(n(c))];

                if ((v1 >= e2) && (v2 >= e2))
                    continue;

                vertex = this.v(opposite[c]);
                if (!distances.ContainsKey(vertex))
                {
                    distances[vertex] = center.DistanceTo2(ref mesh.Points[vertex]);
                    if (distances[vertex]<e2)
                        result.Add(vertex);
                }
                visitedTriangles.Add(t(opposite[c]));

                workQueue.Push(n(opposite[c]));
                workQueue.Push(p(opposite[c]));
            }
            return (result.ToArray());
        }

        /// <summary>
        /// Returns a list of neighbouring edges in given radius.
        /// </summary>
        /// <param name="v">Vertex index</param>
        /// <param name="r">Radius</param>
        /// <param name="visitedTriangles">Output triangles, with at least one vertex inside the radius</param>
        /// <param name="distances">Output squared distances of vertices of visitedTriangles (can be greater than r*r)</param>
        /// <returns>Dictionary of corners corresponding to edges, with attached boolean: TRUE - edge is fully inside the radius, FALSE - edge is partially outside</returns>
        public Dictionary<int, bool> VEE(int v, double r, HashSet<int> visitedTriangles, Dictionary<int, double> distances)
        {
            if (visitedTriangles == null)
                visitedTriangles = new HashSet<int>();

            double rsq = r * r;
            Point3D centerVertex = mesh.Points[v];

            if (distances == null)
                distances = new Dictionary<int, double>();
            Dictionary<int, bool> resultingEdges = new Dictionary<int, bool>();
            int start = incidentCorner[v];
            if (start < 0)
                return (null);
            Stack<int> workQueue = new Stack<int>();

            int vertex = this.v(n(start));
            distances[vertex] = (mesh.Points[vertex] - centerVertex).Abs2();

            visitedTriangles.Add(t(start));
            workQueue.Push(start);
            int current = -1;
            int op = opposite[p(start)];
            if (op >=0)
                current = p(op);
            else
            {
                vertex = this.v(n(start));
                distances[vertex] = (mesh.Points[vertex] - centerVertex).Abs2();
                if (distances[vertex] < rsq)
                    resultingEdges.Add(p(start), false);
                else
                    resultingEdges.Add(p(start), true);
            }
            while ((current != start) && (current >= 0))
            {
                vertex = this.v(n(current));
                distances[vertex] = (mesh.Points[vertex] - centerVertex).Abs2();
                vertex = this.v(p(current));
                if (distances[vertex] < rsq)
                    resultingEdges.Add(n(current), false);
                else
                    resultingEdges.Add(n(current), true);
                
                visitedTriangles.Add(t(current));
                workQueue.Push(current);
                op = opposite[p(current)];
                if (op < 0)
                {
                    vertex = this.v(n(current));
                    if (distances[vertex] < rsq)
                        resultingEdges.Add(p(current), false);
                    else
                        resultingEdges.Add(p(current), true);
                    current = -1;
                }
                else
                    current = p(op);
            }

            // treatment of border. Only two border edges per vertex allowed!
            if (current < 0)
            {
                vertex = this.v(p(start));
                distances[vertex] = (mesh.Points[vertex] - centerVertex).Abs2();
                op = opposite[n(start)];
                if (op >= 0)
                    current = n(op);
                else
                {
                    vertex = this.v(p(start));
                    if (distances[vertex] < rsq)
                        resultingEdges.Add(n(start), false);
                    else
                        resultingEdges.Add(n(start), true);
                }
                while (current >= 0)
                {
                    vertex = this.v(p(current));
                    distances[vertex] = (mesh.Points[vertex] - centerVertex).Abs2();
                    vertex = this.v(n(current));
                    if (distances[vertex] < rsq)
                        resultingEdges.Add(p(current), false);
                    else
                        resultingEdges.Add(p(current), true);
                    
                    
                    visitedTriangles.Add(t(current));
                    workQueue.Push(current);
                    op = opposite[n(current)];
                    if (op < 0)
                    {
                        vertex = this.v(p(current));
                        distances[vertex] = (mesh.Points[vertex] - centerVertex).Abs2();
                        if (distances[vertex] < rsq)
                            resultingEdges.Add(n(current), false);
                        else
                            resultingEdges.Add(n(current), true);
                        current = -1;
                    }
                    else
                        current = n(op);
                }
            }
            else
            {
                vertex = this.v(p(start));
                
                if (distances[vertex] < rsq)
                    resultingEdges.Add(n(start), false);
                else
                    resultingEdges.Add(n(start), true);
            }

            while (workQueue.Count > 0)
            {
                int c = workQueue.Pop();

                if (opposite[c] < 0)
                {
                    double d1 = distances[this.v(p(c))];
                    double d2 = distances[this.v(n(c))];

                    if ((d1 > rsq) && (d2 > rsq))
                        continue;

                    if ((d1 > rsq) || (d2 > rsq))
                        resultingEdges.Add(c, true);
                    else
                        resultingEdges.Add(c, false);
                    continue;
                }

                if (visitedTriangles.Contains(t(opposite[c])))
                {
                    if (!resultingEdges.ContainsKey(opposite[c]))
                    {
                        double d1 = distances[this.v(p(c))];
                        double d2 = distances[this.v(n(c))];                        

                        if ((d1 > rsq) || (d2 > rsq))
                            resultingEdges.Add(c, true);
                        else
                            resultingEdges.Add(c, false);
                    }
                    continue;
                }



                int index = this.v(p(c));
                double v1 = double.MaxValue;
                if (distances.ContainsKey(index)) 
                {
                    v1 = distances[index];
                }
                
                index = this.v(n(c));
                double v2 = double.MaxValue;
                if (distances.ContainsKey(index))
                {
                    v2 = distances[index];
                }
                               
                if ((v1 >rsq) && (v2 > rsq))
                    continue;

                if ((v1 < rsq) && (v2 < rsq))
                    resultingEdges.Add(c, false);
                else
                    resultingEdges.Add(c, true);

                vertex = this.v(opposite[c]);
                distances[vertex] = (mesh.Points[vertex] - centerVertex).Abs2();
                visitedTriangles.Add(t(opposite[c]));

                workQueue.Push(n(opposite[c]));
                workQueue.Push(p(opposite[c]));
            }
            return (resultingEdges);
        }

        /// <summary>
        /// Returns a list of neighbouring edges in given radius.
        /// </summary>
        /// <param name="v">Vertex index</param>
        /// <param name="r">Radius</param>
        /// <param name="distances">Output squared distances of vertices of triangles (can be greater than r*r)</param>
        /// <returns>HashSet of all triangles up to distance specified by radius</returns>
        public HashSet<int> VFE(int v, double r, Dictionary<int, double> distances)
        {
            HashSet<int> visitedTriangles = new HashSet<int>();

            double rsq = r * r;
            Point3D centerVertex = mesh.Points[v];

            if (distances == null)
                distances = new Dictionary<int, double>();
            //TODO for clarity (in HildebrandtEstimator.getFaceParts) it is usefull 
            //to have center vertex in dict too, can I do it this way?
            distances[v] = 0;          
            int start = incidentCorner[v];
            if (start < 0)
                return (null);
            Stack<int> workQueue = new Stack<int>();

            int vertex = this.v(n(start));
            distances[vertex] = (mesh.Points[vertex] - centerVertex).Abs2();

            visitedTriangles.Add(t(start));
            workQueue.Push(start);
            int current = -1;
            int op = opposite[p(start)];
            if (op >= 0)
                current = p(op);
            else
            {
                vertex = this.v(n(start));
                distances[vertex] = (mesh.Points[vertex] - centerVertex).Abs2();                
            }
            while ((current != start) && (current >= 0))
            {
                vertex = this.v(n(current));
                distances[vertex] = (mesh.Points[vertex] - centerVertex).Abs2();
                vertex = this.v(p(current));                

                visitedTriangles.Add(t(current));
                workQueue.Push(current);
                op = opposite[p(current)];
                if (op < 0)
                {
                    vertex = this.v(n(current));                    
                    current = -1;
                }
                else
                    current = p(op);
            }

            // treatment of border. Only two border edges per vertex allowed!
            if (current < 0)
            {
                vertex = this.v(p(start));
                distances[vertex] = (mesh.Points[vertex] - centerVertex).Abs2();
                op = opposite[n(start)];
                if (op >= 0)
                    current = n(op);
                else
                {
                    vertex = this.v(p(start));
                }
                while (current >= 0)
                {
                    vertex = this.v(p(current));
                    distances[vertex] = (mesh.Points[vertex] - centerVertex).Abs2();
                    vertex = this.v(n(current));

                    visitedTriangles.Add(t(current));
                    workQueue.Push(current);
                    op = opposite[n(current)];
                    if (op < 0)
                    {
                        vertex = this.v(p(current));
                        distances[vertex] = (mesh.Points[vertex] - centerVertex).Abs2();
                        current = -1;
                    }
                    else
                        current = n(op);
                }
            }
            else
                vertex = this.v(p(start));

            while (workQueue.Count > 0)
            {
                int c = workQueue.Pop();

                if (opposite[c] < 0)
                    continue;

                if (visitedTriangles.Contains(t(opposite[c])))
                    continue;

                int index = this.v(p(c));
                double v1 = double.MaxValue;
                if (distances.ContainsKey(index))
                {
                    v1 = distances[index];
                }

                index = this.v(n(c));
                double v2 = double.MaxValue;
                if (distances.ContainsKey(index))
                {
                    v2 = distances[index];
                }

                if ((v1 > rsq) && (v2 > rsq))
                    continue;

                vertex = this.v(opposite[c]);
                distances[vertex] = (mesh.Points[vertex] - centerVertex).Abs2();
                visitedTriangles.Add(t(opposite[c]));

                workQueue.Push(n(opposite[c]));
                workQueue.Push(p(opposite[c]));
            }
            return (visitedTriangles);
        }
    }
}
