﻿using System.Collections.Generic;
using Zcu.CurvatureBenchmark.Framework.Estimators;

namespace Zcu.CurvatureBenchmark.Framework.Evaluators
{
    /// <summary>
    /// Represents a module that evaluates the estimated the principal curvatures.
    /// </summary>
    /// <remarks>
    /// As many estimators use some neighborhood, meshes with borders can contain imprecise or invalid
    /// values close to the borders. Therefore the evaluators should call 
    /// <see cref="SetExclusionZone(TriangleMesh)"/> or <see cref="TriangleMesh.excludeVertices(bool, bool, int, double)"/> 
    /// before the evaluation.
    /// </remarks>
    public interface ICurvatureEvaluator : IModule
    {
        /// <summary>
        /// Gets/sets whether to exclude the some border area based on topology distance <see cref="TopologicalExclusionZone"/> (for meshes with borders).
        /// </summary>
        bool ExcludeBorderTopological { get; set; }

        /// <summary>
        /// Gets/sets whether to exclude the some border area based on geometry distance <see cref="GeometricExclusionZone"/> (for meshes with borders).
        /// </summary>
        bool ExcludeBorderGeometric { get; set; }

        /// <summary>
        /// Gets/sets the topological size of exclusion zone.
        /// </summary>
        int TopologicalExclusionZone { get; set; }

        /// <summary>
        /// Gets/sets the geometric size of exclusion zone.
        /// </summary>
        double GeometricExclusionZone { get; set; }

        /// <summary>
        /// Computes the exclusion zone.
        /// </summary>
        /// <param name="mesh">Triangle mesh</param>
        void SetExclusionZone(TriangleMesh mesh);

        /// <summary>
        /// Evaluates the estimated curvatures.
        /// </summary>
        /// <param name="mesh">Mesh</param>
        /// <param name="estimator">Estimator</param>
        /// <returns>Key-value pairs for log output</returns>
        Dictionary<string, object> Evaluate(TriangleMesh mesh, ICurvatureEstimator estimator);
    }
}
