﻿using System;
using System.Xml;

namespace Zcu.CurvatureBenchmark.Framework
{
    /// <summary>
    /// <c>IModule</c> is a base interface for all modules of all types. Each framework module (data source,
    /// estimator, evaluator or distorter) must implement this interface to be recognized by the framework.
    /// </summary>
    /// <remarks>
    /// To save your time, there is an abstract class <see cref="AbstractModule"/> implementing this 
    /// interface and a set of <c>Abstract*</c> classes for sources, distorters...
    /// </remarks>
    public interface IModule
    {
        /// <summary>
        /// Setup module parameters from config xml.
        /// </summary>
        /// <param name="configNode">Part of xml configuration.</param>
        void SetConfig(XmlNode configNode);

        /// <summary>
        /// Checks all prerequisities (usually used for wrapper modules).
        /// </summary>
        /// <returns>TRUE if everything is OK.</returns>
        bool CheckPrerequisities();

        /// <summary>
        /// Progress 0 to 100. The progress property should be used especially for time consuming modules.
        /// </summary>
        byte Progress { get; }

		/// <summary>
		/// Gets/sets whether the module will be used during the processing.
		/// </summary>
		bool Active { get; set; }

        /// <summary>
        /// Progress changed reporter should occur on Progress property change.
        /// </summary>
        event NotifyProgressChanged ProgressChanged;

        /// <summary>
        /// Formatting string for ToString method. It can contain $Property$, which is replaced by the value of corresponding Property.
        /// </summary>
        string ToStringFormat { get; set; }

        /// <summary>
        /// Should return class name.
        /// </summary>
        string Name { get; }

        /// <summary>
        /// Should return fully qualified class name.
        /// </summary>
        string FullName { get; }
    }
   
    /// <summary>
    /// Progress changed reporter
    /// </summary>
    /// <param name="module">Reporting module</param>
    /// <param name="args">Current progress in %</param>
    public delegate void NotifyProgressChanged(IModule module, NotifyProgressChangedEventArgs args);

    /// <summary>
    /// Progress changed reporter argument
    /// </summary>
    public class NotifyProgressChangedEventArgs : EventArgs
    {
        /// <summary>
        /// Current progress (from 0 to 100 %) 
        /// </summary>
        public byte Progress { get; set; }

        /// <summary>
        /// Create event argument with progress.
        /// </summary>
        /// <param name="progress">Current progress</param>
        public NotifyProgressChangedEventArgs(byte progress) { Progress = progress; }
    }

}
