﻿using System;
using System.Collections.Generic;
using System.IO;

namespace Zcu.CurvatureBenchmark.Framework
{
    /// <summary>
    /// Logging system allows to log all information about running test into a separate directory.
    /// </summary>
    /// <remarks>
    /// The following information is stored:
    /// <list type="bullet">
    /// <item>log.txt - SourceType, Source,	MeshIndex, DataIndex, Distorter, Estimator, AvgK1e, AvgK2e, DevK1, DevK2 for each measurement</item>
    /// <item>stats.txt - statistics of EdgeLengths, InnerAngles, DihedralAngles, DihedralAnglesSigned, SolidAngles, VertexDegrees, CircleRatios, MVLaplacians, LaplacianDifferences, DoubleLaplacians for each mesh</item>
    /// <item>summary.txt - summary of avgK1, K2 for all estimators</item>
    /// <item>[config].xml - copy of xml config file</item>
    /// </list>
    /// </remarks>
    public class Log
    {
        List<Dictionary<string, object>> log;
        Dictionary<string, object> currentLine;
        string fileName;
        string summaryFileName;
        string dir;
        Dictionary<string, object>.KeyCollection allKeys;
        /// <summary>
        /// Gets the stats.txt filename and path
        /// </summary>
        public string StatsFileName { get; private set; }

        /// <summary>
        /// Creates directory and prepares Log class.
        /// </summary>
        /// <param name="configFileName">Name of current test config file</param>
        public Log(string configFileName)
        {
            this.log = new List<Dictionary<string, object>>();
            log.Add(new Dictionary<string, object>());
            currentLine = log[0];

            dir = DateTime.Now.ToString("yyMMddHHmmss");
            Directory.CreateDirectory(dir);

            fileName = dir + "\\" + "log.txt";
            StatsFileName = dir + "\\" + "stats.txt";
            summaryFileName = dir + "\\" + "summary.txt";
            TriangleMeshStat.DumpStatsHeaderToFile(StatsFileName);
            string cfn = Path.GetFileNameWithoutExtension(configFileName) + ".xml";
            File.Copy(configFileName, dir + "\\" + cfn);

            //File.Create(fileName);
        }

        /// <summary>
        /// Adds new line to log file.
        /// </summary>
        public void NewLine()
        {
            // copy values that have not been set from the previous line
            if (log.Count>1)
            {
                foreach (string key in log[log.Count - 2].Keys)
                    if (!currentLine.ContainsKey(key))
                        currentLine[key] = log[log.Count - 2][key];
            }
            else
            {
                // this is the first line, let's also dump the column headers
                allKeys = currentLine.Keys;
                string headers = "";
                foreach (string key in currentLine.Keys)
                    headers += key + "\t";

                StreamWriter sw = new StreamWriter(new FileStream(fileName, FileMode.Create));                
                sw.WriteLine(headers);
                sw.Close();
            }

            // dump line to log file
            dumpLine(currentLine);

            // create a new line
            currentLine = new Dictionary<string, object>();
            log.Add(currentLine);
        }

        private void dumpLine(Dictionary<string, object> logLine)
        {
            string line = "";
            foreach (string key in allKeys)
                line += logLine[key].ToString() + "\t";
            StreamWriter sw = new StreamWriter(new FileStream(fileName, FileMode.Append));            
            sw.WriteLine(line);
            sw.Close();
        }

        /// <summary>
        /// Adds a key-value pair to current line.
        /// </summary>
        /// <param name="key">Key</param>
        /// <param name="value">Value</param>
        public void LogValue(string key, object value)
        {
            currentLine[key] = value;
        }

        /// <summary>
        /// Adds key-value pairs to current line.
        /// </summary>
        /// <param name="keyvalues">Dictionary of key-value pairs</param>
        public void LogValues(Dictionary<string,object> keyvalues)
        {
            foreach(var kv in keyvalues)
            {
                currentLine[kv.Key] = kv.Value;
            }
        }

        /// <summary>
        /// Writes summary file.
        /// </summary>
        public void WriteSummary()
        {
            if (log.Count < 1)
                return;
            if (!log[0].ContainsKey("AvgK1e"))
                return;
            if (!log[0].ContainsKey("AvgK2e"))
                return;
            if (log[log.Count - 1].Count == 0)
                log.RemoveAt(log.Count - 1);

            StreamWriter sw = new StreamWriter(new FileStream(summaryFileName, FileMode.Append));            

            List<string> dists = new List<string>();
            List<string> estimators = new List<string>();
            int maxSourceIndex = 0;
            foreach (var line in log)
            {
                if (!dists.Contains(line["Distorter"].ToString()))
                    dists.Add(line["Distorter"].ToString());
                if (((int)line["SourceIndex"]) > maxSourceIndex)
                    maxSourceIndex = (int)line["SourceIndex"];
                if (!estimators.Contains(line["Estimator"].ToString()))
                    estimators.Add(line["Estimator"].ToString());
            }
            string lin = "";
            foreach(string est in estimators)
            {
                lin += "\t" + est;
            }
            sw.WriteLine(lin);

            Dictionary<string, int> estimatorIndices = new Dictionary<string, int>();
            for (int i = 0; i < estimators.Count; i++)
                estimatorIndices[estimators[i]] = i;
            

            foreach(string dist in dists)
            {
                sw.WriteLine(dist);

                for (int i = 0; i < maxSourceIndex + 1; i++)
                {

                    double[] averages = new double[estimators.Count];
                    int[] counts = new int[estimators.Count];


                    foreach (var line in log)
                    {
                        if (line["Distorter"].ToString() == dist)
                        {

                            if (((int)line["SourceIndex"]) == i)
                            {
                                double avg = (double)line["AvgK1e"] + (double)line["AvgK2e"];
                                averages[estimatorIndices[line["Estimator"].ToString()]] += avg;
                                counts[estimatorIndices[line["Estimator"].ToString()]] += 1;
                            }
                        }
                    }

                    lin = "";
                    for (int j = 0; j < estimators.Count; j++)
                        lin += "\t" + (averages[j] / counts[j]).ToString();
                    sw.WriteLine(lin);
                }                 
            }
            sw.Close();
        }
    }
}
