﻿using System;

namespace Zcu.CurvatureBenchmark.Framework
{
    /// <summary>
    /// Simple class representing 4x4 matrix
    /// </summary>
    public class Matrix4x4
    {

        /// <summary>
        /// Matrix
        /// </summary>
        public double[,] M = new double[4, 4];

        /// <summary>
        /// Gets identity matrix.
        /// </summary>
        public static Matrix4x4 Identity 
        {
            get 
            {
                Matrix4x4 m = new Matrix4x4();
                m.M[0, 0] = 1;
                m.M[1, 1] = 1;
                m.M[2, 2] = 1;
                m.M[3, 3] = 1;
                return m;
            }
        }

       
        /// <summary>
        /// Creates rotation matrix.
        /// </summary>
        /// <param name="axis">Axis of rotation</param>
        /// <param name="angle">Angle</param>
        /// <returns>Rotation matrix</returns>
        public static Matrix4x4 Rotation(Point3D axis, double angle)
        {
            double x = axis.X;
            double y = axis.Y;
            double z = axis.Z;

            Matrix4x4 result = new Matrix4x4();

            double cosinus = Math.Cos(angle);
            double sinus = Math.Sin(angle);
            double xx = x * x;
            double yy = y * y;
            double zz = z * z;
            double xy = x * y;
            double xz = x * z;
            double yz = y * z;

            result.M[0, 0] = xx + (cosinus * (1.0 - xx));
            result.M[0, 1] = (xy - (cosinus * xy)) + (sinus * z);
            result.M[0, 2] = (xz - (cosinus * xz)) - (sinus * y);
            result.M[0, 3] = 0.0;
            result.M[1, 0] = (xy - (cosinus * xy)) - (sinus * z);
            result.M[1, 1] = yy + (cosinus * (1.0f - yy));
            result.M[1, 2] = (yz - (cosinus * yz)) + (sinus * x);
            result.M[1, 3] = 0.0;
            result.M[2, 0] = (xz - (cosinus * xz)) + (sinus * y);
            result.M[2, 1] = (yz - (cosinus * yz)) - (sinus * x);
            result.M[2, 2] = zz + (cosinus * (1.0 - zz));
            result.M[2, 3] = 0.0;
            result.M[3, 0] = 0.0;
            result.M[3, 1] = 0.0;
            result.M[3, 2] = 0.0;
            result.M[3, 3] = 1.0;

            return result;

        }

        /// <summary>
        /// Creates translation matrix.
        /// </summary>
        /// <param name="translation">Translation</param>
        /// <returns>Translation matrix</returns>
        public static Matrix4x4 Translation(Point3D translation)
        {
            Matrix4x4 result = Matrix4x4.Identity;
		    result.M[3, 0] = translation.X;
            result.M[3, 1] = translation.Y;
            result.M[3, 2] = translation.Z;
		    return result;

        }

        /// <summary>
        /// Matrix multiplication
        /// </summary>
        /// <param name="left">Left matrix</param>
        /// <param name="right">Right matrix</param>
        /// <returns>Multiplied matrix</returns>
        public static Matrix4x4 operator *(Matrix4x4 left, Matrix4x4 right)
		{
			Matrix4x4 result = new Matrix4x4();

			result.M[0, 0] = (left.M[0, 0] * right.M[0, 0]) + (left.M[0, 1] * right.M[1, 0]) + (left.M[0, 2] * right.M[2, 0]) + (left.M[0, 3] * right.M[3, 0]);
			result.M[0, 1] = (left.M[0, 0] * right.M[0, 1]) + (left.M[0, 1] * right.M[1, 1]) + (left.M[0, 2] * right.M[2, 1]) + (left.M[0, 3] * right.M[3, 1]);
			result.M[0, 2] = (left.M[0, 0] * right.M[0, 2]) + (left.M[0, 1] * right.M[1, 2]) + (left.M[0, 2] * right.M[2, 2]) + (left.M[0, 3] * right.M[3, 2]);
			result.M[0, 3] = (left.M[0, 0] * right.M[0, 3]) + (left.M[0, 1] * right.M[1, 3]) + (left.M[0, 2] * right.M[2, 3]) + (left.M[0, 3] * right.M[3, 3]);
			result.M[1, 0] = (left.M[1, 0] * right.M[0, 0]) + (left.M[1, 1] * right.M[1, 0]) + (left.M[1, 2] * right.M[2, 0]) + (left.M[1, 3] * right.M[3, 0]);
			result.M[1, 1] = (left.M[1, 0] * right.M[0, 1]) + (left.M[1, 1] * right.M[1, 1]) + (left.M[1, 2] * right.M[2, 1]) + (left.M[1, 3] * right.M[3, 1]);
			result.M[1, 2] = (left.M[1, 0] * right.M[0, 2]) + (left.M[1, 1] * right.M[1, 2]) + (left.M[1, 2] * right.M[2, 2]) + (left.M[1, 3] * right.M[3, 2]);
			result.M[1, 3] = (left.M[1, 0] * right.M[0, 3]) + (left.M[1, 1] * right.M[1, 3]) + (left.M[1, 2] * right.M[2, 3]) + (left.M[1, 3] * right.M[3, 3]);
			result.M[2, 0] = (left.M[2, 0] * right.M[0, 0]) + (left.M[2, 1] * right.M[1, 0]) + (left.M[2, 2] * right.M[2, 0]) + (left.M[2, 3] * right.M[3, 0]);
			result.M[2, 1] = (left.M[2, 0] * right.M[0, 1]) + (left.M[2, 1] * right.M[1, 1]) + (left.M[2, 2] * right.M[2, 1]) + (left.M[2, 3] * right.M[3, 1]);
			result.M[2, 2] = (left.M[2, 0] * right.M[0, 2]) + (left.M[2, 1] * right.M[1, 2]) + (left.M[2, 2] * right.M[2, 2]) + (left.M[2, 3] * right.M[3, 2]);
			result.M[2, 3] = (left.M[2, 0] * right.M[0, 3]) + (left.M[2, 1] * right.M[1, 3]) + (left.M[2, 2] * right.M[2, 3]) + (left.M[2, 3] * right.M[3, 3]);
			result.M[3, 0] = (left.M[3, 0] * right.M[0, 0]) + (left.M[3, 1] * right.M[1, 0]) + (left.M[3, 2] * right.M[2, 0]) + (left.M[3, 3] * right.M[3, 0]);
			result.M[3, 1] = (left.M[3, 0] * right.M[0, 1]) + (left.M[3, 1] * right.M[1, 1]) + (left.M[3, 2] * right.M[2, 1]) + (left.M[3, 3] * right.M[3, 1]);
			result.M[3, 2] = (left.M[3, 0] * right.M[0, 2]) + (left.M[3, 1] * right.M[1, 2]) + (left.M[3, 2] * right.M[2, 2]) + (left.M[3, 3] * right.M[3, 2]);
			result.M[3, 3] = (left.M[3, 0] * right.M[0, 3]) + (left.M[3, 1] * right.M[1, 3]) + (left.M[3, 2] * right.M[2, 3]) + (left.M[3, 3] * right.M[3, 3]);


			return result;
		}
    }
}
