﻿using System;
using System.Diagnostics;
using Zcu.CurvatureBenchmark.Framework.Estimators;
using Zcu.CurvatureBenchmark.Framework.DataSources;

namespace Zcu.CurvatureBenchmark.Framework
{

    /// <summary>
    /// Delegate for output event.
    /// </summary>
    /// <param name="args">Output event arguments</param>
    public delegate void OnOutputEvent(OutputEventArguments args);

    /// <summary>
    /// Describes output event.
    /// </summary>
    public class OutputEventArguments
    {
        /// <summary>
        /// Gets/sets output type
        /// </summary>
        public OutputType Type { get; set; }
        /// <summary>
        /// Gets/sets output message
        /// </summary>
        public string Message { get; set; }
        /// <summary>
        /// Gets/sets module that produces this output
        /// </summary>
        public IModule Module { get; set; }
    }

    /// <summary>
    /// Output types
    /// </summary>
    public enum OutputType
    {
        /// <summary>
        /// Informative output
        /// </summary>
        Info,
        /// <summary>
        /// Debug output (used only when DEBUG is set)
        /// </summary>
        Debug,
        /// <summary>
        /// Warnings
        /// </summary>
        Warning,
        /// <summary>
        /// Errors
        /// </summary>
        Error,
        /// <summary>
        /// Critical errors
        /// </summary>
        Critical,
        /// <summary>
        /// Successes
        /// </summary>
        Success,
        /// <summary>
        /// Module start
        /// </summary>
        Start
    }

    /// <summary>
    /// Output class is a set of extension methods used to customize all outputs. 
    /// </summary>
    /// <remarks>
    /// Each module can produce 
    /// varoius outputs types by calling corresponding method. Application can register each of these 
    /// output reporters and treat them separately.
    /// <para>
    /// All modules should use this set of extensions for text outputs (instead of Console and Debug).
    /// Using these extensions allows to format, collate and filter various messages.
    /// </para>
    /// <para>
    /// The usage of these extensions is very similar to Console.Write(format, params...), but it extends IMoudle
    /// itself. 
    /// </para>
    /// <para>
    /// To get the reports, one has to register the listeners. This can be done maually (eg. <c>Output.OnError += MyErrorOutput</c>
    /// or by calling <c>Output.SetDefault()</c> method which provides simple colored console output.
    /// </para>
    /// </remarks>
    /// <example>
    /// In any IModule it is possible to call Output extensions:
    /// <code>
    /// this.Critical("Non-manifold triangles at vertex {0} detected!", nonManifoldVertex);
    /// </code>
    /// </example>
    public static class Output
    {   
        /// <summary>
        /// Set default event reporter - color console
        /// </summary>
        public static void SetDefault()
        {
            OnError += DefaultOutput;
            OnWarning += DefaultOutput;
            OnInfo += DefaultOutput;
            OnCritical += DefaultOutput;
            OnSuccess += DefaultOutput;
            OnStart += DefaultOutput;
            OnDebug += DefaultOutput;
        }

        /// <summary>
        /// Used for system success (loading module)
        /// </summary>
        public static event OnOutputEvent OnSuccess;

        /// <summary>
        /// Used for critical errors - ie. errors leading to unstable or corrupted results or exceptions.
        /// </summary>
        public static event OnOutputEvent OnCritical;

        /// <summary>
        /// Used for errors that are somehow recoverable
        /// </summary>
        public static event OnOutputEvent OnError;

        /// <summary>
        /// Used for warning (eg. NaN values)
        /// </summary>
        public static event OnOutputEvent OnWarning;

        /// <summary>
        /// Important info
        /// </summary>
        public static event OnOutputEvent OnInfo;

        /// <summary>
        /// Debug event - this event is omitted in Release
        /// </summary>
        public static event OnOutputEvent OnDebug;


        /// <summary>
        /// Module start event
        /// </summary>
        public static event OnOutputEvent OnStart;

        /// <summary>
        /// Generic event processor.
        /// </summary>
        /// <typeparam name="T">IModule</typeparam>
        /// <param name="module">Module</param>
        /// <param name="eventListener"></param>
        /// <param name="type"></param>
        /// <param name="format"></param>
        /// <param name="param"></param>
        static void Event<T>(this T module, OnOutputEvent eventListener, OutputType type, string format, params object[] param) where T : IModule
        {
            var listener = eventListener;
            if (listener != null)
            {
                listener(new OutputEventArguments() { Type = type, Message = string.Format(format, param), Module = module });
            }
        }

        /// <summary>
        /// Call to report important success (eg. module loaded).
        /// </summary>
        /// <typeparam name="T">IModule</typeparam>
        /// <param name="module">Module</param>
        /// <param name="format">Message format string</param>
        /// <param name="param">Message parameters</param>
        public static void Success<T>(this T module, string format, params object[] param) where T : IModule
        {
            module.Event(OnSuccess, OutputType.Success, format, param);
        }


        /// <summary>
        /// Call to report critical error that leads to unstable or corrupted results or exceptions. 
        /// </summary>
        /// <typeparam name="T">IModule</typeparam>
        /// <param name="module">Module</param>
        /// <param name="format">Message format string</param>
        /// <param name="param">Message parameters</param>
        public static void Critical<T>(this T module, string format, params object[] param) where T : IModule
        {
            module.Event(OnCritical, OutputType.Critical, format, param);
        }

        /// <summary>
        /// Call to report error that is recoverable.
        /// </summary>
        /// <typeparam name="T">IModule</typeparam>
        /// <param name="module">Module</param>
        /// <param name="format">Message format string</param>
        /// <param name="param">Message parameters</param>
        public static void Error<T>(this T module, string format, params object[] param) where T : IModule
        {
            module.Event(OnError, OutputType.Error, format, param);
        }

        /// <summary>
        /// Call to report warning on some unexpected behavior (eg. NaN).
        /// </summary>
        /// <typeparam name="T">IModule</typeparam>
        /// <param name="module">Module</param>
        /// <param name="format">Message format string</param>
        /// <param name="param">Message parameters</param>
        public static void Warning<T>(this T module, string format, params object[] param) where T : IModule
        {
            module.Event(OnWarning, OutputType.Warning, format, param);
        }

        /// <summary>
        /// Use for debug purpose. Calls are omitted in Release.
        /// </summary>
        /// <typeparam name="T">IModule</typeparam>
        /// <param name="module">Module</param>
        /// <param name="format">Message format string</param>
        /// <param name="param">Message parameters</param>
        [Conditional("DEBUG")]
        public static void Debug<T>(this T module, string format, params object[] param) where T : IModule
        {
            module.Event(OnDebug, OutputType.Debug, format, param);
        }

        /// <summary>
        /// Call to report some important information. Do not use to often.
        /// </summary>
        /// <typeparam name="T">IModule</typeparam>
        /// <param name="module">Module</param>
        /// <param name="format">Message format string</param>
        /// <param name="param">Message parameters</param>
        public static void Info<T>(this T module, string format, params object[] param) where T : IModule
        {
            module.Event(OnInfo, OutputType.Info, format, param);
        }

        /// <summary>
        /// Call to report the start of module. Do not use directly. This methods is called by Start* function.
        /// </summary>
        /// <typeparam name="T">IModule</typeparam>
        /// <param name="module">Module</param>
        /// <param name="format">Message format string</param>
        /// <param name="param">Message parameters</param>
        public static void Start<T>(this T module, string format, params object[] param) where T : IModule
        {
            module.Event(OnStart, OutputType.Start, format, param);
        }

        /// <summary>
        /// Default reporters
        /// </summary>
        /// <param name="args"></param>
        static void DefaultOutput(OutputEventArguments args)
        {
            ConsoleColor color = ConsoleColor.Gray;
            switch (args.Type) 
            {
                case OutputType.Debug: color = ConsoleColor.DarkGray; break;
                case OutputType.Warning: color = ConsoleColor.Yellow; break;
                case OutputType.Error: color = ConsoleColor.Red; break;
                case OutputType.Critical: color = ConsoleColor.DarkRed; break;
                case OutputType.Success: color = ConsoleColor.Green; break;
                case OutputType.Start: color = ConsoleColor.Cyan; break;
                default: break;
            }

            if (Console.CursorLeft != 0) Console.WriteLine();

            ConsoleColor oldcolor = Console.ForegroundColor;
            Console.Write("[");
            Console.ForegroundColor = color;
            Console.Write(args.Type);
            Console.ForegroundColor = oldcolor;
            Console.Write("]");
            if (args.Module != null)
                Console.Write(" {0}:", args.Module.GetType());
            if (args.Message != null)
                Console.Write(" {0}", args.Message);
            Console.WriteLine();
        }

        /// <summary>
        /// Extension method to report start of DataSource module <see cref="IDataSource.GenerateData"/>
        /// </summary>
        /// <typeparam name="T">IModule</typeparam>
        /// <param name="module">Module</param>
        /// <returns>Set of triangle meshes</returns>
        public static TriangleMesh[] StartGenerateData<T>(this T module) where T : IDataSource
        {
            module.Start("");
            return module.GenerateData();
        }

        /// <summary>
        /// Extension method to report start of Estimator module <see cref="ICurvatureEstimator.EstimateCurvatures(TriangleMesh)"/>
        /// </summary>
        /// <typeparam name="T">IModule</typeparam>
        /// <param name="module">Module</param>
        /// <param name="mesh">Mesh</param>
        public static void StartEstimateCurvatures<T>(this T module, TriangleMesh mesh) where T : ICurvatureEstimator
        {
            module.Start("");
            module.EstimateCurvatures(mesh);
        }
    }
}
