﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace Zcu.CurvatureBenchmark.Framework
{
    /// <summary>
    /// Plane structure. A plane is described with general form x*Normal.x + y*Normal.y + z*Normal.z + d = 0.
    /// </summary>
    public struct Plane
    {
        private Point3D n;
        private double d;

        /// <summary>
        /// Gets normal vector.
        /// </summary>
        public Point3D Normal { get { return this.n; } }
        /// <summary>
        /// Gets d.
        /// </summary>
        public double D { get { return this.d; } }


        /// <summary>
        /// Creates plane directly.
        /// </summary>
        /// <param name="A">Normal.x</param>
        /// <param name="B">Normal.y</param>
        /// <param name="C">Normal.z</param>
        /// <param name="D">d</param>
        public Plane(double A, double B, double C, double D) : this()
        {
	        this.n.X = A;
	        this.n.Y = B;
	        this.n.Z = C;
	        this.d = D;
	        this.Normalize();
        }

        /// <summary>
        /// Creates plane directly.
        /// </summary>
        /// <param name="normal">Normal</param>
        /// <param name="d">d</param>
        public Plane(ref Point3D normal, double d)
        {
	        this.n = normal;	
	        this.d = d;
	        this.Normalize();
        }

        /// <summary>
        /// Creates plane using normal and a point in the plane.
        /// </summary>
        /// <param name="normal">Normal</param>
        /// <param name="point">Point in the plane</param>
        public Plane(ref Point3D normal, ref Point3D point)
        {
	        this.n = normal;	
	        this.d = -1 * Point3D.DotProduct(ref point, ref normal);

    	    this.Normalize();
        }

        /// <summary>
        /// Creates plane using 3 points
        /// </summary>
        /// <param name="P1">Point 1</param>
        /// <param name="P2">Point 2</param>
        /// <param name="P3">Point 3</param>
        public Plane(ref Point3D P1, ref Point3D P2, ref Point3D P3)
        {	
	        this.n = Point3D.CrossProduct(P2 - P1, P3 - P1);
	        this.d = -1 * Point3D.DotProduct(ref P1, ref this.n);

	        this.Normalize();
        }



        double Dot(ref Point3D point)
        {
	        return (this.Normal.X * point.X) + (this.Normal.Y * point.Y) +  (this.Normal.Z * point.Z) + (this.D);
        }

        /// <summary>
        /// Normalizes the plane normal and updates d. 
        /// </summary>
        public void Normalize()
        {
	        double magnitude = 1.0 / this.Normal.Abs();
	        this.n.X *= magnitude;
	        this.n.Y *= magnitude;
	        this.n.Z *= magnitude;

	        this.d *= magnitude;
        }

        /// <summary>
        /// Project point onto the plane
        /// </summary>
        /// <param name="point">Point to project</param>
        /// <returns>Projected point</returns>
        public Point3D ProjectPoint(Point3D point)
        {	
	        //Dot(dir, point) + plane.D => plane.Dot(point)
            return this.ProjectPoint(ref point);
	
        }

        /// <summary>
        /// Project point onto the plane
        /// </summary>
        /// <param name="point">Point to project</param>
        /// <returns>Projected point</returns>
        public Point3D ProjectPoint(ref Point3D point)
        {
            //Dot(dir, point) + plane.D => plane.Dot(point)
            double citatel = Point3D.DotProduct(ref this.n, ref point) + this.D;

            //no need to divide, because dir is normalized

            return new Point3D(point.X - this.Normal.X * citatel,
                           point.Y - this.Normal.Y * citatel,
                           point.Z - this.Normal.Z * citatel);

        }

        /// <summary>
        /// Computes intersection with ray.
        /// </summary>
        /// <param name="start">Ray starting point</param>
        /// <param name="dir">Ray direction</param>
        /// <param name="result">Outout intersection</param>
        /// <returns>False if not intersected</returns>
        public bool RayIntersection(ref Point3D start, ref Point3D dir, out Point3D result)
        {
	        // Create Intersection.

            result = new Point3D(0, 0, 0);

	        double cosAlpha = Point3D.DotProduct(ref this.n, ref dir);

	        if ((cosAlpha >= -Utils.EPSILON) && (cosAlpha <= Utils.EPSILON))
	        {
                

		        return false;
	        }

	        double t = (Point3D.DotProduct(ref this.n, ref start) + this.D) / cosAlpha;

		
	        result.X = start.X - dir.X * t;
	        result.Y = start.Y - dir.Y * t;
	        result.Z = start.Z - dir.Z * t;

	        return true;
   
        }


    }
}
