﻿using System;
using System.Collections.Generic;
using System.Text;

//using MathNet.Numerics.LinearAlgebra;

namespace Zcu.CurvatureBenchmark.Framework
{
    /// <summary>
    /// Represents point/vector in E3.
    /// </summary>
    public struct Point3D
    {        
        /// <summary>
        /// Gets/sets x coordinate
        /// </summary>
        public double X { get; set; }

        /// <summary>
        /// Gets/sets y coordinate
        /// </summary>
        public double Y { get; set; }


        /// <summary>
        /// Gets/sets z coordinate
        /// </summary>
        public double Z { get; set; }


        /// <summary>
        /// Coordinate indexer. Try to avoid the indexer due to performance hit.
        /// </summary>
        /// <param name="index">Coordinate index</param>
        /// <returns>Coordinate</returns>
        /// <exception cref="IndexOutOfRangeException">For index out of range</exception>
        [Obsolete("Try to avoid this indexer")]
        public double this[int index]
        {
            get
            {
                switch (index)
                {
                    case 0: return X;
                    case 1: return Y;
                    case 2: return Z;
                    default: throw new IndexOutOfRangeException(); 
                }                
            }
            set
            {
                switch (index)
                {
                    case 0: X = value; break;
                    case 1: Y = value; break;
                    case 2: Z = value; break;
                    default: throw new IndexOutOfRangeException();
                }
                
            }
        }

        /*public static implicit operator Vector<double>(Point3D p)
        {
            Vector<double> X = Vector<double>.Build.Dense(3);
            X[0] = p.X;
            X[1] = p.Y;
            X[2] = p.Z;
            return X;
        }*/


        /// <summary>
        /// Creates a point with x,y,z coordinates.
        /// </summary>
        /// <param name="x">X coordinate</param>
        /// <param name="y">Y coordinate</param>
        /// <param name="z">Z coordinate</param>
        public Point3D(double x, double y, double z) : this()
        {
            X = x;
            Y = y;
            Z = z;
        }

        /// <summary>
        /// Computes vector length
        /// </summary>
        /// <returns>Vector length</returns>
        public double Abs()
        {
            return Math.Sqrt(X * X + Y * Y + Z * Z);
        }

        /// <summary>
        /// Computes vector length
        /// </summary>
        /// <returns>Vector length</returns>
        public double Length()
        {
            return this.Abs();
        }

        /// <summary>
        /// Computes squared vector length
        /// </summary>
        /// <returns>Vector squared length</returns>
        public double Abs2()
        {
            return X * X + Y * Y + Z * Z;
        }

        /// <summary>
        /// Computes squared vector length
        /// </summary>
        /// <returns>Vector squared length</returns>
        public double LengthSquared()
        {
            return this.Abs2();
        }

        /// <summary>
        /// Computes normalized vector
        /// </summary>
        /// <returns>Copy of normalized vector</returns>
        public Point3D Normalize()
        {
            double length = this.Length();
            if (length == 0.0f)
            {
                return this;
            }

            double num = 1.0 / length;
            this.X *= num;
            this.Y *= num;
            this.Z *= num;

            return this;

        }

        /// <summary>
        /// Computes angle between two vectors.
        /// </summary>
        /// <param name="first">First vector</param>
        /// <param name="second">Second vector</param>
        /// <param name="normalize">Normalize dot product (default true)</param>
        /// <returns>Angle between vectors.</returns>
        public static double GetAngle(Point3D first, Point3D second, bool normalize = true)
        {
	        double tmp = Point3D.DotProduct(ref first, ref second);
    
            if(normalize)
        	    tmp /= (first.Length() * second.Length());

	        if (tmp > 1) tmp = 1;
	        else if (tmp < -1) tmp = -1;

	        return Math.Acos(tmp);

        }

        /// <summary>
        /// Transforms point with transformation matrix
        /// </summary>
        /// <param name="point">Point</param>
        /// <param name="transform">Transformation matrix</param>
        /// <returns>Transformed point</returns>
        public static Point3D TransformCoordinate(Point3D point, Matrix4x4 transform)
        {

	        double tX = (((point.X * transform.M[0, 0]) + (point.Y * transform.M[1, 0])) + (point.Z * transform.M[2, 0])) + transform.M[3, 0];
            double tY = (((point.X * transform.M[0, 1]) + (point.Y * transform.M[1, 1])) + (point.Z * transform.M[2, 1])) + transform.M[3, 1];
            double tZ = (((point.X * transform.M[0, 2]) + (point.Y * transform.M[1, 2])) + (point.Z * transform.M[2, 2])) + transform.M[3, 2];
            double tW = (((point.X * transform.M[0, 3]) + (point.Y * transform.M[1, 3])) + (point.Z * transform.M[2, 3])) + transform.M[3, 3];

	        tW = 1.0f / tW;

            return new Point3D(tX * tW, tY * tW, tZ * tW);
        }


        #region Operators and operator functions


        #region Addition
        /// <summary>
        /// Addition of two points.
        /// </summary>
        /// <param name="a">First operand.</param>
        /// <param name="b">Second operand.</param>
        /// <returns>Resulting sum (a + b).</returns>
        public static Point3D operator +(Point3D a, Point3D b)
        {
            return (new Point3D(a.X + b.X, a.Y + b.Y, a.Z + b.Z));
        }

        /// <summary>
        /// Addition of two points.
        /// </summary>
        /// <param name="a">First operand.</param>
        /// <param name="b">Second operand.</param>
        /// <returns>Resulting sum (a + b).</returns>
        public static Point3D Add(Point3D a, Point3D b)
        {
            return (new Point3D(a.X + b.X, a.Y + b.Y, a.Z + b.Z));
        }

        /// <summary>
        /// Addition of two points.
        /// </summary>
        /// <param name="a">First operand.</param>
        /// <param name="b">Second operand.</param>
        /// <returns>Resulting sum (a + b).</returns>
        public static Point3D Add(ref Point3D a, ref Point3D b)
        {
            return (new Point3D(a.X + b.X, a.Y + b.Y, a.Z + b.Z));
        }

        /// <summary>
        /// Addition of two points.
        /// </summary>
        /// <param name="a">First operand.</param>
        /// <param name="b">Second operand.</param>
        /// <returns>Resulting sum (a + b).</returns>
        public static Point3D Add(Point3D a, ref Point3D b)
        {
            return (new Point3D(a.X + b.X, a.Y + b.Y, a.Z + b.Z));
        }

        /// <summary>
        /// Addition of two points.
        /// </summary>
        /// <param name="a">First operand.</param>
        /// <param name="b">Second operand.</param>
        /// <returns>Resulting sum (a + b).</returns>
        public static Point3D Add(ref Point3D a, Point3D b)
        {
            return (new Point3D(a.X + b.X, a.Y + b.Y, a.Z + b.Z));
        }

        /// <summary>
        /// Addition of point to current point.
        /// </summary>
        /// <param name="a">Second operand.</param>
        /// <returns>Resulting sum (this + a).</returns>
        public void Add(Point3D a)
        {
            this.X += a.X;
            this.Y += a.Y;
            this.Z += a.Z;
        }

        /// <summary>
        /// Addition of point to current point.
        /// </summary>
        /// <param name="a">Second operand.</param>
        /// <returns>Resulting sum (this + a).</returns>
        public void Add(ref Point3D a)
        {
            this.X += a.X;
            this.Y += a.Y;
            this.Z += a.Z;
        }
        #endregion

        #region Subtraction
        /// <summary>
        /// Subtraction of two points.
        /// </summary>
        /// <param name="a">First operand.</param>
        /// <param name="b">Second operand.</param>
        /// <returns>Resulting difference (a - b).</returns>
        public static Point3D operator -(Point3D a, Point3D b)
        {
            return (new Point3D(a.X - b.X, a.Y - b.Y, a.Z - b.Z));
        }

        /// <summary>
        /// Subtraction of two points.
        /// </summary>
        /// <param name="a">First operand.</param>
        /// <param name="b">Second operand.</param>
        /// <returns>Resulting difference (a - b).</returns>
        public static Point3D Subtract(Point3D a, Point3D b)
        {
            return (new Point3D(a.X - b.X, a.Y - b.Y, a.Z - b.Z));
        }

    
        /// <summary>
        /// Subtraction of two points.
        /// </summary>
        /// <param name="a">First operand.</param>
        /// <param name="b">Second operand.</param>
        /// <returns>Resulting difference (a - b).</returns>
        public static Point3D Subtract(ref Point3D a, Point3D b)
        {
            return (new Point3D(a.X - b.X, a.Y - b.Y, a.Z - b.Z));
        }

        /// <summary>
        /// Subtraction of two points.
        /// </summary>
        /// <param name="a">First operand.</param>
        /// <param name="b">Second operand.</param>
        /// <returns>Resulting difference (a - b).</returns>
        public static Point3D Subtract(Point3D a, ref Point3D b)
        {
            return (new Point3D(a.X - b.X, a.Y - b.Y, a.Z - b.Z));
        }

        /// <summary>
        /// Subtraction of two points.
        /// </summary>
        /// <param name="a">First operand.</param>
        /// <param name="b">Second operand.</param>
        /// <returns>Resulting difference (a - b).</returns>
        public static Point3D Subtract(ref Point3D a, ref Point3D b)
        {
            return (new Point3D(a.X - b.X, a.Y - b.Y, a.Z - b.Z));
        }

        /// <summary>
        /// Subtraction of point from current point.
        /// </summary>
        /// <param name="a">First operand.</param>
        public void Subtract(ref Point3D a)
        {
            this.X -= a.X;
            this.Y -= a.Y;
            this.Z -= a.Z;
        }

        /// <summary>
        /// Subtraction of point from current point.
        /// </summary>
        /// <param name="a">First operand.</param>
        public void Subtract(Point3D a)
        {
            this.X -= a.X;
            this.Y -= a.Y;
            this.Z -= a.Z;
        }
        /// <summary>
        /// Subtraction of point and double.
        /// </summary>
        /// <param name="a">First operand.</param>
        /// <param name="b">Second operand.</param>
        /// <returns>Resulting difference (a - b).</returns>
        public static Point3D operator -(Point3D a, double b)
        {
            return (new Point3D(a.X - b, a.Y - b, a.Z - b));
        }
        #endregion

        #region Sign inverse
        /// <summary>
        /// Sign inverse of a point.
        /// </summary>
        /// <param name="p">Operand.</param>
        /// <returns>Resulting point.</returns>
        public static Point3D operator -(Point3D p)
        {
            return (new Point3D(-p.X, -p.Y, -p.Z));
        }
        #endregion

        #region Multiplication
        /// <summary>
        /// Multiplication bY constant number.
        /// </summary>
        /// <param name="p">Point to be multiplied.</param>
        /// <param name="m">Multiplicator.</param>
        /// <returns>Resulting point (m*p.X, m*p.Y, m*p.Z)</returns>
        public static Point3D operator *(Point3D p, double m)
        {
            return (new Point3D(p.X * m, p.Y * m, p.Z * m));
        }

        /// <summary>
        /// Multiplication bY constant number.
        /// </summary>
        /// <param name="p">Point to be multiplied.</param>
        /// <param name="m">Multiplicator.</param>
        /// <returns>Resulting point (m*p.X, m*p.Y, m*p.Z)</returns>
        public static Point3D MultiplY(Point3D p, double m)
        {
            return (new Point3D(p.X * m, p.Y * m, p.Z * m));
        }

        /// <summary>
        /// Multiplication bY constant number.
        /// </summary>
        /// <param name="p">Point to be multiplied.</param>
        /// <param name="m">Multiplicator.</param>
        /// <returns>Resulting point (m*p.X, m*p.Y, m*p.Z)</returns>
        public static Point3D MultiplY(ref Point3D p, double m)
        {
            return (new Point3D(p.X * m, p.Y * m, p.Z * m));
        }

        /// <summary>
        /// Multiplication bY constant number.
        /// </summary>
        /// <param name="m">Multiplicator.</param>
        public void MultiplY(double m)
        {
            this.X *= m;
            this.Y *= m;
            this.Z *= m;
        }

        /// <summary>
        /// Multiplication bY a scalar.
        /// </summary>
        /// <param name="p">Point to be multiplied.</param>
        /// <param name="m">Multiplicator.</param>
        /// <returns>Resulting point (m*p.X, m*p.Y, m*p.Z).</returns>
        public static Point3D operator *(double m, Point3D p)
        {
            return (new Point3D(p.X * m, p.Y * m, p.Z * m));
        }
        #endregion

        #region Division
        /// <summary>
        /// Division bY a scalar.
        /// </summary>
        /// <param name="p">Point to be divided.</param>
        /// <param name="m">Divider.</param>
        /// <returns>Resulting point (p.X/m, p.Y/m, p.Z/m).</returns>
        public static Point3D operator /(Point3D p, double m)
        {
            return (new Point3D(p.X / m, p.Y / m, p.Z / m));
        }

        /// <summary>
        /// Division bY a scalar.
        /// </summary>
        /// <param name="p">Point to be divided.</param>
        /// <param name="m">Divider.</param>
        /// <returns>Resulting point (p.X/m, p.Y/m, p.Z/m).</returns>
        public static Point3D Divide(Point3D p, double m)
        {
            return (new Point3D(p.X / m, p.Y / m, p.Z / m));
        }

        /// <summary>
        /// Division bY a scalar.
        /// </summary>
        /// <param name="p">Point to be divided.</param>
        /// <param name="m">Divider.</param>
        /// <returns>Resulting point (p.X/m, p.Y/m, p.Z/m).</returns>
        public static Point3D Divide(ref Point3D p, double m)
        {
            return (new Point3D(p.X / m, p.Y / m, p.Z / m));
        }

        /// <summary>
        /// Division bY a scalar.
        /// </summary>
        /// <param name="m">Divider.</param>
        public void Divide(double m)
        {
            this.X /= m;
            this.Y /= m;
            this.Z /= m;
        }

        #endregion

        #endregion

        #region DistanceTo

        /// <summary>
        /// Returns euclidean distance between two points.
        /// </summary>
        /// <param name="p">Argument point to which distance will be computed.</param>
        /// <returns>Euclidean distance between the points.</returns>
        public double DistanceTo(ref Point3D p)
        {
            return (Math.Sqrt((p.X - this.X) * (p.X - this.X) + (p.Y - this.Y) * (p.Y - this.Y) + (p.Z - this.Z) * (p.Z - this.Z)));
        }

        /// <summary>
        /// Returns euclidean distance between two points.
        /// </summary>
        /// <param name="p">Argument point to which distance will be computed.</param>
        /// <returns>Euclidean distance between the points.</returns>
        public double DistanceTo(Point3D p)
        {
            return (Math.Sqrt((p.X - this.X) * (p.X - this.X) + (p.Y - this.Y) * (p.Y - this.Y) + (p.Z - this.Z) * (p.Z - this.Z)));
        }

        /// <summary>
        /// Returns euclidean distance between two points.
        /// </summary>
        /// <param name="p">Argument point to which distance will be computed.</param>
        /// <returns>Euclidean distance between the points.</returns>
        public double DistanceTo2(ref Point3D p)
        {
            return ((p.X - this.X) * (p.X - this.X) + (p.Y - this.Y) * (p.Y - this.Y) + (p.Z - this.Z) * (p.Z - this.Z));
        }

        /// <summary>
        /// Returns squared euclidean distance between two points
        /// (is slightlY faster than the DistanceTo function as no square root is computed),
        /// </summary>
        /// <param name="p">Argument point to which distance will be computed.</param>
        /// <returns>Square euclidean distance between the points.</returns>
        public double SquareDistanceTo(ref Point3D p)
        {
            return ((p.X - this.X) * (p.X - this.X) + (p.Y - this.Y) * (p.Y - this.Y) + (p.Z - this.Z) * (p.Z - this.Z));
        }

        /// <summary>
        /// Returns squared euclidean distance between two points
        /// (is slightlY faster than the DistanceTo function as no square root is computed),
        /// </summary>
        /// <param name="p">Argument point to which distance will be computed.</param>
        /// <returns>Square euclidean distance between the points.</returns>
        public double SquareDistanceTo(Point3D p)
        {
            return ((p.X - this.X) * (p.X - this.X) + (p.Y - this.Y) * (p.Y - this.Y) + (p.Z - this.Z) * (p.Z - this.Z));
        }
        #endregion

        #region DotProduct
        /// <summary>
        /// Returns the dot product of two points.
        /// </summary>
        /// <param name="arg">Second argument of the dot product operation.</param>
        /// <returns>Scalar value of this.arg.</returns>
        public double DotProduct(ref Point3D arg)
        {
            return (this.X * arg.X + this.Y * arg.Y + this.Z * arg.Z);
        }

        /// <summary>
        /// Returns the dot product of two points.
        /// </summary>
        /// <param name="arg">Second argument of the dot product operation.</param>
        /// <returns>Scalar value of this.arg.</returns>
        public double DotProduct(Point3D arg)
        {
            return (this.X * arg.X + this.Y * arg.Y + this.Z * arg.Z);
        }


        /// <summary>
        /// Returns the dot product of two points.
        /// </summary>
        /// <param name="p1">First operand.</param>
        /// <param name="p2">Second operand.</param>
        /// <returns>Dot product.</returns>
        public static double DotProduct(ref Point3D p1, ref Point3D p2)
        {
            return (p1.X * p2.X + p1.Y * p2.Y + p1.Z * p2.Z);
        }

        /// <summary>
        /// Returns the dot product of two points.
        /// </summary>
        /// <param name="p1">First operand.</param>
        /// <param name="p2">Second operand.</param>
        /// <returns>Dot product.</returns>
        public static double DotProduct(ref Point3D p1, Point3D p2)
        {
            return (p1.X * p2.X + p1.Y * p2.Y + p1.Z * p2.Z);
        }

        /// <summary>
        /// Returns the dot product of two points.
        /// </summary>
        /// <param name="p1">First operand.</param>
        /// <param name="p2">Second operand.</param>
        /// <returns>Dot product.</returns>
        public static double DotProduct(Point3D p1, ref Point3D p2)
        {
            return (p1.X * p2.X + p1.Y * p2.Y + p1.Z * p2.Z);
        }

        /// <summary>
        /// Returns the dot product of two points.
        /// </summary>
        /// <param name="p1">First operand.</param>
        /// <param name="p2">Second operand.</param>
        /// <returns>Dot product.</returns>
        public static double DotProduct(Point3D p1, Point3D p2)
        {
            return (p1.X * p2.X + p1.Y * p2.Y + p1.Z * p2.Z);
        }
        #endregion


        #region CrossProduct
        /// <summary>
        /// Returns the cross product of two points.
        /// </summary>
        /// <param name="arg">Second argument of the cross product operation.</param>
        /// <returns>Point3D representing the resulting vector.</returns>
        public Point3D CrossProduct(ref Point3D arg)
        {
            return (new Point3D(this.Y * arg.Z - this.Z * arg.Y, this.Z * arg.X - this.X * arg.Z, this.X * arg.Y - this.Y * arg.X));
        }

        /// <summary>
        /// Returns the cross product of two points.
        /// </summary>
        /// <param name="arg">Second argument of the cross product operation.</param>
        /// <returns>Point3D representing the resulting vector.</returns>
        public Point3D CrossProduct(Point3D arg)
        {
            return (new Point3D(this.Y * arg.Z - this.Z * arg.Y, this.Z * arg.X - this.X * arg.Z, this.X * arg.Y - this.Y * arg.X));
        }

        /// <summary>
        /// Returns the cross product of two points.
        /// </summary>
        /// <param name="p1">First argument of the cross product operation.</param>
        /// <param name="p2">Second argument of the cross product operation.</param>
        /// <returns>Point3D representing the resulting vector.</returns>
        public static Point3D CrossProduct(Point3D p1, Point3D p2)
        {
            return (new Point3D(p1.Y * p2.Z - p1.Z * p2.Y, p1.Z * p2.X - p1.X * p2.Z, p1.X * p2.Y - p1.Y * p2.X));
        }

        /// <summary>
        /// Returns the cross product of two points.
        /// </summary>
        /// <param name="p1">First argument of the cross product operation.</param>
        /// <param name="p2">Second argument of the cross product operation.</param>
        /// <returns>Point3D representing the resulting vector.</returns>
        public static Point3D CrossProduct(ref Point3D p1, ref Point3D p2)
        {
            return (new Point3D(p1.Y * p2.Z - p1.Z * p2.Y, p1.Z * p2.X - p1.X * p2.Z, p1.X * p2.Y - p1.Y * p2.X));
        }

        /// <summary>
        /// Returns the cross product of two points.
        /// </summary>
        /// <param name="p1">First argument of the cross product operation.</param>
        /// <param name="p2">Second argument of the cross product operation.</param>
        /// <returns>Point3D representing the resulting vector.</returns>
        public static Point3D CrossProduct(ref Point3D p1, Point3D p2)
        {
            return (new Point3D(p1.Y * p2.Z - p1.Z * p2.Y, p1.Z * p2.X - p1.X * p2.Z, p1.X * p2.Y - p1.Y * p2.X));
        }

        /// <summary>
        /// Returns the cross product of two points.
        /// </summary>
        /// <param name="p1">First argument of the cross product operation.</param>
        /// <param name="p2">Second argument of the cross product operation.</param>
        /// <returns>Point3D representing the resulting vector.</returns>
        public static Point3D CrossProduct(Point3D p1, ref Point3D p2)
        {
            return (new Point3D(p1.Y * p2.Z - p1.Z * p2.Y, p1.Z * p2.X - p1.X * p2.Z, p1.X * p2.Y - p1.Y * p2.X));
        }
        #endregion


        #region Overriden methods of the class Object

        /// <inheritdoc/>
        override public string ToString()
        {
            return "Point3D: [" + X + "; " + Y + "; " + Z + "]";
        }

        /// <inheritdoc/>
        public override int GetHashCode()
        {
            //int hc = this.X.GetHashCode() ^ this.Y.GetHashCode() ^ this.Z.GetHashCode();
            int hc = Math.Round(this.X, 5).GetHashCode()^ Math.Round(this.Y, 5).GetHashCode() ^ Math.Round(this.Z, 5).GetHashCode();
            return hc;
        }

        /// <inheritdoc/>
        public override bool Equals(object obj)
        {
            if (obj is Point3D)
            {
                Point3D o = (Point3D)obj;                
                if (Math.Abs(this.X - o.X) > Utils.EPSILON)
                    return (false);
                if (Math.Abs(this.Y - o.Y) > Utils.EPSILON)
                    return (false);
                if (Math.Abs(this.Z - o.Z) > Utils.EPSILON)
                    return (false);
                return true;
            }
            else
                return false;
        }

        #endregion
        

    }
}
