﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace Zcu.CurvatureBenchmark.Framework
{
    /// <summary>
    /// Extension class for random number generator system. The common random generator system allows to control 
    /// the randomness during the tests. 
    /// <para>
    /// The random generator system has a single seed that is common for all modules. Each module has its own 
    /// instance of System.Random. Using this instance, there is no dependency on the order of module calls.
    /// </para>
    /// </summary>
    public static class RandomExt
    {
        /// <summary>
        /// Dictionary of random generators for each instance
        /// </summary>
        static Dictionary<object, Random> random = new Dictionary<object, Random>();

        /// <summary>
        /// Seed for all random generators. If you'd like to change seed, do it before first call of Rand(T)!
        /// </summary>
        public static int Seed { get; set; }

        /// <summary>
        /// Sets a seed for a single instance. Use this method if and only if you are absolutely sure, what you are doing!
        /// </summary>
        /// <param name="instance">Instance of object to change seed</param>
        /// <param name="seed">New seed</param>
        [Obsolete("Use SetSeed() method if and only if you are absolutely sure, what you are doing!")]
        public static void SetSeed(object instance, int seed)
        {
            if (random.ContainsKey(instance))
            {
                random[instance] = new Random(seed);
            }
        }

        /// <summary>
        /// By default a random seed is generated.
        /// </summary>
        static RandomExt()
        {
            Random r = new Random();
            Seed = r.Next();
        }

        /// <summary>
        /// Returns random instance for given instance
        /// </summary>
        /// <typeparam name="T">Class</typeparam>
        /// <param name="instance">Instance</param>
        /// <returns>Random for Instance</returns>
        public static Random Rand<T>(this T instance) where T : IModule 
        {
            if (!random.ContainsKey(instance))
            {				
				random[instance] = new Random(Seed);
            }
            return random[instance];
        }

        /// <summary>
        /// Extends Random with NextDouble from interval min, max <see cref="Random.NextDouble"/>
        /// </summary>
        /// <param name="instance">Random generator</param>
        /// <param name="min">Minimal value</param>
        /// <param name="max">Maximal value</param>
        /// <returns>Random double from interval &lt;min,max)</returns>
        public static double NextDouble(this Random instance, double min, double max)
        {
            return min + instance.NextDouble() * (max - min);
        }
    }

}
