﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.IO;
using System.Reflection;
using System.Xml;
using Zcu.CurvatureBenchmark.Framework.Estimators;
using Zcu.CurvatureBenchmark.Framework.Evaluators;
using Zcu.CurvatureBenchmark.Framework.Distorters;
using Zcu.CurvatureBenchmark.Framework.DataSources;

namespace Zcu.CurvatureBenchmark.Framework
{
    /// <summary>
    /// Delegate for module loaded event.
    /// </summary>
    /// <param name="module"></param>
    public delegate void NotifyModuleLoaded(IModule module);

    /// <summary>
    /// Class is responsible for loading config files and creating and setting all used modules.
    /// </summary>
    public class TestScenario
    {
        /// <summary>
        /// List of data source modules
        /// </summary>
        public List<IDataSource> sources = new List<IDataSource>();

        /// <summary>
        /// List of estimator modules
        /// </summary>
        public List<ICurvatureEstimator> estimators = new List<ICurvatureEstimator>();

        /// <summary>
        /// List of evaluator modules
        /// </summary>
        public List<ICurvatureEvaluator> evaluators = new List<ICurvatureEvaluator>();

        /// <summary>
        /// List of distorter modules
        /// </summary>
        public List<IDistorter> distorters = new List<IDistorter>();

        NotifyModuleLoaded notifyModuleLoaded;

        /// <summary>
        /// Creates instances and set all modules for a config section.
        /// </summary>
        /// <typeparam name="T">Type of module</typeparam>
        /// <param name="nodes">Config section</param>
        /// <returns>List of loaded modules</returns>
        List<T> GetModules<T>(XmlNodeList nodes) where T : class, IModule
        {
            List<T> result = new List<T>();
            foreach (XmlNode node in nodes)
            {
                if (node.NodeType == XmlNodeType.Comment) continue;
                var module = TestScenario.GetModule<T>(node);

                if (module != null)
                {
                    if (module.CheckPrerequisities())
                    {
                        notifyModuleLoaded?.Invoke(module);
                        result.Add(module);
                    }
                }
            }
            return result;
        }

        /// <summary>
        /// Creates lists of modules from xml file. 
        /// </summary>
        /// <remarks>Changes the current culture to en-US!</remarks>
        /// <param name="configFileName">Config xml</param>
        /// <param name="notify">Delegate for module loaded notification</param>
        public TestScenario(string configFileName, NotifyModuleLoaded notify = null)
        {
            System.Threading.Thread.CurrentThread.CurrentCulture = new System.Globalization.CultureInfo("en-US");
            notifyModuleLoaded = notify;
            XmlDocument configXml = new XmlDocument();
            configXml.Load(Path.GetDirectoryName(configFileName)+"\\"+Path.GetFileNameWithoutExtension(configFileName)+".xml");
            XmlNode root = configXml.DocumentElement;
            foreach (XmlNode node in root.ChildNodes)
            {
                switch (node.Name)
                {
                    case "Sources": sources = GetModules<IDataSource>(node.ChildNodes); break;
                    case "Estimators": estimators = GetModules<ICurvatureEstimator>(node.ChildNodes); break;
                    case "Evaluators": evaluators = GetModules<ICurvatureEvaluator>(node.ChildNodes); break;
                    case "Distorters": distorters = GetModules<IDistorter>(node.ChildNodes); break;
                }
            }
			if (distorters.Count == 0)
            {
                distorters = new List<IDistorter>();
                distorters.Add(new NoDistorter());
            }

        }

        #region Static part
        static Dictionary<string, Type> modules = new Dictionary<string, Type>();


        /// <summary>
        /// Used for outputs.
        /// </summary>
        #pragma warning disable 0649
        static IModule ModuleManager = null;
        #pragma warning restore 0649

        /// <summary>
        /// Gets all non abstract types implementing IModule and stores them
        /// into the types dictionary. Fully qualified name is used.
        /// </summary>
        public static void GetAllIModuleTypes()
        {
            foreach (var file in "*.exe;*.dll".Split(';').SelectMany(ext => Directory.EnumerateFiles("./", ext)))
            {
                try
                {
                    var assembly = Assembly.LoadFrom(file);
                    foreach (var type in assembly.GetExportedTypes().Where(t =>
                                        t.GetInterface("IModule") != null &&
                                        !t.IsAbstract &&
                                        !t.IsInterface))
                    {
                        modules.Add("."+type.FullName, type);
                    }

                }
                catch
                {
                    Output.Warning(ModuleManager, "{0} is not valid assembly.", file);
                }
            }
        }

        /// <summary>
        /// Creates instance of a module according to the xmlnode.Name and configures it. If the given name corresponds to multiple modules, the first one is chosen.
        /// </summary>
        /// <typeparam name="T">Interface or type of module to create.</typeparam>
        /// <param name="node">XmlNode representing the current module.</param>
        /// <returns>Instance of module.</returns>
        /// <exception cref="ApplicationException">Thrown when module was not found or the module type does not match the type parameter.</exception>
        public static T GetModule<T>(XmlNode node) where T : class, IModule
        {
            while (node.NodeType == XmlNodeType.Comment) node = node.NextSibling;
            if (modules.Count == 0) GetAllIModuleTypes();

            var modulecandidates = modules.Where(v => v.Key.EndsWith("."+node.Name) && v.Value.GetInterface(typeof(T).FullName)!=null).Select(v => v.Value);
            var count = modulecandidates.Count();
            if (count == 0)
            {
                Output.Error(ModuleManager, "Module {0} implementing {1} was not found. Make sure it is a public class.", node.Name, typeof(T));
                return null;
            }

            // TODO decide whether to throw an exception
            if (count > 1) Output.Warning(ModuleManager, "There is an ambiguity in config file. I can't resolve module {0}, taking {1}", node.Name, modulecandidates.First().FullName);

            T module = (T)Activator.CreateInstance(modulecandidates.First());
            module.SetConfig(node);
            return module;
        }
        #endregion;



    }
}
