﻿using Zcu.CurvatureBenchmark.Framework.Estimators;

namespace Zcu.CurvatureBenchmark.Framework
{
    /// <summary>
    /// Represents triangle mesh.
    /// </summary>
    public class TriangleMesh
    {
        private Point3D[] points;
        private Triangle[] triangles;
        private Point3D[] normals;
        private Point3D[] exactNormals;
        private double[] exactK1;
        private double[] exactK2;
        private ColorRGBA[] colors;

        /// <summary>
        /// Gets/sets used curvature estimator.
        /// This is a hack for visualization purpose. The estimator is used for visualization description.
        /// </summary>
        /// <preliminary/>
        public ICurvatureEstimator Estimator { get; set; }

        /// <summary>
        /// Gets/sets the array of colors associated with vertices.
        /// </summary>
        public ColorRGBA[] Colors
        {
            get { return colors; }
            set { colors = value; }
        }

        /// <summary>
        /// Gets/sets the array of normals associated with vertices. If ExactNormals are null, it copies Normals to ExactNormals. 
        /// </summary>
        public Point3D[] Normals
        {
            get { return normals; }
            set {
                normals = value;
                if (ExactNormals==null)
                {
                    exactNormals = (Point3D[])normals.Clone();
                }
            }
        }

        /// <summary>
        /// Gets exact normals. Exact normals are normals that are set for the first time (i.e. by data source).
        /// </summary>
        /// <remarks>
        /// Do not use exact normals in evaluators. In real-life scenario it's not probable to obtain exact normals.
        /// Usually normals are computed (e.g. by Max).
        /// </remarks>
        public Point3D[] ExactNormals
        {
            get { return exactNormals; }
        }

        /// <summary>
        /// Gets/sets the array of vertices.
        /// </summary>
        public Point3D[] Points
        {
            get { return points; }
            set { points = value; }
        }

        /// <summary>
        /// Gets/sets the array of triangles.
        /// </summary>
        public Triangle[] Triangles
        {
            get { return triangles; }
            set { triangles = value; }
        }

        /// <summary>
        /// Gets/sets the array of exact K1 associated with vertices.
        /// </summary>
        public double[] ExactK1
        {
            get
            {
                return exactK1;
            }

            set
            {
                exactK1 = value;
            }
        }

        /// <summary>
        /// Gets/sets the array of exact K2 associated with vertices.
        /// </summary>
        public double[] ExactK2
        {
            get
            {
                return exactK2;
            }

            set
            {
                exactK2 = value;
            }
        }


        /// <summary>
        /// Creates a shallow copy of triangle mesh with deep copy of colors. For visualization purposes.
        /// </summary>
        /// <returns>Shallow copy with deep color copy</returns>
		public TriangleMesh DeepCopyOfColors() 
		{
			TriangleMesh copy = new TriangleMesh();

			copy.points = this.points;
			copy.normals = this.normals;
			copy.triangles = this.triangles;
			
			if (this.colors != null)
			{
				copy.colors = new ColorRGBA[this.colors.Length];
				for (int i = 0; i < copy.colors.Length; i++) copy.colors[i] = this.colors[i];
			}

			if (this.exactK1 != null)
			{
				copy.exactK1 = this.exactK1;
			}

			if (this.exactK2 != null)
			{
				copy.exactK2 = this.exactK2;
			}

			return copy;
		} 

        /// <summary>
        /// Creates a deep copy of triangle mesh.
        /// </summary>
        /// <returns>Deep copy</returns>
		public TriangleMesh DeepCopy() 
		{
			TriangleMesh copy = new TriangleMesh();

			copy.points = new Point3D[this.points.Length];
			for (int i = 0; i < copy.points.Length; i++) copy.points[i] = this.points[i];

			copy.normals = new Point3D[this.normals.Length];			
			for (int i = 0; i < copy.normals.Length; i++) copy.normals[i] = this.normals[i];

            if (this.triangles != null)
            {
                copy.triangles = new Triangle[this.triangles.Length];
                for (int i = 0; i < copy.triangles.Length; i++) copy.triangles[i] = this.triangles[i];
            }

			if (this.colors != null)
			{
				copy.colors = new ColorRGBA[this.colors.Length];
				for (int i = 0; i < copy.colors.Length; i++) copy.colors[i] = this.colors[i];
			}

			if (this.exactK1 != null)
			{
				copy.exactK1 = new double[this.exactK1.Length];
				for (int i = 0; i < copy.exactK1.Length; i++) copy.exactK1[i] = this.exactK1[i];
			}

			if (this.exactK2 != null)
			{
				copy.exactK2 = new double[this.exactK2.Length];
				for (int i = 0; i < copy.exactK2.Length; i++) copy.exactK2[i] = this.exactK2[i];
			}
						
			return copy;
		}

        /// <summary>
        /// creates a shallow copy of the mesh, i.e. all the fields are references to current mesh data.
        /// </summary>
        /// <returns>Shallow copy of the instence</returns>
        public TriangleMesh ShallowCopy()
        {
            TriangleMesh copy = new TriangleMesh();

            copy.points = this.points;
            copy.triangles = this.triangles;
            copy.normals = this.normals;
            copy.exactNormals = this.exactNormals;
            copy.exactK1 = this.exactK1;
            copy.exactK2 = this.exactK2;

            return copy;
        }

        /// <inheritdoc/>
        public override string ToString()
        {
            string str = this.GetType().FullName + ":\r\n";

            if (points != null)
            {
                str += "points: " + points.Length + "\r\n";
            }

            if (triangles != null)
            {
                str += "triangles: " + triangles.Length + "\r\n";
            }
            return str;
        }

        /*public void ComputeNormals()
        {
            Point3D[] normals = new Point3D[points.Length];
            for (int i = 0; i < triangles.Length; i++)
            {
                Triangle t = triangles[i];
                Point3D p1 = points[t.V1];
                Point3D p2 = points[t.V2];
                Point3D p3 = points[t.V3];
                Point3D v1 = p2 - p1;
                Point3D v2 = p3 - p1;
                Point3D n = v1.CrossProduct(v2);
                if (n.Abs() == 0)
                    continue;
                n *= 1 / n.Abs();
                normals[t.V1] += n;
                normals[t.V2] += n;
                normals[t.V3] += n;
            }
            for (int i = 0; i < normals.Length;i++ )
            {
                double l = normals[i].Abs();
                if (l < 0.00001)
                    normals[i] = new Point3D(1, 0, 0);
                else
                    normals[i] *= 1.0 / l;
            }

                this.normals = normals;
        }*/

        /// <summary>
        /// Creates an array of excluded vertices, based on topological/geometric criteria.
        /// </summary>
        /// <param name="topologic">Should vertices get excluded based on their topological distance from border?</param>
        /// <param name="geometric">Should vertices get excluded based on their euclidean distance from border?</param>
        /// <param name="topoZone">Topologic distance within which vertices get excluded</param>
        /// <param name="geoZone">Geometric distance within which vertices get excluded. The value is used relative to average edge length, i.e. 2.0 means "exclude vertices up to the distance of twice the average edge length".</param>
        /// <returns>Array of vertex flags, true = excluded</returns>
        public bool[] excludeVertices(bool topologic, bool geometric, int topoZone, double geoZone)
        {
            bool[] excluded = new bool[points.Length];
            CornerTable ct = new CornerTable(this);

            if (topologic)
            {
                for (int i = 0; i < Points.Length; i++)
                {
                    if (ct.IsOnBorder(i))
                    {
                        excluded[i] = true;
                        foreach (int n in ct.VVK(i, topoZone))
                            excluded[n] = true;
                    }
                }
            }

            if (geometric)
            {
                double avgEL = AverageEL();
                for (int i = 0; i < Points.Length; i++)
                {                    
                    if (ct.IsOnBorder(i))
                    {
                        excluded[i] = true;
                        var zone = ct.VVE(i, geoZone*avgEL);
                        foreach (int n in zone)
                            excluded[n] = true;
                    }
                }
            }

            return excluded;
        }

        /// <summary>
        /// Computes average edge length in the mesh. Note that inner edges are counted with doubled weight in comarison with border edges.
        /// </summary>
        /// <returns>Average edge length</returns>
        public double AverageEL()
        {
            double avgEL = 0;
            foreach(Triangle t in triangles)
            {
                avgEL += (points[t.V1] - points[t.V2]).Abs();
                avgEL += (points[t.V2] - points[t.V3]).Abs();
                avgEL += (points[t.V1] - points[t.V3]).Abs();
            }
            avgEL /= (triangles.Length * 3);
            return (avgEL);
        }
    }
}
