﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.IO;
using Zcu.CurvatureBenchmark.Framework.Distorters;
using Zcu.CurvatureBenchmark.Framework.DataSources;


namespace Zcu.CurvatureBenchmark.Framework
{
    /// <summary>
    /// Evaluate mesh statistics
    /// </summary>
    public class TriangleMeshStat
    {
        /// <summary>
        /// Gets the lengths of edges. Edges from exclusion zone are excluded.
        /// </summary>
        public double[] EdgeLengths { get; protected set; }

        /// <summary>
        /// Gets the inner angles of all triangles. Triangles from exclusion zone are excluded.
        /// </summary>
        public double[] InnerAngles { get; protected set; }

        /// <summary>
        /// Gets the dihedral angles of all pairs of triangles. Triangles from exclusion zone are excluded.
        /// </summary>
        public double[] DihedralAngles { get; protected set; }

        /// <summary>
        /// Gets the signed dihedral angles of all pairs of triangles. Triangles from exclusion zone are excluded.
        /// </summary>
        public double[] DihedralAnglesSigned { get; protected set; }

        /// <summary>
        /// Gets the degrees of vertices. Vertices from exclusion zone are excluded.
        /// </summary>
        public double[] VertexDegrees { get; protected set; }

        /// <summary>
        /// Gets the circumcircle/incircle ratio. Triangles from exclusion zone are excluded.
        /// </summary>
        public double[] CircleRatios { get; protected set; }

        /// <summary>
        /// Gets the mean value laplacian. 
        /// </summary>
        public double[] MVLaplacians { get; protected set; }

        /// <summary>
        /// Gets the difference of laplacian and mean value laplacian.
        /// </summary>
        public double[] LaplacianDifferences { get; protected set; }

        /// <summary>
        /// Gets the laplacian of lapacians.
        /// </summary>
        public double[] DoubleLaplacians { get; protected set; }

        /// <summary>
        /// Gets the solid angle in vertex. Vertices from exclusion zone are excluded.
        /// </summary>
        public double[] SolidAngles { get; protected set; }

        CornerTable ct;
        TriangleMesh mesh;
        bool[] excludeTriangles;
        bool[] excludeVertices;
        Point3D[] triangleNormals;

        /// <summary>
        /// Statistic properties mapped to a dictionary.
        /// </summary>
        Dictionary<string, double[]> stats = new Dictionary<string, double[]>();

        /// <summary>
        /// number of vertices not excluded
        /// </summary>
        int VertexCount { get; set; }

        /// <summary>
        /// For internal use only!
        /// </summary>
        private TriangleMeshStat()
        {
            MapStatDictionary();
        }

        /// <summary>
        /// number of triangles not excluded
        /// </summary>
        int TriangleCount { get; set; }

        private string sourceDescriptor, distorsionDescriptor;

        /// <summary>
        /// Compute mesh statistics
        /// </summary>
        /// <param name="mesh">Mesh</param>
        /// <param name="topologic">Should vertices get excluded based on their topological distance from border?</param>
        /// <param name="geometric">Should vertices get excluded based on their euclidean distance from border?</param>
        /// <param name="topoZone">Topologic distance within which vertices get excluded</param>
        /// <param name="geoZone">Geometric distance within which vertices get excluded. The value is used realitve to average edge length, i.e. 2.0 means "exclude vertices up to the distance of twice the average edge length".</param>
        /// <param name="source">Datasource (output string)</param>
        /// <param name="distorter">Distorter (output string)</param>
        public TriangleMeshStat(TriangleMesh mesh, bool topologic, bool geometric, int topoZone, double geoZone, IDataSource source, IDistorter distorter)
        {
            sourceDescriptor = source==null?"":source.ToString();
            distorsionDescriptor = distorter==null?"":distorter.ToString();

            this.mesh = mesh;
            ct = new CornerTable(mesh);

            excludeVertices = mesh.excludeVertices(topologic, geometric, topoZone, geoZone);
            VertexCount = 0;
            for (int i = 0; i < excludeVertices.Length; i++) if (!excludeVertices[i]) VertexCount++;

            excludeTriangles = new bool[mesh.Triangles.Length];
            triangleNormals = new Point3D[mesh.Triangles.Length];
            TriangleCount = 0;

            for (int i=0; i<mesh.Triangles.Length; i++)
            {
                triangleNormals[i] = Point3D.CrossProduct(mesh.Points[mesh.Triangles[i].V2] - mesh.Points[mesh.Triangles[i].V1], mesh.Points[mesh.Triangles[i].V1] - mesh.Points[mesh.Triangles[i].V3]).Normalize();

                excludeTriangles[i] = true;
                if (excludeVertices[mesh.Triangles[i].V1]) continue;
                if (excludeVertices[mesh.Triangles[i].V2]) continue;
                if (excludeVertices[mesh.Triangles[i].V3]) continue;
                excludeTriangles[i] = false;
                TriangleCount++;
            }

            ComputeEdgeLengths();
            ComputeInnerAngles();
            ComputeDihedralAngles();
            ComputeVertexDegrees();
            ComputeCircleRatios();
            ComputeLaplacians();
            ComputeCotanLaplacians();
            ComputeSolidAngles();


            MapStatDictionary();
        }


        void MapStatDictionary()
        {
            stats["EL"] = EdgeLengths;
            stats["IA"] = InnerAngles;
            stats["DA"] = DihedralAngles;
            stats["DAS"] = DihedralAnglesSigned;
            stats["SA"] = SolidAngles;
            stats["VD"] = VertexDegrees;
            stats["CR"] = CircleRatios;
            stats["MVL"] = MVLaplacians;
            stats["LD"] = LaplacianDifferences;
            stats["DL"] = DoubleLaplacians;
        }

        /// <summary>
        /// Dumps statistics to a file.
        /// </summary>
        /// <param name="fileName">Filename</param>
        public void DumpStatsToFile(String fileName)
        {
            string line = string.Format("{0}\t{1}\t{2}\t{3}\t", sourceDescriptor, distorsionDescriptor, VertexCount, TriangleCount);
            foreach(var stat in stats)
            {
                DoubleArrayStatistic arrayStat = new DoubleArrayStatistic(stat.Value);
                line += String.Format("{0}\t{1}\t{2}\t{3}\t{4}\t{5}\t", 
                    arrayStat.Min, 
                    arrayStat.Max, 
                    arrayStat.Mean, 
                    arrayStat.Median, 
                    arrayStat.StDev, 
                    arrayStat.Var);
            }
            FileStream fs = new FileStream(fileName, FileMode.Append);
            StreamWriter sw = new StreamWriter(fs);
            sw.WriteLine(line);
            sw.Close();
            fs.Close();
        }

        /// <summary>
        /// Dumps statistics header to a file.
        /// </summary>
        /// <param name="fileName">Filename</param>
        public static void DumpStatsHeaderToFile(String fileName)
        {
            string line = "Source\tDistorter\tVC\tTC\t";
            var stats = new TriangleMeshStat();
            foreach (var stat in stats.stats)
            {
                line += String.Format("{0}.min\t{0}.max\t{0}.mean\t{0}.median\t{0}.stdev\t{0}.var\t", stat.Key);
            }
            FileStream fs = new FileStream(fileName, FileMode.Append);
            StreamWriter sw = new StreamWriter(fs);
            sw.WriteLine(line);
            sw.Close();
            fs.Close();
        }
       
        private void ComputeEdgeLengths()
        {
            List<double> lengths = new List<double>(mesh.Triangles.Length * 3);
            bool[] visited = new bool[mesh.Triangles.Length * 3];
            for (int i = 0; i < visited.Length; i++) visited[i] = false;

            double ael = mesh.AverageEL();

            for (int i = 0; i < visited.Length; i++)
            {
                if (visited[i]) continue;
                if (excludeTriangles[i / 3]) continue;

                int v1 = ct.v(ct.n(i));
                int v2 = ct.v(ct.p(i));

                lengths.Add((mesh.Points[v1] - mesh.Points[v2]).Length()/ael);

                if(ct.opposite[i]>0)
                    visited[ct.opposite[i]] = true;
            }

            EdgeLengths = lengths.ToArray();

        }

        private void ComputeInnerAngles()
        {
            List<double> angles = new List<double>(mesh.Triangles.Length * 3);

            for (int i = 0; i < mesh.Triangles.Length; i++)
            {
                if (excludeTriangles[i]) continue;

                var e1 = (mesh.Points[mesh.Triangles[i].V1] - mesh.Points[mesh.Triangles[i].V2]).Normalize();
                var e2 = (mesh.Points[mesh.Triangles[i].V2] - mesh.Points[mesh.Triangles[i].V3]).Normalize();
                var e3 = (mesh.Points[mesh.Triangles[i].V3] - mesh.Points[mesh.Triangles[i].V1]).Normalize();

                angles.Add(180 * Math.Acos(e1.DotProduct(-e2)) / Math.PI);
                angles.Add(180 * Math.Acos(e2.DotProduct(-e3)) / Math.PI);
                angles.Add(180 * Math.Acos(e3.DotProduct(-e1)) / Math.PI);

            }

            InnerAngles = angles.ToArray();
        }

        private void ComputeDihedralAngles()
        {
            List<double> angles = new List<double>(mesh.Triangles.Length);
            List<double> anglesSigned = new List<double>(mesh.Triangles.Length);
            bool[] visited = new bool[mesh.Triangles.Length * 3];
            for (int i = 0; i < visited.Length; i++) visited[i] = false;

            for (int i = 0; i < visited.Length; i++)
            {
                if (visited[i]) continue;
                int t1 = i / 3;
                int t2 = ct.opposite[i] / 3;
                if (excludeTriangles[t1]) continue;
                if (excludeTriangles[t2]) continue;


                int ci = 3 * t1;
                for (; ci < 3 * t1 + 3; ci++)
                {
                    if (ct.opposite[ci] / 3 == t2)
                        break;
                }

                Point3D edge = (mesh.Points[ct.v(ct.n(ci))] - mesh.Points[ct.v(ct.p(ci))]).Normalize();

                double angle = Math.Acos(Utils.Clamp(-1, 1, triangleNormals[t1].DotProduct(ref triangleNormals[t2]))) * 180 / Math.PI;

                int sign = Math.Sign(triangleNormals[t1].CrossProduct(ref edge).DotProduct(ref triangleNormals[t2]));
                angles.Add(angle);
                anglesSigned.Add(sign * angle);

                /*Output.Debug((IModule)null,
                    "DihedralAngles: tri {0},{1}; \n\t\tedge {2},{3}; \n\t\tmeanK: {4},{5}\n\t\t{6}", 
                    t1, t2, 
                    ct.v(ct.n(ci)), ct.v(ct.p(ci)), 
                    (mesh.ExactK1[ct.v(ct.n(ci))] + mesh.ExactK2[ct.v(ct.n(ci))])/2, (mesh.ExactK1[ct.v(ct.p(ci))] + mesh.ExactK2[ct.v(ct.p(ci))]) / 2,
                    sign*angle);*/

                if(ct.opposite[i]>0)
                    visited[ct.opposite[i]] = true;
            }

            DihedralAngles = angles.ToArray();
            DihedralAnglesSigned = anglesSigned.ToArray();
        }

        private void ComputeVertexDegrees()
        {
            double[] degreesAll = new double[mesh.Points.Length];
            List<double> degrees = new List<double>(mesh.Points.Length);
            bool[] visited = new bool[mesh.Triangles.Length * 3];

            for (int i = 0; i < visited.Length; i++) visited[i] = false;

            for (int i = 0; i < visited.Length; i++)
            {
                if (visited[i]) continue;

                int v1 = ct.v(ct.n(i));
                int v2 = ct.v(ct.p(i));

                degreesAll[v1]++;
                degreesAll[v2]++;

                if (ct.opposite[i] > 0)
                    visited[ct.opposite[i]] = true;
            }


            for(int i=0; i<mesh.Points.Length; i++)
            {

                if (excludeVertices[i]) continue;
                degrees.Add(degreesAll[i]);
            }
            VertexDegrees = degrees.ToArray();

        }

        private void ComputeCircleRatios()
        {
            List<double> ratios = new List<double>(mesh.Triangles.Length * 3);
            for(int i=0; i<mesh.Triangles.Length; i++)
            {
                if (excludeTriangles[i]) continue;

                Point3D A = mesh.Points[mesh.Triangles[i].V1];
                Point3D B = mesh.Points[mesh.Triangles[i].V2];
                Point3D C = mesh.Points[mesh.Triangles[i].V3];

                double a = (B - C).Length();
                double b = (C - A).Length();
                double c = (A - B).Length();
                
                double s = (a + b + c) / 2;

                double r = Math.Sqrt((s - a) * (s - b) * (s - c) / s);
                double R = (a * b * c) / (4 * r * s);
                ratios.Add(R / r);
            }
            CircleRatios = ratios.ToArray();
        }

        private void ComputeLaplacians()
        {
            List<double> mvLaplacians = new List<double>(mesh.Points.Length);
            List<double> laplacianDifferences = new List<double>(mesh.Points.Length);
            
            List<double> areas = new List<double>(mesh.Points.Length);
            for (int i = 0; i < mesh.Points.Length; i++)
            {
                if (excludeVertices[i]) continue;

                if (ct.IsOnBorder(i)) continue;

                int[] nbs = ct.VV(i);

                double[] angles = new double[nbs.Length];
                double avgEL = 0;
                double area = 0;
                for (int j = 0; j < nbs.Length; j++)
                {
                    int n = (j + 1) % nbs.Length;
                    Point3D v1 = mesh.Points[nbs[j]] - mesh.Points[i];
                    Point3D v2 = mesh.Points[nbs[n]] - mesh.Points[i];
                    Point3D v3 = mesh.Points[nbs[n]] - mesh.Points[nbs[j]];
                    area += v1.CrossProduct(v2).Abs() / 2;
                    avgEL += v1.Abs();
                    avgEL += v3.Abs();
                    v1 = v1.Normalize();
                    v2 = v2.Normalize();
                    double dp = v1.DotProduct(v2);
                    angles[j] = Math.Acos(Math.Max(-1, Math.Min(1, dp)));
                }
                avgEL /= (2 * nbs.Length);
                areas.Add(area);

                double[] weights = new double[nbs.Length];
                double weightSum = 0;
                for (int j = 0; j < nbs.Length; j++)
                {
                    int n = (j + 1) % nbs.Length;
                    weights[j] = (Math.Tan(angles[j] / 2) + Math.Tan(angles[n] / 2)) / (mesh.Points[nbs[n]] - mesh.Points[i]).Abs();
                    weightSum += weights[j];
                }

                Point3D mvLap = new Point3D();
                Point3D uniLap = new Point3D();
                for (int j = 0; j < nbs.Length; j++)
                {
                    int n = (j + 1) % nbs.Length;
                    Point3D v = mesh.Points[nbs[n]] - mesh.Points[i];
                    uniLap += v;
                    mvLap += (weights[j] / weightSum) * v;
                }
                uniLap *= 1.0 / nbs.Length;

                mvLaplacians.Add(mvLap.Abs() / avgEL);
                laplacianDifferences.Add((mvLap - uniLap).Abs() / avgEL);                
            }
            MVLaplacians = mvLaplacians.ToArray();
            LaplacianDifferences = laplacianDifferences.ToArray();
        }

        private void ComputeCotanLaplacians()
        {
            // compute laplace vectors for each vertex
            Point3D[] laplaceVectors = new Point3D[mesh.Points.Length];
            double[] mixedAreas = new double[mesh.Points.Length];
            for (int vi = 0; vi < mesh.Points.Length; vi++)
            {
                if (ct.IsOnBorder(vi))
                    continue;

                double mixedArea = 0;
                Point3D laplaceVector = new Point3D(0, 0, 0);
                Point3D edge, a, b, c, d = new Point3D(0, 0, 0);                                
                int[] incidentVertices = ct.VV(vi);
                int nc = incidentVertices.Length;

                for (int i = 0; i < nc; i++)
                {
                    int vj = incidentVertices[i];
                    int Alpha, Beta = -1;
                    
                    Alpha = incidentVertices[(i - 1+nc)%nc];
                    Beta = incidentVertices[(i + 1+nc)%nc];

                    //calculate vectors
                    edge = mesh.Points[vj] - mesh.Points[vi];
                    a = mesh.Points[vj] - mesh.Points[Alpha];
                    b = mesh.Points[vi] - mesh.Points[Alpha];
                    c = mesh.Points[vj] - mesh.Points[Beta];
                    d = mesh.Points[vi] - mesh.Points[Beta];

                    //calculate cot(angle), which is opposite to edge
                    double cotAlpha = Point3D.DotProduct(a, b) / (Point3D.CrossProduct(a, b)).Abs();
                    double cotBeta = Point3D.DotProduct(c, d) / (Point3D.CrossProduct(c, d)).Abs();
                    //calculate sum of mixed area
                    double cosAtVi = (Point3D.DotProduct(edge, -b) / (edge.Abs() * (-b).Abs()));
                    double cosAtVi2 = (Point3D.DotProduct(edge, -d) / (edge.Abs() * (-d).Abs()));
                    double cosAtAplha = (Point3D.DotProduct(b, a) / (b.Abs() * a.Abs()));
                    double cosAtVj = (Point3D.DotProduct(-edge, -a) / ((-edge).Abs() * (-a).Abs()));

                    //add value of actual edge                    
                    laplaceVector += ((cotAlpha + cotBeta) * edge);
                                        
                    // non-obtuse triangle -> voronoi area
                    if ((cosAtVi > 0) && (cosAtAplha > 0) && (cosAtVj > 0))
                    {
                        //cot(angle)
                        double cotAtVj = (Point3D.DotProduct(-edge, -a) / Math.Sqrt((-edge).Abs2() * (-a).Abs2() -
                            Math.Pow((Point3D.DotProduct(-edge, -a)), 2)));
                        cotAtVj = (Point3D.DotProduct(-edge, -a) / (Point3D.CrossProduct((-edge), (-a))).Abs());

                        //calculate sum of voronoi area
                        double voronoiArea = ((1.0f / 8) * (edge.Abs() * edge.Abs() * cotAlpha +
                            (-b).Abs() * (-b).Abs() * cotAtVj));
                        mixedArea += voronoiArea;
                    }
                    else // obtuse triangle
                    {
                        double areaOfTriangle = (Point3D.CrossProduct(edge, -b).Abs() / 2);
                        if (cosAtVi <= 0)
                            mixedArea += (areaOfTriangle / 2);
                        else
                            mixedArea += (areaOfTriangle / 4);
                    }
                }

                mixedAreas[vi] = mixedArea;
                laplaceVectors[vi] = ((1.0f / (4 * mixedArea)) * laplaceVector);
            }

            Point3D[] laplaceVectors2 = new Point3D[mesh.Points.Length];

            // now apply cotan Laplacian to the Laplace vectors themselves
            for (int vi = 0; vi < mesh.Points.Length; vi++)
            {
                if (ct.IsOnBorder(vi))
                    continue;

                Point3D laplaceVector = new Point3D(0, 0, 0);
                Point3D edge, a, b, c, d = new Point3D(0, 0, 0);
                int[] incidentVertices = ct.VV(vi);
                int nc = incidentVertices.Length;

                for (int i = 0; i < nc; i++)
                {
                    int vj = incidentVertices[i];
                    int Alpha, Beta = -1;

                    Alpha = incidentVertices[(i - 1+nc) % nc];
                    Beta = incidentVertices[(i + 1+nc) % nc];

                    //calculate vectors
                    edge = laplaceVectors[vj] - laplaceVectors[vi];
                    a = mesh.Points[vj] - mesh.Points[Alpha];
                    b = mesh.Points[vi] - mesh.Points[Alpha];
                    c = mesh.Points[vj] - mesh.Points[Beta];
                    d = mesh.Points[vi] - mesh.Points[Beta];

                    //calculate cot(angle), which is opposite to edge
                    double cotAlpha = Point3D.DotProduct(a, b) / (Point3D.CrossProduct(a, b)).Abs();
                    double cotBeta = Point3D.DotProduct(c, d) / (Point3D.CrossProduct(c, d)).Abs();

                    //add value of actual edge                    
                    laplaceVector += ((cotAlpha + cotBeta) * edge);                    
                }

                laplaceVectors2[vi] = ((1.0f / (4 * mixedAreas[vi])) * laplaceVector);
            }

            List<double> doubleLaplacians = new List<double>();
            double avgEL = mesh.AverageEL();
            for (int i = 0;i<mesh.Points.Length;i++)
            {
                if (!excludeVertices[i])
                {
                    // Laplacian vectors scale with 1/EL
                    // thus Laplacian of Laplacian scales with 1/(EL*EL)
                    // in order to have scale invariant measure, we must multiply with EL^2
                    doubleLaplacians.Add(laplaceVectors2[i].Abs() * avgEL * avgEL);
                }
            }
            DoubleLaplacians = doubleLaplacians.ToArray();
        }

        private void ComputeSolidAngles()
        {
            List<double> angles = new List<double>(mesh.Points.Length);

            for(int i=0; i<mesh.Points.Length; i++)
            {
                if (excludeVertices[i]) continue;
                if (ct.IsOnBorder(i))
                {
                    angles.Add(double.NaN);
                }
                else
                {
                    double angle = 0;
                    int[] neighbors = ct.VV(i);
                    Point3D n = mesh.Normals[i];
                    for (int j=0; j<neighbors.Length; j++)
                    {
                        Point3D v1 = (mesh.Points[neighbors[j]] - mesh.Points[i]).Normalize();
                        Point3D v2 = (mesh.Points[neighbors[(j+1)%neighbors.Length]] - mesh.Points[i]).Normalize();

                        angle += 2*Math.Atan2((v2.CrossProduct(v1)).DotProduct(n), (1 + v1.DotProduct(v2) + v2.DotProduct(n) + v1.DotProduct(n)));
                    }
                    angles.Add(Math.Abs(angle));
                }
            }
            SolidAngles = angles.ToArray();
        }

    }
}
