﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace Zcu.CurvatureBenchmark.Framework
{
    /// <summary>
    /// Some usable utilities class.
    /// All methods and members are (and should be) static
    /// </summary>
    public static class Utils
    {
        /// <summary>
        /// Epsilon for floating point comparison.
        /// </summary>
        public static double EPSILON = 0.000000001;
      
        /// <summary>
        /// Checks if value IsZero by EPSILON test.
        /// </summary>
        /// <param name="value">Value</param>
        /// <returns>TRUE if almost zero</returns>
        public static bool IsZero(double value)
        {
            if (value < -EPSILON) return false;
            if (value > EPSILON) return false;
            return true;
        }

        /// <summary>
        /// Clamps number to interval &lt;min, max&gt;.
        /// </summary>
        /// <typeparam name="T">IComparable</typeparam>
        /// <param name="min">Minimum</param>
        /// <param name="max">Maximum</param>
        /// <param name="value">Value to clamp</param>
        /// <returns>Clamped value</returns>
        public static T Clamp<T>(T min, T max, T value) where T : System.IComparable<T>
        {
            if (value.CompareTo(max) > 0)
                return max;
            if (value.CompareTo(min) < 0)
                return min;
            return value;
	    }

        /// <summary>
        /// Computes the area of triangle.
        /// </summary>
        /// <param name="mesh">Mesh</param>
        /// <param name="tri">Triangle to compute</param>
        /// <returns>Area of triangle</returns>
        public static double TriangleArea(TriangleMesh mesh, ref Triangle tri) 
        {

            return TriangleArea(ref mesh.Points[tri.V1], ref mesh.Points[tri.V2], ref mesh.Points[tri.V3]);
        }

        /// <summary>
        /// Computes the area of triangle.
        /// </summary>
        /// <param name="A">Vertex A</param>
        /// <param name="B">Vertex B</param>
        /// <param name="C">Vertex C</param>
        /// <returns>Area of triangle</returns>
        public static double TriangleArea(ref Point3D A, ref Point3D B, ref Point3D C)
        {
            Point3D e0 = B - A;
            Point3D e1 = C - A;

            //area of triangle
            return 0.5 * (Point3D.CrossProduct(ref e0, ref e1)).Length();
        }


        /// <summary>
        /// Tests, whether a triangle is an obtuse triangle.
        /// </summary>
        /// <param name="mesh">Mesh</param>
        /// <param name="tri">Triangle to test</param>
        /// <param name="obtuseVertexIndex">Outpu index of obtuse vertex</param>
        /// <returns>TRUE if triangle is an obtuse triangle</returns>
        public static bool IsTriangleObtuse(TriangleMesh mesh, ref Triangle tri, out int obtuseVertexIndex) 
        {
            /*
            //Brute force test to calc angles.. debug mode.. do not use in production
            Point3D a2 = (mesh.Points[tri[1]] - mesh.Points[tri[2]]);
            Point3D b2 = (mesh.Points[tri[2]] - mesh.Points[tri[0]]);
            Point3D c2 = (mesh.Points[tri[0]] - mesh.Points[tri[1]]);

            double a = Utils.RadToDeg(Point3D.GetAngle(a2, -b2));
            double b = Utils.RadToDeg(Point3D.GetAngle(b2, -c2));
            double c = Utils.RadToDeg(Point3D.GetAngle(c2, -a2));

            obtuseVertexIndex = -1;
            if (a > 90) { obtuseVertexIndex = 2; return true; }
            if (b > 90) { obtuseVertexIndex = 0; return true; }
            if (c > 90) { obtuseVertexIndex = 1; return true; }

            return false;
            */
              
            double a2 = (mesh.Points[tri.V2] - mesh.Points[tri.V3]).LengthSquared();
            double b2 = (mesh.Points[tri.V3] - mesh.Points[tri.V1]).LengthSquared();
            double c2 = (mesh.Points[tri.V1] - mesh.Points[tri.V2]).LengthSquared();

            obtuseVertexIndex = -1;

            if ((a2 + b2 - c2) <= 0)
            {
                obtuseVertexIndex = 2;
                return true;
            }
            if ((b2 + c2 - a2) <= 0)
            {
                obtuseVertexIndex = 0;
                return true;
            }
            if ((c2 + a2 - b2) <= 0)
            {
                obtuseVertexIndex = 1;
                return true;
            }

            return false;
            
        }

        /// <summary>
        /// Computes a center of triangle circumcircle.
        /// </summary>
        /// <param name="mesh">Mesh</param>
        /// <param name="tri">Triangle</param>
        /// <returns>Center of circumcircle</returns>
        public static Point3D GetCircumcircleCenter(TriangleMesh mesh, ref Triangle tri)
        {			            				
            double a2 = (mesh.Points[tri.V2] - mesh.Points[tri.V3]).LengthSquared();
            double b2 = (mesh.Points[tri.V3] - mesh.Points[tri.V1]).LengthSquared();
            double c2 = (mesh.Points[tri.V1] - mesh.Points[tri.V2]).LengthSquared();

            double t0 = a2 * (b2 + c2 - a2);
            double t1 = b2 * (c2 + a2 - b2);
            double t2 = c2 * (a2 + b2 - c2);

            Point3D cen = t0 * mesh.Points[tri.V1] +
                        t1 * mesh.Points[tri.V2] + 
                        t2 * mesh.Points[tri.V3];
			
			cen /= (t0 + t1 + t2);


            return cen;            
        }

        /// <summary>
        /// Tests, whether a triangle has any angle smaller than 10°.
        /// </summary>
        /// <param name="mesh">Triangle mesh</param>
        /// <param name="tri">Triangle</param>
        /// <returns>True if triangle is sliver</returns>
        public static bool IsSliver(TriangleMesh mesh, ref Triangle tri) 
        {

            Point3D e0 = (mesh.Points[tri.V2] - mesh.Points[tri.V3]);
            Point3D e1 = (mesh.Points[tri.V3] - mesh.Points[tri.V1]);
            Point3D e2 = (mesh.Points[tri.V1] - mesh.Points[tri.V2]);

            double a0 = Point3D.GetAngle(e0, -e1);
            double a1 = Point3D.GetAngle(e1, -e2);
            double a2 = Point3D.GetAngle(e2, -e0);
            
            double tr = 10 * Math.PI/180;

            if (a0 < tr) return true;
            if (a1 < tr) return true;
            if (a2 < tr) return true;

            return false;
        }
       
    }
}
