﻿using System;
using System.ComponentModel;
using System.Runtime.Serialization;
using System.IO;
using System.Runtime.Serialization.Formatters.Binary;
using SlimDX;

namespace SlimDXRenderSystem.Core
{
    /// <summary>
    /// Camera class based on Quaternions.
    /// </summary>
    public class Camera
    {
        #region Fields
        #region Constants
        private Vector3 WORLD_XAXIS = new Vector3(1.0f, 0.0f, 0.0f);
        private Vector3 WORLD_YAXIS = new Vector3(0.0f, 1.0f, 0.0f);
        private Vector3 WORLD_ZAXIS = new Vector3(0.0f, 0.0f, 1.0f);
        #endregion
        private string name = "";

        private bool targetYAxisOrbiting;
        private float fovx = 90.0f;
        private float aspectRatio;
        private float znear = 0.1f;
        private float zfar = 1000.0f;
        private float orbitMinZoom = 1.1f;
        private float orbitMaxZoom = 500.0f;
        private float orbitOffsetDist = 20.0f;
        private int moveSensitivity = 30;

        private Vector3 eye;
        private Vector3 target;
        private Vector3 targetYAxis;
        private Vector3 xAxis;
        private Vector3 yAxis;
        private Vector3 zAxis;
        private Vector3 viewDir;

        private Quaternion orientation;

        private Matrix viewMatrix;
        private Matrix projMatrix;
        private Matrix viewProjMatrix;

        #region Properties

        public string Name
        {
            get { return name; }
        }

        /// <summary>
        /// Cameras target position.
        /// </summary>
        public Vector3 Target
        {
            get { return target; }
            set { target = value; UnRoll(); }
        }

        /// <summary>
        /// Cameras eye position.
        /// </summary>
        public Vector3 Eye
        {
            get { return eye; }
        }

        /// <summary>
        /// Vector of view direction
        /// </summary>
        public Vector3 ViewDirection
        {
            get { return viewDir; }
        }

        public Vector3 XAxis
        {
            get { return xAxis; }
        }

        public Vector3 YAxis
        {
            get { return yAxis; }
        }

        public Vector3 ZAxis
        {
            get { return zAxis; }
        }

        /// <summary>
        /// Offset distance (zoom).
        /// </summary>
        public float OrbitOffsetDistance
        {
            get { return orbitOffsetDist; }
            set
            {
                if (value < orbitMinZoom)
                    orbitOffsetDist = orbitMinZoom;
                else if (value > orbitMaxZoom)
                    orbitOffsetDist = orbitMaxZoom;
                else
                    orbitOffsetDist = value;
                UpdateViewMatrix();
            }
        }

        /// <summary>
        /// Horizontal field of view.
        /// </summary>
        public float FovX
        {
            get { return fovx; }
            set
            {
                if (value > 180)
                    fovx = 180;
                else if (value < 1)
                    fovx = 1;
                else
                    fovx = value;
                Perspective(fovx, aspectRatio, znear, zfar);
            }
        }

        /// <summary>
        /// Cameras z-near plane.
        /// </summary>
        public float ZNear
        {
            get { return znear; }
            set
            {
                if (value < 0.001f)
                    znear = 0.001f;
                else if (value >= zfar)
                    znear = zfar - 1.0f;
                else
                    znear = value;
                orbitMinZoom = znear;
                Perspective(fovx, aspectRatio, znear, zfar);
            }
        }

        /// <summary>
        /// Cameras z-far plane.
        /// </summary>
        public float ZFar
        {
            get { return zfar; }
            set
            {
                if (value > float.MaxValue)
                    zfar = float.MaxValue;
                else if (value <= znear)
                    zfar = znear + 1.0f;
                else
                    zfar = value;
                orbitMaxZoom = zfar;
                Perspective(fovx, aspectRatio, znear, zfar);
            }
        }

        /// <summary>
        /// Cameras view matrix.
        /// </summary>
        [Browsable(false)]
        public Matrix ViewMatrix
        {
            get { return viewMatrix; }
        }

        /// <summary>
        /// Cameras projection matrix.
        /// </summary>
        [Browsable(false)]
        public Matrix ProjectionMatrix
        {
            get { return projMatrix; }
        }

        /// <summary>
        /// Prefer target Y-axis orbiting. Rotate around its local axis.
        /// </summary>
        public bool PreferTargetYAxisOrbiting
        {
            get { return targetYAxisOrbiting; }
            set { targetYAxisOrbiting = value; UnRoll(); }
        }

        /// <summary>
        /// Sensitivity of camera movement.
        /// </summary>
        public int MoveSensitivity
        {
            get { return moveSensitivity; }
            set
            {
                if (value == 0)
                    return;
                else
                    moveSensitivity = value;
            }
        }

        public float AspectRatio
        {
            get { return aspectRatio; }
            set
            {
                if (value > 0)
                {
                    aspectRatio = value;
                    Perspective(fovx, aspectRatio, znear, zfar);
                }
            }
        }
        #endregion
        #endregion

        /// <summary>
        /// Constructor.
        /// </summary>
        public Camera(string name)
        {
            this.name = name;
            targetYAxisOrbiting = false;
            eye = new Vector3(0.0f, 0.0f, 0.0f);
            target = new Vector3(0.0f, 0.0f, 0.0f);
            xAxis = WORLD_XAXIS;
            yAxis = WORLD_YAXIS;
            zAxis = WORLD_ZAXIS;
            targetYAxis = yAxis;
            viewDir = new Vector3(0.0f, 0.0f, 1.0f);

            orientation = Quaternion.Identity;

            viewMatrix = Matrix.Identity;
            projMatrix = Matrix.Identity;
            viewProjMatrix = Matrix.Identity;
        }

        public Vector3 Position
        {
            get { return eye; }
        }

        /// <summary>
        /// Sets the target witch will camera look at.
        /// </summary>
        /// <param name="target">Position of the target.</param>
        public void LookAt(Vector3 target)
        {
            LookAt(this.eye, target, this.yAxis);
        }

        /// <summary>
        /// Sets the position of the eye, target and the vector pointing to upwards direction.
        /// </summary>
        /// <param name="eye">Position of the camera.</param>
        /// <param name="target">Target witch will camera look at.</param>
        /// <param name="up">Upwards direction for camera.</param>
        public void LookAt(Vector3 eye, Vector3 target, Vector3 up)
        {
            this.eye = eye;
            this.target = target;
            zAxis = target - eye;
            zAxis = Vector3.Normalize(zAxis);

            viewDir = zAxis;

            zAxis = Vector3.Cross(xAxis, up);
            xAxis = Vector3.Normalize(xAxis);

            xAxis = Vector3.Cross(yAxis, zAxis);
            yAxis = Vector3.Normalize(yAxis);
            xAxis = Vector3.Normalize(xAxis);

            viewMatrix = Matrix.Identity;

            viewMatrix.M11 = xAxis.X;
            viewMatrix.M21 = xAxis.Y;
            viewMatrix.M31 = xAxis.Z;
            viewMatrix.M41 = -Vector3.Dot(xAxis, eye);

            viewMatrix.M12 = yAxis.X;
            viewMatrix.M22 = yAxis.Y;
            viewMatrix.M32 = yAxis.Z;
            viewMatrix.M41 = -Vector3.Dot(yAxis, eye);

            viewMatrix.M13 = zAxis.X;
            viewMatrix.M23 = zAxis.Y;
            viewMatrix.M33 = zAxis.Z;
            viewMatrix.M43 = -Vector3.Dot(zAxis, eye);

            viewProjMatrix = viewMatrix * projMatrix;

            orientation = Quaternion.RotationMatrix(viewMatrix);
            UpdateViewMatrix();
        }

        /// <summary>
        /// Sets the perspective of camera.
        /// </summary>
        /// <param name="fovx">Horizontal field of view.</param>
        /// <param name="aspect">Aspect ration of the view.</param>
        /// <param name="znear">Near clipping plane.</param>
        /// <param name="zfar">Far clipping plane.</param>
        public void Perspective(float fovx, float aspect, float znear, float zfar)
        {
            float e = 1.0f / (float)Math.Tan(Utils.ToRadians(fovx) / 2.0f);
            float aspectInv = 1.0f / aspect;
            float fovy = 2.0f * (float)Math.Atan(aspectInv / e);
            float xScale = 1.0f / (float)Math.Tan(0.5f * fovy);
            float yScale = xScale / aspectInv;

            projMatrix.M11 = xScale;
            projMatrix.M21 = 0.0f;
            projMatrix.M31 = 0.0f;
            projMatrix.M41 = 0.0f;

            projMatrix.M12 = 0.0f;
            projMatrix.M22 = yScale;
            projMatrix.M32 = 0.0f;
            projMatrix.M42 = 0.0f;

            projMatrix.M13 = 0.0f;
            projMatrix.M23 = 0.0f;
            projMatrix.M33 = zfar / (zfar - znear);
            projMatrix.M43 = -znear * zfar / (zfar - znear);

            projMatrix.M14 = 0.0f;
            projMatrix.M24 = 0.0f;
            projMatrix.M34 = 1.0f;
            projMatrix.M44 = 0.0f;

            viewProjMatrix = viewMatrix * projMatrix;

            this.fovx = fovx;
            this.aspectRatio = aspect;
            this.znear = znear;
            this.zfar = zfar;
        }

        /// <summary>
        /// Camera movement.
        /// </summary>
        /// <param name="x">Horizontal movement.</param>
        /// <param name="y">Vertical movement.</param>
        /// <param name="z">Depth movement.</param>
        public void Move(float x, float y, float z)
        {
            x /= moveSensitivity;
            y /= moveSensitivity;
            z /= moveSensitivity;

            target -= xAxis * x;
            target += yAxis * y;
            target -= zAxis * z;

            UpdateViewMatrix();
        }

        /// <summary>
        /// Orbit rotation of the camera around the target.
        /// </summary>
        /// <param name="headingDeg">Degrees of heading, witch will be used for rotation.</param>
        /// <param name="pitchDeg">Degrees of pitch, witch will be used for rotation.</param>
        /// <param name="rollDeg">Degrees of roll, witch will be used for rotation.</param>
        public void Rotate(float headingDeg, float pitchDeg, float rollDeg)
        {
            float heading = Utils.ToRadians(-headingDeg);
            float pitch = Utils.ToRadians(-pitchDeg);

            Quaternion rotation;

            if (targetYAxisOrbiting)
            {
                if (heading != 0.0f)
                {
                    rotation = Quaternion.RotationAxis(targetYAxis, heading);
                    orientation = Quaternion.Multiply(rotation, orientation);
                }

                if (pitch != 0.0f)
                {
                    rotation = Quaternion.RotationAxis(WORLD_XAXIS, pitch);
                    orientation = Quaternion.Multiply(orientation, rotation);
                }
            }
            else
            {
                float roll = Utils.ToRadians(rollDeg);

                rotation = Quaternion.RotationYawPitchRoll(heading, pitch, roll);
                orientation = Quaternion.Multiply(orientation, rotation);
            }

            UpdateViewMatrix();
        }

        /// <summary>
        /// Zooms the camera in or out.
        /// </summary>
        /// <param name="zoom">Value to be added to zoom.</param>
        public void Zoom(float zoom)
        {
            Vector3 offset = eye - target;
            orbitOffsetDist += zoom;
            orbitOffsetDist = Math.Min(Math.Max(orbitOffsetDist, orbitMinZoom), orbitMaxZoom);
            offset *= orbitOffsetDist;
            eye = offset + target;

            UpdateViewMatrix();
        }

        /// <summary>
        /// Saves the position of the camera to the file.
        /// </summary>
        /// <param name="file">File name of the camera transformations used for saving.</param>
        public void SavePosition(String file)
        {
            IFormatter formatter = new BinaryFormatter();
            Stream stream = new FileStream(file,
                                             FileMode.Create,
                                             FileAccess.Write, FileShare.None);
            formatter.Serialize(stream, znear);
            formatter.Serialize(stream, zfar);
            formatter.Serialize(stream, orbitOffsetDist);
            formatter.Serialize(stream, eye);
            formatter.Serialize(stream, target);
            formatter.Serialize(stream, targetYAxis);
            formatter.Serialize(stream, viewDir);
            formatter.Serialize(stream, orientation);
            formatter.Serialize(stream, projMatrix);
            stream.Close();
        }

        /// <summary>
        /// Loads the position of the camera to the file.
        /// </summary>
        /// <param name="file">File name of the camera transformations used for loading.</param>
        public void LoadPosition(String file)
        {
            IFormatter formatter = new BinaryFormatter();
            Stream stream = new FileStream(file,
                          FileMode.Open,
                          FileAccess.Read,
                          FileShare.Read);
            znear = (float)formatter.Deserialize(stream);
            zfar = (float)formatter.Deserialize(stream);
            orbitOffsetDist = (float)formatter.Deserialize(stream);
            eye = (Vector3)formatter.Deserialize(stream);
            target = (Vector3)formatter.Deserialize(stream);
            targetYAxis = (Vector3)formatter.Deserialize(stream);
            viewDir = (Vector3)formatter.Deserialize(stream);
            orientation = (Quaternion)formatter.Deserialize(stream);
            projMatrix = (Matrix)formatter.Deserialize(stream);
            stream.Close();

            UpdateViewMatrix();
        }

        /// <summary>
        /// Reset the rotation of the camera.
        /// </summary>
        private void UnRoll()
        {
            LookAt(eye, target, targetYAxis);
        }

        /// <summary>
        /// Update the view matrix by quaternion orientation of the camera.
        /// </summary>
        private void UpdateViewMatrix()
        {
            orientation = Quaternion.Normalize(orientation);
            viewMatrix = Matrix.RotationQuaternion(orientation);

            xAxis = new Vector3(viewMatrix.M11, viewMatrix.M21, viewMatrix.M31);
            yAxis = new Vector3(viewMatrix.M12, viewMatrix.M22, viewMatrix.M32);
            zAxis = new Vector3(viewMatrix.M13, viewMatrix.M23, viewMatrix.M33);
            viewDir = zAxis;
            eye = target + viewDir * -orbitOffsetDist;

            viewMatrix.M41 = -Vector3.Dot(xAxis, eye);
            viewMatrix.M42 = -Vector3.Dot(yAxis, eye);
            viewMatrix.M43 = -Vector3.Dot(zAxis, eye);

            viewProjMatrix = viewMatrix * projMatrix;
        }
    }
}
