﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using SlimDX;
using SlimDXRenderSystem.Direct3DBase;
using SlimDXRenderSystem.Managers;
using System.IO;
using System.Globalization;

namespace SlimDXRenderSystem.Core
{
    public class Object3D : Movable
    {
        #region Fields
        private Group group;
        private Dictionary<String, Object3D> childrens;
        private Object3D parent;

        private bool visible;
        private Mesh mesh;
        private Material material;

        private int vertexCount;
        private int primitiveCount;

        #region Properties
        [Category("Relationships")]
        public Group Group
        {
            get { return group; }
            set { group = value; }
        }

        [Category("Information")]
        public bool Visible
        {
            get { return visible; }
            set { visible = value; }
        }

        [Category("Relationships")]
        public Object3D Parent
        {
            get { return parent; }
            set { parent = value; }
        }

        [Category("Relationships")]
        public Mesh Mesh
        {
            get { return mesh; }
        }

        [Category("Mesh Info")]
        public int Vertices
        {
            get { return vertexCount; }
            set { vertexCount = value; }
        }

        [Category("Mesh Info")]
        public int Primitives
        {
            get { return primitiveCount; }
            set { primitiveCount = value; }
        }

        [Browsable(false)]
        public Material Material
        {
            get { return material; }
            set { material = value; }
        }

        #endregion
        #endregion

        #region Private Methods
        
        #endregion

        #region Public Methods
        public Object3D(String name)
        {
            Name = name;
            parent = null;

            visible = false;
            mesh = null;

            childrens = new Dictionary<string, Object3D>();

            material = new Material();

            CalcMatrices();
        }

        public void AttachObject(ref Object3D child)
        {
            SceneManager.Instance.DetachObject(child.Name);
            childrens.Add(child.Name, child);
            child.parent = this;
        }

        public void AddToGroup(Group gr)
        {
            gr.AddObject(this);
        }

        public void RemoveFromGroup(Group gr)
        {
            gr.RemoveObject(this);
        }

        public Object3D GetChild(String name)
        {
            return childrens[name];
        }

        public Object3D FindChild(String name)
        {
            if (childrens != null)
            {
                if (this.childrens.ContainsKey(name))
                    return childrens[name];

                Object3D temp = null;
                foreach (Object3D obj in childrens.Values)
                {
                    temp = obj.FindChild(name);
                }
                return temp;
            }
            return null;
        }

        public void SetMeshVertices(Vector3[] points, Vector3[] normals, int[] colors)
        {
            if (mesh != null)
            {
                mesh.Dispose();
                mesh = null;
            }
            mesh = new Direct3D9.Mesh9();
            mesh.InitVertexBuffer(points, normals, colors);            
        }

        public void AddMeshIndices(string name, MeshType type, uint[] indices)
        {
            if (mesh != null)
            {
                mesh.AddIndexBuffer(name, type, indices);
            }            
        }

        public void EnableDrawIndices(string name)
        {
            if (mesh != null)
            {
                mesh.EnableDrawIndices(name);
            }
        }

        public void DisableDrawIndices(string name)
        {
            if (mesh != null)
            {
                mesh.DisableDrawIndices(name);
            }
        }

        public void SetMeshSameIndexBuffer(Object3D reference)
        {
            if (mesh != null && reference.mesh != null)
            {
                
                ((Direct3D9.Mesh9)mesh).IndexBuffers = ((Direct3D9.Mesh9)reference.mesh).IndexBuffers;
                ((Direct3D9.Mesh9)mesh).IndexBuffersDrawables = ((Direct3D9.Mesh9)reference.mesh).IndexBuffersDrawables;
                ((Direct3D9.Mesh9)mesh).IndexBuffersLengths = ((Direct3D9.Mesh9)reference.mesh).IndexBuffersLengths;
                ((Direct3D9.Mesh9)mesh).IndexBuffersType = ((Direct3D9.Mesh9)reference.mesh).IndexBuffersType;
            }
        }

        public void LoadFromObj(string filename)
        {
            if (mesh != null)
            {
                mesh.Dispose();
                mesh = null;
            }

            mesh = new Direct3D9.Mesh9();

            // first pass determines the number of vertices and the number of triangles
            StreamReader sr;
            try
            {
                sr = new StreamReader(filename);
            }
            catch (FileNotFoundException fnfe)
            {
                throw new Exception(fnfe.Message);
            }

            int triangleCount = 0;
            int vertexCount = 0;
            int lineCount = 0;
            string line = sr.ReadLine();

            bool normalsFound = false;

            while (line != null)
            {
                line = line.Trim();

                if (line.StartsWith("v ")) vertexCount++;
                if (line.StartsWith("vn ")) normalsFound = true;
                if ((line.StartsWith("f ")) || (line.StartsWith("fo ")) || line.StartsWith("f\t"))
                {
                    triangleCount++;
                    if (line.Split(new char[] { ' ', '\t' }).Length == 5)
                        triangleCount++;
                }
                line = sr.ReadLine();
                lineCount++;
            }

            sr.Close();

            // second pass performs the actual parsing
            sr = new StreamReader(filename);
            Vector3[] vertices = new Vector3[vertexCount];
            Vector3[] normals = null;
            if (normalsFound)
                normals = new Vector3[vertexCount];
            Vector3[] triangles = new Vector3[triangleCount];

            int vi = 0;
            int ti = 0;
            int ni = 0;
            Vector3 point;
            Vector3 normal;
            Vector3 triangle;
            NumberFormatInfo nfi = new NumberFormatInfo();
            nfi.NumberDecimalSeparator = ".";
            nfi.NumberGroupSeparator = ",";

            int linePos = 0;
            line = sr.ReadLine();
            while (line != null)
            {
                line = line.Trim();

                //parsing of a vertex
                if (line.StartsWith("v "))
                {
                    string[] coords = line.Split(new char[] { ' ' }, 4, StringSplitOptions.RemoveEmptyEntries);

                    float c1 = float.Parse(coords[1], nfi);
                    float c2 = float.Parse(coords[2], nfi);
                    float c3 = float.Parse(coords[3], nfi);

                    point = new Vector3(c1, c2, c3);
                    vertices[vi] = point;
                    vi++;
                }

                //parsing of a vertex normal
                if (line.StartsWith("vn "))
                {
                    string[] coords = line.Split(new char[] { ' ' }, 4, StringSplitOptions.RemoveEmptyEntries);
                    float c1, c2, c3;
                    try
                    {
                        c1 = float.Parse(coords[1], nfi);
                        c2 = float.Parse(coords[2], nfi);
                        c3 = float.Parse(coords[3], nfi);
                    }
                    catch (FormatException)
                    {
                        c1 = 0;
                        c2 = 0;
                        c3 = 1;
                    }

                    normal = new Vector3(c1, c2, c3);
                    normals[ni] = normal;
                    ni++;
                }

                // parsing of a triangle
                if ((line.StartsWith("f ")) || (line.StartsWith("fo ")) || (line.StartsWith("f\t")))
                {
                    string[] indices = line.Split(new char[] { ' ', '\t' }, 5, StringSplitOptions.RemoveEmptyEntries);

                    string part = indices[1].Split('/')[0];
                    int v1 = int.Parse(part) - 1;
                    part = indices[2].Split('/')[0];
                    int v2 = int.Parse(part) - 1;
                    part = indices[3].Split('/')[0];
                    int v3 = int.Parse(part) - 1;

                    triangle = new Vector3(v1, v2, v3);
                    triangles[ti] = triangle;
                    ti++;

                    if (indices.Length == 5)
                    {
                        part = indices[4].Split('/')[0];
                        v2 = int.Parse(part) - 1;

                        triangle = new Vector3(v1, v3, v2);
                        triangles[ti] = triangle;
                        ti++;
                    }
                }
                line = sr.ReadLine();
                linePos++;
            }

            int[] colors = new int[vertexCount];
            for (int i = 0; i < vertexCount; i++)
            {
                colors[i] = System.Drawing.Color.Red.ToArgb();
            }

            mesh.InitVertexBuffer(vertices, normals, colors);
            uint[] inds = new uint[triangleCount * 3];
            for (int i = 0; i < triangleCount; i++)
            {
                inds[3 * i] = (uint)triangles[i].X;
                inds[3 * i + 1] = (uint)triangles[i].Y;
                inds[3 * i + 2] = (uint)triangles[i].Z;
            }
            mesh.AddIndexBuffer("base", MeshType.TriangleMesh, inds);
            mesh.EnableDrawIndices("base");
        }

        public void Draw()
        {
            if (Visible && mesh != null)
            {
                if (group != null)
                    SceneManager.Instance.SetWorldTransform(SceneManager.Instance.WorldMatrix * LocalWorld * group.LocalWorld);
                else
                    SceneManager.Instance.SetWorldTransform(SceneManager.Instance.WorldMatrix * LocalWorld);

                EffectManager.Instance.SetParameter(Effect.MATERIAL, material.MaterialStruct);

                EffectManager.Instance.SuperEffect.ApplyChanges();
                mesh.Draw();
            }

            foreach (Object3D obj in childrens.Values)
            {
                obj.Draw();
            }
        }

        public void Dispose()
        {
            if (mesh != null)
            {
                mesh.Dispose();
                mesh = null;
            }

            foreach(Object3D obj in childrens.Values)
            {
                obj.Dispose();
            }
            childrens.Clear();
        }
        #endregion
    }
}
