﻿using System;
using System.Drawing;
using System.Collections.Generic;
using SlimDX;
using SlimDX.Direct3D9;
using RenderMgr = SlimDXRenderSystem.Managers.RenderManager;
using Mesh = SlimDXRenderSystem.Direct3DBase.Mesh;

namespace SlimDXRenderSystem.Direct3D9
{
    /// <summary>
    /// Mesh class using Direct3D9 contains vertex and index buffers for Direct3D9 rendering.
    /// </summary>
    public class Mesh9 : Mesh
    {
        #region Fields
        private Device device;
        private Vertex[] vertices;
        private Material material;

        private VertexBuffer vertexBuffer;
        private Dictionary<string, IndexBuffer> indexBuffers;
        private Dictionary<string, MeshType> ibType;
        private Dictionary<string, int> ibLength;
        private Dictionary<string, bool> ibDraw;

        private VertexElement[] vertexElements;
        private VertexDeclaration vertexDeclaration;

        #region Properties
        public Dictionary<string, IndexBuffer> IndexBuffers
        {
            get { return indexBuffers; }
            set { indexBuffers = value; }
        }

        public Dictionary<string, MeshType> IndexBuffersType
        {
            get { return ibType; }
            set { ibType = value; }
        }

        public Dictionary<string, int> IndexBuffersLengths
        {
            get { return ibLength; }
            set { ibLength = value; }
        }

        public Dictionary<string, bool> IndexBuffersDrawables
        {
            get { return ibDraw; }
            set { ibDraw = value; }
        }

        public override int Vertices
        {
            get { return vertices.Length; }
        }

        public override int Primitives
        {
            get
            {
                int primitives = 0;
                foreach (string name in ibDraw.Keys)
                {
                    switch (ibType[name])
                    {
                        case MeshType.TriangleMesh:
                            primitives += ibLength[name] / 3;
                            break;
                        case MeshType.LineMesh:
                            primitives += ibLength[name] / 2;
                            break;
                    }
                }
                return primitives;
            }
        }

        #endregion
        #endregion

        #region Public Methods
        /// <summary>
        /// Constructor of the mesh.
        /// </summary>
        /// <param name="indices">Vertices indice.</param>
        public Mesh9()
        {
            device = ((Renderer9)RenderMgr.Instance.Renderer).Device;
            indexBuffers = new Dictionary<string, IndexBuffer>();
            ibType = new Dictionary<string, MeshType>();
            ibLength = new Dictionary<string, int>();
            ibDraw = new Dictionary<string, bool>();
            CreateDefaultMaterial();
        }
        
        /// <summary>
        /// This method creates default material for the mesh.
        /// </summary>
        public void CreateDefaultMaterial()
        {
            material = new Material();
            material.Diffuse = Color.Gray;
            material.Ambient = Color.Gray;
            material.Specular = Color.Gray;
            material.Power = 50.0f;
        }

        /// <summary>
        /// Draw as triangle mesh.
        /// </summary>
        public override void Draw()
        {
            try
            {
                Material temp = device.Material;
                device.Material = material;

                device.SetStreamSource(0, vertexBuffer, 0, Vertex.SizeInBytes);
                device.VertexDeclaration = vertexDeclaration;


                foreach (string name in ibDraw.Keys)
                {
                    if (ibDraw[name] == true)
                    {
                        device.Indices = indexBuffers[name];
                        try
                        {
                            switch (ibType[name])
                            {
                                case MeshType.TriangleMesh:
                                    device.DrawIndexedPrimitives(PrimitiveType.TriangleList, 0, 0, vertices.Length, 0, ibLength[name] / 3);
                                    break;
                                case MeshType.LineMesh:
                                    device.DrawIndexedPrimitives(PrimitiveType.LineList, 0, 0, vertices.Length, 0, ibLength[name] / 2);
                                    break;
                            }
                        }
                        catch (SlimDXException sdxe)
                        {
                            System.Windows.Forms.MessageBox.Show(sdxe.Message);
                        }
                    }
                }
                device.Material = temp;
            }
            catch (Direct3D9Exception ex)
            {
                throw ex;
            }
        }

        /// <summary>
        /// Initialize vertex buffers.
        /// </summary>
        /// <param name="points">Vertices position.</param>
        /// <param name="normals">Vertices normal</param>
        /// <param name="colors"></param>
        public override void InitVertexBuffer(Vector3[] points, Vector3[] normals, int[] colors)
        {
            if (normals != null && colors != null)
                if ((points.Length != normals.Length) || (points.Length != colors.Length))
                    return;

            vertices = new Vertex[points.Length];
            for (int i = 0; i < vertices.Length; i++)
            {
                Vector3 point = points[i];
                vertices[i].Position = new Vector4(point.X, point.Y, point.Z, 1.0f);

                if (normals != null)
                {
                    Vector3 normal = normals[i];
                    vertices[i].Normal = new Vector4(-normal.X, -normal.Y, -normal.Z, 1.0f);
                }
                else
                    vertices[i].Normal = new Vector4();

                if (colors != null)
                    vertices[i].Color = colors[i];
                else
                    vertices[i].Color = Color.Gray.ToArgb();
            }

            //Vertex Buffer
            vertexBuffer = new VertexBuffer(device, vertices.Length * Vertex.SizeInBytes, Usage.WriteOnly, VertexFormat.None, Pool.Managed);

            vertexBuffer.Lock(0, vertices.Length * Vertex.SizeInBytes, LockFlags.Discard).WriteRange(vertices);
            vertexBuffer.Unlock();

            vertexElements = new[] {
        		new VertexElement(0, 0, DeclarationType.Float4, DeclarationMethod.Default, DeclarationUsage.Position, 0),
                new VertexElement(0, 16, DeclarationType.Float4, DeclarationMethod.Default, DeclarationUsage.Normal, 0),
        		new VertexElement(0, 32, DeclarationType.Color, DeclarationMethod.Default, DeclarationUsage.Color, 0),
				VertexElement.VertexDeclarationEnd
        	};

            vertexDeclaration = new VertexDeclaration(device, vertexElements);
        }

        public override void AddIndexBuffer(string name, MeshType type, uint[] indices)
        {
            //Index Buffer
            IndexBuffer indexBuffer = new IndexBuffer(device, sizeof(uint) * indices.Length, Usage.WriteOnly, Pool.Managed, false);
            indexBuffer.Lock(0, 0, LockFlags.Discard).WriteRange(indices);
            indexBuffer.Unlock();
            indexBuffers.Add(name, indexBuffer);
            ibType.Add(name, type);
            ibLength.Add(name, indices.Length);
            ibDraw.Add(name, false);
        }

        public override void EnableDrawIndices(string name)
        {
            ibDraw[name] = true;
        }

        public override void DisableDrawIndices(string name)
        {
            ibDraw[name] = false;
        }

        public override string ToString()
        {
            return "Direct3D9 Mesh";
        }

        /// <summary>
        /// Clear all resources.
        /// </summary>
        public override void Dispose()
        {
            if (vertexBuffer != null && !vertexBuffer.Disposed)
            {
                vertexBuffer.Dispose();
                vertexBuffer = null;
            }
            foreach (IndexBuffer ib in indexBuffers.Values)
            {
                if (!ib.Disposed)
                    ib.Dispose();
            }
            indexBuffers.Clear();

            if (vertexDeclaration != null && !vertexDeclaration.Disposed)
            {
                vertexDeclaration.Dispose();
                vertexDeclaration = null;
            }
        }
        #endregion
    }
}
