﻿using System;
using System.Drawing;
using System.ComponentModel;
using SlimDX;
using SlimDXRenderSystem.Core;
using SlimDXRenderSystem.Direct3DBase;
using SlimDX.Direct3D9;
using RenderMgr = SlimDXRenderSystem.Managers.RenderManager;
using SceneMgr = SlimDXRenderSystem.Managers.SceneManager;
using EffectMgr = SlimDXRenderSystem.Managers.EffectManager;
using Effect = SlimDXRenderSystem.Direct3DBase.Effect;
using Light = SlimDXRenderSystem.Core.Light;
using System.Runtime.Serialization;
using System.IO;
using System.Runtime.Serialization.Formatters.Binary;
using System.Reflection;

namespace SlimDXRenderSystem.Direct3D9
{
    /// <summary>
    /// Direct3D9 Renderer
    /// </summary>
    public class Renderer9 : Renderer
    {
        #region Fields
        private Direct3D d3d;
        private Device device;
        private PresentParameters presentParams;
        private Light globalLight;
        private MultisampleQuality multisample;
        private SlimDX.Direct3D9.Font font;
        private Sprite fontSprite;
        private Texture pointTex;
  
        #region Properties
        /// <summary>
        /// Returns Direct3D9 device.
        /// </summary>
        [Browsable(false)]
        public Device Device
        {
            get { return device; }
        }

        /// <summary>
        /// Shade mode of mesh.
        /// </summary>
        public override ShadeMode Shademode
        {
            set
            {
                device.SetRenderState(RenderState.ShadeMode, value);
            }
        }

        /// <summary>
        /// Fill mode of mesh.
        /// </summary>
        public override FillMode Fillmode
        {
            set
            {
                device.SetRenderState(RenderState.FillMode, value);
            }
        }

        /// <summary>
        /// Back face culling.
        /// </summary>
        public override Cull Culling
        {
            set
            {
                device.SetRenderState(RenderState.CullMode, value);
            }
        }

        /// <summary>
        /// Antialiasing multi sample type.
        /// </summary>
        public override MultisampleQuality Multisample
        {
            get { return multisample; }
            set
            {
                if (d3d.CheckDeviceMultisampleType(0, DeviceType.Hardware, Format.X8R8G8B8, true, (MultisampleType)value))
                {
                    multisample = value;
                    RenderMgr.Instance.Reset();
                }
            }
        }

        /// <summary>
        /// Set default lighting the scene
        /// </summary>
        public bool Lighting
        {
            set
            {
                device.SetRenderState(RenderState.Lighting, value);
            }
        }

        /// <summary>
        /// Color of the default ambient light
        /// </summary>
        public Color Ambient
        {
            set
            {
                device.SetRenderState(RenderState.Ambient, value.ToArgb());
            }
        }
        #endregion
        #endregion

        #region Private Methods
        /// <summary>
        /// Sets present parameters for Direct3D9 device.
        /// </summary>
        private void SetPresentParameters()
        {
            presentParams = new PresentParameters()
            {
                SwapEffect = SwapEffect.Discard,
                EnableAutoDepthStencil = true,
                AutoDepthStencilFormat = Format.D16,
                BackBufferFormat = Format.X8R8G8B8,
                BackBufferHeight = RenderMgr.Instance.RenderWindow.Height,
                BackBufferWidth = RenderMgr.Instance.RenderWindow.Width,
                Multisample = (MultisampleType)this.multisample
            };
        }

        /// <summary>
        /// Sets default renderstates.
        /// </summary>
        private void SetRenderStates()
        {
            device.SetRenderState(RenderState.FillMode, RenderMgr.Instance.Fillmode);
            device.SetRenderState(RenderState.CullMode, RenderMgr.Instance.Culling);
            device.SetRenderState(RenderState.ZEnable, true);
            device.SetRenderState(RenderState.ShadeMode, RenderMgr.Instance.Shademode);            
        }

        /// <summary>
        /// Creates global light.
        /// </summary>
        private void SetGlobalLight()
        {
            globalLight = SceneMgr.Instance.CreateLight("Default");
            globalLight.Diffuse = new Vector4(1.0f, 1.0f, 1.0f, 1.0f);
            globalLight.Type = LightType.Directional;
            globalLight.Direction = new Vector3(0, 0, 1);
            globalLight.Specular = new Vector4(1.0f, 1.0f, 1.0f, 1.0f);
        }
        #endregion

        #region Public Methods
        /// <summary>
        /// Constructor.
        /// </summary>
        public Renderer9()
        {
            multisample = RenderMgr.Instance.Multisample;
        }

        /// <summary>
        /// Initialize device and sets all needed states.
        /// </summary>
        public override void Initialize()
        {
            base.Initialize();
            try
            {
                SetPresentParameters();
                d3d = new Direct3D();
                device = new Device(d3d, 0, DeviceType.Hardware, RenderMgr.Instance.RenderWindow.Handle, CreateFlags.HardwareVertexProcessing, presentParams);
                SetRenderStates();
                SetGlobalLight();

                Assembly assembly = this.GetType().Assembly;                
                pointTex = Texture.FromStream(device, assembly.GetManifestResourceStream("SlimDXRenderSystem.Resources.Textures.point.bmp"));
            }
            catch (Exception e)
            {
                throw e;
            }
        }

        

        /// <summary>
        /// Render scene.
        /// </summary>
        public override void Render()
        {
            try
            {
                Effect effect = EffectMgr.Instance.SuperEffect;

                effect.SetParameter(Effect.VIEW, SceneMgr.Instance.Camera.ViewMatrix);
                effect.SetParameter(Effect.PROJECTION, SceneMgr.Instance.Camera.ProjectionMatrix);
                effect.SetParameter(Effect.WORLD, SceneMgr.Instance.WorldMatrix);
                effect.SetParameter(Effect.WORLDINVTRANSPOSE, Matrix.Transpose(Matrix.Invert(SceneMgr.Instance.WorldMatrix)));
                effect.SetParameter(Effect.WORLDVIEWPROJECTION, SceneMgr.Instance.WorldMatrix * SceneMgr.Instance.Camera.ViewMatrix * SceneMgr.Instance.Camera.ProjectionMatrix);
                effect.SetParameter(Effect.CAMERAPOSITION, SceneMgr.Instance.Camera.Position);
                effect.SetParameter(Effect.LIGHT, globalLight.LightStruct);


                device.Clear(ClearFlags.Target | ClearFlags.ZBuffer, BackgroundColor, 1.0f, 0);
                device.BeginScene();

                int passes = ((Effect9)effect).Begin();

                for (int i = 0; i < passes; i++)
                {
                    ((Effect9)effect).BeginPass(i);
                    //Draw meshes
                    SceneMgr.Instance.DrawObjects();

                    ((Effect9)effect).EndPass();
                }


                ((Effect9)effect).End();
                
                //Draw text
                RenderText();
                
                device.EndScene();
                device.Present();
            }
            catch (SlimDXException e)
            {
                try
                {
                    RenderMgr.Instance.Reset();
                }
                catch (Exception ex)
                {
                    throw ex;
                }
            }
        }

        public override void RenderText()
        {
            if (fontSprite == null)
                fontSprite = new Sprite(device);
            if (font == null)
            {
                System.Drawing.Font fnt = new System.Drawing.Font(FontFamily.GenericSansSerif, 12.0f);
                font = new SlimDX.Direct3D9.Font(device, fnt);
            }
            fontSprite.Begin(SpriteFlags.AlphaBlend | SpriteFlags.SortTexture);

            if (RenderMgr.Instance.RenderTextAction != null)
                RenderMgr.Instance.RenderTextAction.Invoke();

            fontSprite.End();
        }

        public override void DrawPoint(Vector3 pos)
        {
            fontSprite.Draw(pointTex, new Vector3(2, 2, 0), pos, Color.White);
        }

        public override void DrawText(int x, int y, string text)
        {
            font.DrawString(fontSprite, text, x, y, Color.White);
        }

        /// <summary>
        /// Reset device and set all needed states.
        /// </summary>
        public override void Reset()
        {
            SetPresentParameters();
            if (font != null)
            {
                font.Dispose();
                font = null;
            }
            if (fontSprite != null)
            {
                fontSprite.Dispose();
                fontSprite = null;
            }
            device.Reset(presentParams);
            SetRenderStates();
            SetGlobalLight();
        }

        /// <summary>
        /// Capture screenshot of window frame
        /// </summary>
        /// <param name="name"></param>
        /// <param name="number"></param>
        public override void CaptureScreenshot(String name, int number)
        {
            Surface backbuffer = device.GetBackBuffer(0, 0);
            Surface.ToFile(backbuffer,  name + "_" + number + ".bmp", ImageFileFormat.Bmp);
            backbuffer.Dispose();
        }

        /// <summary>
        /// Saves the position of the light to the file.
        /// </summary>
        /// <param name="file">File name of the light transformations used for saving.</param>
        public override void SaveLightTransformations(String file)
        {
            IFormatter formatter = new BinaryFormatter();
            Stream stream = new FileStream(file,
                                             FileMode.Create,
                                             FileAccess.Write, FileShare.None);
            formatter.Serialize(stream, globalLight.Direction.X);
            formatter.Serialize(stream, globalLight.Direction.Y);
            formatter.Serialize(stream, globalLight.Direction.Z);
            formatter.Serialize(stream, globalLight.Position.X);
            formatter.Serialize(stream, globalLight.Position.Y);
            formatter.Serialize(stream, globalLight.Position.Z);
            formatter.Serialize(stream, globalLight.Diffuse.X);
            formatter.Serialize(stream, globalLight.Diffuse.Y);
            formatter.Serialize(stream, globalLight.Diffuse.Z);
            formatter.Serialize(stream, globalLight.Diffuse.W);
            formatter.Serialize(stream, globalLight.Ambient.X);
            formatter.Serialize(stream, globalLight.Ambient.Y);
            formatter.Serialize(stream, globalLight.Ambient.Z);
            formatter.Serialize(stream, globalLight.Ambient.W);
            formatter.Serialize(stream, globalLight.Specular.X);
            formatter.Serialize(stream, globalLight.Specular.Y);
            formatter.Serialize(stream, globalLight.Specular.Z);
            formatter.Serialize(stream, globalLight.Specular.W);
            stream.Close();
        }

        /// <summary>
        /// Loads the position of the light to the file.
        /// </summary>
        /// <param name="file">File name of the light transformations used for loading.</param>
        public override void LoadLightTransformations(String file)
        {
            IFormatter formatter = new BinaryFormatter();
            Stream stream = new FileStream(file,
                          FileMode.Open,
                          FileAccess.Read,
                          FileShare.Read);
            float x = (float)formatter.Deserialize(stream);
            float y = (float)formatter.Deserialize(stream);
            float z = (float)formatter.Deserialize(stream);
            globalLight.Direction = new Vector3(x, y, z);

            x = (float)formatter.Deserialize(stream);
            y = (float)formatter.Deserialize(stream);
            z = (float)formatter.Deserialize(stream);
            globalLight.Position = new Vector3(x, y, z);

            x = (float)formatter.Deserialize(stream);
            y = (float)formatter.Deserialize(stream);
            z = (float)formatter.Deserialize(stream);
            float w = (float)formatter.Deserialize(stream);
            globalLight.Diffuse = new Vector4(x, y, z, w);

            x = (float)formatter.Deserialize(stream);
            y = (float)formatter.Deserialize(stream);
            z = (float)formatter.Deserialize(stream);
            w = (float)formatter.Deserialize(stream);
            globalLight.Ambient = new Vector4(x, y, z, w);

            x = (float)formatter.Deserialize(stream);
            y = (float)formatter.Deserialize(stream);
            z = (float)formatter.Deserialize(stream);
            w = (float)formatter.Deserialize(stream);
            globalLight.Specular = new Vector4(x, y, z, w);

            stream.Close();
        }

        /// <summary>
        /// Clear all resources.
        /// </summary>
        public override void Dispose()
        {
            if (d3d != null && !d3d.Disposed)
            {
                d3d.Dispose();
                d3d = null;
            }
            if (device != null && !device.Disposed)
            {
                device.Dispose();
                device = null;
            }
            if (font != null && !font.Disposed)
            {
                font.Dispose();
                font = null;
            }
            if (fontSprite != null && !fontSprite.Disposed)
            {
                fontSprite.Dispose();
                fontSprite = null;
            }
            base.Dispose();
        }
        #endregion
    }
}
