﻿using System;
using System.Collections.Generic;
using SlimDX;
using SlimDX.DirectInput;

namespace SlimDXRenderSystem.Managers
{
    public class InputManager : Core.Singleton<InputManager>
    {
        #region Fields
        //Keyboard key actions
        private Dictionary<Key, Action> keyClickActions;
        private Dictionary<Key, Action> keyIsPressedActions;
        private Dictionary<Key, Action> keyIsReleasedActions;
        //Mouse buttons actions
        private Dictionary<MouseObject, Action> mouseClickActions;
        private Dictionary<MouseObject, Action> mouseIsPressedActions;
        private Dictionary<MouseObject, Action> mouseIsReleasedActions;
        //Mouse move actions
        private Dictionary<Key, Action<int, int, int>> mouseMoveKeyActions;
        private Dictionary<MouseObject, Action<int, int, int>> mouseMoveButActions;
        private Action<int, int, int> mouseMove;

        private bool acquiredMouse = false;
        private bool acquiredKeyboard = false;
        private DirectInput dinput;
        private Keyboard keyboard;
        private KeyboardState oldKeyState;
        private KeyboardState keyState;
        private Mouse mouse;
        private MouseState mouseState;
        private MouseState oldMouseState;

        #region Properties
        /// <summary>
        /// Returns mouse state.
        /// </summary>
        public MouseState Mousestate
        {
            get { return mouseState; }
        }

        public Dictionary<Key, Action> KeyClickActions
        {
            get { return keyClickActions; }
        }

        /// <summary>
        /// Sets or get the input mouse acquire status.
        /// </summary>
        public bool AcquiredMouse
        {
            get { return acquiredMouse; }
            set { acquiredMouse = value; }
        }

        /// <summary>
        /// Sets or get the input keyboard acquire status.
        /// </summary>
        public bool AcquiredKeyboard
        {
            get { return acquiredKeyboard; }
            set { acquiredKeyboard = value; }
        }

        /// <summary>
        /// Returns keyboard state.
        /// </summary>
        public KeyboardState Keystate
        {
            get { return keyState; }
        }
        #endregion
        #endregion

        #region Public methods
        /// <summary>
        /// Initializate input for window.
        /// </summary>
        /// <param name="wndHandle">Handle of main window.</param>
        public void Initialize(IntPtr wndHandle)
        {
            dinput = new DirectInput();

            try
            {
                keyboard = new Keyboard(dinput);
                keyboard.SetCooperativeLevel(wndHandle, CooperativeLevel.Nonexclusive | CooperativeLevel.Foreground);
                mouse = new Mouse(dinput);
                mouse.SetCooperativeLevel(wndHandle, CooperativeLevel.Nonexclusive | CooperativeLevel.Foreground);
            }
            catch (DirectInputException ex)
            {
                System.Windows.Forms.MessageBox.Show(ex.Message);
                return;
            }

            oldKeyState = new KeyboardState();
            oldMouseState = new MouseState();

            keyState = new KeyboardState();
            mouseState = new MouseState();

        }

        #region Keyboard key actions
        /// <summary>
        /// This actions is invoked when the key is clicked.
        /// </summary>
        /// <param name="key">Key on keyboard.</param>
        /// <param name="action">Action witch will be invoked.</param>
        public void AddKeyClickAction(Key key, Action action)
        {
            if (keyClickActions == null)
                keyClickActions = new Dictionary<Key, Action>();
            keyClickActions.Add(key, action);
        }

        /// <summary>
        /// This actions is invoked when the key is pressed.
        /// </summary>
        /// <param name="key">Key on keyboard.</param>
        /// <param name="action">Action witch will be invoked.</param>
        public void AddKeyIsPressedAction(Key key, Action action)
        {
            if (keyIsPressedActions == null)
                keyIsPressedActions = new Dictionary<Key,Action>();
            keyIsPressedActions.Add(key, action);
        }

        /// <summary>
        /// This actions is invoked when the key is released.
        /// </summary>
        /// <param name="key">Key on keyboard.</param>
        /// <param name="action">Action witch will be invoked.</param>
        public void AddKeyIsReleasedAction(Key key, Action action)
        {
            if (keyIsReleasedActions == null)
                keyIsReleasedActions = new Dictionary<Key, Action>();
            keyIsReleasedActions.Add(key, action);
        }
        #endregion

        #region Mouse but actions
        /// <summary>
        /// This actions is invoked when the mouse button is clicked.
        /// </summary>
        /// <param name="mouseBut">Mouse button.</param>
        /// <param name="action">Action witch will be invoked.</param>
        public void AddMouseClickAction(MouseObject mouseBut, Action action)
        {
            if (mouseClickActions == null)
                mouseClickActions = new Dictionary<MouseObject, Action>();
            mouseClickActions.Add(mouseBut, action);
        }

        /// <summary>
        /// This actions is invoked when the mouse button is pressed.
        /// </summary>
        /// <param name="mouseBut">Mouse button.</param>
        /// <param name="action">Action witch will be invoked.</param>
        public void AddMouseIsPressedAction(MouseObject mouseBut, Action action)
        {
            if (mouseIsPressedActions == null)
                mouseIsPressedActions = new Dictionary<MouseObject, Action>();
            mouseIsPressedActions.Add(mouseBut, action);
        }

        /// <summary>
        /// This actions is invoked when the mouse button is released.
        /// </summary>
        /// <param name="mouseBut">Mouse button.</param>
        /// <param name="action">Action witch will be invoked.</param>
        public void AddMouseIsReleasedAction(MouseObject mouseBut, Action action)
        {
            if (mouseIsReleasedActions == null)
                mouseIsReleasedActions = new Dictionary<MouseObject, Action>();
            mouseIsReleasedActions.Add(mouseBut, action);
        }
        #endregion

        #region Mouse move actions
        /// <summary>
        /// This actions is invoked when the keyboard key is clicked and mouse moves.
        /// </summary>
        /// <param name="key">Key on keyboard.</param>
        /// <param name="action">Action witch will be invoked (arg1 is mouse X movement, arg2 is mouse Y movement, arg3 is mouse wheel movement</param>
        public void AddMouseMoveKeyAction(Key key, Action<int, int, int> action)
        {
            if (mouseMoveKeyActions == null)
                mouseMoveKeyActions = new Dictionary<Key, Action<int, int, int>>();
            mouseMoveKeyActions.Add(key, action);
        }

        /// <summary>
        /// This actions is invoked when the mouse button is clicked and mouse moves.
        /// </summary>
        /// <param name="mouseBut">Mouse button.</param>
        /// <param name="action">Action witch will be invoked (arg1 is mouse X movement, arg2 is mouse Y movement, arg3 is mouse wheel movement</param>
        public void AddMouseMoveButAction(MouseObject mouseBut, Action<int, int, int> action)
        {
            if (mouseMoveButActions == null)
                mouseMoveButActions = new Dictionary<MouseObject, Action<int, int, int>>();
            mouseMoveButActions.Add(mouseBut, action);
        }

        /// <summary>
        /// This actions is invoked when mouse moves.
        /// </summary>
        /// <param name="action">Action witch will be invoked (arg1 is mouse X movement, arg2 is mouse Y movement, arg3 is mouse wheel movement</param>
        public void AddMouseMoveAction(Action<int, int, int> action)
        {
            mouseMove = action;
        }
        #endregion

        /// <summary>
        /// Update input states.
        /// </summary>
        public void Update()
        {
            if (acquiredKeyboard)
            {
                UpdateKeyboard();
            }
            if (acquiredMouse)
            {
                UpdateMouse();
            }
        }

        /// <summary>
        /// Dispose all resources.
        /// </summary>
        public void Dispose()
        {
            if(keyClickActions != null)
                keyClickActions.Clear();
            if(keyIsPressedActions != null)
                keyIsPressedActions.Clear();
            if(keyIsReleasedActions != null)
                keyIsReleasedActions.Clear();
            if(mouseClickActions != null)
                mouseClickActions.Clear();
            if(mouseIsPressedActions != null)
                mouseIsPressedActions.Clear();
            if(mouseIsReleasedActions != null)
                mouseIsReleasedActions.Clear();
            if(mouseMoveKeyActions != null)
                mouseMoveKeyActions.Clear();
            if(mouseMoveButActions != null)
                mouseMoveButActions.Clear();
            keyClickActions = null;
            keyIsPressedActions = null;
            keyIsReleasedActions = null;
            mouseClickActions = null;
            mouseIsPressedActions = null;
            mouseIsReleasedActions = null;
            mouseMoveKeyActions = null;
            mouseMoveButActions = null;
            mouseMove = null;

            if (dinput != null && !dinput.Disposed)
            {
                dinput.Dispose();
                dinput = null;
            }
            if (mouse != null && !mouse.Disposed)
            {
                mouse.Dispose();
                mouse = null;
            }
            if (keyboard != null && !keyboard.Disposed)
            {
                keyboard.Dispose();
                keyboard = null;
            }
        }
        #endregion
        #region Private methods
        /// <summary>
        /// Update keyboard states.
        /// </summary>
        private void UpdateKeyboard()
        {
            if (keyboard.Acquire().IsFailure)
                return;

            if (keyboard.Poll().IsFailure)
                return;

            keyState = keyboard.GetCurrentState();
            if (Result.Last.IsFailure)
                return;

            //Key click actions
            if (keyClickActions != null)
            {
                foreach (Key key in keyClickActions.Keys)
                {
                    if (KeypressTest(key))
                    {
                        keyClickActions[key].Invoke();
                    }
                }
            }

            //Key is pressed actions
            if (keyIsPressedActions != null)
            {
                foreach (Key key in keyIsPressedActions.Keys)
                {
                    if (keyState.IsPressed(key))
                    {
                        keyIsPressedActions[key].Invoke();
                    }
                }
            }

            //Key is released actions
            if (keyIsReleasedActions != null)
            {
                foreach (Key key in keyIsReleasedActions.Keys)
                {
                    if (keyState.IsReleased(key))
                    {
                        keyIsReleasedActions[key].Invoke();
                    }
                }
            }

            oldKeyState = keyState;
        }

        /// <summary>
        /// Updates mouse states.
        /// </summary>
        private void UpdateMouse()
        {
            if (mouse.Acquire().IsFailure)
                return;

            if (mouse.Poll().IsFailure)
                return;

            mouseState = mouse.GetCurrentState();
            if (Result.Last.IsFailure)
                return;

            
            //Mouse click actions
            if (mouseClickActions != null)
            {
                foreach (MouseObject buts in mouseClickActions.Keys)
                {
                    if (MousepressTest(buts) && (mouseState.X <= 1) && (mouseState.Y <= 1) && (mouseState.Z <= 1))
                    {
                        mouseClickActions[buts].Invoke();
                    }
                }
            }

            //Mouse is pressed actions
            if (mouseIsPressedActions != null)
            {
                foreach (MouseObject buts in mouseIsPressedActions.Keys)
                {
                    if (mouseState.IsPressed((int)buts))
                    {
                        mouseIsPressedActions[buts].Invoke();
                    }
                }
            }

            //Mouse is released actions
            if (mouseIsReleasedActions != null)
            {
                foreach (MouseObject buts in mouseIsReleasedActions.Keys)
                {
                    if (mouseState.IsReleased((int)buts))
                    {
                        mouseIsReleasedActions[buts].Invoke();
                    }
                }
            }

            //Mouse move with pressed key actions
            if (mouseMoveKeyActions != null)
            {
                if ((mouseState.X != 0) || (mouseState.Y != 0) || (mouseState.Z != 0))
                {
                    foreach (Key key in mouseMoveKeyActions.Keys)
                    {
                        if (keyState.IsPressed(key))
                        {
                            mouseMoveKeyActions[key].Invoke(mouseState.X, mouseState.Y, mouseState.Z);
                        }
                    }
                }
            }

            //Mouse move with pressed button actions
            if (mouseMoveButActions != null)
            {
                if ((mouseState.X != 0) || (mouseState.Y != 0) || (mouseState.Z != 0))
                {
                    foreach (MouseObject buts in mouseMoveButActions.Keys)
                    {
                        if (mouseState.IsPressed((int)buts))
                        {
                            mouseMoveButActions[buts].Invoke(mouseState.X, mouseState.Y, mouseState.Z);
                        }
                    }
                }
            }

            if (mouseMove != null)
            {
                if ((mouseState.X != 0) || (mouseState.Y != 0) || (mouseState.Z != 0))
                {
                    mouseMove.Invoke(mouseState.X, mouseState.Y, mouseState.Z);
                }
            }

            oldMouseState = mouseState;
        }

        /// <summary>
        /// Keypress Test method check if the key is pressed or not.
        /// </summary>
        /// <param name="theKey">What key will be tested</param>
        /// <returns>Returns true if the key is pressed</returns>
        private bool KeypressTest(Key theKey)
        {
            if (keyState.IsPressed(theKey) && oldKeyState.IsReleased(theKey))
                return true;
            return false;
        }

        /// <summary>
        /// Mousepress Test method check if the mouse button is pressed or not.
        /// </summary>
        /// <param name="button">What button will be tested</param>
        /// <returns>Returns true if the key is pressed</returns>
        private bool MousepressTest(MouseObject button)
        {
            if (mouseState.IsPressed((int)button) && oldMouseState.IsReleased((int)button))
                return true;
            return false;
        }
        #endregion
    }
}
