﻿using System;
using System.Drawing;
using System.ComponentModel;
using SlimDX;
using Matrix = SlimDX.Matrix;
using RenderMgr = SlimDXRenderSystem.Managers.RenderManager;
using SceneMgr = SlimDXRenderSystem.Managers.SceneManager;
using InputMgr = SlimDXRenderSystem.Managers.InputManager;
using TimeMgr = SlimDXRenderSystem.Managers.TimeManager;
using System.Threading;
using Zcu.CurvatureBenchmark.Framework;

namespace Zcu.Mve.ZMalik.SlimDXRenderer
{

    public enum PickType
    {
        NONE,
        POINTS,
        PRIMITIVES
    }

    public delegate void PointPickedEvent(Object3D obj, int selectedIndex);

    /// <summary>
    /// This class manage all module stuff like objects, cameras, lights etc.
    /// </summary>
    public sealed class ModuleManager : SlimDXRenderSystem.Core.Singleton<ModuleManager>
    {
        #region Fields
        PickType pick = PickType.NONE;
        private Color4 backColor;
        private int numScreenShots = 0;
        private float progress = 0.0f;
        private bool sameCon;
        private bool showColor;
        private bool showEdgeColor;
        private bool showVector;
        private int currentMesh;
        private int meshCount;

        private int mouseX;
        private int mouseY;

        private int pickedPointId;
        private int pickedPrimId;

        private Size clientSize;

        private SlimDXRenderSystem.Core.Group meshGroup;
        private Object3D[] objects;

        private RenderSetting rendersetting;
        private InputSetting inputsetting;

        #region Properties
        public PickType Pick
        {
            get { return pick; }
            set { pick = value; }
        }

        public event PointPickedEvent PointPicked;

        public int NumScreenShots
        {
            get { return numScreenShots; }
            set { numScreenShots = value; }
        }

        /// <summary>
        /// Progress of buffers initialization.
        /// </summary>
        public float Progress
        {
            get { return progress; }
            set { progress = value; }
        }

        public RenderSetting Rendersetting
        {
            get { return rendersetting; }
        }

        public InputSetting Inputsetting
        {
            get { return inputsetting; }
        }

        public int CurrentMesh
        {
            get { return currentMesh; }
            set
            {
                CurrentObject.Object.Visible = false;
                if (value >= meshCount)
                    currentMesh = 0;
                else if (value < 0)
                    currentMesh = meshCount - 1;
                else
                    currentMesh = value;
                CurrentObject.Object.Visible = true;

			

            }
        }

        public int MeshCount
        {
            get { return meshCount; }
        }

        public SlimDXRenderSystem.Core.Group MeshGroup
        {
            get { return meshGroup; }
        }

        public Object3D CurrentObject
        {
            get { return objects[currentMesh]; }
        }

        public Size ClientSize
        {
            get { return clientSize; }
            set { clientSize = value; }
        }

        public int MouseX
        {
            get { return mouseX; }
            set { mouseX = value; }
        }

        public int MouseY
        {
            get { return mouseY; }
            set { mouseY = value; }
        }

        #endregion
        #endregion

        #region Public Methods
        /// <summary>
        /// This method setup the attributes.
        /// </summary>
        /// <param name="bgColor">Background color of rendering window.</param>
        /// <param name="meshCount">Number of meshes.</param>
        /// <param name="sameConnectivity">Is mesh connected with it selves.</param>
        /// <param name="showColorAttribute">Determines if the color attribute will be shown.</param>
        /// <param name="colorAttributeName">Name of the color attribute.</param>
        /// <param name="showVectorAttribute">Determines if the vector attribute will be shown.</param>
        /// <param name="vectorAttributeName">Determines if the vector attribute will be shown.</param>
        public void Setup(ColorRGBA bgColor, int meshCount, bool sameConnectivity, bool showColorAttribute,
            bool showVectorAttribute, bool showEdgeColorAttribute)
        {
            this.backColor = new Color4(bgColor.AlphaF, bgColor.RedF, bgColor.GreenF, bgColor.BlueF);
            this.sameCon = sameConnectivity;
            this.showColor = showColorAttribute;
            this.showVector = showVectorAttribute;
            this.showEdgeColor = showEdgeColorAttribute;
            this.progress = 0.0f;
            this.currentMesh = 0;
            this.meshCount = meshCount;

            objects = new Object3D[meshCount];

            inputsetting = new InputSetting();
            rendersetting = new RenderSetting();

            SceneMgr.Instance.CreateCamera("Perspective");
            SceneMgr.Instance.Camera.Perspective(45, (float)clientSize.Width / clientSize.Height, 0.1f, 100.0f);
            SceneMgr.Instance.Camera.LookAt(new Vector3(0.0f, 0.0f, 20.0f), new Vector3(0.0f, 0.0f, 0.0f), Vector3.UnitY);
            RenderMgr.Instance.RenderTextAction = WriteText;
        }

        /// <summary>
        /// Initialize buffers for all meshes.
        /// </summary>
        public void Initialize()
        {
            try
            {
                meshGroup = SceneMgr.Instance.CreateGroup("Animation");

                for (int i = 0; i < objects.Length; i++)
                {
                    if (objects[i] != null)
                    {
                        objects[i].InitVertexBuffers();
                        if (sameCon && i != 0)
                            objects[i].Object.SetMeshSameIndexBuffer(objects[0].Object);
                        else
                            objects[i].InitIndexBuffers();

                        meshGroup.AddObject(objects[i].Object); // Add to group
                    }
                    Progress = (float)(i + 1) / objects.Length;
                }
                Progress = 1.0f;
                GC.Collect();
                objects[0].Object.Visible = true;

                RenderMgr.Instance.Renderer.BackgroundColor = backColor;
                
            }
            catch (ThreadAbortException e)
            {
                System.Windows.Forms.MessageBox.Show(e.StackTrace, e.Message);
                Thread.ResetAbort();
                throw new Exception("Thread needs to be aborted", e);
            }
        }

        /// <summary>
        /// Save all points, primitives, colors etc. for the mesh and prepare for index and vertex buffer initialization.
        /// </summary>
        /// <param name="index">Index of mesh.</param>
        /// <param name="mesh">Mesh witch will be used as source.</param>
        public void SetMesh(int index, TriangleMesh mesh, string desc)
        {
            if (index > objects.Length)
                throw new Exception("Index is out of bounds.");

			objects[index] = new Object3D("Mesh" + index.ToString() + ": " + desc, showColor, showVector, showEdgeColor);
            objects[index].SetMesh(mesh);
        }

        /// <summary>
        /// Call the render method of rendermanager.
        /// </summary>
        public void Update()
        {
            
            
        }

        public void Picking()
        {
            if (pick == PickType.NONE)
                return;

            Matrix m = SceneMgr.Instance.WorldMatrix * Matrix.Translation(meshGroup.Position.X, meshGroup.Position.Y, meshGroup.Position.Z)
                * SceneMgr.Instance.Camera.ViewMatrix * SceneMgr.Instance.Camera.ProjectionMatrix;

            switch (pick)
            {
                case PickType.POINTS:
                    double minDist = double.MaxValue;
                    pickedPointId = 0;
                    
                    for (int i = 0; i < CurrentObject.Points.Length; i++)
                    {
                        Point3D p = (Point3D)CurrentObject.Points[i];
                        Vector3 v1 = new Vector3((float)p.X, (float)p.Y, (float)p.Z);
                        Vector4 v = Vector3.Transform(v1, m);

                        if (v.W <= 0)
                            continue;

                        int x = (int)(v.X * RenderMgr.Instance.RenderWindow.Width / 2 / (v.W) + RenderMgr.Instance.RenderWindow.Width / 2);
                        int y = (int)(-v.Y * RenderMgr.Instance.RenderWindow.Height / 2 / (v.W) + RenderMgr.Instance.RenderWindow.Height / 2);

                        if (x > RenderMgr.Instance.RenderWindow.Width || y > RenderMgr.Instance.RenderWindow.Height || x < 0 || y < 0)
                            continue;

                        double distance = (x - mouseX) * (x - mouseX) + (y - mouseY) * (y - mouseY);

                        if (distance < minDist)
                        {
                            pickedPointId = i;
                            minDist = distance;
                        }
                    }
                    var listener = PointPicked;
                    if(listener!=null)
                    {
                        listener(CurrentObject, pickedPointId);
                    }
                    break;
                case PickType.PRIMITIVES:
                    minDist = double.MaxValue;
                    pickedPrimId = 0;

                    for (int i = 0; i < CurrentObject.Primitives.Length; i++)
                    {
                        Point3D p = new Point3D();
                        //if (CurrentObject.Primitives.ElemType == typeof(Triangle))
                        {
                            Triangle t = (Triangle)CurrentObject.Primitives[i];
                            Point3D p1 = (Point3D)CurrentObject.Points[t.V1];
                            Point3D p2 = (Point3D)CurrentObject.Points[t.V2];
                            Point3D p3 = (Point3D)CurrentObject.Points[t.V3];
                            p = (p1 + p2 + p3) * (1 / 3.0);
                        }
                        /*else if (CurrentObject.Primitives.ElemType == typeof(PolyLine))
                        {
                            PolyLine l = (PolyLine)CurrentObject.Primitives[i];
                            p = (Point3D)CurrentObject.Points[l[l.PointCount / 2]];
                        }
                        else if (CurrentObject.Primitives.ElemType == typeof(Line))
                        {
                            Line l = (Line)CurrentObject.Primitives[i];
                            p = (Point3D)CurrentObject.Points[l.V1];
                        }*/

                        Vector3 v1 = new Vector3((float)p.X, (float)p.Y, (float)p.Z);
                        Vector4 v = Vector3.Transform(v1, m);

                        if (v.W <= 0)
                            continue;

                        int x = (int)(v.X * RenderMgr.Instance.RenderWindow.Width / 2 / (v.W) + RenderMgr.Instance.RenderWindow.Width / 2);
                        int y = (int)(-v.Y * RenderMgr.Instance.RenderWindow.Height / 2 / (v.W) + RenderMgr.Instance.RenderWindow.Height / 2);

                        if (x > RenderMgr.Instance.RenderWindow.Width || y > RenderMgr.Instance.RenderWindow.Height || x < 0 || y < 0)
                            continue;

                        double distance = (x - mouseX) * (x - mouseX) + (y - mouseY) * (y - mouseY);

                        if (distance < minDist)
                        {
                            pickedPrimId = i;
                            minDist = distance;
                        }
                    }
                    break;
            }
        }

        public void WriteText()
        {
            //Display Indices
            Matrix m = SceneMgr.Instance.WorldMatrix * Matrix.Translation(meshGroup.Position.X, meshGroup.Position.Y, meshGroup.Position.Z)
                * SceneMgr.Instance.Camera.ViewMatrix * SceneMgr.Instance.Camera.ProjectionMatrix;

            if (rendersetting.DisplayIndices)
            {
                for (int i = 0; i < CurrentObject.Points.Length; i++)
                {
                    Point3D p = (Point3D)CurrentObject.Points[i];
                    Vector3 v1 = new Vector3((float)p.X, (float)p.Y, (float)p.Z);
                    Vector4 v = Vector3.Transform(v1, m);

                    int x = (int)(v.X * RenderMgr.Instance.RenderWindow.Width / 2 / (v.W) + RenderMgr.Instance.RenderWindow.Width / 2);
                    int y = (int)(-v.Y * RenderMgr.Instance.RenderWindow.Height / 2 / (v.W) + RenderMgr.Instance.RenderWindow.Height / 2);

                    RenderMgr.Instance.DrawText(x, y, i.ToString());

                }
            }

            if (rendersetting.DisplayTriangleIndices)
            {
                for (int i = 0; i < CurrentObject.Primitives.Length; i++)
                {
                    Point3D p = new Point3D();
                    //if (CurrentObject.Primitives.ElemType == typeof(Triangle))
                    {
                        Triangle t = (Triangle)CurrentObject.Primitives[i];
                        Point3D p1 = (Point3D)CurrentObject.Points[t.V1];
                        Point3D p2 = (Point3D)CurrentObject.Points[t.V2];
                        Point3D p3 = (Point3D)CurrentObject.Points[t.V3];
                        p = (p1 + p2 + p3) * (1 / 3.0);
                    }
                    /*else if (CurrentObject.Primitives.ElemType == typeof(PolyLine))
                    {
                        PolyLine l = (PolyLine)CurrentObject.Primitives[i];
                        p = (Point3D)CurrentObject.Points[l[l.PointCount / 2]];
                    }
                    else if (CurrentObject.Primitives.ElemType == typeof(Line))
                    {
                        Line l = (Line)CurrentObject.Primitives[i];
                        p = (Point3D)CurrentObject.Points[l.V1];
                    }*/
                    Vector3 v1 = new Vector3((float)p.X, (float)p.Y, (float)p.Z);
                    Vector4 v = Vector3.Transform(v1, m);

                    int x = (int)(v.X * RenderMgr.Instance.RenderWindow.Width / 2 / (v.W) + RenderMgr.Instance.RenderWindow.Width / 2);
                    int y = (int)(-v.Y * RenderMgr.Instance.RenderWindow.Height / 2 / (v.W) + RenderMgr.Instance.RenderWindow.Height / 2);

                    RenderMgr.Instance.DrawText(x, y, i.ToString());
                }
            }

            if (!rendersetting.DisplayIndices && !rendersetting.DisplayTriangleIndices)
            {
                switch (pick)
                {
                    case PickType.POINTS:
                        Point3D p = (Point3D)CurrentObject.Points[pickedPointId];
                        Vector3 v1 = new Vector3((float)p.X, (float)p.Y, (float)p.Z);
                        Vector4 v = Vector3.Transform(v1, m);

                        int x = (int)(v.X * RenderMgr.Instance.RenderWindow.Width / 2 / (v.W) + RenderMgr.Instance.RenderWindow.Width / 2);
                        int y = (int)(-v.Y * RenderMgr.Instance.RenderWindow.Height / 2 / (v.W) + RenderMgr.Instance.RenderWindow.Height / 2);

                        RenderMgr.Instance.DrawPoint(new Vector3(x, y, 0));
                        RenderMgr.Instance.DrawText(x, y, pickedPointId.ToString());
                        break;

                    case PickType.PRIMITIVES:
                        p = new Point3D();
                        //if (CurrentObject.Primitives.ElemType == typeof(Triangle))
                        {
                            Triangle t = (Triangle)CurrentObject.Primitives[pickedPrimId];
                            Point3D p1 = (Point3D)CurrentObject.Points[t.V1];
                            Point3D p2 = (Point3D)CurrentObject.Points[t.V2];
                            Point3D p3 = (Point3D)CurrentObject.Points[t.V3];
                            p = (p1 + p2 + p3) * (1 / 3.0);

                            v1 = new Vector3((float)p1.X, (float)p1.Y, (float)p1.Z);
                            v = Vector3.Transform(v1, m);

                            x = (int)(v.X * RenderMgr.Instance.RenderWindow.Width / 2 / (v.W) + RenderMgr.Instance.RenderWindow.Width / 2);
                            y = (int)(-v.Y * RenderMgr.Instance.RenderWindow.Height / 2 / (v.W) + RenderMgr.Instance.RenderWindow.Height / 2);

                            RenderMgr.Instance.DrawPoint(new Vector3(x, y, 0));

                            v1 = new Vector3((float)p2.X, (float)p2.Y, (float)p2.Z);
                            v = Vector3.Transform(v1, m);

                            x = (int)(v.X * RenderMgr.Instance.RenderWindow.Width / 2 / (v.W) + RenderMgr.Instance.RenderWindow.Width / 2);
                            y = (int)(-v.Y * RenderMgr.Instance.RenderWindow.Height / 2 / (v.W) + RenderMgr.Instance.RenderWindow.Height / 2);

                            RenderMgr.Instance.DrawPoint(new Vector3(x, y, 0));

                            v1 = new Vector3((float)p3.X, (float)p3.Y, (float)p3.Z);
                            v = Vector3.Transform(v1, m);

                            x = (int)(v.X * RenderMgr.Instance.RenderWindow.Width / 2 / (v.W) + RenderMgr.Instance.RenderWindow.Width / 2);
                            y = (int)(-v.Y * RenderMgr.Instance.RenderWindow.Height / 2 / (v.W) + RenderMgr.Instance.RenderWindow.Height / 2);

                            RenderMgr.Instance.DrawPoint(new Vector3(x, y, 0));

                        }
                        /*else if (CurrentObject.Primitives.ElemType == typeof(PolyLine))
                        {
                            PolyLine l = (PolyLine)CurrentObject.Primitives[pickedPrimId];
                            p = (Point3D)CurrentObject.Points[l[l.PointCount / 2]];
                        }
                        else if (CurrentObject.Primitives.ElemType == typeof(Line))
                        {
                            Line l = (Line)CurrentObject.Primitives[pickedPrimId];
                            p = (Point3D)CurrentObject.Points[l.V1];
                        }*/

                        v1 = new Vector3((float)p.X, (float)p.Y, (float)p.Z);
                        v = Vector3.Transform(v1, m);

                        x = (int)(v.X * RenderMgr.Instance.RenderWindow.Width / 2 / (v.W) + RenderMgr.Instance.RenderWindow.Width / 2);
                        y = (int)(-v.Y * RenderMgr.Instance.RenderWindow.Height / 2 / (v.W) + RenderMgr.Instance.RenderWindow.Height / 2);

                        RenderMgr.Instance.DrawText(x, y, pickedPrimId.ToString());
                        break;
                }
            }
			
			//this.CurrentMesh
			RenderMgr.Instance.DrawText(10, 50, this.objects[this.currentMesh].Name);
            RenderMgr.Instance.DrawText(10, 10, TimeMgr.Instance.FPS.ToString());
        }

        /// <summary>
        /// Calculate optimal size and position of mesh
        /// </summary>
        /// <param name="mesh">Mesh witch transformations will be calculated</param>
        public void SetTransforms(TriangleMesh mesh)
        {
            Matrix toCenter, scale;
            scale = Matrix.Identity;

            Point3D minPoint = new Point3D(double.MaxValue, double.MaxValue, double.MaxValue);
            Point3D maxPoint = new Point3D(double.MinValue, double.MinValue, double.MinValue);
            for (int i = 0; i < mesh.Points.Length; i++)
            {
                if (mesh.Points[i].X < minPoint.X) minPoint.X = mesh.Points[i].X;
                if (mesh.Points[i].Y < minPoint.Y) minPoint.Y = mesh.Points[i].Y;
                if (mesh.Points[i].Z < minPoint.Z) minPoint.Z = mesh.Points[i].Z;
                if (mesh.Points[i].X > maxPoint.X) maxPoint.X = mesh.Points[i].X;
                if (mesh.Points[i].Y > maxPoint.Y) maxPoint.Y = mesh.Points[i].Y;
                if (mesh.Points[i].Z > maxPoint.Z) maxPoint.Z = mesh.Points[i].Z;
            }
            Point3D MidPoint = 0.5 * (minPoint + maxPoint);

            toCenter = Matrix.Identity * Matrix.Translation((float)-MidPoint.X, (float)-MidPoint.Y, (float)-MidPoint.Z);
            float w = (float)(maxPoint.X - minPoint.Y);
            float ar = (float)clientSize.Width / clientSize.Height;
            float k2 = ar / w;
            scale *= Matrix.Scaling(k2 * 2, k2 * 2, k2 * 2);
            SceneMgr.Instance.WorldMatrix = toCenter * scale;
        }

        /// <summary>
        /// Dispose all resources.
        /// </summary>
        public void Dispose()
        {

        }
        #endregion
    }

    
}
