﻿using System;
using System.Drawing;
using System.Windows.Forms;
using SlimDXRenderSystem;
using SlimDXRenderSystem.Managers;
using Zcu.CurvatureBenchmark.Framework;


namespace Zcu.Mve.ZMalik.SlimDXRenderer
{
    /// <summary>
    /// Main window form that contains all components to work with.
    /// </summary>
    public partial class SlimDXRendererWindow : Form
    {
        /// <summary>
        /// Contructor of the main window.
        /// </summary>
        /// <param name="d3dVersion">Version of Direct3D to be used.</param>
        /// <param name="width">Width of window.</param>
        /// <param name="height">Height of window.</param>
        /// <param name="bgColor">Background color for rendering frame.</param>
        /// <param name="meshCount">Number of meshes to be loaded.</param>
        /// <param name="sameConnectivity">Same connectivity.</param>
        /// <param name="showColorAttribute">Color attribute use.</param>
        /// <param name="colorAttributeName">Color attribute name.</param>
        /// <param name="showVectorAttribute">Vector attribute use.</param>
        /// <param name="vectorAttributeName">Vector attribute name.</param>
        public SlimDXRendererWindow(D3DVersion d3dVersion, Size windowSize, ColorRGBA bgColor, int meshCount, bool sameConnectivity, bool showColorAttribute,
            bool showVectorAttribute, bool showEdgeColorAttribute)
        {
            RenderManager.Instance.D3dVersion = d3dVersion;
            InitializeComponent();
            this.ClientSize = windowSize;
            this.timer1.Interval = 40;
            this.timer1.Start();
            ModuleManager.Instance.ClientSize = windowSize;
            ModuleManager.Instance.Setup(bgColor, meshCount, sameConnectivity, showColorAttribute, showVectorAttribute, showEdgeColorAttribute);
            this.propertyGrid2.SelectedObject = ModuleManager.Instance.Rendersetting;
            this.rendererControl1.UpdateAction = ModuleManager.Instance.Update;
            ModuleManager.Instance.PointPicked += Instance_PointPicked;
        }

        private void Instance_PointPicked(Object3D obj, int selectedIndex)
        {
            obj.pointAttribs.Index = selectedIndex;
            PointInfoGrid.SelectedObject = obj.pointAttribs;
            PointInfoGrid.Refresh();
        }

        protected override void OnCreateControl()
        {
            base.OnCreateControl();
            if (!DesignMode)
            {
                try
                {
                    Initialize();
                }
                catch (Exception e)
                {
                    System.Windows.Forms.MessageBox.Show(e.StackTrace, e.Message);
                    this.Dispose();
                }
            }
        }

        public void Initialize()
        {
            ModuleManager.Instance.Initialize();

            //Scene Tree view setting
            TreeNode tn = this.treeSceneView.Nodes.Add("Meshes");
            foreach (SlimDXRenderSystem.Core.Group grp in SceneManager.Instance.SceneGroups.Values)
            {
                if (grp.NumberOfObjects > 0)
                {
                    TreeNode gtn = tn.Nodes.Add(grp.Name);
                    gtn.Tag = grp;
                    foreach (SlimDXRenderSystem.Core.Object3D obj in grp.Objects.Values)
                    {
                        TreeNode obtn = gtn.Nodes.Add(obj.Name);
                        obtn.Tag = obj;
                        obtn.Nodes.Add("Material").Tag = obj.Material;
                    }
                }
            }

            tn = this.treeSceneView.Nodes.Add("Cameras");
            foreach (SlimDXRenderSystem.Core.Camera cam in SceneManager.Instance.SceneCameras.Values)
            {
                tn.Nodes.Add(cam.Name).Tag = cam;
            }

            tn = this.treeSceneView.Nodes.Add("Lights");
            foreach (SlimDXRenderSystem.Core.Light light in SceneManager.Instance.SceneLights.Values)
            {
                tn.Nodes.Add(light.Name).Tag = light;
            }
        }

        private void exitToolStripMenuItem_Click(object sender, EventArgs e)
        {
            this.Close();
        }

        private void showNormals_Click(object sender, EventArgs e)
        {
            ModuleManager.Instance.Rendersetting.ToggleShowNormals();
        }

        private void tButSettings_Click(object sender, EventArgs e)
        {
            this.propertyGrid2.SelectedObject = ModuleManager.Instance.Rendersetting;
        }

        private void tButCamera_Click(object sender, EventArgs e)
        {
            this.propertyGrid2.SelectedObject = SceneManager.Instance.Camera;
        }

        private void tbLight_Click(object sender, EventArgs e)
        {

        }

        private void timer1_Tick(object sender, EventArgs e)
        {
            if (ModuleManager.Instance.Rendersetting.PlayAnimation)
                ModuleManager.Instance.CurrentMesh++;
        }

        private void aboutToolStripMenuItem_Click(object sender, EventArgs e)
        {
            new AboutDialog().Show();
        }

        private void settingsToolStripMenuItem_Click(object sender, EventArgs e)
        {
            new SettingsWindow().ShowDialog();
        }

        private void treeSceneView_AfterSelect(object sender, TreeViewEventArgs e)
        {
            this.propertyGrid2.SelectedObject = e.Node.Tag;
        }

        private void cameraSaveBut_Click(object sender, EventArgs e)
        {
            ModuleManager.Instance.Rendersetting.SaveCameraPosition();
        }

        private void cameraLoadBut_Click(object sender, EventArgs e)
        {
            ModuleManager.Instance.Rendersetting.LoadCameraPosition();
        }

        private void screenshotButton_Click(object sender, EventArgs e)
        {
            ModuleManager.Instance.Rendersetting.SaveSingleFrame();
        }

        private void animationButton_Click(object sender, EventArgs e)
        {
            ModuleManager.Instance.Rendersetting.SaveAnimationFrames();
        }

        private void lightingButton_Click(object sender, EventArgs e)
        {
            ModuleManager.Instance.Rendersetting.ToggleLighting();
        }

        private void pickPoints_Click(object sender, EventArgs e)
        {
            ModuleManager.Instance.Pick = PickType.POINTS;
        }

        private void manipulate_Click(object sender, EventArgs e)
        {
            ModuleManager.Instance.Pick = PickType.NONE;
        }

        private void pickPrimitives_Click(object sender, EventArgs e)
        {
            ModuleManager.Instance.Pick = PickType.PRIMITIVES;
        }

        private void rendererControl1_MouseMove(object sender, MouseEventArgs e)
        {
            ModuleManager.Instance.MouseX = e.X;
            ModuleManager.Instance.MouseY = e.Y;
        }
    }
}
